/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#define __STDC_FORMAT_MACROS
#include <stdint.h>
#include <inttypes.h>
#include <limits.h>
#include <map>

#include "astnode.hpp"

#include "../util.hpp"
#include "../codegen/llvm/codegen.hpp"
#include "../codegen/opencl/codegen.hpp"
#include "../codegen/pseudopc/codegen.hpp"

CodeGenerationException::CodeGenerationException(CodeGeneratorType type, AstNode *node, const char *format, ...) :
	m_type(type), m_node(node)
{
	char b[65536];
	
	va_list argp;
	va_start(argp, format);
	vsnprintf(b, 65536, format, argp);
	va_end(argp);
	
	m_str = b;
}

CodeGenerationException::CodeGenerationException(CodeGeneratorType type, AstNode *node) :
	CodeGenerationException(type, node, "No code generator for this node type and and generator type: %s",
		(CODEGEN_NONE == type) ? "None" :
		(CODEGEN_LLVM == type) ? "LLVM" :
		(CODEGEN_PSEUDOPC == type) ? "Pseudopc" :
			/* Default */ "Unknown")
{
}

std::string CodeGenerationException::GetMessage()
{
	return m_str;
}

AstNode::AstNode() :
	m_generator(0),
	m_id(GetUniqueId()),
	m_filename(""),
	m_line(0)
{
}

AstNode::~AstNode()
{
	if (m_generator){
		delete m_generator;
	}
}

void AstNode::SetTreeCodeGenerator(CodeGeneratorType type)
{
	SetCodeGenerator(type);
	std::list<AstNode *> children = GetChildren();
	for (AstNode *child: children){
		if (child){
			child->SetTreeCodeGenerator(type);
		}
	}
}

std::string AstNode::GetDot()
{
	char b[1024];
	snprintf(b, 1024, "ap%p [label=\"%s\" shape=%s style=filled fillcolor=\"#%s\"];\n",
		(void *)this, GetDotName().c_str(), GetDotShape().c_str(), GetDotBgColour().c_str());
	std::string r = b;
	
	for (AstNode *node: GetChildren()){
		if (node){
			r += node->GetDot();
			snprintf(b, 1024, "ap%p -> ap%p\n", (void *)this, (void *)node);
			r += b;
		}
	}
	
	return r;
}

std::string AstNode::GetTextTree()
{
	std::set<AstNode *> previous_nodes;
	return GetTextTree(previous_nodes);
}

std::string AstNode::GetTextTree(std::set<AstNode *> previous_nodes)
{
	if (previous_nodes.count(this)){
		return "[*] Cycle\n";
	}
	
	previous_nodes.insert(this);
	
	std::string r = PrintString("[*] %s ---- %s\n", GetNormalName().c_str(), GetSourceLocation().c_str());
	
	for (AstNode *node: GetChildren()){
		if (node){
			r += Indent(node->GetTextTree(previous_nodes), 1);
		}
	}
	
	if (GetLinks().size()){
		r += "\t[*] Links\n";
		for (AstNode *node: GetLinks()){
			if (node){
				if (dynamic_cast<AstNodeFunctionDeclaration *>(node)){
					r += PrintString("\t\t[*] Function Declaration Link: %s ---- %s\n", node->GetNormalName().c_str(), node->GetSourceLocation().c_str());
				}
				else {
					r += Indent(node->GetTextTree(previous_nodes), 2);
				}
			}
		}
	}
	
	return r;
}


unsigned int AstNode::GetId()
{
	return m_id;
}

std::string AstNode::GetDotShape()
{
	return "box";
}

std::string AstNode::GetDotBgColour()
{
	return "ffffff";
}

std::list<AstNode **> AstNode::GetChildrenPointers()
{
	return {};
}

std::list<AstNode **> AstNode::GetLinksPointers()
{
	return {};
}

void AstNode::AddCodeGenerator(CodeGenerator *generator)
{
	if (m_generator){
		delete m_generator;
	}
	m_generator = generator;
}

void AstNode::AddToMap(AstNode *node, std::map<std::string, AstNode *> &node_map)
{
	std::string identifier;
	AstNodeStatementVariableDeclaration *variable_declaration = dynamic_cast<AstNodeStatementVariableDeclaration *>(node);
	AstNodeFunctionDeclaration *function_declaration = dynamic_cast<AstNodeFunctionDeclaration *>(node);
	AstNodeStructDeclaration *struct_declaration = dynamic_cast<AstNodeStructDeclaration *>(node);
	
	if (variable_declaration){
		identifier = variable_declaration->GetIdentifier();
	}
	else if (function_declaration){
		identifier = function_declaration->GetIdentifier();
	}
	else if (struct_declaration){
		identifier = struct_declaration->GetIdentifier();
	}
	else {
		throw FormatError("Node is not a variable declaration, function declaration, or struct declaration");
	}
	
	
	if (node_map.count(identifier)){
		//throw FormatError("\"%s\" declared multiple times", identifier.c_str());
		throw FormatCompilerError(node, "\"%s\" declared multiple times", identifier.c_str());
	}
	node_map[identifier] = node;
}

void AstNode::CopyAstNode(AstNode *dst)
{
	if (m_generator){
		dst->m_generator = m_generator->Copy();
	}
	else {
		dst->m_generator = 0;
	}
	
	dst->m_filename = m_filename;
	dst->m_line = m_line;
}

uint64_t AstNode::GetStaticComplexityEstimate(ComputeSpec &spec)
{
	return 0;
}

AstNodeExpression *AstNode::GetDynamicComplexityEstimate(ComputeSpec &spec)
{
	return 0;
}

void AstNode::SetSourceLocation(std::string filename, unsigned int line)
{
	m_filename = filename;
	m_line = line;
}

std::string AstNode::GetSourceLocation()
{
	if (!m_line){
		return "Compiler generated";
	}
	return PrintString("At line %u", m_line);
}

std::string AstNode::GetFilename()
{
	return m_filename;
}

unsigned int AstNode::GetLine()
{
	return m_line;
}

unsigned int AstNode::GetColumn()
{
	return 0;
}

std::string AstNode::GetNormalName()
{
	std::string name = GetDotName();
	std::string result;
	for (char c: name){
		if ('\\' != c){
			result += c;
		}
	}
	return result;
}

void AstNode::GetChildren(std::list<AstNode *> &r, bool children, bool links)
{
	r.push_back(this);
	
	if (children){
		std::list<AstNode *> l = GetChildren();
		for (AstNode *node: l){
			if (node){
				node->GetChildren(r, children, links);
			}
		}
	}
	
	if (links){
		std::list<AstNode *> l = GetLinks();
		for (AstNode *node: l){
			if (node && !dynamic_cast<AstNodeFunctionDeclaration *>(node)){
				node->GetChildren(r, children, links);
			}
		}
	}
}

void AstNode::ReplaceNode(AstNode *original, AstNode *replacement,
	bool replace_children, bool replace_links,
	bool recurse_children, bool recurse_links)
{
	std::map<AstNode *, AstNode *> replacements;
	replacements[original] = replacement;
	
	ReplaceNodes(replacements,
		replace_children, replace_links,
		recurse_children, recurse_links);
}

bool AstNode::HasChild(AstNode *child, bool children, bool links)
{
	int distance = 0;
	bool found = false;
	PreOrderTraversalDistance(child, children, links, distance, found);
	return found;
}

int AstNode::CompareChildren(AstNode *a, AstNode *b, bool children, bool links)
{
	if (a == b){
		return 0;
	}
	
	int a_distance = 0;
	bool a_found = false;
	PreOrderTraversalDistance(a, children, links, a_distance, a_found);
	if (!a_found){
		throw FormatError("Could not find A");
	}
	
	int b_distance = 0;
	bool b_found = false;
	PreOrderTraversalDistance(b, children, links, b_distance, b_found);
	if (!b_found){
		throw FormatError("Could not find B");
	}
	
	return b_distance - a_distance;
}

void AstNode::PreOrderTraversalDistance(AstNode *child, bool include_children, bool include_links, int &distance, bool &found)
{
	if (this == child){
		found = true;
		return;
	}
	
	if (!found){
		distance++;
	}
	
	std::list<AstNode *> children;
	if (include_children){
		children = GetChildren();
	}
	
	if (include_links){
		std::list<AstNode *> links = GetLinks();
		for (auto node: links){
			children.push_back(node);
		}
	}
	
	for (AstNode *node: children){
		if (!node){
			continue;
		}
		if (!dynamic_cast<AstNodeFunctionDeclaration *>(node)){
			node->PreOrderTraversalDistance(child, include_children, include_links, distance, found);
			if (found){
				return;
			}
		}
	}
}


static void SubstituteNodes(std::map<AstNode *, AstNode *> replacements, std::list<AstNode **> nodes)
{
	for (auto replacement_it: replacements){
		for (AstNode **node_pointer: nodes){
			if (*node_pointer == replacement_it.first){
				*node_pointer = replacement_it.second;
			}
		}
	}
}

void AstNode::ReplaceNodes(std::map<AstNode *, AstNode *> replacements,
	bool replace_children, bool replace_links,
	bool recurse_children, bool recurse_links)
{
	if (replace_children){
		std::list<AstNode **> children = GetChildrenPointers();
		SubstituteNodes(replacements, children);
	}
	
	if (replace_links){
		std::list<AstNode **> links = GetLinksPointers();
		SubstituteNodes(replacements, links);
	}
	
	if (recurse_children){
		std::list<AstNode *> children = GetChildren();
		for (AstNode *child: children){
			child->ReplaceNodes(replacements, replace_children, replace_links, recurse_children, recurse_links);
		}
	}
	
	if (recurse_links){
		std::list<AstNode *> links = GetLinks();
		for (AstNode *link: links){
			link->ReplaceNodes(replacements, replace_children, replace_links, recurse_children, recurse_links);
		}
	}
}

std::list<AstNode *> AstNode::GetChildren()
{
	std::list<AstNode *> children;
	std::list<AstNode **> child_pointers = GetChildrenPointers();
	
	for (AstNode **child_pointer: child_pointers){
		children.push_back(*child_pointer);
	}
	
	return children;
}

std::list<AstNode *> AstNode::GetLinks()
{
	std::list<AstNode *> children;
	std::list<AstNode **> child_pointers = GetLinksPointers();
	
	for (AstNode **child_pointer: child_pointers){
		children.push_back(*child_pointer);
	}
	
	return children;
}

void AstNode::TracebackDebugMessage()
{
	throw FormatCompilerError(this, "TracebackDebugMessage()");
}


AstNodeRoot::AstNodeRoot(std::list<AstNode *> root_items)
{
	SetCodeGenerator(CODEGEN_LLVM);
	
	for (auto node: root_items){
		AstNodeStructDeclaration *strct = dynamic_cast<AstNodeStructDeclaration *>(node);
		AstNodeStatementVariableDeclaration *var = dynamic_cast<AstNodeStatementVariableDeclaration *>(node);
		AstNodeFunctionDeclaration *fun = dynamic_cast<AstNodeFunctionDeclaration *>(node);
		
		if (strct){
			m_struct_declarations.push_back(strct);
		}
		else if (var){
			var->SetGlobal();
			m_variable_declarations.push_back(var);
		}
		else if (fun){
			m_function_declarations.push_back(fun);
		}
		else {
			throw FormatError("Somehow parsed an incorrect root item");
		}
	}
}

AstNodeRoot::AstNodeRoot(std::list<AstNodeStructDeclaration *> struct_declarations,
	std::list<AstNodeStatementVariableDeclaration *> variable_declarations,
	std::list<AstNodeFunctionDeclaration *> function_declarations) :
	m_struct_declarations(struct_declarations),
	m_variable_declarations(variable_declarations),
	m_function_declarations(function_declarations)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeRoot::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorRootLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorRootOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorRootPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}


AstNodeRoot::~AstNodeRoot()
{
	/*
	 * A common theme throughout the compiler: child nodes must be deleted in
	 * order to avoid memory leaks. AstNodes are tracked classes, and so a
	 * report of undeleted nodes on program termination is available in
	 * debugging mode.
	 */
	for (AstNodeStatementVariableDeclaration *declaration: m_variable_declarations){
		delete declaration;
	}
	
	for (AstNodeFunctionDeclaration *declaration: m_function_declarations){
		delete declaration;
	}
	
	for (AstNodeStructDeclaration *declaration: m_struct_declarations){
		delete declaration;
	}
}

AstNodeRoot *AstNodeRoot::Copy()
{
	std::list<AstNodeStructDeclaration *> struct_declarations;
	std::list<AstNodeStatementVariableDeclaration *> variable_declarations;
	std::list<AstNodeFunctionDeclaration *> function_declarations;
	
	for (auto struct_declaration: m_struct_declarations){
		struct_declarations.push_back(struct_declaration->Copy());
	}
	
	for (auto variable_declaration: m_variable_declarations){
		variable_declarations.push_back(variable_declaration->Copy());
	}
	
	for (auto function_declaration: m_function_declarations){
		function_declarations.push_back(function_declaration->Copy());
	}
	
	return new AstNodeRoot(struct_declarations, variable_declarations, function_declarations);
}

std::string AstNodeRoot::GetDotName()
{
	return "Root";
}

std::list<AstNode **> AstNodeRoot::GetChildrenPointers()
{
	std::list<AstNode **> primary_list; //Create an empty list
	AppendList(primary_list, GetPointers(m_struct_declarations)); //add the structure declarations to the list of children
	AppendList(primary_list, GetPointers(m_variable_declarations)); //add the variable declarations to the list of children
	AppendList(primary_list, GetPointers(m_function_declarations)); //add the function declarations to the list of children
	return primary_list;
}

void AstNodeRoot::Link()
{
	std::map<std::string, AstNode *> node_map;
	
	//add all the structs to the map
	for (auto struct_declaration: m_struct_declarations){
		struct_declaration->LinkToMap(node_map);
	}
	
	//link all the global structs with an empty (struct) map
	for (auto struct_declaration: m_struct_declarations){
		struct_declaration->LinkToOthers(node_map);
	}
	
	//link all the global variables with an empty (variable) map
	for (auto variable_declaration: m_variable_declarations){
		variable_declaration->LinkToOthers(node_map);
	}
	
	//add all the variables to the map
	for (auto variable_declaration: m_variable_declarations){
		variable_declaration->LinkToMap(node_map);
	}
	
	//add the functions to the map
	for (auto function_declaration: m_function_declarations){
		function_declaration->LinkToMap(node_map);
	}
	
	//link the functions to the map
	for (auto function_declaration: m_function_declarations){
		function_declaration->LinkToOthers(node_map);
	}
}

AstNodeRoot *AstNodeRoot::Normalize(CompileContext &context)
{
	std::list<AstNodeStatementVariableDeclaration *> variable_declarations;
	std::list<AstNodeFunctionDeclaration *> function_declarations;
	
	for (auto variable_declaration: m_variable_declarations){
		variable_declarations.push_back(variable_declaration->Normalize(context));
	}
	
	for (auto function_declaration: m_function_declarations){
		function_declarations.push_back(function_declaration->Normalize(context));
	}
	
	m_variable_declarations = variable_declarations;
	m_function_declarations = function_declarations;
	
	return this;
}

void AstNodeRoot::AddFunction(AstNodeFunctionDeclaration *function_declaration)
{
	m_function_declarations.push_front(function_declaration); //does this need to be front?
}

std::list<AstNodeStructDeclaration *> AstNodeRoot::GetStructDeclarations()
{
	return m_struct_declarations;
}

std::list<AstNodeStatementVariableDeclaration *> AstNodeRoot::GetVariableDeclarations()
{
	return m_variable_declarations;
}

std::list<AstNodeFunctionDeclaration *> AstNodeRoot::GetFunctionDeclarations()
{
	return m_function_declarations;
}
