/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef AST_ASTNODE_HPP
#define AST_ASTNODE_HPP

#include <list>
#include <map>
#include <set>
#include <string>
#include <tuple>

#include "../error.hpp"
#include "../computespec/computespec.hpp"

#include "../codegen/codegen.hpp"

class AstNodeStatementVariableDeclaration;
class AstNodeFunctionDeclaration;

class AstNode;

/* I would expect this to be thrown if, for example, an attempt was made to
   generate OpenCL code on a tree that was not supported in OpenCL
 */
class CodeGenerationException : public TrackedClass
{
	public:
		CodeGenerationException(CodeGeneratorType type, AstNode *node);
		CodeGenerationException(CodeGeneratorType type, AstNode *node, const char *format, ...);
		
		std::string GetMessage();
		
	private:
		CodeGeneratorType m_type;
		AstNode *m_node;
		std::string m_str;
};

class CodeGenerator;
class AstNodeExpression;
class AstNode : public TrackedClass
{
	public:
		AstNode();
		virtual ~AstNode();
		
		/*
		 * This returns a pointer to a deep copy of this node, and must be
		 * defined for every AstNode. Since a pointer to the specific node type
		 * should be returned, this must be implemented by leaf classes.
		 * 
		 * It should:
		 *   [*] Create a new instance of the node. This should be a deep copy.
		 *       This might involve calling the classes's normal constructor,
		 *       with copies of the children stored in that class and its
		 *       parents as the arguments to the constructor.
		 *   [*] Call CopyAstNode(). This is needed, for example, to copy the
		 *       code generators rather than be left with the defaults that
		 *       would be filled in by the constructor. If this is not done,
		 *       optimisations that expect different code generators may break
		 *       when the class is copied.
		 */
		virtual AstNode *Copy() = 0;
		
		virtual std::list<AstNode **> GetChildrenPointers();
		virtual std::list<AstNode **> GetLinksPointers();
		std::list<AstNode *> GetChildren();
		std::list<AstNode *> GetLinks();
		std::string GetDot();
		std::string GetTextTree();
		unsigned int GetId();
		
		void SetSourceLocation(std::string filename, unsigned int line);
		std::string GetSourceLocation();
		std::string GetFilename();
		unsigned int GetLine();
		unsigned int GetColumn();
		
		/*
		 * This adds a code generator of type "type". If such a code generator
		 * already exists, it is deleted and replaced with the new generator.
		 */
		void AddCodeGenerator(CodeGenerator *generator);
		
		//Sets the code generator for this and its children
		void SetTreeCodeGenerator(CodeGeneratorType type);
		
		/*
		 * This attempts to find a code generator of type "type", and class type
		 * "T". If no such code generator exists, an error is produced. Such an
		 * error indicates an implementation error in the compiler.
		 */
		template <typename T> T *GetCodeGenerator()
		{
			try {
				return tracked_cast<T>(m_generator);
			}
			catch (Error e){
				#ifdef DEBUG
					throw FormatCompilerError(this, "Internal error: could not find code correct generator: generator is %s, node is %s",
						m_generator ? m_generator->GetClassName().c_str() : "null", this ? GetClassName().c_str() : "null");
				#else
					throw FormatCompilerError(this, "Internal error: could not find code correct generator: generator is %s, node is %s",
						m_generator ? "non-null" : "null", this ? "non-null" : "null");
				#endif
			}
		}
		
		virtual void SetCodeGenerator(CodeGeneratorType type) = 0;
		
		virtual std::string GetDotName() = 0;
		std::string GetNormalName();
		
		//This estimates the complexity of the children at compile time
		virtual uint64_t GetStaticComplexityEstimate(ComputeSpec &spec);
		
		//This gets an expression to estimate the complexity of the children at run time
		virtual AstNodeExpression *GetDynamicComplexityEstimate(ComputeSpec &spec);
		
		//This is used to replace nodes (optimisers for example, use them)
		void ReplaceNode(AstNode *original, AstNode *replacement,
			bool replace_children, bool replace_links,
			bool recurse_children, bool recurse_links); //Convenience to call ReplaceNodes()
		void ReplaceNodes(std::map<AstNode *, AstNode *> replacements,
			bool replace_children, bool replace_links,
			bool recurse_children, bool recurse_links);
		
		//Returns > 0 if b comes after a in an in order traversal, < 0 if a comes after b, and 0 if a and b are equal
		int CompareChildren(AstNode *a, AstNode *b, bool children, bool links);
		bool HasChild(AstNode *child, bool children, bool links);
		void GetChildren(std::list<AstNode *> &r, bool children, bool links);
		
	protected:
		template <typename T> void AppendList(std::list<AstNode **> &primary_list, std::list<T **> append_list)
		{
			for (T **node: append_list){
				primary_list.push_back((AstNode **)node);
			}
		}
		
		template <typename T> std::list<AstNode **> ConvertList(std::list<T **> l)
		{
			std::list<AstNode **> r;
			for (T **node: l){
				r.push_back((AstNode **)node);
			}
			return r;
		}
		
		template <typename T> std::list<T **> GetPointers(std::list<T *> &l)
		{
			std::list<T **> r;
			for (T *&i: l){
				r.push_back(&i);
			}
			return r;
		}
		
		/*
		 * This has to be called in each child node's Copy(). It does stuff like
		 * copy the code generators, replacing the defaults.
		 */
		void CopyAstNode(AstNode *dst);
		
		static void AddToMap(AstNode *node, std::map<std::string, AstNode *> &node_map);
		
	private:
		//We're doing a lot of pointery stuff (for polymorphism)... don't allow
		//copying directly
		AstNode(const AstNode &) = delete;
		AstNode &operator=(const AstNode &) = delete;
		
		virtual std::string GetDotShape();
		virtual std::string GetDotBgColour();
		
		void TracebackDebugMessage();
		std::string GetTextTree(std::set<AstNode *> previous_nodes);
		
		void PreOrderTraversalDistance(AstNode *child, bool children, bool links, int &distance, bool &found);
		
		/*
		 * On deep copy, we'll need to copy the generators too. Perhaps have a
		 * virtual copy method that takes a pointer to the AstNode object?
		 */
		CodeGenerator *m_generator;
		
		/*
		 * Node ID
		 * Each node has a unique ID associated with it (copies have different IDs).
		 * The node ID can be used to generate things like allocas that have known
		 * names to be referenced later on
		 */
		unsigned int m_id;
		
		std::string m_filename;
		unsigned int m_line;
};

#include "type.hpp"
#include "expression.hpp"
#include "statement.hpp"
#include "function.hpp"
#include "dynvect.hpp"
#include "reference.hpp"

class AstNodeRoot : public AstNode
{
	public:
		AstNodeRoot(std::list<AstNode *> root_items);
		AstNodeRoot(std::list<AstNodeStructDeclaration *> struct_declarations,
			std::list<AstNodeStatementVariableDeclaration *> variable_declarations,
			std::list<AstNodeFunctionDeclaration *> function_declarations);
		virtual ~AstNodeRoot();
		virtual AstNodeRoot *Copy();
		virtual AstNodeRoot *Normalize(CompileContext &context);
		
		virtual void SetCodeGenerator(CodeGeneratorType type);
		
		//See the comment for AstNodeStatement::Link()
		void Link();
		
		std::list<AstNodeStructDeclaration *> GetStructDeclarations();
		std::list<AstNodeStatementVariableDeclaration *> GetVariableDeclarations();
		std::list<AstNodeFunctionDeclaration *> GetFunctionDeclarations();
		
		void AddFunction(AstNodeFunctionDeclaration *function_declaration);
		
	private:
		virtual std::string GetDotName();
		virtual std::list<AstNode **> GetChildrenPointers();
		
		std::list<AstNodeStructDeclaration *> m_struct_declarations;
		std::list<AstNodeStatementVariableDeclaration *> m_variable_declarations;
		std::list<AstNodeFunctionDeclaration *> m_function_declarations;
};

#endif
