/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "astnode.hpp"
#include "../codegen/llvm/codegen.hpp"
#include "../codegen/opencl/codegen.hpp"
#include "../codegen/pseudopc/codegen.hpp"

#define LOOP_FACTOR 32
#define BLOCK_FACTOR 8


AstNodeStatementCtrlFlow::AstNodeStatementCtrlFlow(AstNodeExpression *condition, AstNodeStatement *statement) :
	m_condition(condition), m_statement(statement)
{
}

AstNodeStatementCtrlFlow::~AstNodeStatementCtrlFlow()
{
	delete m_condition;
	delete m_statement;
}

void AstNodeStatementCtrlFlow::Link(std::map<std::string, AstNode *> &node_map)
{
	m_condition->Link(node_map);
	m_statement->Link(node_map);
}

AstNodeStatement *AstNodeStatementCtrlFlow::Normalize(CompileContext &context)
{
	m_condition = m_condition->Normalize(context);
	m_statement = m_statement->Normalize(context);
	
	if (!dynamic_cast<AstNodeTypeBool *>(m_condition->GetType())){
		throw FormatCompilerError(this, "Condition not bool");
	}
	
	return this;
}

AstNodeExpression *AstNodeStatementCtrlFlow::GetCondition()
{
	return m_condition;
}

AstNodeStatement *AstNodeStatementCtrlFlow::GetStatement()
{
	return m_statement;
}

std::list<AstNodeStatementVariableDeclaration *> AstNodeStatementCtrlFlow::GetVariableDeclarations(AstNode *stop)
{
	if (this == stop){
		return {};
	}
	return m_statement->GetVariableDeclarations(stop);
}

void AstNodeStatementCtrlFlow::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	m_condition->ReplaceVariableReferences(replacements);
	m_statement->ReplaceVariableReferences(replacements);
}

std::list<AstNode **> AstNodeStatementCtrlFlow::GetChildrenPointers()
{
	return {(AstNode**)&m_condition, (AstNode**)&m_statement};
}



AstNodeStatementFor::AstNodeStatementFor(AstNodeType *type, std::string identifier, AstNodeExpression *expression, AstNodeStatement *statement) :
	m_variable_declaration(new AstNodeStatementVariableDeclaration(type, identifier)), m_expression(expression), m_statement(statement)
{
	m_variable_declaration->SetCopyOnSpawn();
	SetCodeGenerator(CODEGEN_LLVM);
}

AstNodeStatementFor::AstNodeStatementFor(AstNodeStatementVariableDeclaration *variable_declaration, AstNodeExpression *expression, AstNodeStatement *statement) :
	m_variable_declaration(variable_declaration), m_expression(expression), m_statement(statement)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

AstNodeStatementFor::~AstNodeStatementFor()
{
	delete m_variable_declaration;
	delete m_expression;
	delete m_statement;
}

void AstNodeStatementFor::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorStatementForLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorStatementForOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorStatementForPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeStatement *AstNodeStatementFor::Normalize(CompileContext &context)
{
	m_variable_declaration = m_variable_declaration->Normalize(context);
	m_expression = m_expression->Normalize(context);
	m_statement = m_statement->Normalize(context);
	
	if (!dynamic_cast<AstNodeTypeInt *>(m_expression->GetType())){
		throw FormatCompilerError(this, "For to expression not integer");
	}
	
	m_expression = m_expression->Normalize(context);
	if (m_expression->GetType()->Compare(m_variable_declaration->GetType(), context)){
		m_expression = new AstNodeExpressionCast(m_variable_declaration->GetType()->Copy(), m_expression);
		m_expression = m_expression->Normalize(context);
	}
	
	return this;
}

std::string AstNodeStatementFor::GetDotName()
{
	return "for";
}

AstNodeStatementVariableDeclaration *AstNodeStatementFor::GetVariableDeclaration()
{
	return m_variable_declaration;
}

AstNodeExpression *AstNodeStatementFor::GetExpression()
{
	return m_expression;
}

AstNodeStatement *AstNodeStatementFor::GetStatement()
{
	return m_statement;
}

AstNodeStatementFor *AstNodeStatementFor::Copy()
{
	return new AstNodeStatementFor(m_variable_declaration->Copy(), m_expression->Copy(), m_statement->Copy());
}

void AstNodeStatementFor::Link(std::map<std::string, AstNode *> &node_map)
{
	std::map<std::string, AstNode *> node_map_copy = node_map;
	
	m_variable_declaration->Link(node_map_copy);
	m_expression->Link(node_map_copy);
	m_statement->Link(node_map_copy);
}

std::list<AstNode **> AstNodeStatementFor::GetChildrenPointers()
{
	return {(AstNode**)&m_variable_declaration, (AstNode**)&m_expression, (AstNode**)&m_statement};
}

std::list<AstNodeStatementVariableDeclaration *> AstNodeStatementFor::GetVariableDeclarations(AstNode *stop)
{
	if (this == stop){
		return {};
	}
	std::list<AstNodeStatementVariableDeclaration *> declarations = m_statement->GetVariableDeclarations(stop);
	declarations.push_front(m_variable_declaration);
	return declarations;
}

void AstNodeStatementFor::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	if (replacements.count(m_variable_declaration)){
		throw FormatCompilerError(this, "Cannot replace the variable declaration in a for loop");
	}
	
	m_expression->ReplaceVariableReferences(replacements);
	m_statement->ReplaceVariableReferences(replacements);
}

uint64_t AstNodeStatementFor::GetStaticComplexityEstimate(ComputeSpec &spec)
{
	AstNodeExpressionInt *constant = dynamic_cast<AstNodeExpressionInt *>(m_expression);
	if (constant){
		return (uint64_t)(constant->GetValue()) * m_statement->GetStaticComplexityEstimate(spec);
	}
	else {
		return (uint64_t)LOOP_FACTOR * m_statement->GetStaticComplexityEstimate(spec);
	}
}

AstNodeExpression *AstNodeStatementFor::GetDynamicComplexityEstimate(ComputeSpec &spec)
{
	AstNodeExpression *statement_complexity = m_statement->GetDynamicComplexityEstimate(spec);
	AstNodeExpression *mul = m_expression->GetDynamicComplexityEstimate(spec);
	
	if (!statement_complexity && !mul){ //we have no information
		return 0;
	}
	
	if (!mul){ //this means statement_complexity has a value
		AstNodeExpressionInt *constant = dynamic_cast<AstNodeExpressionInt *>(m_expression);
		if (constant){
			mul = constant->Copy();
		}
		else {
			mul = new AstNodeExpressionInt(LOOP_FACTOR);
		}
	}
	else if (!statement_complexity){
		statement_complexity = new AstNodeExpressionInt(BLOCK_FACTOR);
	}
	return new AstNodeExpressionBinaryMultiplication(mul, statement_complexity);
}



AstNodeStatementWhile::AstNodeStatementWhile(AstNodeExpression *condition, AstNodeStatement *statement) :
	AstNodeStatementCtrlFlow(condition, statement)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeStatementWhile::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorStatementWhileLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorStatementWhileOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorStatementWhilePseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeStatementWhile *AstNodeStatementWhile::Copy()
{
	return new AstNodeStatementWhile(GetCondition()->Copy(), GetStatement()->Copy());
}

std::string AstNodeStatementWhile::GetDotName()
{
	return "While";
}

uint64_t AstNodeStatementWhile::GetStaticComplexityEstimate(ComputeSpec &spec)
{
	return (uint64_t)LOOP_FACTOR * (GetStatement()->GetStaticComplexityEstimate(spec) + GetCondition()->GetStaticComplexityEstimate(spec));
}

AstNodeExpression *AstNodeStatementWhile::GetDynamicComplexityEstimate(ComputeSpec &spec)
{
	AstNodeExpression *condition_complexity = GetCondition()->GetDynamicComplexityEstimate(spec);
	AstNodeExpression *statement_complexity = GetStatement()->GetDynamicComplexityEstimate(spec);
	if (condition_complexity && statement_complexity){
		return new AstNodeExpressionBinaryMultiplication(new AstNodeExpressionBinaryAddition(condition_complexity, statement_complexity), new AstNodeExpressionInt(LOOP_FACTOR));
	}
	else if (condition_complexity){
		return new AstNodeExpressionBinaryMultiplication(new AstNodeExpressionBinaryAddition(condition_complexity, new AstNodeExpressionInt(BLOCK_FACTOR)), new AstNodeExpressionInt(LOOP_FACTOR));
	}
	else if (statement_complexity){
		return new AstNodeExpressionBinaryMultiplication(new AstNodeExpressionBinaryAddition(statement_complexity, new AstNodeExpressionInt(BLOCK_FACTOR)), new AstNodeExpressionInt(LOOP_FACTOR));
	}
	return 0;
}

AstNodeStatementIf::AstNodeStatementIf(AstNodeExpression *condition, AstNodeStatement *if_statement, AstNodeStatement *else_statement) :
	m_condition(condition), m_if_statement(if_statement), m_else_statement(else_statement)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeStatementIf::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorStatementIfLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorStatementIfOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorStatementIfPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeStatementIf::~AstNodeStatementIf()
{
	delete m_condition;
	delete m_if_statement;
	if (m_else_statement){
		delete m_else_statement;
	}
}

AstNodeStatementIf *AstNodeStatementIf::Copy()
{
	return new AstNodeStatementIf(m_condition->Copy(), m_if_statement->Copy(), m_else_statement ? m_else_statement->Copy() : 0);
}

AstNodeStatement *AstNodeStatementIf::Normalize(CompileContext &context)
{
	m_condition = m_condition->Normalize(context);
	m_if_statement = m_if_statement->Normalize(context);
	if (m_else_statement){
		m_else_statement = m_else_statement->Normalize(context);
	}
	
	if (!dynamic_cast<AstNodeTypeBool *>(m_condition->GetType())){
		throw FormatCompilerError(this, "Condition not bool");
	}
	
	return this;
}

void AstNodeStatementIf::Link(std::map<std::string, AstNode *> &node_map)
{
	m_condition->Link(node_map);
	m_if_statement->Link(node_map);
	if (m_else_statement){
		m_else_statement->Link(node_map);
	}
}

AstNodeExpression *AstNodeStatementIf::GetCondition()
{
	return m_condition;
}

AstNodeStatement *AstNodeStatementIf::GetIfStatement()
{
	return m_if_statement;
}

AstNodeStatement *AstNodeStatementIf::GetElseStatement()
{
	return m_else_statement;
}

std::string AstNodeStatementIf::GetDotName()
{
	return "If";
}

std::list<AstNode **> AstNodeStatementIf::GetChildrenPointers()
{
	return {(AstNode**)&m_condition, (AstNode**)&m_if_statement, (AstNode**)&m_else_statement};
}

std::list<AstNodeStatementVariableDeclaration *> AstNodeStatementIf::GetVariableDeclarations(AstNode *stop)
{
	if (this == stop){
		return {};
	}
	
	std::list<AstNodeStatementVariableDeclaration *> declarations;
	for (AstNodeStatement *statement: {m_if_statement, m_else_statement}){
		if (statement){
			if (statement == stop){
				return declarations;
			}
			std::list<AstNodeStatementVariableDeclaration *> child_declarations = statement->GetVariableDeclarations(stop);
			for (auto declaration: child_declarations){
				declarations.push_back(declaration);
			}
		}
	}
	return declarations;
}

void AstNodeStatementIf::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	m_condition->ReplaceVariableReferences(replacements);
	m_if_statement->ReplaceVariableReferences(replacements);
	if (m_else_statement){
		m_else_statement->ReplaceVariableReferences(replacements);
	}
}

uint64_t AstNodeStatementIf::GetStaticComplexityEstimate(ComputeSpec &spec)
{
	uint64_t if_estimate = m_if_statement->GetStaticComplexityEstimate(spec);
	uint64_t else_estimate = m_else_statement ? m_else_statement->GetStaticComplexityEstimate(spec) : 0;
	return m_condition->GetStaticComplexityEstimate(spec) + ((if_estimate < else_estimate) ? else_estimate : if_estimate);
}

AstNodeExpression *AstNodeStatementIf::GetDynamicComplexityEstimate(ComputeSpec &spec)
{
	AstNodeExpression *condition_complexity = m_condition->GetDynamicComplexityEstimate(spec);
	AstNodeExpression *if_statement_complexity = m_if_statement->GetDynamicComplexityEstimate(spec);
	AstNodeExpression *else_statement_complexity = m_else_statement ? m_else_statement->GetDynamicComplexityEstimate(spec) : 0;
	
	if (if_statement_complexity && else_statement_complexity){
		return new AstNodeExpressionBinaryAddition(if_statement_complexity, else_statement_complexity);
	}
	else if (condition_complexity && if_statement_complexity && else_statement_complexity){
		return new AstNodeExpressionBinaryAddition(condition_complexity, new AstNodeExpressionBinaryAddition(if_statement_complexity, else_statement_complexity));
	}
	else if (condition_complexity && if_statement_complexity){
		return new AstNodeExpressionBinaryAddition(condition_complexity, if_statement_complexity);
	}
	else if (condition_complexity){
		return new AstNodeExpressionBinaryAddition(condition_complexity, new AstNodeExpressionInt(BLOCK_FACTOR));
	}
	else if (if_statement_complexity){
		return new AstNodeExpressionBinaryAddition(if_statement_complexity, new AstNodeExpressionInt(BLOCK_FACTOR));
	}
	return 0;
}

