/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "astnode.hpp"
#include "../codegen/llvm/codegen.hpp"
#include "../codegen/pseudopc/codegen.hpp"


AstNodeTypeDynamicVector::AstNodeTypeDynamicVector(AstNodeType *type) :
	AstNodeTypeVector(type)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeTypeDynamicVector::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorTypeDynamicVectorLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorTypeDynamicVectorPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeTypeDynamicVector *AstNodeTypeDynamicVector::Copy()
{
	return new AstNodeTypeDynamicVector(GetType()->Copy());
}

int AstNodeTypeDynamicVector::Compare(AstNodeType *rhs, CompileContext &context)
{
	if (dynamic_cast<AstNodeTypeVoid *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeBool *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeInt *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeFloat *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypePointer *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeStaticVector *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeDynamicVector *>(rhs)){
		return GetType()->Compare(dynamic_cast<AstNodeTypeDynamicVector *>(rhs)->GetType(), context);
	}
	else {
		return 1;
	}
}

std::string AstNodeTypeDynamicVector::GetName()
{
	char b[1024];
	snprintf(b, 1024, "%s []", GetType()->GetName().c_str());
	return b;
}

std::string AstNodeTypeDynamicVector::GetDotName()
{
	return "Dynamic Sized Vector";
}



AstNodeExpressionDynamicVector::AstNodeExpressionDynamicVector(AstNodeType *type, AstNodeExpression *expression) :
	AstNodeExpressionValue(new AstNodeTypeDynamicVector(type)), m_expression(expression)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionDynamicVector::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionDynamicVectorLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionDynamicVectorPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionDynamicVector::~AstNodeExpressionDynamicVector()
{
	delete m_expression;
}

AstNodeExpressionDynamicVector *AstNodeExpressionDynamicVector::Copy()
{
	return new AstNodeExpressionDynamicVector(GetType()->Copy(), m_expression->Copy());
}

AstNodeExpression *AstNodeExpressionDynamicVector::Normalize(CompileContext &context)
{
	m_expression = m_expression->Normalize(context);
	return this;
}

AstNodeExpression *AstNodeExpressionDynamicVector::GetExpression()
{
	return m_expression;
}

std::string AstNodeExpressionDynamicVector::GetDotName()
{
	return "Dynamic Vector";
}

std::list<AstNode **> AstNodeExpressionDynamicVector::GetChildrenPointers()
{
	return {GetTypePointer(), (AstNode **)&m_expression};
}

void AstNodeExpressionDynamicVector::Link(std::map<std::string, AstNode *> &node_map)
{
	m_expression->Link(node_map);
	GetType()->Link(node_map);
}

void AstNodeExpressionDynamicVector::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	m_expression->ReplaceVariableReferences(replacements);
}






AstNodeStatementDynamicVectorOperation::AstNodeStatementDynamicVectorOperation(AstNodeExpression *dynvect, std::string dotname) :
	m_dynvect(dynvect), m_dotname(dotname)
{
}

AstNodeStatementDynamicVectorOperation::~AstNodeStatementDynamicVectorOperation()
{
	delete m_dynvect;
}

AstNodeStatement *AstNodeStatementDynamicVectorOperation::Normalize(CompileContext &context)
{
	m_dynvect = m_dynvect->Normalize(context);
	return this;
}

void AstNodeStatementDynamicVectorOperation::Link(std::map<std::string, AstNode *> &node_map)
{
	m_dynvect->Link(node_map);
}

void AstNodeStatementDynamicVectorOperation::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	m_dynvect->ReplaceVariableReferences(replacements);
}

uint64_t AstNodeStatementDynamicVectorOperation::GetStaticComplexityEstimate(ComputeSpec &spec)
{
	return m_dynvect->GetStaticComplexityEstimate(spec);
}

AstNodeExpression *AstNodeStatementDynamicVectorOperation::GetDynamicComplexityEstimate(ComputeSpec &spec)
{
	return m_dynvect->GetDynamicComplexityEstimate(spec);
}

std::string AstNodeStatementDynamicVectorOperation::GetDotName()
{
	return m_dotname;
}

std::list<AstNode **> AstNodeStatementDynamicVectorOperation::GetChildrenPointers()
{
	return {(AstNode**)&m_dynvect};
}

AstNodeExpression *AstNodeStatementDynamicVectorOperation::GetExpression()
{
	return m_dynvect;
}


AstNodeDynamicVectorCommitBefore::AstNodeDynamicVectorCommitBefore(AstNodeExpression *dynvect) :
	AstNodeStatementDynamicVectorOperation(dynvect, "Dynamic Vector Commit Before")
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeDynamicVectorCommitBefore::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorDynamicVectorCommitBeforeLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorDynamicVectorOperationPseudopc(this, "commit.before")); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeDynamicVectorCommitBefore *AstNodeDynamicVectorCommitBefore::Copy()
{
	return new AstNodeDynamicVectorCommitBefore(GetExpression()->Copy());
}

AstNodeDynamicVectorCommitAfter::AstNodeDynamicVectorCommitAfter(AstNodeExpression *dynvect) :
	AstNodeStatementDynamicVectorOperation(dynvect, "Dynamic Vector Commit After")
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeDynamicVectorCommitAfter::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorDynamicVectorCommitAfterLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorDynamicVectorOperationPseudopc(this, "commit.after")); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeDynamicVectorCommitAfter *AstNodeDynamicVectorCommitAfter::Copy()
{
	return new AstNodeDynamicVectorCommitAfter(GetExpression()->Copy());
}





AstNodeDynamicVectorGetPointer::AstNodeDynamicVectorGetPointer(AstNodeExpressionLValue *dynvect, bool read, bool write) :
	m_dynvect(dynvect), m_type(new AstNodeTypePointer(   tracked_cast<AstNodeTypeDynamicVector>(dynvect->GetType())->GetType()->Copy()   )), m_read(read), m_write(write)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeDynamicVectorGetPointer::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorDynamicVectorGetPointerLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorDynamicVectorGetPointerPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeDynamicVectorGetPointer::~AstNodeDynamicVectorGetPointer()
{
	delete m_dynvect;
	delete m_type;
}

AstNodeDynamicVectorGetPointer *AstNodeDynamicVectorGetPointer::Copy()
{
	return new AstNodeDynamicVectorGetPointer(m_dynvect->Copy(), m_read, m_write);
}

AstNodeExpression *AstNodeDynamicVectorGetPointer::Normalize(CompileContext &context)
{
	m_dynvect = m_dynvect->Normalize(context);
	return this;
}

AstNodeType *AstNodeDynamicVectorGetPointer::GetType()
{
	return m_type;
}

void AstNodeDynamicVectorGetPointer::Link(std::map<std::string, AstNode *> &node_map)
{
	m_dynvect->Link(node_map);
	m_type->Link(node_map);
}

void AstNodeDynamicVectorGetPointer::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	m_dynvect->ReplaceVariableReferences(replacements);
}

uint64_t AstNodeDynamicVectorGetPointer::GetStaticComplexityEstimate(ComputeSpec &spec)
{
	return m_dynvect->GetStaticComplexityEstimate(spec);
}

AstNodeExpression *AstNodeDynamicVectorGetPointer::GetDynamicComplexityEstimate(ComputeSpec &spec)
{
	return m_dynvect->GetDynamicComplexityEstimate(spec);
}

std::string AstNodeDynamicVectorGetPointer::GetDotName()
{
	return "Dynamic Vector to Pointer";
}

std::list<AstNode **> AstNodeDynamicVectorGetPointer::GetChildrenPointers()
{
	return {(AstNode **)&m_dynvect, (AstNode **)&m_type};
}

AstNodeExpressionLValue *AstNodeDynamicVectorGetPointer::GetLValue()
{
	return m_dynvect;
}

bool AstNodeDynamicVectorGetPointer::GetRead()
{
	return m_read;
}

bool AstNodeDynamicVectorGetPointer::GetWrite()
{
	return m_write;
}


