/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "astnode.hpp"
#include "../codegen/llvm/codegen.hpp"
#include "../codegen/opencl/codegen.hpp"
#include "../codegen/pseudopc/codegen.hpp"

AstNodeExpression *AstNodeExpression::Normalize(CompileContext &context)
{
	return this;
}

AstNodeExpressionLValue *AstNodeExpressionLValue::Normalize(CompileContext &context)
{
	return this;
}

void AstNodeExpression::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
}

void AstNodeExpression::Link(std::map<std::string, AstNode *> &node_map)
{
}

AstNodeExpressionValue::AstNodeExpressionValue(AstNodeType *type) :
	m_type(type)
{
}

AstNodeExpressionValue::~AstNodeExpressionValue()
{
	if (m_type){
		delete m_type;
	}
}

AstNodeType *AstNodeExpressionValue::GetType()
{
	return m_type;
}

AstNode **AstNodeExpressionValue::GetTypePointer()
{
	return (AstNode**)&m_type;
}

std::list<AstNode **> AstNodeExpressionValue::GetChildrenPointers()
{
	return {(AstNode **)&m_type};
}

void AstNodeExpressionValue::Link(std::map<std::string, AstNode *> &node_map)
{
	m_type->Link(node_map);
}

AstNodeExpressionLValue::AstNodeExpressionLValue() :
	m_dorefcount(true)
{
}

bool AstNodeExpressionLValue::GetDoRefCount()
{
	return m_dorefcount;
}

void AstNodeExpressionLValue::SetDoRefCount(bool dorefcount)
{
	m_dorefcount = dorefcount;
}

AstNodeExpressionVoid::AstNodeExpressionVoid() :
	AstNodeExpressionValue(new AstNodeTypeVoid)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionVoid::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionVoidLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionVoidPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionVoid *AstNodeExpressionVoid::Copy()
{
	return new AstNodeExpressionVoid();
}

std::string AstNodeExpressionVoid::GetDotName()
{
	return "Void Expression";
}

AstNodeExpressionBool::AstNodeExpressionBool(bool value) :
	AstNodeExpressionValue(new AstNodeTypeBool), //TODO
	m_value(value)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionBool::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionBoolLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionBoolOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionBoolPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionBool *AstNodeExpressionBool::Copy()
{
	return new AstNodeExpressionBool(m_value);
}

bool AstNodeExpressionBool::GetValue()
{
	return m_value;
}

std::string AstNodeExpressionBool::GetDotName()
{
	char b[1024];
	snprintf(b, 1024, "Boolean Expression: %s", m_value ? "true" : "false");
	return b;
}

AstNodeExpressionInt::AstNodeExpressionInt(unsigned long long value) :
	AstNodeExpressionValue(new AstNodeTypeInt(32, false, false)), //TODO
	m_value(value)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionInt::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionIntLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionIntOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionIntPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionInt *AstNodeExpressionInt::Copy()
{
	return new AstNodeExpressionInt(m_value);
}

unsigned long long AstNodeExpressionInt::GetValue()
{
	return m_value;
}

std::string AstNodeExpressionInt::GetDotName()
{
	char b[1024];
	snprintf(b, 1024, "Integer Expression: %llu", m_value);
	return b;
}

AstNodeExpressionFloat::AstNodeExpressionFloat(double value) :
	AstNodeExpressionValue(new AstNodeTypeFloat(32, false)),
	m_value(value)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionFloat::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionFloatLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionFloatOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionFloatPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionFloat *AstNodeExpressionFloat::Copy()
{
	return new AstNodeExpressionFloat(m_value);
}

double AstNodeExpressionFloat::GetValue()
{
	return m_value;
}

std::string AstNodeExpressionFloat::GetDotName()
{
	char b[1024];
	snprintf(b, 1024, "Float Expression: %f", m_value);
	return b;
}

AstNodeExpressionLength::AstNodeExpressionLength(AstNodeExpression *expression) :
	AstNodeExpressionValue(new AstNodeTypeInt(0, true, false)), m_expression(expression)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionLength::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionLengthLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionLengthPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionLength::~AstNodeExpressionLength()
{
	delete m_expression;
}

AstNodeExpressionLength *AstNodeExpressionLength::Copy()
{
	return new AstNodeExpressionLength(m_expression->Copy());
}

AstNodeExpression *AstNodeExpressionLength::Normalize(CompileContext &context)
{
	m_expression = m_expression->Normalize(context);
	return this;
}

AstNodeExpression *AstNodeExpressionLength::GetExpression()
{
	return m_expression;
}

std::string AstNodeExpressionLength::GetDotName()
{
	return "Length";
}

std::list<AstNode **> AstNodeExpressionLength::GetChildrenPointers()
{
	return {(AstNode**)&m_expression};
}

void AstNodeExpressionLength::Link(std::map<std::string, AstNode *> &node_map)
{
	m_expression->Link(node_map);
	GetType()->Link(node_map);
}

void AstNodeExpressionLength::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	m_expression->ReplaceVariableReferences(replacements);
}


AstNodeExpressionLValueIdentifier::AstNodeExpressionLValueIdentifier(std::string name) :
	m_name(name), m_function_link(0), m_variable_link(0)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionLValueIdentifier::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionLValueIdentifierLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionLValueIdentifierOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionLValueIdentifierPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionLValueIdentifier *AstNodeExpressionLValueIdentifier::Copy()
{
	AstNodeExpressionLValueIdentifier *copy = new AstNodeExpressionLValueIdentifier(m_name);
	copy->m_function_link = m_function_link;
	copy->m_variable_link = m_variable_link;
	copy->SetDoRefCount(GetDoRefCount());
	return copy;
}

std::string AstNodeExpressionLValueIdentifier::GetIdentifier()
{
	return m_name;
}

AstNodeStatementVariableDeclaration *AstNodeExpressionLValueIdentifier::GetLink()
{
	if (!m_variable_link){
		throw FormatCompilerError(this, "Variable link null");
	}
	return m_variable_link;
}

AstNodeFunctionDeclaration *AstNodeExpressionLValueIdentifier::GetFunctionLink()
{
	if (!m_function_link){
		throw FormatCompilerError(this, "Function link null");
	}
	return m_function_link;
}

bool AstNodeExpressionLValueIdentifier::GetIsFunction()
{
	return m_function_link;
}

void AstNodeExpressionLValueIdentifier::Link(std::map<std::string, AstNode *> &node_map)
{
	if (!node_map.count(m_name)){
		throw FormatCompilerError(this, "\"%s\" is not declared here", m_name.c_str());
	}
	
	AstNodeStatementVariableDeclaration *variable_node = dynamic_cast<AstNodeStatementVariableDeclaration *>(node_map[m_name]);
	AstNodeFunctionDeclaration *function_node = dynamic_cast<AstNodeFunctionDeclaration *>(node_map[m_name]);
	
	if (variable_node){
		m_variable_link = variable_node;
	}
	else if (function_node){
		m_function_link = function_node;
	}
	else {
		throw FormatCompilerError(this, "\"%s\" is not a variable", m_name.c_str());
	}
}

AstNodeType *AstNodeExpressionLValueIdentifier::GetType()
{
	if (m_variable_link){
		return m_variable_link->GetType();
	}
	else if (m_function_link){
		return m_function_link->GetType();
	}
	else {
		throw FormatCompilerError(this, "Link is null: \"%s\"", m_name.c_str());
	}
}

std::list<AstNode **> AstNodeExpressionLValueIdentifier::GetLinksPointers()
{
	return {(AstNode**)&m_variable_link, (AstNode**)&m_function_link};
}

void AstNodeExpressionLValueIdentifier::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	//fprintf(stderr, "[*] AstNodeExpressionLValueIdentifier::ReplaceVariableReferences %p\n",
	//	(void *)m_variable_link);
	if (m_variable_link){
		//fprintf(stderr, "[*] Checking replacement reference for %p: %s\n",
		//	(void *)m_variable_link, m_variable_link->GetIdentifier().c_str());
		if (replacements.count(m_variable_link)){
			//fprintf(stderr, "[*] Replacing reference for %p: %s with %p: %s\n",
			//	(void *)m_variable_link, m_variable_link->GetIdentifier().c_str(),
			//	(void *)replacements[m_variable_link], replacements[m_variable_link]->GetIdentifier().c_str());
			m_variable_link = replacements[m_variable_link];
		}
	}
}

std::string AstNodeExpressionLValueIdentifier::GetDotName()
{
	return std::string("Identifier Expression: \\\"") + m_name + std::string("\\\"");
}


AstNodeExpressionLValueVector::AstNodeExpressionLValueVector(AstNodeExpressionLValue *lvalue, AstNodeExpression *expression) :
	m_lvalue(lvalue), m_expression(expression)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionLValueVector::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionLValueVectorLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionLValueVectorOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionLValueVectorPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionLValueVector::~AstNodeExpressionLValueVector()
{
	delete m_lvalue;
	delete m_expression;
}

AstNodeExpressionLValueVector *AstNodeExpressionLValueVector::Copy()
{
	AstNodeExpressionLValueVector *copy = new AstNodeExpressionLValueVector(m_lvalue->Copy(), m_expression->Copy());
	copy->SetDoRefCount(GetDoRefCount());
	return copy;
}

AstNodeExpressionLValue *AstNodeExpressionLValueVector::Normalize(CompileContext &context)
{
	m_lvalue = m_lvalue->Normalize(context);
	m_expression = m_expression->Normalize(context);
	return this;
}

AstNodeType *AstNodeExpressionLValueVector::GetType()
{
	AstNodeTypeVector *type = dynamic_cast<AstNodeTypeVector *>(m_lvalue->GetType());
	if (!type){
		#ifdef DEBUG
			throw FormatCompilerError(this, "Vector LValue type is not vector: %p %p %p %s %s: %s",
				(void *)type, (void *)m_lvalue->GetType(), (void *)m_lvalue, TypeIdToString(typeid(*m_lvalue->GetType())).c_str(), m_lvalue->GetDotName().c_str(), GetSourceLocation().c_str());
		#else
			throw FormatCompilerError(this, "Vector LValue type is not vector");
		#endif
	}
	
	return type->GetType();
}

void AstNodeExpressionLValueVector::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	m_lvalue->ReplaceVariableReferences(replacements);
	m_expression->ReplaceVariableReferences(replacements);
}

void AstNodeExpressionLValueVector::Link(std::map<std::string, AstNode *> &node_map)
{
	m_lvalue->Link(node_map);
	m_expression->Link(node_map);
}

AstNodeExpressionLValue *AstNodeExpressionLValueVector::GetLValue()
{
	return m_lvalue;
}

AstNodeExpression *AstNodeExpressionLValueVector::GetExpression()
{
	return m_expression;
}

std::string AstNodeExpressionLValueVector::GetDotName()
{
	return "Vector LValue";
}

std::list<AstNode **> AstNodeExpressionLValueVector::GetChildrenPointers()
{
	return {(AstNode**)&m_lvalue, (AstNode**)&m_expression};
}

uint64_t AstNodeExpressionLValueVector::GetStaticComplexityEstimate(ComputeSpec &spec)
{
	return m_lvalue->GetStaticComplexityEstimate(spec) + m_expression->GetStaticComplexityEstimate(spec);
}

AstNodeExpression *AstNodeExpressionLValueVector::GetDynamicComplexityEstimate(ComputeSpec &spec)
{
	AstNodeExpression *a = m_expression->GetDynamicComplexityEstimate(spec);
	AstNodeExpression *b = m_lvalue->GetDynamicComplexityEstimate(spec);
	
	if (a && b){
		return new AstNodeExpressionBinaryAddition(a, b);
	}
	else if (a){
		return new AstNodeExpressionBinaryAddition(a, new AstNodeExpressionInt(b->GetStaticComplexityEstimate(spec)));
	}
	else if (b){
		return new AstNodeExpressionBinaryAddition(b, new AstNodeExpressionInt(a->GetStaticComplexityEstimate(spec)));
	}
	return 0;
}






AstNodeExpressionBlock::AstNodeExpressionBlock(std::list<AstNodeStatement *> statements, AstNodeExpression *expression) :
	m_statements(statements), m_expression(expression)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionBlock::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionBlockLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionBlockPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionBlock::~AstNodeExpressionBlock()
{
	for (AstNodeStatement *statement: m_statements){
		delete statement;
	}
	
	delete m_expression;
}

AstNodeExpressionBlock *AstNodeExpressionBlock::Copy()
{
	std::list<AstNodeStatement *> statements;
	for (auto statement: m_statements){
		statements.push_back(statement->Copy());
	}
	
	return new AstNodeExpressionBlock(statements, m_expression->Copy());
}

AstNodeExpression *AstNodeExpressionBlock::Normalize(CompileContext &context)
{
	std::list<AstNodeStatement *> statements;
	for (auto statement: m_statements){
		statements.push_back(statement->Normalize(context));
	}
	m_statements = statements;
	
	m_expression = m_expression->Normalize(context);
	
	return this;
}


std::list<AstNodeStatementVariableDeclaration *> AstNodeExpressionBlock::GetVariableDeclarations(AstNode *stop)
{
	if (this == stop){
		return {};
	}
	
	std::list<AstNodeStatementVariableDeclaration *> declarations;
	for (auto statement: m_statements){
		if (statement == stop){
			return declarations;
		}
		std::list<AstNodeStatementVariableDeclaration *> child_declarations = statement->GetVariableDeclarations(stop);
		for (auto declaration: child_declarations){
			declarations.push_back(declaration);
		}
	}
	return declarations;
}

void AstNodeExpressionBlock::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	for (auto statement: m_statements){
		statement->ReplaceVariableReferences(replacements);
	}
	m_expression->ReplaceVariableReferences(replacements);
}

AstNodeType *AstNodeExpressionBlock::GetType()
{
	return m_expression->GetType();
}

void AstNodeExpressionBlock::Link(std::map<std::string, AstNode *> &node_map)
{
	//We create a copy of the maps
	std::map<std::string, AstNode *> node_map_copy = node_map;
	
	for (auto statement: m_statements){
		statement->Link(node_map_copy);
	}
	
	m_expression->Link(node_map_copy);
}

uint64_t AstNodeExpressionBlock::GetStaticComplexityEstimate(ComputeSpec &spec)
{
	uint64_t sum = 0;
	for (auto statement: m_statements){
		sum += statement->GetStaticComplexityEstimate(spec);
	}
	sum += m_expression->GetStaticComplexityEstimate(spec);
	
	return sum;
}

AstNodeExpression *AstNodeExpressionBlock::GetDynamicComplexityEstimate(ComputeSpec &spec)
{
	AstNodeExpression *sum = 0;
	unsigned int dynamic_count = 0;
	std::list<AstNode *> children = GetChildren();
	
	for (auto child: children){
		AstNodeExpression *e = child->GetDynamicComplexityEstimate(spec);
		if (e){
			dynamic_count++;
		}
		
		if (sum && e){
			sum = new AstNodeExpressionBinaryAddition(e, sum);
		}
		else if (sum){
			sum = new AstNodeExpressionBinaryAddition(sum, new AstNodeExpressionInt(child->GetStaticComplexityEstimate(spec)));
		}
		else {
			sum = e;
		}
	}
	
	if (dynamic_count){
		return sum;
	}
	else {
		delete sum; //avoid memory leak
		return 0;
	}
}

std::list<AstNodeStatement *> AstNodeExpressionBlock::GetStatements()
{
	return m_statements;
}

AstNodeExpression *AstNodeExpressionBlock::GetExpression()
{
	return m_expression;
}

std::string AstNodeExpressionBlock::GetDotName()
{
	return "Expression with Statements";
}

std::list<AstNode **> AstNodeExpressionBlock::GetChildrenPointers()
{
	std::list<AstNode** > r = ConvertList(GetPointers(m_statements));
	AppendList<AstNode>(r, {(AstNode**)&m_expression});
	return r;
}

