/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef AST_EXPRESSION_HPP
#define AST_EXPRESSION_HPP

#include "../compilecontext.hpp"

/*
 * Literal type resolution idea:
 *     Take an expression, and determine the type of all literals based on non-literal types
 *     If all subexpressions are literal, then similar for the parent
 *     If all literal, then wherever the value gets used (e.g: assignment or function call)
 *
 *     The function call and assignment link methods can call a literal type inference method on expressions
 *         (with the expected expression type)
 *     Expressions pass their type down to eventually reach the literals
 */

class AstNodeExpression : public AstNode
{
	public:
		virtual AstNodeExpression *Copy() = 0;
		virtual AstNodeExpression *Normalize(CompileContext &context);
		
		virtual AstNodeType *GetType() = 0;
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		
		virtual void ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements);
};

class AstNodeExpressionLValue : public AstNodeExpression
{
	public:
		AstNodeExpressionLValue();
		
		virtual AstNodeExpressionLValue *Copy() = 0;
		virtual AstNodeExpressionLValue *Normalize(CompileContext &context);
		
		bool GetDoRefCount();
		void SetDoRefCount(bool dorefcount);
		
	private:
		bool m_dorefcount;
};

class AstNodeExpressionValue : public AstNodeExpression
{
	public:
		AstNodeExpressionValue(AstNodeType *type);
		virtual ~AstNodeExpressionValue();
		virtual AstNodeType *GetType();
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		virtual std::list<AstNode **> GetChildrenPointers();
		
	protected:
		AstNode **GetTypePointer();
		
	private:
		AstNodeType *m_type;
};

class AstNodeExpressionVoid : public AstNodeExpressionValue
{
	public:
		AstNodeExpressionVoid();
		virtual AstNodeExpressionVoid *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
		
	private:
		virtual std::string GetDotName();
};

class AstNodeExpressionBool : public AstNodeExpressionValue
{
	public:
		AstNodeExpressionBool(bool value);
		virtual AstNodeExpressionBool *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
		
		bool GetValue();
		
	private:
		virtual std::string GetDotName();
		
		bool m_value;
};

class AstNodeExpressionInt : public AstNodeExpressionValue
{
	public:
		AstNodeExpressionInt(unsigned long long value);
		virtual AstNodeExpressionInt *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
		
		unsigned long long GetValue();
		
	private:
		virtual std::string GetDotName();
		
		unsigned long long m_value;
};

class AstNodeExpressionFloat : public AstNodeExpressionValue
{
	public:
		AstNodeExpressionFloat(double value);
		virtual AstNodeExpressionFloat *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
		
		double GetValue();
		
	private:
		virtual std::string GetDotName();
		
		double m_value;
};

class AstNodeExpressionLength : public AstNodeExpressionValue
{
	public:
		AstNodeExpressionLength(AstNodeExpression *expression);
		virtual ~AstNodeExpressionLength();
		virtual AstNodeExpressionLength *Copy();
		virtual AstNodeExpression *Normalize(CompileContext &context);
		virtual void SetCodeGenerator(CodeGeneratorType type);
		
		AstNodeExpression *GetExpression();
		
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		virtual void ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements);
		
	private:
		virtual std::string GetDotName();
		virtual std::list<AstNode **> GetChildrenPointers();
		
		AstNodeExpression *m_expression;
};

class AstNodeExpressionLValueIdentifier : public AstNodeExpressionLValue
{
	public:
		AstNodeExpressionLValueIdentifier(std::string name);
		virtual AstNodeExpressionLValueIdentifier *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
		
		virtual AstNodeType *GetType();
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		
		std::string GetIdentifier();
		AstNodeStatementVariableDeclaration *GetLink();
		AstNodeFunctionDeclaration *GetFunctionLink();
		bool GetIsFunction();
		
		virtual void ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements);
		
	private:
		virtual std::string GetDotName();
		virtual std::list<AstNode **> GetLinksPointers();
		
		std::string m_name;
		AstNodeFunctionDeclaration *m_function_link;
		AstNodeStatementVariableDeclaration *m_variable_link;
};

class AstNodeExpressionLValueVector : public AstNodeExpressionLValue
{
	public:
		AstNodeExpressionLValueVector(AstNodeExpressionLValue *lvalue, AstNodeExpression *expression);
		virtual ~AstNodeExpressionLValueVector();
		virtual AstNodeExpressionLValueVector *Copy();
		virtual AstNodeExpressionLValue *Normalize(CompileContext &context);
		virtual void SetCodeGenerator(CodeGeneratorType type);
		
		virtual AstNodeType *GetType();
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		
		AstNodeExpressionLValue *GetLValue();
		AstNodeExpression *GetExpression();
		
		virtual void ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements);
		
		virtual uint64_t GetStaticComplexityEstimate(ComputeSpec &spec);
		virtual AstNodeExpression *GetDynamicComplexityEstimate(ComputeSpec &spec);
		
	private:
		virtual std::string GetDotName();
		virtual std::list<AstNode **> GetChildrenPointers();
		
		AstNodeExpressionLValue *m_lvalue;
		AstNodeExpression *m_expression;
		
};





/*
 * This doesn't occur in real programs, but it's useful in optimisations.
 * It performs the statements, and then the expression
 */
class AstNodeStatement;
class AstNodeExpressionBlock : public AstNodeExpression
{
	public:
		AstNodeExpressionBlock(std::list<AstNodeStatement *> statements, AstNodeExpression *expression);
		virtual ~AstNodeExpressionBlock();
		virtual AstNodeExpressionBlock *Copy();
		virtual AstNodeExpression *Normalize(CompileContext &context);
		virtual void SetCodeGenerator(CodeGeneratorType type);
		
		virtual AstNodeType *GetType();
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		
		virtual std::list<AstNodeStatementVariableDeclaration *> GetVariableDeclarations(AstNode *stop = 0);
		virtual void ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements);
		
		virtual uint64_t GetStaticComplexityEstimate(ComputeSpec &spec);
		virtual AstNodeExpression *GetDynamicComplexityEstimate(ComputeSpec &spec);
		
		std::list<AstNodeStatement *> GetStatements();
		AstNodeExpression *GetExpression();
		
	private:
		virtual std::string GetDotName();
		virtual std::list<AstNode **> GetChildrenPointers();
		
		std::list<AstNodeStatement *> m_statements;
		AstNodeExpression *m_expression;
};


#include "struct.hpp"
#include "op.hpp"

#endif
 
