/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "astnode.hpp"
#include "../codegen/llvm/codegen.hpp"
#include "../codegen/opencl/codegen.hpp"
#include "../codegen/pseudopc/codegen.hpp"

AstNodeFunctionDeclaration::AstNodeFunctionDeclaration(AstNodeType *return_type, std::string name,
	std::list<AstNodeStatementVariableDeclaration *> args, AstNodeStatement *statement, bool is_export, bool is_ccall, bool is_lock) :
	m_type(new AstNodeTypeFunctionPointer(return_type, args, is_ccall)),
	m_return_type(return_type), m_name(name), m_args(args), m_statement(statement), m_is_export(is_export), m_is_ccall(is_ccall), m_is_lock(is_lock), m_is_compiler(false)
{
	SetCodeGenerator(CODEGEN_LLVM);
	
	unsigned int i = 1;
	for (auto arg: m_args){
		arg->SetArgument(i);
		i++;
	}
}

void AstNodeFunctionDeclaration::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorFunctionDeclarationLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorFunctionDeclarationOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorFunctionDeclarationPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeFunctionDeclaration::~AstNodeFunctionDeclaration()
{
	if (m_return_type){
		delete m_return_type;
	}
	
	for (AstNodeStatementVariableDeclaration *declaration: m_args){
		if (declaration){
			delete declaration;
		}
	}
	
	if (m_statement){
		delete m_statement;
	}
	
	delete m_type;
}

AstNodeFunctionDeclaration *AstNodeFunctionDeclaration::Copy()
{
	std::list<AstNodeStatementVariableDeclaration *> args;
	for (auto arg: m_args){
		args.push_back(arg->Copy());
	}
	
	return new AstNodeFunctionDeclaration(m_return_type->Copy(), m_name, args, m_statement->Copy(), m_is_export, m_is_ccall, m_is_lock);
}

AstNodeFunctionDeclaration *AstNodeFunctionDeclaration::Normalize(CompileContext &context)
{
	if (m_statement){
		m_statement = m_statement->Normalize(context);
	}
	
	std::list<AstNodeStatementVariableDeclaration *> args;
	for (AstNodeStatementVariableDeclaration *declaration: m_args){
		if (declaration){
			args.push_back(declaration->Normalize(context));
		}
		else {
			args.push_back(0);
		}
	}
	m_args = args;
	
	if (m_is_lock){
		throw FormatCompilerError(this, "Lock functions not implemented yet");
	}
	
	return this;
}

bool AstNodeFunctionDeclaration::GetIsCompiler()
{
	return m_is_compiler;
}

void AstNodeFunctionDeclaration::SetIsCompiler()
{
	m_is_compiler = true;
}

std::string AstNodeFunctionDeclaration::GetDotName()
{
	std::string r = std::string("Function Declaration: \\\"") + m_name + std::string("\\\"");
	if (!m_statement){
		r += " (imported)";
	}
	if (m_is_export){
		r += " (exported)";
	}
	if (m_is_ccall){
		r += " (ccall)";
	}
	return r;
}


std::list<AstNode **> AstNodeFunctionDeclaration::GetChildrenPointers()
{
	std::list<AstNode **> l;
	l.push_back((AstNode**)&m_type);
	l.push_back((AstNode**)&m_return_type);
	AppendList<AstNodeStatementVariableDeclaration>(l, GetPointers(m_args));
	if (m_statement){
		l.push_back((AstNode**)&m_statement);
	}
	
	return l;
}

bool AstNodeFunctionDeclaration::GetCCall()
{
	return m_is_ccall;
}

bool AstNodeFunctionDeclaration::GetExport()
{
	return m_is_export;
}

bool AstNodeFunctionDeclaration::GetLock()
{
	return m_is_lock;
}


AstNodeTypeFunctionPointer *AstNodeFunctionDeclaration::GetType()
{
	return m_type;
}

AstNodeType *AstNodeFunctionDeclaration::GetReturnType()
{
	return m_return_type;
}

AstNodeStatement *AstNodeFunctionDeclaration::GetStatement()
{
	return m_statement;
}

std::list<AstNodeType *> AstNodeFunctionDeclaration::GetArgumentTypes()
{
	std::list<AstNodeType *> types;
	for (auto arg: m_args){
		types.push_back(arg->GetType());
	}
	return types;
}

std::list<AstNodeStatementVariableDeclaration *> AstNodeFunctionDeclaration::GetArguments()
{
	return m_args;
}

std::string AstNodeFunctionDeclaration::GetIdentifier()
{
	return m_name;
}

void AstNodeFunctionDeclaration::LinkToOthers(std::map<std::string, AstNode *> &node_map)
{
	std::map<std::string, AstNode *> node_map_copy = node_map;
	
	for (auto arg: m_args){
		arg->LinkToMap(node_map_copy);
		arg->GetType()->Link(node_map_copy);
	}
	
	if (m_statement){
		m_statement->Link(node_map_copy);
	}
	
	m_type->Link(node_map);
	m_return_type->Link(node_map);
}

void AstNodeFunctionDeclaration::LinkToMap(std::map<std::string, AstNode *> &node_map)
{
	AddToMap(this, node_map);
}

std::list<AstNodeStatementVariableDeclaration *> AstNodeFunctionDeclaration::GetVariableDeclarations(AstNode *stop)
{
	if (this == stop){
		return {};
	}
	std::list<AstNodeStatementVariableDeclaration *> declarations;
	for (AstNodeStatementVariableDeclaration *arg: m_args){
		if (arg == stop){
			return declarations;
		}
		declarations.push_back(arg);
	}
	
	std::list<AstNodeStatementVariableDeclaration *> child_declarations = m_statement->GetVariableDeclarations(stop);
	for (auto declaration: child_declarations){
		declarations.push_back(declaration);
	}
	
	return declarations;
}

void AstNodeFunctionDeclaration::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	m_statement->ReplaceVariableReferences(replacements);
}

uint64_t AstNodeFunctionDeclaration::GetStaticComplexityEstimate(ComputeSpec &spec)
{
	if (m_statement){
		return m_statement->GetStaticComplexityEstimate(spec);
	}
	else {
		return 1000; //Meh - why not?:p
	}
}

AstNodeExpression *AstNodeFunctionDeclaration::GetDynamicComplexityEstimate(ComputeSpec &spec)
{
	if (m_statement){
		return m_statement->GetDynamicComplexityEstimate(spec);
	}
	else {
		return 0; //None
	}
}

unsigned int AstNodeFunctionDeclaration::ReplaceArgument(AstNodeStatementVariableDeclaration *argument, std::list<AstNodeStatementVariableDeclaration *> replacements)
{
	unsigned int i = 0;
	for (auto arg: m_args){
		if (arg == argument){
			break;
		}
		i++;
	}
	unsigned int r = i;
	
	m_args.remove(argument);
	for (AstNodeStatementVariableDeclaration *replacement: replacements){
		auto it = m_args.begin();
		for (unsigned int j = 0; j < i; j++){
			it++;
		}
		m_args.insert(it, replacement);
	}
	
	i = 0;
	for (auto arg: m_args){
		i++;
		arg->SetArgument(i);
	}
	
	delete m_type;
	m_type = new AstNodeTypeFunctionPointer(m_return_type, m_args, m_is_ccall);
	
	return r;
}

AstNodeFunctionCall::AstNodeFunctionCall(AstNodeExpressionLValue *lvalue, std::list<AstNodeExpression *> expressions) :
	m_lvalue(lvalue), m_expressions(expressions)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeFunctionCall::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionFunctionCallLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionFunctionCallOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionFunctionCallPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeFunctionCall::~AstNodeFunctionCall()
{
	for (AstNodeExpression *expression: m_expressions){
		if (expression){
			delete expression;
		}
	}
	delete m_lvalue;
}

AstNodeFunctionCall *AstNodeFunctionCall::Copy()
{
	std::list<AstNodeExpression *> expressions;
	for (auto expression: m_expressions){
		expressions.push_back(expression->Copy());
	}
	
	return new AstNodeFunctionCall(m_lvalue->Copy(), expressions);
}

AstNodeExpression *AstNodeFunctionCall::Normalize(CompileContext &context)
{
	m_lvalue = m_lvalue->Normalize(context);
	
	std::list<AstNodeType *> types = GetArgumentTypes();
	if (types.size() != m_expressions.size()){
		throw FormatCompilerError(this, "Function expects %u arguments; %u given",
			(unsigned int)types.size(), (unsigned int)m_expressions.size());
	}
	//The lists are the same size if we get here
	std::list<AstNodeType *>::iterator argument_type_it = types.begin();
	
	std::list<AstNodeExpression *> expressions;
	for (auto expression_it: m_expressions){
		AstNodeExpression *expression = expression_it;
		AstNodeType *type = *argument_type_it;
		
		//cast if we need to
		if (expression->GetType()->Compare(type, context)){
			expression = new AstNodeExpressionCast(type->Copy(), expression);
		}
		
		expressions.push_back(expression->Normalize(context));
		argument_type_it++;
	}
	m_expressions = expressions;
	return this;
}

std::list<AstNodeExpression *> AstNodeFunctionCall::GetExpressions()
{
	return m_expressions;
}

std::string AstNodeFunctionCall::GetDotName()
{
	return PrintString("Function Call (%s)", m_lvalue->GetDotName().c_str());
}

std::list<AstNode **> AstNodeFunctionCall::GetChildrenPointers()
{
	std::list<AstNode **> r = ConvertList(GetPointers(m_expressions));
	r.push_back((AstNode **)&m_lvalue); //TODO: Get rid of the cycles this causes
	return r;
}

AstNodeType *AstNodeFunctionCall::GetType()
{
	return GetFunctionPointerType()->GetReturnType();
}

bool AstNodeFunctionCall::GetCCall()
{
	return GetFunctionPointerType()->GetCCall();
}

std::list<AstNodeType *> AstNodeFunctionCall::GetArgumentTypes()
{
	return GetFunctionPointerType()->GetArguments();
}

AstNodeExpressionLValue *AstNodeFunctionCall::GetLValue()
{
	return m_lvalue;
}

void AstNodeFunctionCall::Link(std::map<std::string, AstNode *> &node_map)
{
	for (auto expression: m_expressions){
		expression->Link(node_map);
	}
	
	m_lvalue->Link(node_map);
}

void AstNodeFunctionCall::ReplaceArgument(AstNodeExpression *argument, std::list<AstNodeExpression *> replacements)
{
	unsigned int i = 0;
	for (auto arg: m_expressions){
		if (arg == argument){
			break;
		}
		i++;
	}
	
	m_expressions.remove(argument);
	for (AstNodeExpression *replacement: replacements){
		auto it = m_expressions.begin();
		for (unsigned int j = 0; j < i; j++){
			it++;
		}
		m_expressions.insert(it, replacement);
	}
}

void AstNodeFunctionCall::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	for (auto expression: m_expressions){
		expression->ReplaceVariableReferences(replacements);
	}
	m_lvalue->ReplaceVariableReferences(replacements);
}

uint64_t AstNodeFunctionCall::GetStaticComplexityEstimate(ComputeSpec &spec)
{
	return GetFunctionPointerType()->GetStaticComplexityEstimate(spec);
}

AstNodeExpression *AstNodeFunctionCall::GetDynamicComplexityEstimate(ComputeSpec &spec)
{
	if (m_lvalue){
		return GetFunctionPointerType()->GetDynamicComplexityEstimate(spec);
	}
	else {
		throw FormatError("Link is null");
	}
}

AstNodeTypeFunctionPointer *AstNodeFunctionCall::GetFunctionPointerType()
{
	AstNodeTypeFunctionPointer *type = dynamic_cast<AstNodeTypeFunctionPointer *>(m_lvalue->GetType());
	if (!type){
		throw FormatError("lvalue is not a function pointer");
	}
	return type;
}
