/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "astnode.hpp"
#include "../codegen/llvm/codegen.hpp"
#include "../codegen/opencl/codegen.hpp"
#include "../codegen/pseudopc/codegen.hpp"

AstNodeExpressionCast::AstNodeExpressionCast(AstNodeType *type, AstNodeExpression *expression) :
	m_type(type), m_expression(expression)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionCast::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionCastLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionCastOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionCastPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionCast::~AstNodeExpressionCast()
{
	delete m_type;
	delete m_expression;
}

void AstNodeExpressionCast::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	m_expression->ReplaceVariableReferences(replacements);
}

AstNodeExpressionCast *AstNodeExpressionCast::Copy()
{
	return new AstNodeExpressionCast(m_type->Copy(), m_expression->Copy());
}

AstNodeExpression *AstNodeExpressionCast::Normalize(CompileContext &context)
{
	m_expression = m_expression->Normalize(context);
	return this;
}

uint64_t AstNodeExpressionCast::GetStaticComplexityEstimate(ComputeSpec &spec)
{
	return m_expression->GetStaticComplexityEstimate(spec);
}

AstNodeExpression *AstNodeExpressionCast::GetDynamicComplexityEstimate(ComputeSpec &spec)
{
	return m_expression->GetDynamicComplexityEstimate(spec);
}

AstNodeExpression *AstNodeExpressionCast::GetExpression()
{
	return m_expression;
}

std::string AstNodeExpressionCast::GetDotName()
{
	char b[1024];
	snprintf(b, 1024, "Cast from %s to %s", m_expression->GetType()->GetName().c_str(), m_type->GetName().c_str());
	return b;
}

std::list<AstNode **> AstNodeExpressionCast::GetChildrenPointers()
{
	return {(AstNode**)&m_type, (AstNode**)&m_expression};
}

AstNodeType *AstNodeExpressionCast::GetType()
{
	return m_type;
}

void AstNodeExpressionCast::Link(std::map<std::string, AstNode *> &node_map)
{
	m_expression->Link(node_map);
	m_type->Link(node_map);
}


AstNodeExpressionBinary::AstNodeExpressionBinary(std::string name, AstNodeExpression *a, AstNodeExpression *b) :
	m_name(name), m_a(a), m_b(b)
{
}

AstNodeExpressionBinary::~AstNodeExpressionBinary()
{
	delete m_a;
	delete m_b;
}

AstNodeExpression *AstNodeExpressionBinary::Normalize(CompileContext &context)
{
	m_a = m_a->Normalize(context);
	m_b = m_b->Normalize(context);
	
	
	bool is_strict = false;
	if (dynamic_cast<AstNodeTypeInt *>(m_a->GetType())){
		is_strict = is_strict | tracked_cast<AstNodeTypeInt>(m_a->GetType())->GetIsStrict();
	}
	if (dynamic_cast<AstNodeTypeInt *>(m_b->GetType())){
		is_strict = is_strict | tracked_cast<AstNodeTypeInt>(m_b->GetType())->GetIsStrict();
	}
	if (dynamic_cast<AstNodeTypeFloat *>(m_a->GetType())){
		is_strict = is_strict | tracked_cast<AstNodeTypeFloat>(m_a->GetType())->GetIsStrict();
	}
	if (dynamic_cast<AstNodeTypeFloat *>(m_b->GetType())){
		is_strict = is_strict | tracked_cast<AstNodeTypeFloat>(m_b->GetType())->GetIsStrict();
	}
	
	
	int compare = m_a->GetType()->Compare(m_b->GetType(), context);
	if (compare < 0){
		m_a = new AstNodeExpressionCast(m_b->GetType()->GetCastType(), m_a);
		m_a = m_a->Normalize(context);
	}
	else if (compare > 0){
		m_b = new AstNodeExpressionCast(m_a->GetType()->GetCastType(), m_b);
		m_b = m_b->Normalize(context);
	}
	
	if (is_strict){
		if (dynamic_cast<AstNodeTypeInt *>(m_a->GetType())){
			tracked_cast<AstNodeTypeInt>(m_a->GetType())->SetIsStrict(true);
		}
		if (dynamic_cast<AstNodeTypeInt *>(m_b->GetType())){
			tracked_cast<AstNodeTypeInt>(m_b->GetType())->SetIsStrict(true);
		}
		if (dynamic_cast<AstNodeTypeFloat *>(m_a->GetType())){
			tracked_cast<AstNodeTypeFloat>(m_a->GetType())->SetIsStrict(true);
		}
		if (dynamic_cast<AstNodeTypeFloat *>(m_b->GetType())){
			tracked_cast<AstNodeTypeFloat>(m_b->GetType())->SetIsStrict(true);
		}
		
		m_a = m_a->Normalize(context);
		m_b = m_b->Normalize(context);
	}
	
	return this;
}

void AstNodeExpressionBinary::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	m_a->ReplaceVariableReferences(replacements);
	m_b->ReplaceVariableReferences(replacements);
}

std::string AstNodeExpressionBinary::GetDotName()
{
	return m_name;
}

std::list<AstNode **> AstNodeExpressionBinary::GetChildrenPointers()
{
	return {(AstNode**)&m_a, (AstNode**)&m_b};
}

void AstNodeExpressionBinary::Link(std::map<std::string, AstNode *> &node_map)
{
	m_a->Link(node_map);
	m_b->Link(node_map);
}

AstNodeType *AstNodeExpressionBinary::GetType()
{
	return m_a->GetType(); //TODO
}

AstNodeExpression *AstNodeExpressionBinary::GetA()
{
	return m_a;
}

AstNodeExpression *AstNodeExpressionBinary::GetB()
{
	return m_b;
}

std::string AstNodeExpressionBinary::GetName()
{
	return m_name;
}

uint64_t AstNodeExpressionBinary::GetStaticComplexityEstimate(ComputeSpec &spec)
{
	return m_a->GetStaticComplexityEstimate(spec) + m_b->GetStaticComplexityEstimate(spec) + 1;
}

AstNodeExpression *AstNodeExpressionBinary::GetDynamicComplexityEstimate(ComputeSpec &spec)
{
	AstNodeExpression *a = m_a->GetDynamicComplexityEstimate(spec);
	AstNodeExpression *b = m_b->GetDynamicComplexityEstimate(spec);
	
	if (a && b){
		return new AstNodeExpressionBinaryAddition(new AstNodeExpressionBinaryAddition(a, b), new AstNodeExpressionInt(1));
	}
	else if (a){
		return new AstNodeExpressionBinaryAddition(a, new AstNodeExpressionInt(b->GetStaticComplexityEstimate(spec) + 1));
	}
	else if (b){
		return new AstNodeExpressionBinaryAddition(b, new AstNodeExpressionInt(a->GetStaticComplexityEstimate(spec) + 1));
	}
	return 0;
}

AstNodeExpressionBinaryAddition::AstNodeExpressionBinaryAddition(AstNodeExpression *a, AstNodeExpression *b) :
	AstNodeExpressionBinary("+", a, b)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionBinaryAddition::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionBinaryAdditionLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionBinaryOpencl(this, "+")); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionBinaryPseudopc(this, "+")); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionBinaryAddition *AstNodeExpressionBinaryAddition::Copy()
{
	return new AstNodeExpressionBinaryAddition(GetA()->Copy(), GetB()->Copy());
}

AstNodeExpressionBinarySubtraction::AstNodeExpressionBinarySubtraction(AstNodeExpression *a, AstNodeExpression *b) :
	AstNodeExpressionBinary("-", a, b)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionBinarySubtraction::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionBinarySubtractionLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionBinaryOpencl(this, "-")); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionBinaryPseudopc(this, "-")); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionBinarySubtraction *AstNodeExpressionBinarySubtraction::Copy()
{
	return new AstNodeExpressionBinarySubtraction(GetA()->Copy(), GetB()->Copy());
}

AstNodeExpressionBinaryMultiplication::AstNodeExpressionBinaryMultiplication(AstNodeExpression *a, AstNodeExpression *b) :
	AstNodeExpressionBinary("*", a, b)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionBinaryMultiplication::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionBinaryMultiplicationLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionBinaryOpencl(this, "*")); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionBinaryPseudopc(this, "*")); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionBinaryMultiplication *AstNodeExpressionBinaryMultiplication::Copy()
{
	return new AstNodeExpressionBinaryMultiplication(GetA()->Copy(), GetB()->Copy());
}

AstNodeExpressionBinaryDivision::AstNodeExpressionBinaryDivision(AstNodeExpression *a, AstNodeExpression *b) :
	AstNodeExpressionBinary("/", a, b)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionBinaryDivision::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionBinaryDivisionLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionBinaryOpencl(this, "/")); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionBinaryPseudopc(this, "/")); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionBinaryDivision *AstNodeExpressionBinaryDivision::Copy()
{
	return new AstNodeExpressionBinaryDivision(GetA()->Copy(), GetB()->Copy());
}

AstNodeExpressionBinaryModulo::AstNodeExpressionBinaryModulo(AstNodeExpression *a, AstNodeExpression *b) :
	AstNodeExpressionBinary("%", a, b)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionBinaryModulo::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionBinaryModuloLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionBinaryOpencl(this, "%")); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionBinaryPseudopc(this, "%")); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionBinaryModulo *AstNodeExpressionBinaryModulo::Copy()
{
	return new AstNodeExpressionBinaryModulo(GetA()->Copy(), GetB()->Copy());
}

AstNodeExpressionBinaryAnd::AstNodeExpressionBinaryAnd(AstNodeExpression *a, AstNodeExpression *b) :
	AstNodeExpressionBinary("&", a, b)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionBinaryAnd::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionBinaryAndLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionBinaryOpencl(this, "&")); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionBinaryPseudopc(this, "&")); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionBinaryAnd *AstNodeExpressionBinaryAnd::Copy()
{
	return new AstNodeExpressionBinaryAnd(GetA()->Copy(), GetB()->Copy());
}

AstNodeExpressionBinaryOr::AstNodeExpressionBinaryOr(AstNodeExpression *a, AstNodeExpression *b) :
	AstNodeExpressionBinary("|", a, b)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionBinaryOr::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionBinaryOrLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionBinaryOpencl(this, "|")); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionBinaryPseudopc(this, "|")); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionBinaryOr *AstNodeExpressionBinaryOr::Copy()
{
	return new AstNodeExpressionBinaryOr(GetA()->Copy(), GetB()->Copy());
}

AstNodeExpressionBinaryXor::AstNodeExpressionBinaryXor(AstNodeExpression *a, AstNodeExpression *b) :
	AstNodeExpressionBinary("^", a, b)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionBinaryXor::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionBinaryXorLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionBinaryOpencl(this, "^")); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionBinaryPseudopc(this, "^")); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionBinaryXor *AstNodeExpressionBinaryXor::Copy()
{
	return new AstNodeExpressionBinaryXor(GetA()->Copy(), GetB()->Copy());
}

AstNodeExpressionBinaryShl::AstNodeExpressionBinaryShl(AstNodeExpression *a, AstNodeExpression *b) :
	AstNodeExpressionBinary("<<", a, b)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionBinaryShl::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionBinaryShlLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionBinaryOpencl(this, "<<")); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionBinaryPseudopc(this, "<<")); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionBinaryShl *AstNodeExpressionBinaryShl::Copy()
{
	return new AstNodeExpressionBinaryShl(GetA()->Copy(), GetB()->Copy());
}

AstNodeExpressionBinaryShr::AstNodeExpressionBinaryShr(AstNodeExpression *a, AstNodeExpression *b) :
	AstNodeExpressionBinary(">>", a, b)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionBinaryShr::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionBinaryShrLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionBinaryOpencl(this, ">>")); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionBinaryPseudopc(this, ">>")); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionBinaryShr *AstNodeExpressionBinaryShr::Copy()
{
	return new AstNodeExpressionBinaryShr(GetA()->Copy(), GetB()->Copy());
}


/* Comparison operations */
AstNodeExpressionBinaryCompare::AstNodeExpressionBinaryCompare(std::string name, AstNodeExpression *a, AstNodeExpression *b) :
	AstNodeExpressionBinary(name, a, b)
{
}

AstNodeType *AstNodeExpressionBinaryCompare::GetType()
{
	return &m_type;
}


AstNodeExpressionBinaryLessThan::AstNodeExpressionBinaryLessThan(AstNodeExpression *a, AstNodeExpression *b) :
	AstNodeExpressionBinaryCompare("<", a, b)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionBinaryLessThan::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionBinaryCompareLlvm(this, "lt")); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionBinaryOpencl(this, "<")); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionBinaryPseudopc(this, "<")); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionBinaryLessThan *AstNodeExpressionBinaryLessThan::Copy()
{
	return new AstNodeExpressionBinaryLessThan(GetA()->Copy(), GetB()->Copy());
}

AstNodeExpressionBinaryLessThanEqual::AstNodeExpressionBinaryLessThanEqual(AstNodeExpression *a, AstNodeExpression *b) :
	AstNodeExpressionBinaryCompare("<=", a, b)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionBinaryLessThanEqual::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionBinaryCompareLlvm(this, "le")); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionBinaryOpencl(this, "<=")); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionBinaryPseudopc(this, "<=")); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionBinaryLessThanEqual *AstNodeExpressionBinaryLessThanEqual::Copy()
{
	return new AstNodeExpressionBinaryLessThanEqual(GetA()->Copy(), GetB()->Copy());
}

AstNodeExpressionBinaryGreaterThan::AstNodeExpressionBinaryGreaterThan(AstNodeExpression *a, AstNodeExpression *b) :
	AstNodeExpressionBinaryCompare(">", a, b)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionBinaryGreaterThan::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionBinaryCompareLlvm(this, "gt")); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionBinaryOpencl(this, ">")); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionBinaryPseudopc(this, ">")); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionBinaryGreaterThan *AstNodeExpressionBinaryGreaterThan::Copy()
{
	return new AstNodeExpressionBinaryGreaterThan(GetA()->Copy(), GetB()->Copy());
}

AstNodeExpressionBinaryGreaterThanEqual::AstNodeExpressionBinaryGreaterThanEqual(AstNodeExpression *a, AstNodeExpression *b) :
	AstNodeExpressionBinaryCompare(">=", a, b)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionBinaryGreaterThanEqual::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionBinaryCompareLlvm(this, "ge")); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionBinaryOpencl(this, ">/")); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionBinaryPseudopc(this, ">=")); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionBinaryGreaterThanEqual *AstNodeExpressionBinaryGreaterThanEqual::Copy()
{
	return new AstNodeExpressionBinaryGreaterThanEqual(GetA()->Copy(), GetB()->Copy());
}

AstNodeExpressionBinaryEqual::AstNodeExpressionBinaryEqual(AstNodeExpression *a, AstNodeExpression *b) :
	AstNodeExpressionBinaryCompare("==", a, b)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionBinaryEqual::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionBinaryCompareLlvm(this, "eq", false)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionBinaryOpencl(this, "==")); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionBinaryPseudopc(this, "==")); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionBinaryEqual *AstNodeExpressionBinaryEqual::Copy()
{
	return new AstNodeExpressionBinaryEqual(GetA()->Copy(), GetB()->Copy());
}

AstNodeExpressionBinaryNotEqual::AstNodeExpressionBinaryNotEqual(AstNodeExpression *a, AstNodeExpression *b) :
	AstNodeExpressionBinaryCompare("!=", a, b)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionBinaryNotEqual::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionBinaryCompareLlvm(this, "ne", false)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionBinaryOpencl(this, "!=")); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionBinaryPseudopc(this, "!=")); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionBinaryNotEqual *AstNodeExpressionBinaryNotEqual::Copy()
{
	return new AstNodeExpressionBinaryNotEqual(GetA()->Copy(), GetB()->Copy());
}
