/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef AST_OP_HPP
#define AST_OP_HPP

#include "../compilecontext.hpp"

class AstNodeExpressionCast : public AstNodeExpression
{
	public:
		AstNodeExpressionCast(AstNodeType *type, AstNodeExpression *expression);
		virtual AstNodeExpressionCast *Copy();
		virtual ~AstNodeExpressionCast();
		virtual AstNodeExpression *Normalize(CompileContext &context);
		virtual void SetCodeGenerator(CodeGeneratorType type);
		
		virtual AstNodeType *GetType();
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		
		AstNodeExpression *GetExpression();
		
		virtual void ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements);
		
		virtual uint64_t GetStaticComplexityEstimate(ComputeSpec &spec);
		virtual AstNodeExpression *GetDynamicComplexityEstimate(ComputeSpec &spec);
		
	private:
		virtual std::string GetDotName();
		virtual std::list<AstNode **> GetChildrenPointers();
		
		AstNodeType *m_type;
		AstNodeExpression *m_expression;
};

class AstNodeExpressionBinary : public AstNodeExpression
{
	public:
		AstNodeExpressionBinary(std::string name, AstNodeExpression *a, AstNodeExpression *b);
		virtual ~AstNodeExpressionBinary();
		virtual AstNodeExpression *Normalize(CompileContext &context);
		
		virtual AstNodeType *GetType();
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		
		AstNodeExpression *GetA();
		AstNodeExpression *GetB();
		std::string GetName();
		
		virtual void ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements);
		
		virtual uint64_t GetStaticComplexityEstimate(ComputeSpec &spec);
		virtual AstNodeExpression *GetDynamicComplexityEstimate(ComputeSpec &spec);
		
	private:
		virtual std::string GetDotName();
		virtual std::list<AstNode **> GetChildrenPointers();
		
		std::string m_name;
		AstNodeExpression *m_a;
		AstNodeExpression *m_b;
};

class AstNodeExpressionBinaryAddition : public AstNodeExpressionBinary
{
	public:
		AstNodeExpressionBinaryAddition(AstNodeExpression *a, AstNodeExpression *b);
		virtual AstNodeExpressionBinaryAddition *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
};

class AstNodeExpressionBinarySubtraction : public AstNodeExpressionBinary
{
	public:
		AstNodeExpressionBinarySubtraction(AstNodeExpression *a, AstNodeExpression *b);
		virtual AstNodeExpressionBinarySubtraction *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
};

class AstNodeExpressionBinaryMultiplication : public AstNodeExpressionBinary
{
	public:
		AstNodeExpressionBinaryMultiplication(AstNodeExpression *a, AstNodeExpression *b);
		virtual AstNodeExpressionBinaryMultiplication *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
};

class AstNodeExpressionBinaryDivision : public AstNodeExpressionBinary
{
	public:
		AstNodeExpressionBinaryDivision(AstNodeExpression *a, AstNodeExpression *b);
		virtual AstNodeExpressionBinaryDivision *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
};

class AstNodeExpressionBinaryModulo : public AstNodeExpressionBinary
{
	public:
		AstNodeExpressionBinaryModulo(AstNodeExpression *a, AstNodeExpression *b);
		virtual AstNodeExpressionBinaryModulo *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
};

/* Bitwise binary operations */

class AstNodeExpressionBinaryAnd : public AstNodeExpressionBinary
{
	public:
		AstNodeExpressionBinaryAnd(AstNodeExpression *a, AstNodeExpression *b);
		virtual AstNodeExpressionBinaryAnd *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
};

class AstNodeExpressionBinaryOr : public AstNodeExpressionBinary
{
	public:
		AstNodeExpressionBinaryOr(AstNodeExpression *a, AstNodeExpression *b);
		virtual AstNodeExpressionBinaryOr *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
};

class AstNodeExpressionBinaryXor : public AstNodeExpressionBinary
{
	public:
		AstNodeExpressionBinaryXor(AstNodeExpression *a, AstNodeExpression *b);
		virtual AstNodeExpressionBinaryXor *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
};

class AstNodeExpressionBinaryShl : public AstNodeExpressionBinary
{
	public:
		AstNodeExpressionBinaryShl(AstNodeExpression *a, AstNodeExpression *b);
		virtual AstNodeExpressionBinaryShl *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
};

class AstNodeExpressionBinaryShr : public AstNodeExpressionBinary
{
	public:
		AstNodeExpressionBinaryShr(AstNodeExpression *a, AstNodeExpression *b);
		virtual AstNodeExpressionBinaryShr *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
};


/* Boolean operators */

class AstNodeExpressionBinaryCompare : public AstNodeExpressionBinary
{
	public:
		AstNodeExpressionBinaryCompare(std::string name, AstNodeExpression *a, AstNodeExpression *b);
		
		virtual AstNodeType *GetType();
		
	private:
		AstNodeTypeBool m_type;
};

class AstNodeExpressionBinaryLessThan : public AstNodeExpressionBinaryCompare
{
	public:
		AstNodeExpressionBinaryLessThan(AstNodeExpression *a, AstNodeExpression *b);
		virtual AstNodeExpressionBinaryLessThan *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
};

class AstNodeExpressionBinaryLessThanEqual : public AstNodeExpressionBinaryCompare
{
	public:
		AstNodeExpressionBinaryLessThanEqual(AstNodeExpression *a, AstNodeExpression *b);
		virtual AstNodeExpressionBinaryLessThanEqual *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
};

class AstNodeExpressionBinaryGreaterThan : public AstNodeExpressionBinaryCompare
{
	public:
		AstNodeExpressionBinaryGreaterThan(AstNodeExpression *a, AstNodeExpression *b);
		virtual AstNodeExpressionBinaryGreaterThan *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
};

class AstNodeExpressionBinaryGreaterThanEqual : public AstNodeExpressionBinaryCompare
{
	public:
		AstNodeExpressionBinaryGreaterThanEqual(AstNodeExpression *a, AstNodeExpression *b);
		virtual AstNodeExpressionBinaryGreaterThanEqual *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
};

class AstNodeExpressionBinaryEqual : public AstNodeExpressionBinaryCompare
{
	public:
		AstNodeExpressionBinaryEqual(AstNodeExpression *a, AstNodeExpression *b);
		virtual AstNodeExpressionBinaryEqual *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
};

class AstNodeExpressionBinaryNotEqual : public AstNodeExpressionBinaryCompare
{
	public:
		AstNodeExpressionBinaryNotEqual(AstNodeExpression *a, AstNodeExpression *b);
		virtual AstNodeExpressionBinaryNotEqual *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
};

#endif
 
 
