/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "astnode.hpp"
#include "../codegen/llvm/codegen.hpp"
#include "../codegen/pseudopc/codegen.hpp"

AstNodeStatementParallel::AstNodeStatementParallel(AstNodeStatement *statement) :
	m_statement(statement)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeStatementParallel::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorStatementParallelLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorStatementParallelPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeStatementParallel::~AstNodeStatementParallel()
{
	if (m_statement){
		delete m_statement;
	}
}

AstNodeStatement *AstNodeStatementParallel::Normalize(CompileContext &context)
{
	m_statement = m_statement->Normalize(context);
	return this;
}

AstNodeStatement *AstNodeStatementParallel::GetStatement()
{
	return m_statement;
}

void AstNodeStatementParallel::SetStatement(AstNodeStatement *statement)
{
	m_statement = statement;
}

std::list<AstNode **> AstNodeStatementParallel::GetChildrenPointers()
{
	if (m_statement){
		return {(AstNode**)&m_statement};
	}
	else {
		return {};
	}
}

std::list<AstNodeStatementVariableDeclaration *> AstNodeStatementParallel::GetVariableDeclarations(AstNode *stop)
{
	if ((this == stop) || !m_statement){
		return {};
	}
	
	return m_statement->GetVariableDeclarations(stop);
}

void AstNodeStatementParallel::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	if (m_statement){
		m_statement->ReplaceVariableReferences(replacements);
	}
}

void AstNodeStatementParallel::Link(std::map<std::string, AstNode *> &node_map)
{
	if (m_statement){
		m_statement->Link(node_map);
	}
}

uint64_t AstNodeStatementParallel::GetStaticComplexityEstimate(ComputeSpec &spec)
{
	return m_statement->GetStaticComplexityEstimate(spec);
}

AstNodeExpression *AstNodeStatementParallel::GetDynamicComplexityEstimate(ComputeSpec &spec)
{
	return m_statement->GetDynamicComplexityEstimate(spec);
}


AstNodeStatementParallelSpawn::AstNodeStatementParallelSpawn(AstNodeStatement *statement) :
	AstNodeStatementParallel(statement)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeStatementParallelSpawn::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorStatementParallelLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorStatementParallelSpawnPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeStatementParallelSpawn *AstNodeStatementParallelSpawn::Copy()
{
	return new AstNodeStatementParallelSpawn(GetStatement() ? GetStatement()->Copy() : 0);
}

void AstNodeStatementParallelSpawn::RemoveStatement()
{
	SetStatement(0);
}

std::string AstNodeStatementParallelSpawn::GetDotName()
{
	return "Spawn";
}

AstNodeStatementParallelSync::AstNodeStatementParallelSync(AstNodeStatement *statement) :
	AstNodeStatementParallel(statement)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeStatementParallelSync::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorStatementParallelSyncLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorStatementParallelSyncPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeStatementParallelSync *AstNodeStatementParallelSync::Copy()
{
	return new AstNodeStatementParallelSync(GetStatement() ? GetStatement()->Copy() : 0);
}

std::string AstNodeStatementParallelSync::GetDotName()
{
	return "Sync";
}


AstNodeStatementExtracted::AstNodeStatementExtracted(AstNodeStatement *statement, std::list<AstNodeStatementVariableDeclaration *> variables, unsigned int linkid) :
	m_statement(statement), m_variables(variables), m_linkid(linkid)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeStatementExtracted::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorStatementExtractedLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorStatementExtractedPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeStatementExtracted::~AstNodeStatementExtracted()
{
	delete m_statement;
}

AstNodeStatementExtracted *AstNodeStatementExtracted::Copy()
{
	return new AstNodeStatementExtracted(m_statement->Copy(), m_variables, m_linkid);
}

AstNodeStatement *AstNodeStatementExtracted::Normalize(CompileContext &context)
{
	m_statement = m_statement->Normalize(context);
	return this;
}

void AstNodeStatementExtracted::Link(std::map<std::string, AstNode *> &node_map)
{
	m_statement->Link(node_map);
	std::map<std::string, AstNode *> node_map_copy = node_map;
}

std::list<AstNodeStatementVariableDeclaration *> AstNodeStatementExtracted::GetVariableDeclarations(AstNode *stop)
{
	return m_statement->GetVariableDeclarations(stop);
}

void AstNodeStatementExtracted::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	m_statement->ReplaceVariableReferences(replacements);
}

uint64_t AstNodeStatementExtracted::GetStaticComplexityEstimate(ComputeSpec &spec)
{
	return m_statement->GetStaticComplexityEstimate(spec);
}

AstNodeExpression *AstNodeStatementExtracted::GetDynamicComplexityEstimate(ComputeSpec &spec)
{
	return m_statement->GetDynamicComplexityEstimate(spec);
}

AstNodeStatement *AstNodeStatementExtracted::GetStatement()
{
	return m_statement;
}

std::list<AstNodeStatementVariableDeclaration *> AstNodeStatementExtracted::GetVariables()
{
	return m_variables;
}

unsigned int AstNodeStatementExtracted::GetLinkId()
{
	return m_linkid;
}

std::string AstNodeStatementExtracted::GetDotName()
{
	return "Extracted Spawn";
}

std::list<AstNode **> AstNodeStatementExtracted::GetChildrenPointers()
{
	return {(AstNode **)&m_statement};
}



AstNodeTypeLock::AstNodeTypeLock(AstNodeType *type) :
	m_type(type)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

AstNodeTypeLock::~AstNodeTypeLock()
{
	delete m_type;
}

AstNodeTypeLock *AstNodeTypeLock::Copy()
{
	return new AstNodeTypeLock(m_type->Copy());
}

void AstNodeTypeLock::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorTypeLockLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorTypeLockPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

std::string AstNodeTypeLock::GetName()
{
	return "lock";
}

int AstNodeTypeLock::Compare(AstNodeType *rhs, CompileContext &context)
{
	return m_type->Compare(rhs, context);
}

AstNodeType *AstNodeTypeLock::GetType()
{
	return m_type;
}

std::string AstNodeTypeLock::GetDotName()
{
	return "Lock";
}

std::list<AstNode **> AstNodeTypeLock::GetChildrenPointers()
{
	return {(AstNode**)&m_type};
}

AstNodeType *AstNodeTypeLock::GetCastType()
{
	return GetType()->GetCastType();
}



AstNodeStatementParallelAtomic::AstNodeStatementParallelAtomic(AstNodeStatement *statement) :
	AstNodeStatementParallel(statement)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

AstNodeStatementParallelAtomic *AstNodeStatementParallelAtomic::Copy()
{
	return new AstNodeStatementParallelAtomic(GetStatement()->Copy());
}

void AstNodeStatementParallelAtomic::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorStatementParallelAtomicLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorStatementParallelAtomicPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

std::string AstNodeStatementParallelAtomic::GetDotName()
{
	return "Atomic";
}



AstNodeExpressionLValueLock::AstNodeExpressionLValueLock(AstNodeExpressionLValue *lvalue) :
	m_lvalue(lvalue)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

AstNodeExpressionLValueLock::~AstNodeExpressionLValueLock()
{
	if (m_lvalue){ //Normalize might make m_lvalue 0
		delete m_lvalue;
	}
}

AstNodeExpressionLValueLock *AstNodeExpressionLValueLock::Copy()
{
	return new AstNodeExpressionLValueLock(m_lvalue->Copy());
}

AstNodeExpressionLValue *AstNodeExpressionLValueLock::Normalize(CompileContext &context)
{
	m_lvalue = m_lvalue->Normalize(context);
	if (dynamic_cast<AstNodeTypeLock *>(m_lvalue->GetType())){ //It *is* a lock
		return this;
	}
	else { //Normalize this away
		AstNodeExpressionLValue *lvalue = m_lvalue;
		m_lvalue = 0;
		delete this;
		return lvalue;
	}
}

void AstNodeExpressionLValueLock::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionLValueLockLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionLValueLockPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeType *AstNodeExpressionLValueLock::GetType()
{
	if (dynamic_cast<AstNodeTypeLock *>(m_lvalue->GetType())){
		return tracked_cast<AstNodeTypeLock>(m_lvalue->GetType())->GetType();
	}
	else {
		return m_lvalue->GetType();
	}
}

AstNodeExpressionLValue *AstNodeExpressionLValueLock::GetLValue()
{
	return m_lvalue;
}

std::string AstNodeExpressionLValueLock::GetDotName()
{
	return "Lock";
}

void AstNodeExpressionLValueLock::Link(std::map<std::string, AstNode *> &node_map)
{
	m_lvalue->Link(node_map);
}

void AstNodeExpressionLValueLock::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	m_lvalue->ReplaceVariableReferences(replacements);
}

uint64_t AstNodeExpressionLValueLock::GetStaticComplexityEstimate(ComputeSpec &spec)
{
	return m_lvalue->GetStaticComplexityEstimate(spec);
}

AstNodeExpression *AstNodeExpressionLValueLock::GetDynamicComplexityEstimate(ComputeSpec &spec)
{
	return m_lvalue->GetDynamicComplexityEstimate(spec);
}

std::list<AstNode **> AstNodeExpressionLValueLock::GetChildrenPointers()
{
	return {(AstNode**)&m_lvalue};
}




