/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "astnode.hpp"
#include "../codegen/llvm/codegen.hpp"
#include "../codegen/pseudopc/codegen.hpp"


AstNodeTypeReference::AstNodeTypeReference(AstNodeType *type) :
	m_type(type)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

AstNodeTypeReference::~AstNodeTypeReference()
{
	delete m_type;
}

void AstNodeTypeReference::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorTypeReferenceLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorTypeReferencePseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeTypeReference *AstNodeTypeReference::Copy()
{
	return new AstNodeTypeReference(m_type->Copy());
}

void AstNodeTypeReference::Link(std::map<std::string, AstNode *> &node_map)
{
	m_type->Link(node_map);
}

int AstNodeTypeReference::Compare(AstNodeType *rhs, CompileContext &context)
{
	if (dynamic_cast<AstNodeTypeVoid *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeBool *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeInt *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeFloat *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypePointer *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeStaticVector *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeDynamicVector *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeFunctionPointer *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeStruct *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeReference *>(rhs)){
		return 0;
	}
	else {
		return 1;
	}
}

std::string AstNodeTypeReference::GetName()
{
	return PrintString("%s ref", m_type->GetName().c_str());
}

AstNodeType *AstNodeTypeReference::GetType()
{
	return m_type;
}

std::string AstNodeTypeReference::GetDotName()
{
	return "ref type";
}

std::list<AstNode **> AstNodeTypeReference::GetChildrenPointers()
{
	return {(AstNode **)&m_type};
}





AstNodeExpressionReference::AstNodeExpressionReference(AstNodeExpressionLValue *lvalue) :
	m_lvalue(lvalue), m_type(0)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

AstNodeExpressionReference::~AstNodeExpressionReference()
{
	delete m_lvalue;
	if (m_type){
		delete m_type;
	}
}

void AstNodeExpressionReference::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionReferenceLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionReferencePseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionReference *AstNodeExpressionReference::Copy()
{
	return new AstNodeExpressionReference(m_lvalue->Copy());
}

AstNodeExpressionReference *AstNodeExpressionReference::Normalize(CompileContext &context)
{
	m_lvalue = m_lvalue->Normalize(context);
	return this;
}

AstNodeTypeReference *AstNodeExpressionReference::GetType()
{
	return m_type;
}

void AstNodeExpressionReference::Link(std::map<std::string, AstNode *> &node_map)
{
	m_lvalue->Link(node_map);
	
	if (m_type){
		delete m_type;
	}
	m_type = new AstNodeTypeReference(m_lvalue->GetType()->Copy());
}

void AstNodeExpressionReference::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	m_lvalue->ReplaceVariableReferences(replacements);
}

AstNodeExpressionLValue *AstNodeExpressionReference::GetLValue()
{
	return m_lvalue;
}

std::string AstNodeExpressionReference::GetDotName()
{
	return "ref";
}

std::list<AstNode **> AstNodeExpressionReference::GetChildrenPointers()
{
	return {(AstNode **)&m_lvalue, (AstNode **)&m_type};
}





AstNodeExpressionLValueDereference::AstNodeExpressionLValueDereference(AstNodeExpression *expression) :
	m_expression(expression)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

AstNodeExpressionLValueDereference::~AstNodeExpressionLValueDereference()
{
	delete m_expression;
}

void AstNodeExpressionLValueDereference::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionLValueDereferenceLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionLValueDereferencePseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionLValueDereference *AstNodeExpressionLValueDereference::Copy()
{
	return new AstNodeExpressionLValueDereference(m_expression->Copy());
}

AstNodeExpressionLValueDereference *AstNodeExpressionLValueDereference::Normalize(CompileContext &context)
{
	m_expression = m_expression->Normalize(context);
	
	if (!dynamic_cast<AstNodeTypeReference *>(m_expression->GetType())){
		throw FormatCompilerError(this, "Expression to dereference is not a reference type");
	}
	
	return this;
}

AstNodeType *AstNodeExpressionLValueDereference::GetType()
{
	if (!dynamic_cast<AstNodeTypeReference *>(m_expression->GetType())){
		throw FormatCompilerError(this, "Expression to dereference is not a reference type");
	}
	return tracked_cast<AstNodeTypeReference>(m_expression->GetType())->GetType();
}

void AstNodeExpressionLValueDereference::Link(std::map<std::string, AstNode *> &node_map)
{
	m_expression->Link(node_map);
}

void AstNodeExpressionLValueDereference::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	m_expression->ReplaceVariableReferences(replacements);
}

AstNodeExpression *AstNodeExpressionLValueDereference::GetExpression()
{
	return m_expression;
}

std::string AstNodeExpressionLValueDereference::GetDotName()
{
	return "deref";
}

std::list<AstNode **> AstNodeExpressionLValueDereference::GetChildrenPointers()
{
	return {(AstNode **)&m_expression};
}


AstNodeExpressionAlloc::AstNodeExpressionAlloc(AstNodeType *type) :
	m_type(new AstNodeTypeReference(type))
{
	SetCodeGenerator(CODEGEN_LLVM);
}

AstNodeExpressionAlloc::~AstNodeExpressionAlloc()
{
	delete m_type;
}

AstNodeExpressionAlloc *AstNodeExpressionAlloc::Copy()
{
	return new AstNodeExpressionAlloc(m_type->GetType()->Copy());
}

AstNodeExpressionAlloc *AstNodeExpressionAlloc::Normalize(CompileContext &context)
{
	return this;
}

void AstNodeExpressionAlloc::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionAllocLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionAllocPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

void AstNodeExpressionAlloc::Link(std::map<std::string, AstNode *> &node_map)
{
	m_type->Link(node_map);
}

AstNodeTypeReference *AstNodeExpressionAlloc::GetType()
{
	return m_type;
}

std::string AstNodeExpressionAlloc::GetDotName()
{
	return "Alloc";
}

std::list<AstNode **> AstNodeExpressionAlloc::GetChildrenPointers()
{
	return {(AstNode**)&m_type};
}


AstNodeStatementDealloc::AstNodeStatementDealloc(AstNodeExpression *expression) :
	m_expression(expression)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

AstNodeStatementDealloc::~AstNodeStatementDealloc()
{
	delete m_expression;
}

AstNodeStatementDealloc *AstNodeStatementDealloc::Copy()
{
	return new AstNodeStatementDealloc(m_expression->Copy());
}

AstNodeStatementDealloc *AstNodeStatementDealloc::Normalize(CompileContext &context)
{
	m_expression = m_expression->Normalize(context);
	return this;
}

void AstNodeStatementDealloc::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorStatementDeallocLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorStatementDeallocPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

void AstNodeStatementDealloc::Link(std::map<std::string, AstNode *> &node_map)
{
	m_expression->Link(node_map);
}

void AstNodeStatementDealloc::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	m_expression->ReplaceVariableReferences(replacements);
}

AstNodeExpression *AstNodeStatementDealloc::GetExpression()
{
	return m_expression;
}

std::string AstNodeStatementDealloc::GetDotName()
{
	return "Dealloc";
}

std::list<AstNode **> AstNodeStatementDealloc::GetChildrenPointers()
{
	return {(AstNode**)&m_expression};
}

