/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef AST_REFERENCE_HPP
#define AST_REFERENCE_HPP 

/*
 * ref ---- Takes an lvalue, and returns an expression that is a reference
 * deref ---- Takes an expression that is a reference, and returns an lvalue
 * alloc ---- Allocates a variable on the heap, returning a reference to that variable
 * dealloc ---- Deallocates a variable allocated with alloc when supplied with a reference to it
 */

class AstNodeTypeReference : public AstNodeType
{
	public:
		AstNodeTypeReference(AstNodeType *type);
		virtual ~AstNodeTypeReference();
		
		virtual void SetCodeGenerator(CodeGeneratorType type);
		virtual AstNodeTypeReference *Copy();
		
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		
		virtual int Compare(AstNodeType *rhs, CompileContext &context);
		virtual std::string GetName();
		
		AstNodeType *GetType();
		
	private:
		virtual std::string GetDotName();
		virtual std::list<AstNode **> GetChildrenPointers();
		
		AstNodeType *m_type;
};

class AstNodeExpressionReference : public AstNodeExpression
{
	public:
		AstNodeExpressionReference(AstNodeExpressionLValue *lvalue);
		virtual ~AstNodeExpressionReference();
		
		virtual void SetCodeGenerator(CodeGeneratorType type);
		virtual AstNodeExpressionReference *Copy();
		virtual AstNodeExpressionReference *Normalize(CompileContext &context);
		virtual AstNodeTypeReference *GetType();
		
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		virtual void ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements);
		
		AstNodeExpressionLValue *GetLValue();
		
	private:
		virtual std::string GetDotName();
		virtual std::list<AstNode **> GetChildrenPointers();
		
		AstNodeExpressionLValue *m_lvalue;
		AstNodeTypeReference *m_type; //This must be updated when we link
};

class AstNodeExpressionLValueDereference : public AstNodeExpressionLValue
{
	public:
		AstNodeExpressionLValueDereference(AstNodeExpression *expression);
		virtual ~AstNodeExpressionLValueDereference();
		
		virtual void SetCodeGenerator(CodeGeneratorType type);
		virtual AstNodeExpressionLValueDereference *Copy();
		virtual AstNodeExpressionLValueDereference *Normalize(CompileContext &context);
		virtual AstNodeType *GetType();
		
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		virtual void ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements);
		
		AstNodeExpression *GetExpression();
		
	private:
		virtual std::string GetDotName();
		virtual std::list<AstNode **> GetChildrenPointers();
		
		AstNodeExpression *m_expression;
};


class AstNodeExpressionAlloc : public AstNodeExpression
{
	public:
		AstNodeExpressionAlloc(AstNodeType *type);
		virtual ~AstNodeExpressionAlloc();
		
		virtual AstNodeExpressionAlloc *Copy();
		virtual AstNodeExpressionAlloc *Normalize(CompileContext &context);
		virtual void SetCodeGenerator(CodeGeneratorType type);
		
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		
		virtual AstNodeTypeReference *GetType();
		
	private:
		virtual std::string GetDotName();
		virtual std::list<AstNode **> GetChildrenPointers();
		
		AstNodeTypeReference *m_type;
};

class AstNodeStatementDealloc : public AstNodeStatement
{
	public:
		AstNodeStatementDealloc(AstNodeExpression *expression);
		virtual ~AstNodeStatementDealloc();
		
		virtual AstNodeStatementDealloc *Copy();
		virtual AstNodeStatementDealloc *Normalize(CompileContext &context);
		virtual void SetCodeGenerator(CodeGeneratorType type);
		
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		virtual void ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements);
		
		virtual AstNodeExpression *GetExpression();
		
	private:
		virtual std::string GetDotName();
		virtual std::list<AstNode **> GetChildrenPointers();
		
		AstNodeExpression *m_expression;
};


#endif
