/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "astnode.hpp"
#include "../codegen/llvm/codegen.hpp"
#include "../codegen/opencl/codegen.hpp"
#include "../codegen/pseudopc/codegen.hpp"

void AstNodeStatement::Link(std::map<std::string, AstNode *> &node_map)
{
}

std::list<AstNodeStatementVariableDeclaration *> AstNodeStatement::GetVariableDeclarations(AstNode *stop)
{
	return {};
}

AstNodeStatementBlock::AstNodeStatementBlock(std::list<AstNodeStatement *> statements) :
	m_statements(statements)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeStatementBlock::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorStatementBlockLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorStatementBlockOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorStatementBlockPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeStatementBlock::~AstNodeStatementBlock()
{
	for (AstNodeStatement *statement: m_statements){
		if (statement){
			delete statement;
		}
	}
}

AstNodeStatementBlock *AstNodeStatementBlock::Copy()
{
	std::list<AstNodeStatement *> statements;
	for (auto statement: m_statements){
		statements.push_back(statement->Copy());
	}
	
	return new AstNodeStatementBlock(statements);
}

AstNodeStatement *AstNodeStatementBlock::Normalize(CompileContext &context)
{
	std::list<AstNodeStatement *> statements;
	for (auto statement: m_statements){
		statements.push_back(statement->Normalize(context));
	}
	m_statements = statements;
	return this;
}


std::list<AstNodeStatementVariableDeclaration *> AstNodeStatementBlock::GetVariableDeclarations(AstNode *stop)
{
	if (this == stop){
		return {};
	}
	
	std::list<AstNodeStatementVariableDeclaration *> declarations;
	for (auto statement: m_statements){
		if (statement == stop){
			return declarations;
		}
		std::list<AstNodeStatementVariableDeclaration *> child_declarations = statement->GetVariableDeclarations(stop);
		for (auto declaration: child_declarations){
			declarations.push_back(declaration);
		}
	}
	return declarations;
}

void AstNodeStatementBlock::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	for (auto statement: m_statements){
		statement->ReplaceVariableReferences(replacements);
	}
}

std::list<AstNodeStatement *> AstNodeStatementBlock::GetStatements()
{
	return m_statements;
}

void AstNodeStatementBlock::Link(std::map<std::string, AstNode *> &node_map)
{
	//We create a copy of the maps
	std::map<std::string, AstNode *> node_map_copy = node_map;
	
	for (auto statement: m_statements){
		statement->Link(node_map_copy);
	}
}

std::list<AstNode **> AstNodeStatementBlock::GetChildrenPointers()
{
	return ConvertList(GetPointers(m_statements));
}

std::string AstNodeStatementBlock::GetDotName()
{
	return "Statement Block";
}

uint64_t AstNodeStatementBlock::GetStaticComplexityEstimate(ComputeSpec &spec)
{
	uint64_t sum = 0;
	for (auto statement: m_statements){
		sum += statement->GetStaticComplexityEstimate(spec);
	}
	return sum;
}

AstNodeExpression *AstNodeStatementBlock::GetDynamicComplexityEstimate(ComputeSpec &spec)
{
	AstNodeExpression *sum = 0;
	unsigned int dynamic_count = 0;
	for (auto statement: m_statements){
		AstNodeExpression *e = statement->GetDynamicComplexityEstimate(spec);
		if (e){
			dynamic_count++;
		}
		
		if (sum && e){
			sum = new AstNodeExpressionBinaryAddition(e, sum);
		}
		else if (sum){
			sum = new AstNodeExpressionBinaryAddition(sum, new AstNodeExpressionInt(statement->GetStaticComplexityEstimate(spec)));
		}
		else {
			sum = e;
		}
	}
	
	if (dynamic_count){
		return sum;
	}
	else {
		delete sum; //avoid memory leak
		return 0;
	}
}

AstNodeStatementExpression::AstNodeStatementExpression(AstNodeExpression *expression) :
	m_expression(expression)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeStatementExpression::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorStatementExpressionLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorStatementExpressionOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorStatementExpressionPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeStatementExpression::~AstNodeStatementExpression()
{
	delete m_expression;
}

AstNodeStatementExpression *AstNodeStatementExpression::Copy()
{
	return new AstNodeStatementExpression(m_expression->Copy());
}

AstNodeStatement *AstNodeStatementExpression::Normalize(CompileContext &context)
{
	m_expression = m_expression->Normalize(context);
	return this;
}

AstNodeExpression *AstNodeStatementExpression::GetExpression()
{
	return m_expression;
}

void AstNodeStatementExpression::Link(std::map<std::string, AstNode *> &node_map)
{
	m_expression->Link(node_map);
}

std::string AstNodeStatementExpression::GetDotName()
{
	return "Expression Statement";
}

std::list<AstNode **> AstNodeStatementExpression::GetChildrenPointers()
{
	return {(AstNode**)&m_expression};
}

void AstNodeStatementExpression::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	m_expression->ReplaceVariableReferences(replacements);
}

uint64_t AstNodeStatementExpression::GetStaticComplexityEstimate(ComputeSpec &spec)
{
	return m_expression->GetStaticComplexityEstimate(spec);
}

AstNodeExpression *AstNodeStatementExpression::GetDynamicComplexityEstimate(ComputeSpec &spec)
{
	return m_expression->GetDynamicComplexityEstimate(spec);
}


AstNodeStatementVariableDeclaration::AstNodeStatementVariableDeclaration
	(AstNodeType *type, std::string identifier) :
	m_type(type), m_identifier(identifier), m_global(false), m_argument(0), m_copy_on_spawn(false)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeStatementVariableDeclaration::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorStatementVariableDeclarationLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorStatementVariableDeclarationOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorStatementVariableDeclarationPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeStatementVariableDeclaration::~AstNodeStatementVariableDeclaration()
{
	if (m_type){
		delete m_type;
	}
}

AstNodeStatementVariableDeclaration *AstNodeStatementVariableDeclaration::Copy()
{
	AstNodeStatementVariableDeclaration *declaration = new AstNodeStatementVariableDeclaration (m_type->Copy(), m_identifier);
	if (m_global){
		declaration->SetGlobal();
	}
	declaration->SetArgument(m_argument);
	if (m_copy_on_spawn){
		declaration->SetCopyOnSpawn();
	}
	return declaration;
}

AstNodeStatementVariableDeclaration *AstNodeStatementVariableDeclaration::Normalize(CompileContext &context)
{
	return this;
}

void AstNodeStatementVariableDeclaration::SetGlobal()
{
	throw FormatCompilerError(this, "Global variables are now somwhat out of date and untested. You probably don't want to use them.");
	m_global = true;
}

void AstNodeStatementVariableDeclaration::SetArgument(unsigned int n)
{
	m_argument = n;
}

void AstNodeStatementVariableDeclaration::SetCopyOnSpawn()
{
	m_copy_on_spawn = true;
}

std::string AstNodeStatementVariableDeclaration::GetDotName()
{
	return std::string("Variable Declaration: \\\"") + m_identifier + std::string("\\\"");
}

std::list<AstNode **> AstNodeStatementVariableDeclaration::GetChildrenPointers()
{
	return {(AstNode**)&m_type};
}

std::string AstNodeStatementVariableDeclaration::GetIdentifier()
{
	return m_identifier;
}

AstNodeType *AstNodeStatementVariableDeclaration::GetType()
{
	return m_type;
}

bool AstNodeStatementVariableDeclaration::GetGlobal()
{
	return m_global;
}

unsigned int AstNodeStatementVariableDeclaration::GetArgument()
{
	return m_argument;
}

bool AstNodeStatementVariableDeclaration::GetCopyOnSpawn()
{
	return m_copy_on_spawn;
}

std::list<AstNodeStatementVariableDeclaration *> AstNodeStatementVariableDeclaration::GetVariableDeclarations(AstNode *stop)
{
	if (this == stop){
		return {};
	}
	else {
		return {this};
	}
}

void AstNodeStatementVariableDeclaration::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	if (replacements.count(this)){
		throw FormatCompilerError(this, "Cannot replace this with something else as a variable reference replacement");
	}
}

void AstNodeStatementVariableDeclaration::LinkToOthers(std::map<std::string, AstNode *> &node_map)
{
	m_type->Link(node_map);
}

void AstNodeStatementVariableDeclaration::LinkToMap(std::map<std::string, AstNode *> &node_map)
{
	AddToMap(this, node_map);
}

void AstNodeStatementVariableDeclaration::Link(std::map<std::string, AstNode *> &node_map)
{
	AddToMap(this, node_map);
	m_type->Link(node_map);
}

AstNodeStatementVariableAssignment::AstNodeStatementVariableAssignment
	(AstNodeExpressionLValue *lvalue, AstNodeExpression *expression) :
	m_lvalue(lvalue), m_expression(expression)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeStatementVariableAssignment::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorStatementVariableAssignmentLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorStatementVariableAssignmentOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorStatementVariableAssignmentPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeStatementVariableAssignment::~AstNodeStatementVariableAssignment()
{
	delete m_lvalue;
	delete m_expression;
}

AstNodeStatementVariableAssignment *AstNodeStatementVariableAssignment::Copy()
{
	return new AstNodeStatementVariableAssignment (m_lvalue->Copy(), m_expression->Copy());
}

AstNodeStatement *AstNodeStatementVariableAssignment::Normalize(CompileContext &context)
{
	m_lvalue = m_lvalue->Normalize(context);
	
	m_expression = m_expression->Normalize(context);
	if (m_expression->GetType()->Compare(m_lvalue->GetType(), context)){
		m_expression = new AstNodeExpressionCast(m_lvalue->GetType()->GetCastType(), m_expression);
		m_expression = m_expression->Normalize(context);
	}
	
	return this;
}

AstNodeExpressionLValue *AstNodeStatementVariableAssignment::GetLValue()
{
	return m_lvalue;
}

AstNodeExpression *AstNodeStatementVariableAssignment::GetExpression()
{
	return m_expression;
}

std::string AstNodeStatementVariableAssignment::GetDotName()
{
	return "Variable Assignment";
}

std::list<AstNode **> AstNodeStatementVariableAssignment::GetChildrenPointers()
{
	return {(AstNode**)&m_lvalue, (AstNode**)&m_expression};
}

void AstNodeStatementVariableAssignment::Link(std::map<std::string, AstNode *> &node_map)
{
	m_lvalue->Link(node_map);
	m_expression->Link(node_map);
}

void AstNodeStatementVariableAssignment::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	m_lvalue->ReplaceVariableReferences(replacements);
	m_expression->ReplaceVariableReferences(replacements);
}

uint64_t AstNodeStatementVariableAssignment::GetStaticComplexityEstimate(ComputeSpec &spec)
{
	return m_lvalue->GetStaticComplexityEstimate(spec) + m_expression->GetStaticComplexityEstimate(spec);
}

AstNodeExpression *AstNodeStatementVariableAssignment::GetDynamicComplexityEstimate(ComputeSpec &spec)
{
	AstNodeExpression *a = m_expression->GetDynamicComplexityEstimate(spec);
	AstNodeExpression *b = m_lvalue->GetDynamicComplexityEstimate(spec);
	
	if (a && b){
		return new AstNodeExpressionBinaryAddition(a, b);
	}
	else if (a){
		return new AstNodeExpressionBinaryAddition(a, new AstNodeExpressionInt(b->GetStaticComplexityEstimate(spec)));
	}
	else if (b){
		return new AstNodeExpressionBinaryAddition(b, new AstNodeExpressionInt(a->GetStaticComplexityEstimate(spec)));
	}
	return 0;
}

AstNodeStatementReturn::AstNodeStatementReturn(AstNodeExpression *expression) :
	m_expression(expression)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeStatementReturn::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorStatementReturnLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorStatementReturnOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorStatementReturnPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeStatementReturn::~AstNodeStatementReturn()
{
	if (m_expression){
		delete m_expression;
	}
}

AstNodeStatementReturn *AstNodeStatementReturn::Copy()
{
	return new AstNodeStatementReturn(m_expression->Copy());
}

AstNodeStatement *AstNodeStatementReturn::Normalize(CompileContext &context)
{
	m_expression = m_expression->Normalize(context);
	return this;
}

std::string AstNodeStatementReturn::GetDotName()
{
	return "Return";
}

std::list<AstNode **> AstNodeStatementReturn::GetChildrenPointers()
{
	return {(AstNode**)&m_expression};
}

void AstNodeStatementReturn::Link(std::map<std::string, AstNode *> &node_map)
{
	if (m_expression){
		m_expression->Link(node_map);
	}
}

AstNodeExpression *AstNodeStatementReturn::GetExpression()
{
	return m_expression;
}

void AstNodeStatementReturn::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	m_expression->ReplaceVariableReferences(replacements);
}

uint64_t AstNodeStatementReturn::GetStaticComplexityEstimate(ComputeSpec &spec)
{
	return m_expression->GetStaticComplexityEstimate(spec);
}

AstNodeExpression *AstNodeStatementReturn::GetDynamicComplexityEstimate(ComputeSpec &spec)
{
	return m_expression->GetDynamicComplexityEstimate(spec);
}

AstNodeStatementMarkedGroup::AstNodeStatementMarkedGroup(std::list<AstNodeStatement *> statements) :
	m_statements(statements)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeStatementMarkedGroup::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorStatementMarkedGroupLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorStatementMarkedGroupPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeStatementMarkedGroup::~AstNodeStatementMarkedGroup()
{
	for (AstNodeStatement *statement: m_statements){
		if (statement){
			delete statement;
		}
	}
}

std::list<AstNodeStatement *> AstNodeStatementMarkedGroup::CopyStatements()
{
	std::list<AstNodeStatement *> statements;
	for (auto statement: m_statements){
		statements.push_back(statement->Copy());
	}
	return statements;
}

AstNodeStatement *AstNodeStatementMarkedGroup::Normalize(CompileContext &context)
{
	std::list<AstNodeStatement *> statements;
	for (auto statement: m_statements){
		statements.push_back(statement->Normalize(context));
	}
	m_statements = statements;
	return this;
}


std::list<AstNodeStatementVariableDeclaration *> AstNodeStatementMarkedGroup::GetVariableDeclarations(AstNode *stop)
{
	if (this == stop){
		return {};
	}
	
	std::list<AstNodeStatementVariableDeclaration *> declarations;
	for (auto statement: m_statements){
		if (statement == stop){
			return declarations;
		}
		std::list<AstNodeStatementVariableDeclaration *> child_declarations = statement->GetVariableDeclarations(stop);
		for (auto declaration: child_declarations){
			declarations.push_back(declaration);
		}
	}
	return declarations;
}

void AstNodeStatementMarkedGroup::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	for (auto statement: m_statements){
		statement->ReplaceVariableReferences(replacements);
	}
}

std::list<AstNodeStatement *> AstNodeStatementMarkedGroup::GetStatements()
{
	return m_statements;
}

void AstNodeStatementMarkedGroup::Link(std::map<std::string, AstNode *> &node_map)
{
	//The difference between a marked group and a block is that the node map isn't copied in the link stage
	for (auto statement: m_statements){
		statement->Link(node_map);
	}
}

std::list<AstNode **> AstNodeStatementMarkedGroup::GetChildrenPointers()
{
	return ConvertList(GetPointers(m_statements));
}

std::string AstNodeStatementMarkedGroup::GetDotName()
{
	return PrintString("Statement Marked Group of %u Items: %s", (unsigned int)m_statements.size(), GetMarkName().c_str());
}

uint64_t AstNodeStatementMarkedGroup::GetStaticComplexityEstimate(ComputeSpec &spec)
{
	uint64_t sum = 0;
	for (auto statement: m_statements){
		sum += statement->GetStaticComplexityEstimate(spec);
	}
	return sum;
}

AstNodeExpression *AstNodeStatementMarkedGroup::GetDynamicComplexityEstimate(ComputeSpec &spec)
{
	AstNodeExpression *sum = 0;
	unsigned int dynamic_count = 0;
	for (auto statement: m_statements){
		AstNodeExpression *e = statement->GetDynamicComplexityEstimate(spec);
		if (e){
			dynamic_count++;
		}
		
		if (sum && e){
			sum = new AstNodeExpressionBinaryAddition(e, sum);
		}
		else if (sum){
			sum = new AstNodeExpressionBinaryAddition(sum, new AstNodeExpressionInt(statement->GetStaticComplexityEstimate(spec)));
		}
		else {
			sum = e;
		}
	}
	
	if (dynamic_count){
		return sum;
	}
	else {
		delete sum; //avoid memory leak
		return 0;
	}
}
