/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef AST_STATEMENT_HPP
#define AST_STATEMENT_HPP

#include "../compilecontext.hpp"

class AstNodeStatement : public AstNode
{
	public:
		virtual AstNodeStatement *Copy() = 0;
		virtual AstNodeStatement *Normalize(CompileContext &context) = 0;
		
		/*
		 * Link() must be called whenever anything is done that might mess up 
		 * he linking e.g: deep copying a node.
		 * Link() is responsible for propogating function and variable names to
		 * nodes that need to reference them. For example, the function call
		 * node needs to be able to find the function definition that
		 * corresponds to the name in the function call node. Similarly, an
		 * identifier value is a node that gets its value from that stored in a
		 * memory location. The memory location is created when the variable is
		 * declared. So the identifier value needs a pointer to the variable
		 * declaration with the same name as the identifier value.
		 * It is this method that allows nodes to find the pointer that
		 * corresponds with the name.
		 * 
		 * All AstNodes that require linking have a similar method. These
		 * methods are called recursively. It is often necessary to ensure the
		 * maps contain only the names up to the node being linked.
		 * AstNodeStatementBlock::Link() is a good example of this. The maps are
		 * references to higher level maps in order to allow a child node to add
		 * its name and pointer to the map to be used by other nodes. Where this
		 * behaviour is not desired, a copy of the map must be made.
		 * AstNodeStatementBlock::Link() is a good example of this too, where
		 * declarations should not be visible to its parent. Link() is often
		 * virtual in order to allow linking even where the type of the node is
		 * not availabe until runtime.
		 */
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		
		virtual std::list<AstNodeStatementVariableDeclaration *> GetVariableDeclarations(AstNode *stop = 0);
		virtual void ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements) = 0;
		
	private:
};

class AstNodeStatementBlock : public AstNodeStatement
{
	public:
		AstNodeStatementBlock(std::list<AstNodeStatement *> statements);
		virtual ~AstNodeStatementBlock();
		virtual AstNodeStatementBlock *Copy();
		virtual AstNodeStatement *Normalize(CompileContext &context);
		virtual void SetCodeGenerator(CodeGeneratorType type);
		
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		
		std::list<AstNodeStatement *> GetStatements();
		
		virtual std::list<AstNodeStatementVariableDeclaration *> GetVariableDeclarations(AstNode *stop = 0);
		virtual void ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements);
		
		virtual uint64_t GetStaticComplexityEstimate(ComputeSpec &spec);
		virtual AstNodeExpression *GetDynamicComplexityEstimate(ComputeSpec &spec);
		
	private:
		virtual std::string GetDotName();
		virtual std::list<AstNode **> GetChildrenPointers();
		
		std::list<AstNodeStatement *> m_statements;
};

class AstNodeStatementExpression : public AstNodeStatement
{
	public:
		AstNodeStatementExpression(AstNodeExpression *expression);
		virtual ~AstNodeStatementExpression();
		virtual AstNodeStatementExpression *Copy();
		virtual AstNodeStatement *Normalize(CompileContext &context);
		virtual void SetCodeGenerator(CodeGeneratorType type);
		
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		
		AstNodeExpression *GetExpression();
		
		virtual void ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements);
		
		virtual uint64_t GetStaticComplexityEstimate(ComputeSpec &spec);
		virtual AstNodeExpression *GetDynamicComplexityEstimate(ComputeSpec &spec);
		
	private:
		virtual std::string GetDotName();
		virtual std::list<AstNode **> GetChildrenPointers();
		
		AstNodeExpression *m_expression;
};


class AstNodeStatementVariableDeclaration : public AstNodeStatement
{
	public:
		AstNodeStatementVariableDeclaration(AstNodeType *type, std::string identifier);
		virtual ~AstNodeStatementVariableDeclaration();
		virtual AstNodeStatementVariableDeclaration *Copy();
		virtual AstNodeStatementVariableDeclaration *Normalize(CompileContext &context);
		virtual void SetCodeGenerator(CodeGeneratorType type);
		
		void SetGlobal();
		void SetArgument(unsigned int n);
		void SetCopyOnSpawn();
		
		bool GetGlobal();
		unsigned int GetArgument();
		bool GetCopyOnSpawn();
		std::string GetIdentifier();
		AstNodeType *GetType();
		
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		void LinkToMap(std::map<std::string, AstNode *> &node_map);
		void LinkToOthers(std::map<std::string, AstNode *> &node_map);
		
		virtual std::list<AstNodeStatementVariableDeclaration *> GetVariableDeclarations(AstNode *stop = 0);
		virtual void ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements);
		
	private:
		virtual std::string GetDotName();
		virtual std::list<AstNode **> GetChildrenPointers();
		
		AstNodeType *m_type;
		std::string m_identifier;
		bool m_global;
		unsigned int m_argument;
		bool m_copy_on_spawn;
};


class AstNodeStatementVariableAssignment : public AstNodeStatement
{
	public:
		AstNodeStatementVariableAssignment(AstNodeExpressionLValue *lvalue, AstNodeExpression *expression);
		virtual AstNodeStatementVariableAssignment *Copy();
		virtual ~AstNodeStatementVariableAssignment();
		virtual AstNodeStatement *Normalize(CompileContext &context);
		virtual void SetCodeGenerator(CodeGeneratorType type);
		
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		
		AstNodeExpressionLValue *GetLValue();
		AstNodeExpression *GetExpression();
		
		virtual void ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements);
		
		virtual uint64_t GetStaticComplexityEstimate(ComputeSpec &spec);
		virtual AstNodeExpression *GetDynamicComplexityEstimate(ComputeSpec &spec);
		
	private:
		virtual std::string GetDotName();
		virtual std::list<AstNode **> GetChildrenPointers();
		
		AstNodeExpressionLValue *m_lvalue;
		AstNodeExpression *m_expression;
};

class AstNodeStatementReturn : public AstNodeStatement
{
	public:
		AstNodeStatementReturn(AstNodeExpression *expression);
		virtual AstNodeStatementReturn *Copy();
		virtual ~AstNodeStatementReturn();
		virtual AstNodeStatement *Normalize(CompileContext &context);
		virtual void SetCodeGenerator(CodeGeneratorType type);
		
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		
		AstNodeExpression *GetExpression();
		
		virtual void ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements);
		
		virtual uint64_t GetStaticComplexityEstimate(ComputeSpec &spec);
		virtual AstNodeExpression *GetDynamicComplexityEstimate(ComputeSpec &spec);
		
	private:
		virtual std::string GetDotName();
		virtual std::list<AstNode **> GetChildrenPointers();
		
		AstNodeExpression *m_expression;
};


class AstNodeStatementMarkedGroup : public AstNodeStatement
{
	public:
		AstNodeStatementMarkedGroup(std::list<AstNodeStatement *> statements);
		virtual ~AstNodeStatementMarkedGroup();
		virtual AstNodeStatement *Normalize(CompileContext &context);
		virtual void SetCodeGenerator(CodeGeneratorType type);
		
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		
		std::list<AstNodeStatement *> GetStatements();
		
		virtual std::list<AstNodeStatementVariableDeclaration *> GetVariableDeclarations(AstNode *stop = 0);
		virtual void ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements);
		
		virtual uint64_t GetStaticComplexityEstimate(ComputeSpec &spec);
		virtual AstNodeExpression *GetDynamicComplexityEstimate(ComputeSpec &spec);
		
		virtual std::string GetMarkName() = 0;
		
		//Finds the first of T
		template <typename T> T *FindByType()
		{
			for (AstNodeStatement *statement: m_statements){
				if (dynamic_cast<T *>(statement)){
					return tracked_cast<T>(statement);
				}
			}
			
			#ifdef DEBUG
				throw FormatCompilerError("Optimisation error: Cannot find statement of the specified type\n");
			#else
				return 0;
			#endif
		}
		
	protected:
		std::list<AstNodeStatement *> CopyStatements();
		
	private:
		virtual std::string GetDotName();
		virtual std::list<AstNode **> GetChildrenPointers();
		
		std::list<AstNodeStatement *> m_statements;
};

#define DefineAstNodeStatementMarkedGroup(c, n) \
	class c : public AstNodeStatementMarkedGroup \
	{ \
		public: \
			c(std::list<AstNodeStatement *> statements) : AstNodeStatementMarkedGroup(statements){} \
			virtual c *Copy(){return new c(CopyStatements());} \
			virtual std::string GetMarkName(){return n;} \
	}

DefineAstNodeStatementMarkedGroup(AstNodeStatementVariableDeclarationAssignment, "Variable Declaration and Assignment");

#include "ctrlflow.hpp"

#endif
