/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "astnode.hpp"
#include "../codegen/llvm/codegen.hpp"
#include "../codegen/opencl/codegen.hpp"
#include "../codegen/pseudopc/codegen.hpp"


AstNodeStructItem::AstNodeStructItem(AstNodeType *type, std::string identifier) :
	m_type(type), m_identifier(identifier)
{
}

void AstNodeStructItem::SetCodeGenerator(CodeGeneratorType type)
{
	//Actually, do nothing
}

AstNodeStructItem::~AstNodeStructItem()
{
	delete m_type;
}

AstNodeStructItem *AstNodeStructItem::Copy()
{
	return new AstNodeStructItem(m_type->Copy(), m_identifier);
}

std::string AstNodeStructItem::GetIdentifier()
{
	return m_identifier;
}

AstNodeType *AstNodeStructItem::GetType()
{
	return m_type;
}

std::string AstNodeStructItem::GetDotName()
{
	return PrintString("Structure declaration item: \\\"%s\\\"", m_identifier.c_str());
}

std::list<AstNode **> AstNodeStructItem::GetChildrenPointers()
{
	return {(AstNode**)&m_type};
}



AstNodeStructDeclaration::AstNodeStructDeclaration(std::string identifier, std::list<AstNodeStructItem *> items) :
	m_identifier(identifier), m_items(items)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeStructDeclaration::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorStructDeclarationLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorStructDeclarationOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorStructDeclarationPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeStructDeclaration::~AstNodeStructDeclaration()
{
	for (auto item: m_items){
		delete item;
	}
}

AstNodeStructDeclaration *AstNodeStructDeclaration::Copy()
{
	std::list<AstNodeStructItem *> items;
	for (auto item: m_items){
		items.push_back(item->Copy());
	}
	
	return new AstNodeStructDeclaration(m_identifier, items);
}

std::string AstNodeStructDeclaration::GetIdentifier()
{
	return m_identifier;
}

void AstNodeStructDeclaration::LinkToOthers(std::map<std::string, AstNode *> &node_map)
{
	for (auto item: m_items){
		item->GetType()->Link(node_map);
	}
}

void AstNodeStructDeclaration::LinkToMap(std::map<std::string, AstNode *> &node_map)
{
	AddToMap(this, node_map);
}

void AstNodeStructDeclaration::Link(std::map<std::string, AstNode *> &node_map)
{
	AddToMap(this, node_map);
	LinkToOthers(node_map);
}

std::list<AstNodeStructItem *> AstNodeStructDeclaration::GetItems()
{
	return m_items;
}

std::string AstNodeStructDeclaration::GetDotName()
{
	return PrintString("Structure declaration: \\\"%s\\\"", m_identifier.c_str());
}

std::list<AstNode **> AstNodeStructDeclaration::GetChildrenPointers()
{
	return ConvertList(GetPointers(m_items));
}


AstNodeTypeStruct::AstNodeTypeStruct(std::string identifier) :
	m_identifier(identifier), m_link(0)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeTypeStruct::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorTypeStructLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorTypeStructOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorTypeStructPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeTypeStruct *AstNodeTypeStruct::Copy()
{
	return new AstNodeTypeStruct(m_identifier);
}

int AstNodeTypeStruct::Compare(AstNodeType *rhs, CompileContext &context)
{
	if (dynamic_cast<AstNodeTypeVoid *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeBool *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeInt *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeFloat *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypePointer *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeStaticVector *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeDynamicVector *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeFunctionPointer *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeStruct *>(rhs)){
		return 0;
	}
	else {
		return 1;
	}
}

std::string AstNodeTypeStruct::GetIdentifier()
{
	return m_identifier;
}

AstNodeStructDeclaration *AstNodeTypeStruct::GetLink()
{
	if (!m_link){
		throw FormatCompilerError(this, "Link is null: \"%s\": %s", m_identifier.c_str(), GetSourceLocation().c_str());
	}
	return m_link;
}

std::string AstNodeTypeStruct::GetName()
{
	return PrintString("struct %s", m_identifier.c_str());
}

std::string AstNodeTypeStruct::GetDotName()
{
	return PrintString("Structure type: \\\"%s\\\"", m_identifier.c_str());
}

void AstNodeTypeStruct::Link(std::map<std::string, AstNode *> &node_map)
{
	if (!node_map.count(m_identifier)){
		throw FormatCompilerError(this, "\"%s\" is not declared: %s", m_identifier.c_str(), GetSourceLocation().c_str());
	}
	AstNodeStructDeclaration *node = dynamic_cast<AstNodeStructDeclaration *>(node_map[m_identifier]);
	if (!node){
		throw FormatCompilerError(this, "\"%s\" is not a struct", m_identifier.c_str());
	}
	m_link = node;
}

std::list<AstNode **> AstNodeTypeStruct::GetLinksPointers()
{
	return {(AstNode **)&m_link};
}


AstNodeExpressionLValueStructElement::AstNodeExpressionLValueStructElement(AstNodeExpressionLValue *lvalue, std::string identifier) :
	m_lvalue(lvalue), m_identifier(identifier)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeExpressionLValueStructElement::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorExpressionLValueStructElementLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorExpressionLValueStructElementOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorExpressionLValueStructElementPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeExpressionLValueStructElement::~AstNodeExpressionLValueStructElement()
{
	delete m_lvalue;
}

AstNodeExpressionLValueStructElement *AstNodeExpressionLValueStructElement::Copy()
{
	AstNodeExpressionLValueStructElement *copy = new AstNodeExpressionLValueStructElement(m_lvalue->Copy(), m_identifier);
	copy->SetDoRefCount(GetDoRefCount());
	return copy;
}

AstNodeExpressionLValue *AstNodeExpressionLValueStructElement::Normalize(CompileContext &context)
{
	m_lvalue = m_lvalue->Normalize(context);
	
	AstNodeTypeStruct *struct_type = dynamic_cast<AstNodeTypeStruct *>(m_lvalue->GetType());
	if (!struct_type){
		throw FormatCompilerError(this, "LValue type is not a struct");
	}
	
	return this;
}

AstNodeType *AstNodeExpressionLValueStructElement::GetType()
{
	AstNodeStructDeclaration *struct_declaration = GetStructType()->GetLink();
	std::list<AstNodeStructItem *> items = struct_declaration->GetItems();
	
	for (auto item: items){
		if (item->GetIdentifier() == m_identifier){
			return item->GetType();
		}
	}
	
	throw FormatCompilerError(this, "\"%s\" is not in struct \"%s\"", m_identifier.c_str(), struct_declaration->GetIdentifier().c_str());
}

AstNodeTypeStruct *AstNodeExpressionLValueStructElement::GetStructType()
{
	AstNodeTypeStruct *struct_type = dynamic_cast<AstNodeTypeStruct *>(m_lvalue->GetType());
	if (!struct_type){
		throw FormatCompilerError(this, "LValue type is not a struct");
	}
	return struct_type;
}

void AstNodeExpressionLValueStructElement::Link(std::map<std::string, AstNode *> &node_map)
{
	m_lvalue->Link(node_map);
}

AstNodeExpressionLValue *AstNodeExpressionLValueStructElement::GetLValue()
{
	return m_lvalue;
}

std::string AstNodeExpressionLValueStructElement::GetIdentifier()
{
	return m_identifier;
}

void AstNodeExpressionLValueStructElement::ReplaceVariableReferences(std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements)
{
	m_lvalue->ReplaceVariableReferences(replacements);
}

uint64_t AstNodeExpressionLValueStructElement::GetStaticComplexityEstimate(ComputeSpec &spec)
{
	return m_lvalue->GetStaticComplexityEstimate(spec);
}

AstNodeExpression *AstNodeExpressionLValueStructElement::GetDynamicComplexityEstimate(ComputeSpec &spec)
{
	return m_lvalue->GetDynamicComplexityEstimate(spec);
}

std::string AstNodeExpressionLValueStructElement::GetDotName()
{
	return PrintString("Struct element access: \\\"%s\\\"", m_identifier.c_str());
}

std::list<AstNode **> AstNodeExpressionLValueStructElement::GetChildrenPointers()
{
	return {(AstNode**)&m_lvalue};
}



