/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "astnode.hpp"
#include "../codegen/llvm/codegen.hpp"
#include "../codegen/opencl/codegen.hpp"
#include "../codegen/pseudopc/codegen.hpp"

void AstNodeType::Link(std::map<std::string, AstNode *> &node_map)
{
}

AstNodeType *AstNodeType::GetCastType()
{
	return Copy();
}

AstNodeTypeVoid::AstNodeTypeVoid()
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeTypeVoid::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorTypeVoidLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorTypeVoidOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorTypeVoidPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeTypeVoid *AstNodeTypeVoid::Copy()
{
	return new AstNodeTypeVoid();
}

int AstNodeTypeVoid::Compare(AstNodeType *rhs, CompileContext &context)
{
	return dynamic_cast<AstNodeTypeVoid *>(rhs) ? 0 : -1;
}

std::string AstNodeTypeVoid::GetDotName()
{
	return "Void Type";
}

std::string AstNodeTypeVoid::GetName()
{
	return "void";
}

AstNodeTypeBool::AstNodeTypeBool()
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeTypeBool::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorTypeBoolLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorTypeBoolOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorTypeBoolPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

int AstNodeTypeBool::Compare(AstNodeType *rhs, CompileContext &context)
{
	if (dynamic_cast<AstNodeTypeVoid *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeBool *>(rhs)){
		return 0;
	}
	else {
		return 1;
	}
}

AstNodeTypeBool *AstNodeTypeBool::Copy()
{
	return new AstNodeTypeBool();
}

std::string AstNodeTypeBool::GetDotName()
{
	return "Boolean Type";
}

std::string AstNodeTypeBool::GetName()
{
	return "bool";
}

AstNodeTypeInt::AstNodeTypeInt(unsigned int width, bool strict, bool is_signed) :
	m_width(width), m_strict(strict), m_is_signed(is_signed)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeTypeInt::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorTypeIntLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorTypeIntOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorTypeIntPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeTypeInt *AstNodeTypeInt::Copy()
{
	return new AstNodeTypeInt(m_width, m_strict, m_is_signed);
}

int AstNodeTypeInt::Compare(AstNodeType *rhs, CompileContext &context)
{
	if (dynamic_cast<AstNodeTypeVoid *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeBool *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeInt *>(rhs)){
		AstNodeTypeInt *type = dynamic_cast<AstNodeTypeInt *>(rhs);
		if (GetWidth(context) < type->GetWidth(context)){
			return -1;
		}
		else if (GetWidth(context) > type->GetWidth(context)){
			return 1;
		}
		else {
			return 0;
		}
	}
	else {
		return 1;
	}
}

bool AstNodeTypeInt::GetIsStrict()
{
	return m_strict;
}

void AstNodeTypeInt::SetIsStrict(bool strict)
{
	m_strict = strict;
}

unsigned int AstNodeTypeInt::GetWidth()
{
	return m_width;
}

unsigned int AstNodeTypeInt::GetWidth(CompileContext &context)
{
	return m_width ? m_width : context.GetComputeSpec().GetCpu().GetSizeT();
}

bool AstNodeTypeInt::GetSigned()
{
	return m_is_signed;
}

std::string AstNodeTypeInt::GetDotName()
{
	char b[1024];
	snprintf(b, 1024, "%s%sInteger Type: %u bit",
		m_strict ? "Strict " : "", m_is_signed ? "Signed " : "", m_width);
	return b;
}

std::string AstNodeTypeInt::GetName()
{
	char b[1024];
	snprintf(b, 1024, "int %u", m_width);
	return b;
}

AstNodeTypeFloat::AstNodeTypeFloat(unsigned int width, bool strict) :
	m_width(width), m_strict(strict)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeTypeFloat::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorTypeFloatLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorTypeFloatOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorTypeFloatPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeTypeFloat *AstNodeTypeFloat::Copy()
{
	return new AstNodeTypeFloat(m_width, m_strict);
}

int AstNodeTypeFloat::Compare(AstNodeType *rhs, CompileContext &context)
{
	if (dynamic_cast<AstNodeTypeVoid *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeBool *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeInt *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeFloat *>(rhs)){
		AstNodeTypeFloat *type = dynamic_cast<AstNodeTypeFloat *>(rhs);
		if (GetWidth() < type->GetWidth()){
			return -1;
		}
		else if (GetWidth() > type->GetWidth()){
			return 1;
		}
		else {
			return 0;
		}
	}
	else {
		return 1;
	}
}

bool AstNodeTypeFloat::GetIsStrict()
{
	return m_strict;
}

void AstNodeTypeFloat::SetIsStrict(bool strict)
{
	m_strict = strict;
}

unsigned int AstNodeTypeFloat::GetWidth()
{
	return m_width;
}

std::string AstNodeTypeFloat::GetDotName()
{
	char b[1024];
	snprintf(b, 1024, "%sFloat Type: %u bit", m_strict ? "Strict " : "", m_width);
	return b;
}

std::string AstNodeTypeFloat::GetName()
{
	char b[1024];
	snprintf(b, 1024, "float %u", m_width);
	return b;
}


AstNodeTypeVector::AstNodeTypeVector(AstNodeType *type) :
	m_type(type)
{
}

AstNodeTypeVector::~AstNodeTypeVector()
{
	delete m_type;
}

AstNodeType *AstNodeTypeVector::GetType()
{
	return m_type;
}

std::list<AstNode **> AstNodeTypeVector::GetChildrenPointers()
{
	return {(AstNode**)&m_type};
}

void AstNodeTypeVector::Link(std::map<std::string, AstNode *> &node_map)
{
	m_type->Link(node_map);
}

AstNodeTypePointer::AstNodeTypePointer(AstNodeType *type) :
	AstNodeTypeVector(type)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeTypePointer::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorTypePointerLlvm(this)); break;
		case CODEGEN_OPENCL: AddCodeGenerator(new CodeGeneratorTypePointerOpencl(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorTypePointerPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

int AstNodeTypePointer::Compare(AstNodeType *rhs, CompileContext &context)
{
	if (dynamic_cast<AstNodeTypeVoid *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeBool *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeInt *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeFloat *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypePointer *>(rhs)){
		return GetType()->Compare(dynamic_cast<AstNodeTypePointer *>(rhs)->GetType(), context);
	}
	else {
		return 1;
	}
}

AstNodeTypePointer *AstNodeTypePointer::Copy()
{
	return new AstNodeTypePointer(GetType()->Copy());
}

std::string AstNodeTypePointer::GetName()
{
	char b[1024];
	snprintf(b, 1024, "%s pointer", GetType()->GetName().c_str());
	return b;
}

std::string AstNodeTypePointer::GetDotName()
{
	return "Pointer";
}


AstNodeTypeStaticVector::AstNodeTypeStaticVector(AstNodeType *type, unsigned int count) :
	AstNodeTypeVector(type),
	m_count(count)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

void AstNodeTypeStaticVector::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorTypeStaticVectorLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorTypeStaticVectorPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeTypeStaticVector *AstNodeTypeStaticVector::Copy()
{
	return new AstNodeTypeStaticVector(GetType()->Copy(), m_count);
}

int AstNodeTypeStaticVector::Compare(AstNodeType *rhs, CompileContext &context)
{
	if (dynamic_cast<AstNodeTypeVoid *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeBool *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeInt *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeFloat *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypePointer *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeStaticVector *>(rhs)){
		return GetType()->Compare(dynamic_cast<AstNodeTypeStaticVector *>(rhs)->GetType(), context);
	}
	else {
		return 1;
	}
}

std::string AstNodeTypeStaticVector::GetName()
{
	char b[1024];
	snprintf(b, 1024, "%s [%u]", GetType()->GetName().c_str(), m_count);
	return b;
}

unsigned int AstNodeTypeStaticVector::GetCount()
{
	return m_count;
}

std::string AstNodeTypeStaticVector::GetDotName()
{
	char b[1024];
	snprintf(b, 1024, "Static Sized Vector: %u elements", m_count);
	return b;
}

AstNodeTypeFunctionPointer::AstNodeTypeFunctionPointer(AstNodeType *return_type, std::list<AstNodeType *> arguments, bool ccall) :
	m_return_type(return_type), m_arguments(arguments), m_ccall(ccall)
{
	SetCodeGenerator(CODEGEN_LLVM);
}

AstNodeTypeFunctionPointer::AstNodeTypeFunctionPointer(AstNodeType *return_type, std::list<AstNodeStatementVariableDeclaration *> arguments, bool ccall) :
	m_return_type(return_type->Copy()), m_ccall(ccall)
{
	SetCodeGenerator(CODEGEN_LLVM);
	for (auto argument: arguments){
		m_arguments.push_back(argument->GetType()->Copy());
	}
}

void AstNodeTypeFunctionPointer::SetCodeGenerator(CodeGeneratorType type)
{
	switch (type)
	{
		case CODEGEN_LLVM: AddCodeGenerator(new CodeGeneratorTypeFunctionPointerLlvm(this)); break;
		case CODEGEN_PSEUDOPC: AddCodeGenerator(new CodeGeneratorTypeFunctionPointerPseudopc(this)); break;
		default: throw CodeGenerationException(type, this);
	}
}

AstNodeTypeFunctionPointer::~AstNodeTypeFunctionPointer()
{
	delete m_return_type;
	for (auto type: m_arguments){
		delete type;
	}
}

AstNodeTypeFunctionPointer *AstNodeTypeFunctionPointer::Copy()
{
	std::list<AstNodeType *> args;
	for (auto arg: m_arguments){
		args.push_back(arg->Copy());
	}
	
	return new AstNodeTypeFunctionPointer(m_return_type, args, m_ccall);
}

void AstNodeTypeFunctionPointer::Link(std::map<std::string, AstNode *> &node_map)
{
	m_return_type->Link(node_map);
	for (auto argument: m_arguments){
		argument->Link(node_map);
	}
}

int AstNodeTypeFunctionPointer::Compare(AstNodeType *rhs, CompileContext &context)
{
	if (dynamic_cast<AstNodeTypeVoid *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeBool *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeInt *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeFloat *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypePointer *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeStaticVector *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeDynamicVector *>(rhs)){
		return -1;
	}
	else if (dynamic_cast<AstNodeTypeFunctionPointer *>(rhs)){
		return 0;
	}
	else {
		return 1;
	}
}

AstNodeType *AstNodeTypeFunctionPointer::GetReturnType()
{
	return m_return_type;
}

std::list<AstNodeType *> AstNodeTypeFunctionPointer::GetArguments()
{
	return m_arguments;
}

std::string AstNodeTypeFunctionPointer::GetName()
{
	std::string args;
	for (auto arg: m_arguments){
		if (args.size()){
			args += ", ";
		}
		args += arg->GetName();
	}
	
	return PrintString("%s (%s)", m_return_type->GetName().c_str(), args.c_str());
}

std::string AstNodeTypeFunctionPointer::GetDotName()
{
	return "Function Pointer";
}

std::list<AstNode **> AstNodeTypeFunctionPointer::GetChildrenPointers()
{
	std::list<AstNode **> l;
	l.push_back((AstNode**)&m_return_type);
	for (AstNodeType *&type: m_arguments){
		l.push_back((AstNode**)&type);
	}
	return l;
}

bool AstNodeTypeFunctionPointer::GetCCall()
{
	return m_ccall;
}
