/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef AST_TYPE_HPP
#define AST_TYPE_HPP

#include "../compilecontext.hpp"

class AstNodeType : public AstNode
{
	public:
		virtual AstNodeType *Copy() = 0;
		virtual std::string GetName() = 0;
		
		/*
		 * -1 if *this <  *rhs
		 *  0 if *this == *ths
		 *  1 if *this >  *rhs
		 */
		virtual int Compare(AstNodeType *rhs, CompileContext &context) = 0;
		virtual AstNodeType *GetCastType();
		
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		
	private:
};

class AstNodeTypeVoid : public AstNodeType
{
	public:
		AstNodeTypeVoid();
		virtual AstNodeTypeVoid *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
		virtual int Compare(AstNodeType *rhs, CompileContext &context);
		
		virtual std::string GetName();
		
	private:
		virtual std::string GetDotName();
};

class AstNodeTypeBool : public AstNodeType
{
	public:
		AstNodeTypeBool();
		virtual AstNodeTypeBool *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
		virtual int Compare(AstNodeType *rhs, CompileContext &context);
		
		virtual std::string GetName();
		
	private:
		virtual std::string GetDotName();
};

class AstNodeTypeInt : public AstNodeType
{
	public:
		AstNodeTypeInt(unsigned int width, bool strict, bool is_signed);
		virtual AstNodeTypeInt *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
		virtual int Compare(AstNodeType *rhs, CompileContext &context);
		
		virtual std::string GetName();
		unsigned int GetWidth(); //returns 0 if it's %t.size
		unsigned int GetWidth(CompileContext &context); //returns the size of %t.size if it's of type %t.size
		bool GetSigned();
		
		bool GetIsStrict();
		void SetIsStrict(bool strict);
		
	private:
		virtual std::string GetDotName();
		
		unsigned int m_width;
		bool m_strict;
		bool m_is_signed;
};

class AstNodeTypeFloat : public AstNodeType
{
	public:
		AstNodeTypeFloat(unsigned int width, bool strict);
		virtual AstNodeTypeFloat *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
		virtual int Compare(AstNodeType *rhs, CompileContext &context);
		
		virtual std::string GetName();
		unsigned int GetWidth();
		bool GetIsStrict();
		void SetIsStrict(bool strict);
		
	private:
		virtual std::string GetDotName();
		
		unsigned int m_width;
		bool m_strict;
};

class AstNodeTypeVector : public AstNodeType
{
	public:
		AstNodeTypeVector(AstNodeType *type);
		virtual ~AstNodeTypeVector();
		AstNodeType *GetType();
		
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		
	private:
		virtual std::list<AstNode **> GetChildrenPointers();
		
		AstNodeType *m_type;
};

class AstNodeTypePointer : public AstNodeTypeVector
{
	public:
		AstNodeTypePointer(AstNodeType *type);
		virtual AstNodeTypePointer *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
		virtual int Compare(AstNodeType *rhs, CompileContext &context);
		
		virtual std::string GetName();
		
	private:
		virtual std::string GetDotName();
};


class AstNodeTypeStaticVector : public AstNodeTypeVector
{
	public:
		AstNodeTypeStaticVector(AstNodeType *type, unsigned int count);
		virtual AstNodeTypeStaticVector *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
		virtual int Compare(AstNodeType *rhs, CompileContext &context);
		
		virtual std::string GetName();
		unsigned int GetCount();
		
	private:
		virtual std::string GetDotName();
		
		unsigned int m_count;
};

class AstNodeTypeFunctionPointer : public AstNodeType
{
	public:
		AstNodeTypeFunctionPointer(AstNodeType *return_type, std::list<AstNodeType *> arguments, bool ccall);
		AstNodeTypeFunctionPointer(AstNodeType *return_type, std::list<AstNodeStatementVariableDeclaration *> arguments, bool ccall);
		virtual ~AstNodeTypeFunctionPointer();
		virtual AstNodeTypeFunctionPointer *Copy();
		virtual void SetCodeGenerator(CodeGeneratorType type);
		
		virtual int Compare(AstNodeType *rhs, CompileContext &context);
		virtual void Link(std::map<std::string, AstNode *> &node_map);
		virtual std::string GetName();
		
		AstNodeType *GetReturnType();
		std::list<AstNodeType *> GetArguments();
		bool GetCCall();
		
	private:
		virtual std::string GetDotName();
		virtual std::list<AstNode **> GetChildrenPointers();
		
		AstNodeType *m_return_type;
		std::list<AstNodeType *> m_arguments;
		bool m_ccall;
};

#endif
