/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "cl.hpp"
#include "../util.hpp"

namespace OpenCL
{
	static std::string SampleKernelSource =
		"kernel void Zero(global float *v){unsigned int i = get_global_id(0); v[i] = 0;}\n";
	
	static void ClWrapperError(int e)
	{
		if (CL_SUCCESS != e){
			throw FormatError("e != CL_SUCCESS in CL wrapper: %i", e);
		}
	}
	
	static std::string SetToString(std::set<std::string> strings)
	{
		std::string result;
		for (std::string string: strings){
			result += PrintString("%s\n", string.c_str());
		}
		return result;
	}
	
	static Kernel GetSampleKernel(Device device, std::string name)
	{
		Context context(device);
		Program program(context, SampleKernelSource);
		return Kernel(program, name);
	}
	
	std::set<std::string> StringToSet(std::string string, char sep)
	{
		std::set<std::string> strings;
		std::string current;
		
		for (char c: string){
			if (' ' == c){
				if (current.size()){
					strings.insert(current);
				}
				current = "";
			}
			else {
				current += c;
			}
		}
		if (current.size()){
			strings.insert(current);
		}
		
		return strings;
	}
	
	Device::Device(cl_device_id id) :
		m_id(id)
	{
	}
	
	bool Device::operator< (const Device &rhs) const
	{
		return m_id < rhs.m_id;
	}
	
	template <typename T> T Device::GetDeviceInfo(cl_platform_info param_name)
	{
		T param_value;
		GetDeviceInfo(param_name, sizeof(T), &param_value);
		return param_value;
	}
	
	template <> std::string Device::GetDeviceInfo<std::string>(cl_platform_info param_name)
	{
		char param_value[65536];
		GetDeviceInfo(param_name, 65536, param_value);
		return param_value;
	}
	
	cl_device_id Device::GetDeviceId()
	{
		return m_id;
	}
	
	cl_device_type Device::GetType()
	{
		return GetDeviceInfo<cl_device_type>(CL_DEVICE_TYPE);
	}
	
	uint64_t Device::GetVendorId()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_VENDOR_ID);
	}
	
	uint64_t Device::GetMaxComputeUnits()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_MAX_COMPUTE_UNITS);
	}
	
	uint64_t Device::GetMaxWorkItemDimensions()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS);
	}
	
	std::vector<uint64_t> Device::GetMaxWorkItemSizes()
	{
		unsigned int dimensions = GetMaxWorkItemDimensions();
		
		size_t *sizes = new size_t[dimensions];
		GetDeviceInfo(CL_DEVICE_MAX_WORK_ITEM_SIZES, dimensions * sizeof(size_t), sizes);
		
		std::vector<uint64_t> r;
		for (unsigned int i = 0; i < dimensions; i++){
			r.push_back(sizes[i]);
		}
		
		delete []sizes;
		return r;
	}
	
	uint64_t Device::GetMaxWorkGroupSize()
	{
		return GetDeviceInfo<size_t>(CL_DEVICE_MAX_WORK_GROUP_SIZE);
	}
	
	uint64_t Device::GetPreferredVectorWidthChar()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_PREFERRED_VECTOR_WIDTH_CHAR);
	}
	
	uint64_t Device::GetPreferredVectorWidthShort()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_PREFERRED_VECTOR_WIDTH_SHORT);
	}
	
	uint64_t Device::GetPreferredVectorWidthInt()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_PREFERRED_VECTOR_WIDTH_INT);
	}
	
	uint64_t Device::GetPreferredVectorWidthLong()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_PREFERRED_VECTOR_WIDTH_LONG);
	}
	
	uint64_t Device::GetPreferredVectorWidthFloat()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_PREFERRED_VECTOR_WIDTH_FLOAT);
	}
	
	uint64_t Device::GetPreferredVectorWidthDouble()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_PREFERRED_VECTOR_WIDTH_DOUBLE);
	}
	
	uint64_t Device::GetPreferredVectorWidthHalf()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_PREFERRED_VECTOR_WIDTH_HALF);
	}
	
	uint64_t Device::GetNativeVectorWidthChar()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_NATIVE_VECTOR_WIDTH_CHAR);
	}
	
	uint64_t Device::GetNativeVectorWidthShort()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_NATIVE_VECTOR_WIDTH_SHORT);
	}
	
	uint64_t Device::GetNativeVectorWidthInt()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_NATIVE_VECTOR_WIDTH_INT);
	}
	
	uint64_t Device::GetNativeVectorWidthLong()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_NATIVE_VECTOR_WIDTH_LONG);
	}
	
	uint64_t Device::GetNativeVectorWidthFloat()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_NATIVE_VECTOR_WIDTH_FLOAT);
	}
	
	uint64_t Device::GetNativeVectorWidthDouble()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_NATIVE_VECTOR_WIDTH_DOUBLE);
	}
	
	uint64_t Device::GetNativeVectorWidthHalf()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_NATIVE_VECTOR_WIDTH_HALF);
	}
	
	uint64_t Device::GetMaxClockFrequency()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_MAX_CLOCK_FREQUENCY);
	}
	
	uint64_t Device::GetAddressBits()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_ADDRESS_BITS);
	}
	
	uint64_t Device::GetMaxMemAllocSize()
	{
		return GetDeviceInfo<cl_ulong>(CL_DEVICE_MAX_MEM_ALLOC_SIZE);
	}
	
	bool Device::GetImageSupport()
	{
		return GetDeviceInfo<cl_bool>(CL_DEVICE_IMAGE_SUPPORT);
	}
	
	uint64_t Device::GetMaxReadImageArgs()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_MAX_READ_IMAGE_ARGS);
	}
	
	uint64_t Device::GetMaxWriteImageArgs()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_MAX_WRITE_IMAGE_ARGS);
	}
	
	uint64_t Device::GetImage2DMaxWidth()
	{
		return GetDeviceInfo<size_t>(CL_DEVICE_IMAGE2D_MAX_WIDTH);
	}
	
	uint64_t Device::GetImage2DMaxHeight()
	{
		return GetDeviceInfo<size_t>(CL_DEVICE_IMAGE2D_MAX_HEIGHT);
	}
	
	uint64_t Device::GetImage3DMaxWidth()
	{
		return GetDeviceInfo<size_t>(CL_DEVICE_IMAGE3D_MAX_WIDTH);
	}
	
	uint64_t Device::GetImage3DMaxHeight()
	{
		return GetDeviceInfo<size_t>(CL_DEVICE_IMAGE3D_MAX_HEIGHT);
	}
	
	uint64_t Device::GetImage3DMaxDepth()
	{
		return GetDeviceInfo<size_t>(CL_DEVICE_IMAGE3D_MAX_DEPTH);
	}
	
	uint64_t Device::GetMaxSamples()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_MAX_SAMPLERS);
	}
	
	uint64_t Device::GetMaxParameterSize()
	{
		return GetDeviceInfo<size_t>(CL_DEVICE_MAX_PARAMETER_SIZE);
	}
	
	uint64_t Device::GetMemBaseAddrAlign()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_MEM_BASE_ADDR_ALIGN);
	}
	
	cl_device_fp_config Device::GetSingleFpConfig()
	{
		return GetDeviceInfo<cl_device_fp_config>(CL_DEVICE_SINGLE_FP_CONFIG);
	}
	
	cl_device_mem_cache_type Device::GetGlobalMemCacheType()
	{
		return GetDeviceInfo<cl_device_mem_cache_type>(CL_DEVICE_GLOBAL_MEM_CACHE_TYPE);
	}
	
	uint64_t Device::GetGlobalMemCacheLineSize()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_GLOBAL_MEM_CACHELINE_SIZE);
	}
	
	uint64_t Device::GetGlobalMemCacheSize()
	{
		return GetDeviceInfo<cl_ulong>(CL_DEVICE_GLOBAL_MEM_CACHE_SIZE);
	}
	
	uint64_t Device::GetGlobalMemSize()
	{
		return GetDeviceInfo<cl_ulong>(CL_DEVICE_GLOBAL_MEM_SIZE);
	}
	
	uint64_t Device::GetMaxConstantBufferSize()
	{
		return GetDeviceInfo<cl_ulong>(CL_DEVICE_MAX_CONSTANT_BUFFER_SIZE);
	}
	
	uint64_t Device::GetMaxConstantArgs()
	{
		return GetDeviceInfo<cl_uint>(CL_DEVICE_MAX_CONSTANT_ARGS);
	}
	
	cl_device_local_mem_type Device::GetLocalMemType()
	{
		return GetDeviceInfo<cl_device_local_mem_type>(CL_DEVICE_LOCAL_MEM_TYPE);
	}
	
	uint64_t Device::GetLocalMemSize()
	{
		return GetDeviceInfo<cl_ulong>(CL_DEVICE_LOCAL_MEM_SIZE);
	}
	
	bool Device::GetErrorCorrectionSupport()
	{
		return GetDeviceInfo<cl_bool>(CL_DEVICE_ERROR_CORRECTION_SUPPORT);
	}
	
	bool Device::GetHostUnifiedMemory()
	{
		return GetDeviceInfo<cl_bool>(CL_DEVICE_HOST_UNIFIED_MEMORY);
	}
	
	uint64_t Device::GetProfilingTimerResolution()
	{
		return GetDeviceInfo<size_t>(CL_DEVICE_PROFILING_TIMER_RESOLUTION);
	}
	
	bool Device::GetEndianLittle()
	{
		return GetDeviceInfo<cl_bool>(CL_DEVICE_ENDIAN_LITTLE);
	}
	
	bool Device::GetAvailable()
	{
		return GetDeviceInfo<cl_bool>(CL_DEVICE_AVAILABLE);
	}
	
	bool Device::GetCompilerAvailable()
	{
		return GetDeviceInfo<cl_bool>(CL_DEVICE_COMPILER_AVAILABLE);
	}
	
	cl_device_exec_capabilities Device::GetExecutionCapabilities()
	{
		return GetDeviceInfo<cl_device_exec_capabilities>(CL_DEVICE_EXECUTION_CAPABILITIES);
	}
	
	cl_command_queue_properties Device::GetQueueProperties()
	{
		return GetDeviceInfo<cl_command_queue_properties>(CL_DEVICE_QUEUE_PROPERTIES);
	}
	
	std::string Device::GetName()
	{
		return GetDeviceInfo<std::string>(CL_DEVICE_NAME);
	}
	
	std::string Device::GetVendor()
	{
		return GetDeviceInfo<std::string>(CL_DEVICE_VENDOR);
	}
	
	std::string Device::GetProfile()
	{
		return GetDeviceInfo<std::string>(CL_DEVICE_PROFILE);
	}
	
	std::string Device::GetVersion()
	{
		return GetDeviceInfo<std::string>(CL_DEVICE_VERSION);
	}
	
	std::string Device::GetOpenclCVersion()
	{
		return GetDeviceInfo<std::string>(CL_DEVICE_OPENCL_C_VERSION);
	}
	
	std::set<std::string> Device::GetExtensions()
	{
		return StringToSet(GetDeviceInfo<std::string>(CL_DEVICE_EXTENSIONS), ' ');
	}
	
	uint64_t Device::GetKernelWorkGroupSize()
	{
		return GetSampleKernel(*this, "Zero").GetWorkGroupSize();
	}
	
	uint64_t Device::GetKernelPreferredWorkGroupSizeMultiple()
	{
		return GetSampleKernel(*this, "Zero").GetPreferredWorkGroupSizeMultiple();
	}
	
	size_t Device::GetDeviceInfo(cl_device_info param_name, size_t param_value_size, void *param_value)
	{
		size_t param_value_size_ret = 0;
		ClWrapperError(clGetDeviceInfo(m_id, param_name, param_value_size, param_value, &param_value_size_ret));
		return param_value_size_ret;
	}
	
	std::string Device::TypeToString(cl_device_type type)
	{
		std::string result;
		if (type & CL_DEVICE_TYPE_CPU){
			result = "CPU";
		}
		if (type & CL_DEVICE_TYPE_GPU){
			result += result.size() ? " | " : "";
			result += "GPU";
		}
		if (type & CL_DEVICE_TYPE_ACCELERATOR){
			result += result.size() ? " | " : "";
			result += "Accelerator";
		}
		if (type & CL_DEVICE_TYPE_DEFAULT){
			result += result.size() ? " | " : "";
			result += "Default";
		}
		/*if (type & CL_DEVICE_TYPE_CUSTOM){
			result += result.size() ? " | " : "";
			result += "Custom";
		}*/
		return result;
	}
	
	std::string Device::IntVectorToString(std::vector<uint64_t> vector)
	{
		std::string result;
		for (uint64_t i: vector){
			result += PrintString("%s%" PRIu64, result.size() ? ", " : "", i);
		}
		return result;
	}
	
	std::string Device::FpConfigToString(cl_device_fp_config config)
	{
		std::string result;
		if (config & CL_FP_DENORM){
			result = "Denorm";
		}
		if (config & CL_FP_INF_NAN){
			result += result.size() ? " | " : "";
			result += "Infinity and NaN";
		}
		if (config & CL_FP_ROUND_TO_NEAREST){
			result += result.size() ? " | " : "";
			result += "Round to Nearest";
		}
		if (config & CL_FP_ROUND_TO_ZERO){
			result += result.size() ? " | " : "";
			result += "Round to Zero";
		}
		if (config & CL_FP_ROUND_TO_INF){
			result += result.size() ? " | " : "";
			result += "Round to Infinity";
		}
		/*if (config & CP_FP_FMA){
			result += result.size() ? " | " : "";
			result += "Fuse Multiply Add";
		}*/
		if (config & CL_FP_SOFT_FLOAT){
			result += result.size() ? " | " : "";
			result += "Software Float";
		}
		/*if (config & CL_FP_CORRECTLY_ROUNDED_DIVIDE_SQRT){
			result += result.size() ? " | " : "";
			result += "Correctly Rounded Divide Square Root";
		}*/
		return result;
	}
	
	std::string Device::MemCacheTypeToString(cl_device_mem_cache_type type)
	{
		switch (type)
		{
			case CL_NONE: return "None";
			case CL_READ_ONLY_CACHE: return "Read Only Cache";
			case CL_READ_WRITE_CACHE: return "Read/Write Cache";
			default: return "Unknown";
		}
	}
	
	std::string Device::LocalMemoryTypeToString(cl_device_local_mem_type type)
	{
		switch (type)
		{
			case CL_NONE: return "None";
			case CL_LOCAL: return "Dedicated Local Memory";
			case CL_GLOBAL: return "Global Memory";
			default: return "Unknown";
		}
	}
	
	std::string Device::ExecutionCapabilitiesToString(cl_device_exec_capabilities capabilities)
	{
		std::string result;
		if (capabilities & CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE){
			result = "Out of Order Execution Mode";
		}
		if (capabilities & CL_QUEUE_PROFILING_ENABLE){
			result += result.size() ? " | " : "";
			result += "Profiling";
		}
		return result;
	}
	
	std::string Device::QueuePropertiesToString(cl_command_queue_properties properties)
	{
		std::string result;
		if (properties & CL_EXEC_KERNEL){
			result = "Kernel";
		}
		if (properties & CL_EXEC_NATIVE_KERNEL){
			result += result.size() ? " | " : "";
			result += "Native Kernel";
		}
		return result;
	}
	
	std::string Device::GetHumanReadableString()
	{
		return PrintString(
			"Device:\n"
				"\tType: %s\n"
				"\tVendor ID: %" PRIu64 "\n"
				
				"\tMaximum Compute Units: %" PRIu64 "\n"
				"\tMaximum Work Item Dimensions: %" PRIu64 "\n"
				"\tMaximum Work Item Sizes: %s\n"
				"\tMaximum Work Group Size: %" PRIu64 "\n"
				
				"\tPreferred Char Vector Width: %" PRIu64 "\n"
				"\tPreferred Short Vector Width: %" PRIu64 "\n"
				"\tPreferred Int Vector Width: %" PRIu64 "\n"
				"\tPreferred Long Vector Width: %" PRIu64 "\n"
				"\tPreferred Float Vector Width: %" PRIu64 "\n"
				"\tPreferred Double Vector Width: %" PRIu64 "\n"
				"\tPreferred Half Vector Width: %" PRIu64 "\n"
				
				"\tNative Char Vector Width: %" PRIu64 "\n"
				"\tNative Short Vector Width: %" PRIu64 "\n"
				"\tNative Int Vector Width: %" PRIu64 "\n"
				"\tNative Long Vector Width: %" PRIu64 "\n"
				"\tNative Float Vector Width: %" PRIu64 "\n"
				"\tNative Double Vector Width: %" PRIu64 "\n"
				"\tNative Half Vector Width: %" PRIu64 "\n"
				
				"\tMaximum Clock Frequency: %" PRIu64 "\n"
				"\tAddress Bits: %" PRIu64 "\n"
				"\tMaximum Memory Allocation Size: %" PRIu64 "\n"
				
				"\tImage Support: %s\n"
				"\tMaximum Read Image Arguments: %" PRIu64 "\n"
				"\tMaximum Write Image Arguments: %" PRIu64 "\n"
				"\tImage 2D Maximum Width: %" PRIu64 "\n"
				"\tImage 2D Maximum Height: %" PRIu64 "\n"
				"\tImage 3D Maximum Width: %" PRIu64 "\n"
				"\tImage 3D Maximum Height: %" PRIu64 "\n"
				"\tImage 3D Maximum Depth: %" PRIu64 "\n"
				"\tGet Samples: %" PRIu64 "\n"
				
				"\tMaximum Parameter Size: %" PRIu64 "\n"
				"\tMemory Base Address Align: %" PRIu64 "\n"
				
				"\tSingle Floating Point Configuration: %s\n"
				
				"\tGlobal Memory Cache Type: %s\n"
				"\tGlobal Memory Cache Line Size: %" PRIu64 "\n"
				"\tGlobal Memory Cache Size: %" PRIu64 "\n"
				"\tGlobal Memory Size: %" PRIu64 "\n"
				
				"\tMaximum Constant Buffer Size: %" PRIu64 "\n"
				"\tMaximum Constant Arguments: %" PRIu64 "\n"
				
				"\tLocal Memory Type: %s\n"
				"\tLocal Memory Size: %" PRIu64 "\n"
				
				"\tCorrection Support: %s\n"
				"\tHost Unified Memory: %s\n"
				"\tProfiling Timer Resolution: %" PRIu64 "\n"
				"\tEndianness: %s\n"
				"\tAvailable: %s\n"
				"\tCompiler Available: %s\n"
				
				"\tExecution Capabilities: %s\n"
				"\tQueue Properties: %s\n"
				
				"\tName: %s\n"
				"\tVendor: %s\n"
				"\tProfile: %s\n"
				"\tVersion: %s\n"
				"\tOpenCL C Version: %s\n"
				"\tExtensions:\n%s\n"
				
				"\tKernel Work Group Size: %" PRIu64 "\n"
				"\tKernel Preferred Work Group Size Multiple: %" PRIu64 "\n",
				
				
				TypeToString(GetType()).c_str(),
				GetVendorId(),
				
				GetMaxComputeUnits(),
				GetMaxWorkItemDimensions(),
				IntVectorToString(GetMaxWorkItemSizes()).c_str(),
				GetMaxWorkGroupSize(),
				
				GetPreferredVectorWidthChar(),
				GetPreferredVectorWidthShort(),
				GetPreferredVectorWidthInt(),
				GetPreferredVectorWidthLong(),
				GetPreferredVectorWidthFloat(),
				GetPreferredVectorWidthDouble(),
				GetPreferredVectorWidthHalf(),
				
				GetNativeVectorWidthChar(),
				GetNativeVectorWidthShort(),
				GetNativeVectorWidthInt(),
				GetNativeVectorWidthLong(),
				GetNativeVectorWidthFloat(),
				GetNativeVectorWidthDouble(),
				GetNativeVectorWidthHalf(),
				
				GetMaxClockFrequency(),
				GetAddressBits(),
				GetMaxMemAllocSize(),
				
				GetImageSupport() ? "Yes" : "No",
				GetMaxReadImageArgs(),
				GetMaxWriteImageArgs(),
				GetImage2DMaxWidth(),
				GetImage2DMaxHeight(),
				GetImage3DMaxWidth(),
				GetImage3DMaxHeight(),
				GetImage3DMaxDepth(),
				GetMaxSamples(),
				
				GetMaxParameterSize(),
				GetMemBaseAddrAlign(),
				
				FpConfigToString(GetSingleFpConfig()).c_str(),
				
				MemCacheTypeToString(GetGlobalMemCacheType()).c_str(),
				GetGlobalMemCacheLineSize(),
				GetGlobalMemCacheSize(),
				GetGlobalMemSize(),
				
				GetMaxConstantBufferSize(),
				GetMaxConstantArgs(),
				
				LocalMemoryTypeToString(GetLocalMemType()).c_str(),
				GetLocalMemSize(),
				
				GetErrorCorrectionSupport() ? "Yes" : "No",
				GetHostUnifiedMemory() ? "Yes" : "No",
				GetProfilingTimerResolution(),
				GetEndianLittle() ? "Little" : "Big",
				GetAvailable() ? "Yes" : "No",
				GetCompilerAvailable() ? "Yes" : "No",
				
				ExecutionCapabilitiesToString(GetExecutionCapabilities()).c_str(),
				QueuePropertiesToString(GetQueueProperties()).c_str(),
				
				GetName().c_str(),
				GetVendor().c_str(),
				GetProfile().c_str(),
				GetVersion().c_str(),
				GetOpenclCVersion().c_str(),
				Indent(SetToString(GetExtensions()), 2).c_str(),
				
				GetKernelWorkGroupSize(),
				GetKernelPreferredWorkGroupSizeMultiple()
		);
		
	}
	
	
	
	
	
	Platform::Platform(cl_platform_id id) :
		m_id(id)
	{
		cl_uint num_devices = 0;
		ClWrapperError(clGetDeviceIDs(m_id, CL_DEVICE_TYPE_ALL, 0, 0, &num_devices));
		
		cl_device_id *devices = new cl_device_id[num_devices];
		ClWrapperError(clGetDeviceIDs(m_id, CL_DEVICE_TYPE_ALL, num_devices, devices, 0));
		
		for (unsigned int i = 0; i < num_devices; i++){
			m_devices.insert(Device(devices[i]));
		}
		
		delete []devices;
	}
	
	bool Platform::operator< (const Platform &rhs) const
	{
		return m_id < rhs.m_id;
	}
	
	std::set<Platform> Platform::GetPlatforms()
	{
		cl_uint num_platforms = 0;
		ClWrapperError(clGetPlatformIDs(0, 0, &num_platforms));
		
		cl_platform_id *platforms = new cl_platform_id[num_platforms];
		ClWrapperError(clGetPlatformIDs(num_platforms, platforms, 0));
		
		std::set<Platform> r;
		for (unsigned int i = 0; i < num_platforms; i++){
			r.insert(Platform(platforms[i]));
		}
		
		delete []platforms;
		return r;
	}
	
	std::set<Device> Platform::GetDevices()
	{
		return m_devices;
	}
	
	std::string Platform::GetProfile()
	{
		return GetPlatformInfoString(CL_PLATFORM_PROFILE);
	}
	
	std::string Platform::GetVersion()
	{
		return GetPlatformInfoString(CL_PLATFORM_VERSION);
	}
	
	std::string Platform::GetPlatformName()
	{
		return GetPlatformInfoString(CL_PLATFORM_NAME);
	}
	
	std::string Platform::GetVendor()
	{
		return GetPlatformInfoString(CL_PLATFORM_VENDOR);
	}
	
	std::set<std::string> Platform::GetExtensions()
	{
		return StringToSet(GetPlatformInfoString(CL_PLATFORM_EXTENSIONS), ' ');
	}
	
	std::string Platform::GetHumanReadableString()
	{
		std::string result = PrintString(
			"Platform\n"
				"\tProfile: %s\n"
				"\tVersion: %s\n"
				"\tPlatform Name: %s\n"
				"\tVendor: %s\n"
				"\tExtensions:\n%s\n"
				"\tDevices:\n",
			
			GetProfile().c_str(),
				GetVersion().c_str(),
				GetPlatformName().c_str(),
				GetVendor().c_str(),
				Indent(SetToString(GetExtensions()), 2).c_str()
		);
		
		std::set<Device> devices = GetDevices();
		for (Device device: devices){
			result += Indent(device.GetHumanReadableString(), 2);
		}
		
		return result;
	}
	
	size_t Platform::GetPlatformInfo(cl_platform_info param_name, size_t param_value_size, void *param_value)
	{
		size_t param_value_size_ret = 0;
		ClWrapperError(clGetPlatformInfo(m_id, param_name, param_value_size, param_value, &param_value_size_ret));
		return param_value_size_ret;
	}
	
	std::string Platform::GetPlatformInfoString(cl_platform_info param_name)
	{
		char result[65536];
		GetPlatformInfo(param_name, 65536, result);
		return result;
	}
	
	
	
	Context::Context(Device device) :
		m_device(device)
	{
		cl_platform_id platform_id;
		cl_device_id device_id = device.GetDeviceId();
		ClWrapperError(clGetDeviceInfo(device_id, CL_DEVICE_PLATFORM, sizeof(cl_platform_id), &platform_id, 0));
		
		cl_context_properties properties[] = {CL_CONTEXT_PLATFORM, (cl_context_properties)platform_id, 0};
		
		cl_int e = 0;
		m_context = clCreateContext(properties, 1, &device_id, 0, 0, &e);
		ClWrapperError(e);
	}
	
	Context::~Context()
	{
		ClWrapperError(clReleaseContext (m_context));
	}
	
	Context::Context(const Context &context) :
		m_context(context.m_context), m_device(context.m_device)
	{
		ClWrapperError(clRetainContext(m_context));
	}
	
	Context &Context::operator=(const Context &context)
	{
		if (this != &context){
			m_context = context.m_context;
			m_device = context.m_device;
			ClWrapperError(clRetainContext(m_context));
		}
		return *this;
	}
	
	cl_context Context::GetClContext()
	{
		return m_context;
	}
	
	Device Context::GetDevice()
	{
		return m_device;
	}
	
	
	
	Program::Program(Context context, std::string source) :
		m_context(context)
	{
		char source_c_str[1046876];
		snprintf(source_c_str, 1046876, "%s", source.c_str());
		const char *source_c_str_ptr = source_c_str;
		
		cl_int e = 0;
		m_program = clCreateProgramWithSource(m_context.GetClContext(), 1, &source_c_str_ptr, 0, &e);
		ClWrapperError(e);
		
		cl_device_id device_id = context.GetDevice().GetDeviceId();
		ClWrapperError(clBuildProgram(m_program, 1, &device_id, "", 0, 0));
	}
	
	Program::~Program()
	{
		clReleaseProgram(m_program);
	}
	
	Program::Program(const Program &program) :
		m_program(program.m_program), m_context(program.m_context)
	{
		ClWrapperError(clRetainProgram(m_program));
	}
	
	Program &Program::operator=(const Program &program)
	{
		if (this != &program){
			m_program = program.m_program;
			m_context = program.m_context;
			ClWrapperError(clRetainProgram(m_program));
		}
		return *this;
	}
	
	cl_program Program::GetClProgram()
	{
		return m_program;
	}
	
	Context Program::GetContext()
	{
		return m_context;
	}
	
	
	
	Kernel::Kernel(Program program, std::string name) :
		m_kernel(), m_program(program)
	{
		cl_int e = 0;
		m_kernel = clCreateKernel(program.GetClProgram(), name.c_str(), &e);
		ClWrapperError(e);
	}
	
	Kernel::~Kernel()
	{
		 ClWrapperError(clReleaseKernel(m_kernel));
	}
	
	Kernel::Kernel(const Kernel &kernel) :
		m_kernel(kernel.m_kernel), m_program(kernel.m_program)
	{
		ClWrapperError(clRetainKernel(m_kernel));
	}
	
	Kernel &Kernel::operator=(const Kernel &kernel)
	{
		if (this != &kernel){
			m_kernel = kernel.m_kernel;
			m_program = kernel.m_program;
			ClWrapperError(clRetainKernel(m_kernel));
		}
		return *this;
	}
	
	uint64_t Kernel::GetWorkGroupSize()
	{
		return GetKernelWorkGroupInfo<size_t>(CL_KERNEL_WORK_GROUP_SIZE);
	}
	
	uint64_t Kernel::GetPreferredWorkGroupSizeMultiple()
	{
		return GetKernelWorkGroupInfo<size_t>(CL_KERNEL_PREFERRED_WORK_GROUP_SIZE_MULTIPLE);
	}
	
	size_t Kernel::GetKernelWorkGroupInfo(cl_kernel_work_group_info param_name, size_t param_value_size, void *param_value)
	{
		size_t param_value_size_ret = 0;
		ClWrapperError(clGetKernelWorkGroupInfo(m_kernel, m_program.GetContext().GetDevice().GetDeviceId(), param_name, param_value_size, param_value, &param_value_size_ret));
		return param_value_size_ret;
	}
	
	template <typename T> T Kernel::GetKernelWorkGroupInfo(cl_kernel_work_group_info param_name)
	{
		T param_value;
		GetKernelWorkGroupInfo(param_name, sizeof(T), &param_value);
		return param_value;
	}
}


