/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../error.hpp"

#include <stdint.h>
#include <set>
#include <vector>

#include <CL/cl.h>

namespace OpenCL
{
	class Platform;
	
	class Device : public TrackedClass
	{
		public:
			Device(cl_device_id id);
			bool operator< (const Device &rhs) const;
			
			cl_device_id GetDeviceId();
			
			static std::string TypeToString(cl_device_type type);
			static std::string IntVectorToString(std::vector<uint64_t> vector);
			static std::string FpConfigToString(cl_device_fp_config config);
			static std::string MemCacheTypeToString(cl_device_mem_cache_type type);
			static std::string LocalMemoryTypeToString(cl_device_local_mem_type type);
			static std::string ExecutionCapabilitiesToString(cl_device_exec_capabilities capabilities);
			static std::string QueuePropertiesToString(cl_command_queue_properties properties);
			
			cl_device_type GetType();
			uint64_t GetVendorId();
			uint64_t GetMaxComputeUnits();
			uint64_t GetMaxWorkItemDimensions();
			std::vector<uint64_t> GetMaxWorkItemSizes();
			uint64_t GetMaxWorkGroupSize();
			uint64_t GetPreferredVectorWidthChar();
			uint64_t GetPreferredVectorWidthShort();
			uint64_t GetPreferredVectorWidthInt();
			uint64_t GetPreferredVectorWidthLong();
			uint64_t GetPreferredVectorWidthFloat();
			uint64_t GetPreferredVectorWidthDouble();
			uint64_t GetPreferredVectorWidthHalf();
			uint64_t GetNativeVectorWidthChar();
			uint64_t GetNativeVectorWidthShort();
			uint64_t GetNativeVectorWidthInt();
			uint64_t GetNativeVectorWidthLong();
			uint64_t GetNativeVectorWidthFloat();
			uint64_t GetNativeVectorWidthDouble();
			uint64_t GetNativeVectorWidthHalf();
			uint64_t GetMaxClockFrequency();
			uint64_t GetAddressBits();
			uint64_t GetMaxMemAllocSize();
			bool GetImageSupport();
			uint64_t GetMaxReadImageArgs();
			uint64_t GetMaxWriteImageArgs();
			uint64_t GetImage2DMaxWidth();
			uint64_t GetImage2DMaxHeight();
			uint64_t GetImage3DMaxWidth();
			uint64_t GetImage3DMaxHeight();
			uint64_t GetImage3DMaxDepth();
			//uint64_t GetImageMaxBufferSize();
			//uint64_t GetImageMaxArraySize();
			uint64_t GetMaxSamples();
			uint64_t GetMaxParameterSize();
			uint64_t GetMemBaseAddrAlign();
			cl_device_fp_config GetSingleFpConfig();
			//cl_device_fp_config GetDoubleFpConfig();
			cl_device_mem_cache_type GetGlobalMemCacheType();
			uint64_t GetGlobalMemCacheLineSize();
			uint64_t GetGlobalMemCacheSize();
			uint64_t GetGlobalMemSize();
			uint64_t GetMaxConstantBufferSize();
			uint64_t GetMaxConstantArgs();
			cl_device_local_mem_type GetLocalMemType();
			uint64_t GetLocalMemSize();
			bool GetErrorCorrectionSupport();
			bool GetHostUnifiedMemory();
			uint64_t GetProfilingTimerResolution();
			bool GetEndianLittle();
			bool GetAvailable();
			bool GetCompilerAvailable();
			//bool GetLinkerAvailable();
			cl_device_exec_capabilities GetExecutionCapabilities();
			cl_command_queue_properties GetQueueProperties();
			//std::set<std::string> GetBuiltInKernels();
			std::string GetName();
			//std::string GetDriverVersion();
			std::string GetVendor();
			std::string GetProfile();
			std::string GetVersion();
			std::string GetOpenclCVersion();
			std::set<std::string> GetExtensions();
			//uint64_t GetPrintfBufferSize();
			//bool GetPreferredInteropUserSync();
			//uint64_t GetPartitionMaxSubdivides();
			//std::set<cl_device_partition_property> GetPartitionProperties();
			//cl_device_affinity_domain GetPartitionAffinityDomain();
			//std::set<cl_device_partition_property> GetPartitionType();
			//uint64_t GetReferenceCount();
			
			uint64_t GetKernelWorkGroupSize();
			uint64_t GetKernelPreferredWorkGroupSizeMultiple();
			
			std::string GetHumanReadableString();
			
		private:
			size_t GetDeviceInfo(cl_device_info param_name, size_t param_value_size, void *param_value);
			template <typename T> T GetDeviceInfo(cl_platform_info param_name);
			
			cl_device_id m_id;
	};
	
	class Platform : public TrackedClass
	{
		public:
			Platform(cl_platform_id id);
			bool operator< (const Platform &rhs) const;
			static std::set<Platform> GetPlatforms();
			std::set<Device> GetDevices();
			
			
			std::string GetProfile();
			std::string GetVersion();
			
			std::string GetPlatformName();
			std::string GetVendor();
			std::set<std::string> GetExtensions();
			
			std::string GetHumanReadableString();
			
		private:
			size_t GetPlatformInfo(cl_platform_info param_name, size_t param_value_size, void *param_value);
			std::string GetPlatformInfoString(cl_platform_info param_name);
			
			cl_platform_id m_id;
			std::set<Device> m_devices;
	};
	
	class Context : public TrackedClass
	{
		public:
			Context(Device device);
			virtual ~Context();
			
			Context(const Context &context);
			Context &operator=(const Context &context);
			
			cl_context GetClContext();
			Device GetDevice();
			
			
		private:
			cl_context m_context;
			Device m_device;
	};
	
	class Program : public TrackedClass
	{
		public:
			Program(Context context, std::string source);
			virtual ~Program();
			
			Program(const Program &program);
			Program &operator=(const Program &program);
			
			cl_program GetClProgram();
			Context GetContext();
			
		private:
			cl_program m_program;
			Context m_context;
	};
	
	class Kernel : public TrackedClass
	{
		public:
			Kernel(Program program, std::string name);
			virtual ~Kernel();
			
			Kernel(const Kernel &kernel);
			Kernel &operator=(const Kernel &kernel);
			
			uint64_t GetWorkGroupSize();
			uint64_t GetPreferredWorkGroupSizeMultiple();
			
		private:
			size_t GetKernelWorkGroupInfo(cl_kernel_work_group_info param_name, size_t param_value_size, void *param_value);
			template <typename T> T GetKernelWorkGroupInfo(cl_kernel_work_group_info param_name);
			
			cl_kernel m_kernel;
			Program m_program;
	};
}
