/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cstdio>

#include "../../ast/astnode.hpp"
#include "codegen.hpp"

CodeGeneratorLlvm::CodeGeneratorLlvm(AstNode *node) :
	CodeGenerator(node)
{
}

std::string CodeGeneratorLlvm::GetDebug(CompileContext &context, unsigned int d)
{
	return context.GetDebug() ? PrintString(", !dbg !{i32 %u, i32 0, metadata !%u, null}", GetNode()->GetLine(), d).c_str() : "";
}

CodeGeneratorRootLlvm::CodeGeneratorRootLlvm(AstNodeRoot *node) :
	CodeGeneratorLlvm(node)
{
}

CodeGeneratorRootLlvm *CodeGeneratorRootLlvm::Copy()
{
	return new CodeGeneratorRootLlvm(GetNode());
}

static std::list<std::string> GetLines(std::string text)
{
	std::string line;
	std::list<std::string> lines;
	
	for (char c: text){
		if ('\n' == c){
			lines.push_back(line);
			line = "";
		}
		else {
			line += c;
		}
	}
	if (line.size()){
		lines.push_back(line);
	}
	
	return lines;
}

static std::string GetOpcode(std::string text)
{
	size_t p = text.find_first_of(";");
	text = text.substr(0, p);
	
	p = text.find_first_of("=");
	if ((std::string::npos != p) && (p < text.size() - 1)){
		text = text.substr(p + 1, std::string::npos);
	}
	
	p = text.find_first_not_of("\t ");
	if (std::string::npos != p){
		text = text.substr(p, std::string::npos);
	}
	
	p = text.find_first_of("\t ");
	if (std::string::npos != p){
		text = text.substr(0, p);
	}
	
	return text;
}

static std::string InsertDebugToLlvm(std::string text, unsigned int line)
{
	size_t p = text.find_first_of(";");
	std::string insertion = PrintString(", !dbg !{i32 %u, i32 0, metadata !2, null}", line);
	
	return (std::string::npos == p) ? text + insertion : text.insert(p, insertion);
}

static std::string AddHeaderDebugging(std::string header)
{
	std::list<std::string> opcodes = {"alloca", "br", "call", "load", "ret", "store"};
	
	std::string result;
	std::list<std::string> lines = GetLines(header);
	
	unsigned int line_number = 1;
	for (std::string line: lines){
		bool do_debug = false;
		std::string opcode = GetOpcode(line);
		
		for (std::string candidate: opcodes){
			if (candidate == opcode){
				do_debug = true;
			}
		}
		
		if (do_debug){
			result += InsertDebugToLlvm(line, line_number);
		}
		else {
			result += line;
		}
		
		result += '\n';
		line_number++;
	}
	
	return result;
}

static std::string StripDebuggingInformation(std::string header)
{
	return header;
}

static std::string ReplaceDbgWithLlvmLine(std::string line, unsigned int n)
{
	size_t p = line.find(", !dbg !{i32 ");
	if (std::string::npos == p){
		return line;
	}
	p += 13;
	
	size_t q = line.find(" ", p);
	if (std::string::npos == q){
		throw FormatError("Debug line is broken: \"%s\"", line.c_str());
	}
	
	return PrintString("%s%u,%s", line.substr(0, p).c_str(), n, line.substr(q, std::string::npos).c_str());
}

static std::string ReplaceDbgWithLlvmLines(std::string text)
{
	std::string result;
	unsigned int n = 1;
	std::list<std::string> lines = GetLines(text);
	for (auto line: lines){
		result += ReplaceDbgWithLlvmLine(line, n) + "\n";
		n++;
	}
	
	return result;
}

std::string CodeGeneratorRootLlvm::Generate(std::string directory, CompileContext &context)
{
	std::string filename = directory + "header.ll";
	FILE *f = fopen(filename.c_str(), "r");
	if (!f){
		throw FormatError("Could not open \"%s\"", filename.c_str());
	}
	std::string lib;
	while (!feof(f)){
		lib += getc(f);
	}
	fclose(f);
	
	lib = lib.substr(0, lib.size() - 1);
	
	if (context.GetDebugLlvmHeader()){
		lib = AddHeaderDebugging(lib);
	}
	else {
		lib = StripDebuggingInformation(lib);
	}
	
	unsigned int t_size_size = context.GetComputeSpec().GetCpu().GetSizeT();
	
	/*
	 * This is the polycute private library - the methods that are included in
	 * each source file so that they may be inlined and optimised by LLVM
	 */
	std::string source = PrintString("; ======== Runtime Types ======== ;\n"
		"%%t.size = type i%u ; like C's size_t\n", t_size_size);
	
	source += "\n\n";
	
	source += PrintString("; ======== Runtime Functions ======== ;\n"
		"declare void @llvm.memcpy.p0i8.p0i8.i%u(i8*, i8*, i%u, i32, i1)\n"
		"define fastcc void @.memcpy (i8* %%dst, i8* %%src, %%t.size %%length) alwaysinline\n"
		"{\n"
		"\tcall void (i8*, i8*, %%t.size, i32, i1)* @llvm.memcpy.p0i8.p0i8.i%u(i8* %%dst, i8* %%src, i%u %%length, i32 1, i1 0)\n"
		"\tret void\n"
		"}\n\n\n"
		
		"define fastcc void @.memcpy_align (i8* %%dst, i8* %%src, %%t.size %%length) alwaysinline\n"
		"{\n"
		"\tcall void (i8*, i8*, %%t.size, i32, i1)* @llvm.memcpy.p0i8.p0i8.i%u(i8* %%dst, i8* %%src, i%u %%length, i32 64, i1 0)\n"
		"\tret void\n"
		"}\n\n\n",
		
		t_size_size, t_size_size,
		t_size_size, t_size_size,
		t_size_size, t_size_size
 	);
	
	source += lib;
	
	source += "\n\n\n; ======== Structure Types ======== ;\n\n";
	for (auto struct_declaration: GetNode()->GetStructDeclarations()){
		source += struct_declaration->GetCodeGenerator<CodeGeneratorStructDeclarationLlvm>()->Generate(context, 1);
	}
	
	source += "\n\n\n; ======== Global Variables ======== ;\n\n";
	/*
	 * Our global variables are stored separately to the functions, so we're
	 * writing these out first
	 */
	/*for (auto variable_declaration: GetNode()->GetVariableDeclarations()){
		source += variable_declaration->GetCodeGenerator<CodeGeneratorStatementVariableDeclarationLlvm>()->Generate(context, 1);
		source += variable_declaration->GetCodeGenerator<CodeGeneratorStatementVariableDeclarationLlvm>()->GenerateTopLevel(context, 1);
	}*/
	if (GetNode()->GetVariableDeclarations().size()){
		throw CodeGenerationException(CODEGEN_OPENCL, GetNode(), "Global variables not yet supported for LLVM");
	}
	
	/*
	 * Functions are next - output all the functions in this source file
	 */
	source += "\n\n\n; ======== Functions ======== ;\n\n";
	std::string functions_debug; //contains the metadata list (requires that function declarations create their metadata with the same ID as their node ID)
	for (auto function_declaration: GetNode()->GetFunctionDeclarations()){
		source += function_declaration->GetCodeGenerator<CodeGeneratorFunctionDeclarationLlvm>()->Generate(context, 1);
		functions_debug += PrintString("%smetadata !%u", functions_debug.size() ? ", " : "", function_declaration->GetId());
	}
	
	if (context.GetDebug()){
		//Main compilation unit
		source += PrintString(
			"declare void @llvm.dbg.value(metadata, i64, metadata)"
			"declare void @llvm.dbg.declare(metadata, metadata) nounwind readnone\n"
			"!llvm.dbg.cu = !{!0}\n"
			"!0 = metadata !{\n"
				"\ti32 %u, ; Tag = 17 + LLVMDebugVersion (DW_TAG_compile_unit)\n"
				"\ti32 0, ; Unused field.\n"
				"\ti32 12, ; DWARF language identifier (ex. DW_LANG_C89)\n"
				"\tmetadata !\"%s\", ; Source file name\n"
				"\tmetadata !\"%s\", ; Source file directory (includes trailing slash)\n"
				"\tmetadata !\"Polycute\", ; Producer\"\n"
				"\ti1 true, ; True if this is a main compile unit.\n"
				"\ti1 false, ; True if this is optimized.\n"
				"\tmetadata !\"\", ; Flags\n"
				"\ti32 0, ; Runtime version\n"
				"\tmetadata !{metadata !{i32 0}}, ; List of enums types\n"
				"\tmetadata !{metadata !{i32 0}}, ; List of retained types\n"
				"\tmetadata !{metadata !{%s}}, ; List of subprograms\n"
				"\tmetadata !{metadata !{i32 0}}  ; List of global variables\n"
			"}\n",
			LlvmDebugVersion + 17,
			"test.pc",//GetNode()->GetFilename().c_str(),
			"/home/nicholas/programming/polycute/test", //pwd
			functions_debug.c_str()
		);
		
		//File
		source += PrintString(
			"!1 = metadata !{\n"
				"\ti32 %u, ; Tag = 41 + LLVMDebugVersion (DW_TAG_file_type)\n"
				"\tmetadata !\"%s\", ; Source file name\n"
				"\tmetadata !\"%s\", ; Source file directory (includes trailing slash)\n"
				"\tnull ; Unused\n"
			"}\n",
			LlvmDebugVersion + 41,
			"test.pc", //GetNode()->GetFilename().c_str(),
			"/home/nicholas/programming/polycute/test" //pwd
		);
		
		if (context.GetDebugLlvmHeader()){
			source += PrintString(
				"!2 = metadata !{\n"
					"\ti32 %u, ; Tag = 41 + LLVMDebugVersion (DW_TAG_file_type)\n"
					"\tmetadata !\"header.ll\", ; Source file name\n"
					"\tmetadata !\"%s\", ; Source file directory (includes trailing slash)\n"
					"\tnull ; Unused\n"
				"}\n",
				LlvmDebugVersion + 41,
				"/home/nicholas/programming/polycute/src/lib"
			);
			
			source += PrintString("!3 = %s", CodeGeneratorTypeDynamicVectorLlvm::GenerateDebugGeneric(context, 5).c_str());
			source += PrintString("!4 = %s", CodeGeneratorTypeIntLlvm::GenerateDebugGeneric("Blob", 1, false).c_str());
			source += CodeGeneratorTypeDynamicVectorLlvm::GenerateDebugTopLevelGeneric(context, 5, "metadata !4");
		}
	}
	
	if (context.GetDebugLlvmLines()){
		source = ReplaceDbgWithLlvmLines(source);
	}
	
	return source;
}
