/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../../ast/astnode.hpp"
#include "codegen.hpp"

CodeGeneratorExpressionLlvm::CodeGeneratorExpressionLlvm(AstNodeExpression *node) :
	CodeGeneratorLlvm(node)
{
}

std::string CodeGeneratorExpressionLlvm::GenerateFnBegin(CompileContext &context, unsigned int d)
{
	return "";
}

std::string CodeGeneratorExpressionLlvm::GenerateFnEnd(CompileContext &context, unsigned int d)
{
	return "";
}

std::string CodeGeneratorExpressionLlvm::GenerateTopLevel(CompileContext &context, unsigned int d)
{
	return "";
}

std::string CodeGeneratorExpressionLlvm::GenerateCopy(CompileContext &context, unsigned int d, unsigned int n)
{
	unsigned int tmp = GetUniqueId();
	std::string copy_code = GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateCopy(context, d, tmp, n);
	
	if (copy_code.size()){
		return Generate(context, d, tmp) + copy_code;
	}
	else {
		return Generate(context, d, n);
	}
}

bool CodeGeneratorExpressionLlvm::CopyIfCopyable()
{
	return true;
}

CodeGeneratorExpressionLValueLlvm::CodeGeneratorExpressionLValueLlvm(AstNodeExpressionLValue *node) :
	CodeGeneratorExpressionLlvm(node)
{
}


CodeGeneratorExpressionVoidLlvm::CodeGeneratorExpressionVoidLlvm(AstNodeExpressionVoid *node) :
	CodeGeneratorExpressionLlvm(node)
{
}

CodeGeneratorExpressionVoidLlvm *CodeGeneratorExpressionVoidLlvm::Copy()
{
	return new CodeGeneratorExpressionVoidLlvm(GetNode());
}

std::string CodeGeneratorExpressionVoidLlvm::Generate(CompileContext &context, unsigned int d, unsigned int n)
{
	throw CodeGenerationException(CODEGEN_LLVM, GetNode(), "Void value not supported in LLVM");
}

CodeGeneratorExpressionBoolLlvm::CodeGeneratorExpressionBoolLlvm(AstNodeExpressionBool *node) :
	CodeGeneratorExpressionLlvm(node)
{
}

CodeGeneratorExpressionBoolLlvm *CodeGeneratorExpressionBoolLlvm::Copy()
{
	return new CodeGeneratorExpressionBoolLlvm(GetNode());
}

std::string CodeGeneratorExpressionBoolLlvm::Generate(CompileContext &context, unsigned int d, unsigned int n)
{
	return PrintString("%%.%u = or i1 %u, 0\n", n, GetNode()->GetValue() ? 1 : 0);
}

CodeGeneratorExpressionIntLlvm::CodeGeneratorExpressionIntLlvm(AstNodeExpressionInt *node) :
	CodeGeneratorExpressionLlvm(node)
{
}

CodeGeneratorExpressionIntLlvm *CodeGeneratorExpressionIntLlvm::Copy()
{
	return new CodeGeneratorExpressionIntLlvm(GetNode());
}

std::string CodeGeneratorExpressionIntLlvm::Generate(CompileContext &context, unsigned int d, unsigned int n)
{
	return PrintString("%%.%u = or i32 %u, 0\n", n, (unsigned int)GetNode()->GetValue());
}

CodeGeneratorExpressionFloatLlvm::CodeGeneratorExpressionFloatLlvm(AstNodeExpressionFloat *node) :
	CodeGeneratorExpressionLlvm(node)
{
}

CodeGeneratorExpressionFloatLlvm *CodeGeneratorExpressionFloatLlvm::Copy()
{
	return new CodeGeneratorExpressionFloatLlvm(GetNode());
}

std::string CodeGeneratorExpressionFloatLlvm::Generate(CompileContext &context, unsigned int d, unsigned int n)
{
	return PrintString("%%.%u = fadd float %f, 0.0\n", n, GetNode()->GetValue());
}

CodeGeneratorExpressionLengthLlvm::CodeGeneratorExpressionLengthLlvm(AstNodeExpressionLength *node) :
	CodeGeneratorExpressionLlvm(node)
{
}

CodeGeneratorExpressionLengthLlvm *CodeGeneratorExpressionLengthLlvm::Copy()
{
	return new CodeGeneratorExpressionLengthLlvm(GetNode());
}

std::string CodeGeneratorExpressionLengthLlvm::Generate(CompileContext &context, unsigned int d, unsigned int n)
{
	std::string source;
	unsigned int expression = GetUniqueId(); //The expression is of vector type
	
	//Calculate the expression that represents the length of the vector
	source += GetNode()->GetExpression()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->Generate(context, d, expression);
	
	source += PrintString("%%.%u = call fastcc %%t.size (%%t.dynvect)* @.dynvect.element_count(%%t.dynvect %%.%u)%s ; Get the length\n",
		n, expression, GetDebug(context, d).c_str());
	
	return source;
}

std::string CodeGeneratorExpressionLengthLlvm::GenerateFnBegin(CompileContext &context, unsigned int d)
{
	return GetNode()->GetExpression()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateFnBegin(context, d);
}

std::string CodeGeneratorExpressionLengthLlvm::GenerateFnEnd(CompileContext &context, unsigned int d)
{
	return GetNode()->GetExpression()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateFnEnd(context, d);
}


CodeGeneratorExpressionLValueIdentifierLlvm::CodeGeneratorExpressionLValueIdentifierLlvm(AstNodeExpressionLValueIdentifier *node) :
	CodeGeneratorExpressionLValueLlvm(node)
{
}

CodeGeneratorExpressionLValueIdentifierLlvm *CodeGeneratorExpressionLValueIdentifierLlvm::Copy()
{
	return new CodeGeneratorExpressionLValueIdentifierLlvm(GetNode());
}

std::string CodeGeneratorExpressionLValueIdentifierLlvm::Generate(CompileContext &context, unsigned int d, unsigned int n)
{
	std::string debug = GetDebug(context, d);
	if (GetNode()->GetIsFunction()){
		unsigned int tmp = GetUniqueId();
		std::string type = GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
		
		return PrintString(
			"%%.%u = ptrtoint %s @%s to %%t.size\n"
			"%%.%u = inttoptr %%t.size %%.%u to %s\n",
			tmp, type.c_str(), GetNode()->GetIdentifier().c_str(),
			n, tmp, type.c_str());
	}
	else {
		return PrintString(
			"%%.%u = load %s* %%.%u%s ; Load %s\n",
			n, GetNode()->GetLink()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(),
				GetNode()->GetLink()->GetId(), debug.c_str(), GetNode()->GetIdentifier().c_str()
		);
	}
}

std::string CodeGeneratorExpressionLValueIdentifierLlvm::GenerateStore(CompileContext &context, unsigned int d, unsigned int n)
{
	std::string source;
	std::string debug = GetDebug(context, d);
	
	//NOTE: It's important that we always copy BEFORE we destroy, in case we're doing self assignment
	/*unsigned int old_value = GetUniqueId();
	source += Generate(context, d, old_value).c_str();
	source += GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateCopy(context, d, n);
	source += GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateDestroy(context, d, old_value);*/
	
	std::string type = GetNode()->GetLink()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
	source += PrintString("store %s %%.%u, %s* %%.%u%s ; variable assignment\n",
		type.c_str(), n, type.c_str(), GetNode()->GetLink()->GetId(), debug.c_str());
	
	return source;
}

std::string CodeGeneratorExpressionLValueIdentifierLlvm::GenerateReference(CompileContext &context, unsigned int d, unsigned int n)
{
	if (GetNode()->GetIsFunction()){
		throw FormatCompilerError(GetNode(), "Reference to function taken, should be function pointer");
	}
	
	std::string debug = GetDebug(context, d);
	std::string type = GetNode()->GetLink()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
	
	unsigned int pointer_type = GetNode()->GetLink()->GetId();
	unsigned int pointer_i8 = GetUniqueId();
	
	return PrintString(
		"; Generating reference to identifier %s\n"
			"\t%%.%u = bitcast %s* %%.%u to i8* ; Bitcast the pointer\n"
			"\t%%.%u = call fastcc %%t.ref (i8*)* @.ref.create(i8* %%.%u)%s ; Generate the reference\n",
		
		GetNode()->GetSourceLocation().c_str(),
		pointer_i8, type.c_str(), pointer_type,
		n, pointer_i8, debug.c_str()
	);
}

CodeGeneratorExpressionLValueVectorLlvm::CodeGeneratorExpressionLValueVectorLlvm(AstNodeExpressionLValueVector *node) :
	CodeGeneratorExpressionLValueLlvm(node)
{
}

CodeGeneratorExpressionLValueVectorLlvm *CodeGeneratorExpressionLValueVectorLlvm::Copy()
{
	return new CodeGeneratorExpressionLValueVectorLlvm(GetNode());
}

std::string CodeGeneratorExpressionLValueVectorLlvm::Generate(CompileContext &context, unsigned int d, unsigned int n)
{
	std::string source;
	unsigned int lvalue = GetUniqueId();
	unsigned int expression = GetUniqueId();
	std::string debug = GetDebug(context, d);
	
	//The expression is of type %t.size
	source += GetNode()->GetExpression()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->Generate(context, d, expression).c_str();
	source += GetNode()->GetLValue()->GetCodeGenerator<CodeGeneratorExpressionLValueLlvm>()->Generate(context, d, lvalue).c_str();
	
	if (dynamic_cast<AstNodeTypeStaticVector *>(GetNode()->GetLValue()->GetType())){
		unsigned int expression32 = expression;
		if (context.GetComputeSpec().GetCpu().GetSizeT() > 32){ //we're assuming the minimum pointer length is 32 bits
			expression32 = GetUniqueId();
			source += PrintString("%%.%u = trunc %s %%.%u to i32%s ; extractelement is strange and needs i32\n",
				expression32, GetNode()->GetExpression()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), expression,
				debug.c_str());
		}
		
		source += PrintString("%%.%u = extractelement %s %%.%u, i32 %%.%u%s; Vector loading\n",
			n, GetNode()->GetLValue()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), lvalue, expression32,
			debug.c_str());
	}
	else if (dynamic_cast<AstNodeTypeDynamicVector *>(GetNode()->GetLValue()->GetType())){
		AstNodeTypeDynamicVector *ast_type = tracked_cast<AstNodeTypeDynamicVector>(GetNode()->GetLValue()->GetType());
		CodeGeneratorTypeDynamicVectorLlvm *cg_type = ast_type->GetCodeGenerator<CodeGeneratorTypeDynamicVectorLlvm>();
		
		//lvalue is (a pointer to) a vector
		unsigned int buffer_i8 = GetUniqueId();
		unsigned int buffer_type = GetUniqueId();
		std::string type = ast_type->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
		
		source += PrintString("%%.%u = call fastcc i8* (%%t.dynvect, %%t.size, %%t.size)* @.dynvect.read (%%t.dynvect %%.%u, %%t.size %%.%u, %%t.size 1)%s ; Get the buffer to read from\n"
			"call fastcc void (%%t.dynvect)* @.dynvect.commit.before(%%t.dynvect %%.%u)%s ; Commit the read\n"
			"%%.%u = bitcast i8* %%.%u to %s*%s ; Convert the type of the pointer\n"
			"%%.%u = load %s* %%.%u, align %u%s ; Load the value\n"
			"call fastcc void (%%t.dynvect)* @.dynvect.commit.after(%%t.dynvect %%.%u)%s ; Commit the read\n",
			buffer_i8, lvalue, expression, debug.c_str(),
			lvalue, debug.c_str(),
			buffer_type, buffer_i8, type.c_str(), debug.c_str(),
			n, type.c_str(), buffer_type, cg_type->GetAlignment(context), debug.c_str(),
			lvalue, debug.c_str());
	}
	else if (dynamic_cast<AstNodeTypePointer *>(GetNode()->GetLValue()->GetType())){
		unsigned int pointer = GetUniqueId();
		
		std::string type = GetNode()->GetLValue()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
		source += PrintString("%%.%u = getelementptr %s %%.%u, %%t.size %%.%u%s ; calculate the pointer to the element\n"
			"%%.%u = load %s %%.%u%s ; load from the calculated element\n",
			pointer, type.c_str(), lvalue, expression, debug.c_str(),
			n, type.c_str(), pointer, debug.c_str());
	}
	else {
		throw FormatError("LValue generate on unrecognised type");
	}
	
	return source;
}

/* WARNING TODO: We need to make it so things can copy dynamic vectors; including optimisers 
 TODO: Make it so we cast the integer to %t.size (if it's a static vector, then cast back to i32?) */
std::string CodeGeneratorExpressionLValueVectorLlvm::GenerateStore(CompileContext &context, unsigned int d, unsigned int n)
{
	std::string source;
	unsigned int lvalue = GetUniqueId();
	unsigned int expression = GetUniqueId();
	std::string debug = GetDebug(context, d);
	
	source += GetNode()->GetExpression()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->Generate(context, d, expression).c_str();
	source += GetNode()->GetLValue()->GetCodeGenerator<CodeGeneratorExpressionLValueLlvm>()->Generate(context, d, lvalue).c_str();
	
	//NOTE: It's important that we always copy BEFORE we destroy, in case we're doing self assignment
	/*unsigned int old_value = GetUniqueId();
	source += Generate(context, d, old_value).c_str();
	source += GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateCopy(context, d, n);
	source += GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateDestroy(context, d, old_value);*/
	
	if (dynamic_cast<AstNodeTypeStaticVector *>(GetNode()->GetLValue()->GetType())){
		unsigned int lvalue_store = GetUniqueId();
		
		unsigned int expression32 = expression;
		if (context.GetComputeSpec().GetCpu().GetSizeT() > 32){ //we're assuming the minimum pointer length is 32 bits
			expression32 = GetUniqueId();
			source += PrintString("%%.%u = trunc %s %%.%u to i32%s ; extractelement is strange and needs i32\n",
				expression32, GetNode()->GetExpression()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), expression,
				debug.c_str());
		}
		
		source += PrintString("%%.%u = insertelement %s %%.%u, %s %%.%u, i32 %%.%u%s; Vector storing\n",
			lvalue_store, GetNode()->GetLValue()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), lvalue,
			GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), n, expression32,
			debug.c_str());
		
		source += GetNode()->GetLValue()->GetCodeGenerator<CodeGeneratorExpressionLValueLlvm>()->GenerateStore(context, d, lvalue_store).c_str();
	}
	else if (dynamic_cast<AstNodeTypeDynamicVector *>(GetNode()->GetLValue()->GetType())){
		//lvalue is (a pointer to) a vector
		AstNodeTypeDynamicVector *ast_type = tracked_cast<AstNodeTypeDynamicVector>(GetNode()->GetLValue()->GetType());
		CodeGeneratorTypeDynamicVectorLlvm *cg_type = ast_type->GetCodeGenerator<CodeGeneratorTypeDynamicVectorLlvm>();
		
		
		unsigned int buffer_i8 = GetUniqueId(); //the buffer we write into
		unsigned int buffer_type = GetUniqueId(); //the buffer we write into, but of the correct pointer type
		std::string type = ast_type->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
		
		source += PrintString("%%.%u = call fastcc i8* @.dynvect.write (%%t.dynvect %%.%u, %%t.size %%.%u, %%t.size 1)%s ; Get the vector/buffer\n"
			"call fastcc void (%%t.dynvect)* @.dynvect.commit.before(%%t.dynvect %%.%u)%s ; Commit the write\n"
			"%%.%u = bitcast i8* %%.%u to %s*%s ; Convert the pointer type\n"
			"store %s %%.%u, %s* %%.%u, align %u%s ; store the new value\n"
			"call fastcc void (%%t.dynvect)* @.dynvect.commit.after(%%t.dynvect %%.%u)%s ; Commit the write\n",
			buffer_i8, lvalue, expression, debug.c_str(),
			lvalue, debug.c_str(),
			buffer_type, buffer_i8, type.c_str(), debug.c_str(),
			type.c_str(), n, type.c_str(), buffer_type, cg_type->GetAlignment(context), debug.c_str(),
			lvalue, debug.c_str());
	}
	else if (dynamic_cast<AstNodeTypePointer *>(GetNode()->GetLValue()->GetType())){
		unsigned int pointer = GetUniqueId();
		std::string type = GetNode()->GetLValue()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
		source += PrintString("%%.%u = getelementptr %s %%.%u, %%t.size %%.%u%s ; calculate the pointer to the element\n"
			"store %s %%.%u, %s %%.%u%s ; load from the array\n",
			pointer, type.c_str(), lvalue, expression, debug.c_str(),
			tracked_cast<AstNodeTypePointer>(GetNode()->GetLValue()->GetType())->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(),
				n, type.c_str(), pointer, debug.c_str());
	}
	else {
		throw FormatError("LValue generate store on unrecognised type");
	}
	
	return source;
}

std::string CodeGeneratorExpressionLValueVectorLlvm::GenerateReference(CompileContext &context, unsigned int d, unsigned int n)
{
	throw FormatCompilerError(GetNode(), "Reference to vector element not currently supported");
}

std::string CodeGeneratorExpressionLValueVectorLlvm::GenerateFnBegin(CompileContext &context, unsigned int d)
{
	return GetNode()->GetExpression()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateFnBegin(context, d) +
		GetNode()->GetLValue()->GetCodeGenerator<CodeGeneratorExpressionLValueLlvm>()->GenerateFnBegin(context, d);
}

std::string CodeGeneratorExpressionLValueVectorLlvm::GenerateFnEnd(CompileContext &context, unsigned int d)
{
	return GetNode()->GetExpression()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateFnEnd(context, d) +
		GetNode()->GetLValue()->GetCodeGenerator<CodeGeneratorExpressionLValueLlvm>()->GenerateFnEnd(context, d);
}

CodeGeneratorExpressionBlockLlvm::CodeGeneratorExpressionBlockLlvm(AstNodeExpressionBlock *node) :
	CodeGeneratorExpressionLlvm(node)
{
}

CodeGeneratorExpressionBlockLlvm *CodeGeneratorExpressionBlockLlvm::Copy()
{
	return new CodeGeneratorExpressionBlockLlvm(GetNode());
}

std::string CodeGeneratorExpressionBlockLlvm::Generate(CompileContext &context, unsigned int d, unsigned int n)
{
	std::string source = PrintString("; ==== Expression Block at line %u ==== ;\n", GetNode()->GetLine());
	for (auto statement: GetNode()->GetStatements()){
		source += Indent(statement->GetCodeGenerator<CodeGeneratorStatementLlvm>()->Generate(context, GetNode()->GetId()), 1);
	}
	source += "\n";
	
	source += PrintString("\t; ==== Expression for expression block at line %u ==== ;\n", GetNode()->GetExpression()->GetLine());
	source += Indent(GetNode()->GetExpression()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->Generate(context, d, n), 1);
	
	return source;
}

std::string CodeGeneratorExpressionBlockLlvm::GenerateFnBegin(CompileContext &context, unsigned int d)
{
	std::string source = PrintString("; ==== Expression Block at line %u, beginning of function ==== ;\n", GetNode()->GetLine());
	for (auto statement: GetNode()->GetStatements()){
		source += Indent(statement->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateFnBegin(context, GetNode()->GetId()), 1);
	}
	return source + "\n";
}

std::string CodeGeneratorExpressionBlockLlvm::GenerateFnEnd(CompileContext &context, unsigned int d)
{
	std::string source = PrintString("; ==== Expression Block at line %u, end of function ==== ;\n", GetNode()->GetLine());
	for (auto statement: GetNode()->GetStatements()){
		source += Indent(statement->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateFnEnd(context, GetNode()->GetId()), 1);
	}
	return source + "\n";
}

std::string CodeGeneratorExpressionBlockLlvm::GenerateTopLevel(CompileContext &context, unsigned int d)
{
	std::string source;
	for (auto statement: GetNode()->GetStatements()){
		source += statement->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateTopLevel(context, GetNode()->GetId());
	}
	return source + "\n";
}





/* 
 * Exponentially weighted dynamic optimisation of data location
 */

