/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../../ast/astnode.hpp"
#include "codegen.hpp"

CodeGeneratorFunctionDeclarationLlvm::CodeGeneratorFunctionDeclarationLlvm(AstNodeFunctionDeclaration *node) :
	CodeGeneratorLlvm(node)
{
}

CodeGeneratorFunctionDeclarationLlvm *CodeGeneratorFunctionDeclarationLlvm::Copy()
{
	return new CodeGeneratorFunctionDeclarationLlvm(GetNode());
}

std::string CodeGeneratorFunctionDeclarationLlvm::Generate(CompileContext &context, unsigned int d)
{
	unsigned int function_context = GetNode()->GetId();
	std::string return_type = GetNode()->GetReturnType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, function_context);
	
	//start of function definition
	std::string source = PrintString("%s %s %s %s @%s (",
		GetNode()->GetStatement() ? "define" : "declare", (GetNode()->GetExport() | !GetNode()->GetStatement()) ? "" : "private", GetNode()->GetCCall() ? "ccc" : "fastcc",
		return_type.c_str(),
		GetNode()->GetIdentifier().c_str());
	
	//arguments
	unsigned int i = 0;
	for (auto arg: GetNode()->GetArguments()){
		if (i){
			source += ", ";
		}
		source += arg->GetCodeGenerator<CodeGeneratorStatementVariableDeclarationLlvm>()->Generate(context, function_context);
		i++;
	}
	source += PrintString(") ; Function declaration at line %u \n", GetNode()->GetLine());
	
	if (GetNode()->GetStatement()){
		//top level stuff - We put this at the top in case we need to reference it in the function
		source = GetNode()->GetStatement()->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateTopLevel(context, function_context) + source;
		
		source += "{\n";
		std::string body;
		
		if (!dynamic_cast<AstNodeTypeVoid *>(GetNode()->GetReturnType())){
			body += "; ======== Return value ======== ;\n" +
				Indent(PrintString("%%.function_return_value = alloca %s ; Allocate space for the return value\n",
					return_type.c_str()), 1);
		}
		
		//main body
		
		//load the argument variables into suitable places
		body += "; ======== Variable declaration function beginning ======== ;\n";
		for (auto arg: GetNode()->GetArguments()){
			body += Indent(arg->GetCodeGenerator<CodeGeneratorStatementVariableDeclarationLlvm>()->GenerateFnBegin(context, function_context), 1);
		}
		
		
		//function header (e.g: allocate space for local variables)
		body += "; ======== Statement function beginning ======== ;\n";
		body += Indent(GetNode()->GetStatement()->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateFnBegin(context, function_context), 1);
		
		//statement
		body += "; ======== Statement ======== ;\n" +
			Indent(GetNode()->GetStatement()->GetCodeGenerator<CodeGeneratorStatementLlvm>()->Generate(context, function_context), 1);
		
		
		/* End of function */
		body += "; ================ End of Function ================\n"
			"br label %.function_end\n"
			".function_end:\n";
		
		
		unsigned int return_value_original = GetUniqueId();
		unsigned int return_value_copy = GetUniqueId();
		std::string copycode = GetNode()->GetReturnType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateCopy(context, d, return_value_original, return_value_copy);
		
		if (copycode.size()){
			body += PrintString(
				"\n\t; ======== Copy the return value ========\n"
					"\t\t%%.%u = load %s* %%.function_return_value ; Load the original return value\n"
					"\t\t; Copy the return value\n"
						"%s"
					"\t\tstore %s %%.%u, %s* %%.function_return_value\n\n",
					
				return_value_original, return_type.c_str(),
				Indent(copycode, 3).c_str(),
				return_type.c_str(), return_value_copy, return_type.c_str()
			);
		}
		
		
		//End of the function (function footer, e.g deallocate heap allocated stuff)
		body += "\t; ======== Statement function end ======== ;\n" +
			Indent(GetNode()->GetStatement()->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateFnEnd(context, function_context), 2);
		
		
		//End of the function - destroy the variable declarations
		body += "\t; ======== Variable declaration function end ======== ;\n";
		for (auto arg: GetNode()->GetArguments()){
			body += Indent(arg->GetCodeGenerator<CodeGeneratorStatementVariableDeclarationLlvm>()->GenerateFnEnd(context, function_context), 2);
		}
		
		//if it's a void type, we should return void
		if (dynamic_cast<AstNodeTypeVoid *>(GetNode()->GetReturnType())){
			body += "\tret void ; end of function\n";
		}
		else { //if it's not a void type, we should return the return value
			unsigned int return_load = GetUniqueId();
			body += PrintString("\t%%.%u = load %s* %%.function_return_value\nret %s %%.%u\n",
				return_load, return_type.c_str(), return_type.c_str(), return_load);
		}
		
		source += Indent(body, 1);
		source += "}\n";
	}
	
	if (context.GetDebug()){
		std::string function_type_args;
		std::string debug_arg_types;
		std::string arg_tops;
		for (auto arg: GetNode()->GetArguments()){
			if (function_type_args.size()){
				function_type_args += ", ";
			}
			function_type_args += arg->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, function_context);
			
			debug_arg_types += ", " + arg->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateDebug(context, function_context);
			
			arg_tops += arg->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateTopLevel(context, function_context);
		}
		arg_tops += GetNode()->GetReturnType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateTopLevel(context, function_context);
		
		std::string function_type_name = PrintString("%s (%s)* @%s",
			GetNode()->GetReturnType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, GetNode()->GetId()).c_str(),
			function_type_args.c_str(),
			GetNode()->GetIdentifier().c_str());
		
		source += arg_tops;
		
		source += PrintString("!%u = metadata !{\n"
			"\ti32 %u, ; Tag = 46 + LLVMDebugVersion (DW_TAG_subprogram)\n"
			"\ti32 0, ; Unused field.\n"
			"\tmetadata !%u, ; Reference to context descriptor\n"
			"\tmetadata !\"%s\", ; Name\n"
			"\tmetadata !\"%s\", ; Display name (fully qualified C++ name)\n"
			"\tmetadata !\"\", ; MIPS linkage name (for C++)\n"
			"\tmetadata !1, ; Reference to file where defined\n"
			"\ti32 %u, ; Line number where defined\n"
			
			"\tmetadata !{\n"
			"\t\ti32 %u, ; DW_TAG_subroutine_type\n"
			"\t\ti32 0,\n"
			"\t\tmetadata !\"\",\n"
			"\t\ti32 0,\n"
			"\t\ti32 0,\n"
			"\t\ti64 0,\n"
			"\t\ti64 0,\n"
			"\t\ti64 0,\n"
			"\t\ti32 0,\n"
			"\t\tnull,\n"
			"\t\tmetadata !{\n%s\n%s\n\t\t},\n"
			"\t\ti32 0,\n"
			"\t\ti32 0\n"
			"\t},\n"
			
			"\ti1 %s, ; True if the global is local to compile unit (static)\n"
			"\ti1 %s, ; True if the global is defined in the compile unit (not extern)\n"
			"\ti32 0, ; Line number where the scope of the subprogram begins\n"
			"\ti32 0, ; Virtuality, e.g. dwarf::DW_VIRTUALITY__virtual\n"
			//"\ti32 0, ; Index into a virtual function\n" // looking at what clang generated, one of these last three is missing
			"\tnull, ; indicates which base type contains the vtable pointer for the derived class\n"
			"\ti32 256, ; Flags - Artifical, Private, Protected, Explicit, Prototyped.\n" //or 256
			"\ti1 false, ; isOptimized\n"
			"\t%s, ; Pointer to LLVM function\n"
			"\tnull, ; Lists function template parameters\n"
			"\tnull, ; Function declaration descriptor\n"
			"\tmetadata !{metadata !{i32 0}}, ; List of function variables\n"
			"\ti32 31\n"
			"}\n",
			GetNode()->GetId(),
			LlvmDebugVersion + 46,
			d,
			GetNode()->GetIdentifier().c_str(),
			GetNode()->GetIdentifier().c_str(),
			GetNode()->GetLine(),
			
			LlvmDebugVersion + 21,
			Indent(GetNode()->GetReturnType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateDebug(context, GetNode()->GetId()), 3).c_str(), //The return type
			Indent(debug_arg_types, 3).c_str(),
			
			
			GetNode()->GetExport() ? "false" : "true",
			GetNode()->GetStatement() ? "true" : "false",
			function_type_name.c_str()
		);
	}
	
	source += "\n";
	return source;
}

CodeGeneratorExpressionFunctionCallLlvm::CodeGeneratorExpressionFunctionCallLlvm(AstNodeFunctionCall *node) :
	CodeGeneratorExpressionLlvm(node)
{
}

CodeGeneratorExpressionFunctionCallLlvm *CodeGeneratorExpressionFunctionCallLlvm::Copy()
{
	return new CodeGeneratorExpressionFunctionCallLlvm(GetNode());
}

std::string CodeGeneratorExpressionFunctionCallLlvm::Generate(CompileContext &context, unsigned int d, unsigned int n)
{
	std::string source;
	unsigned int function = GetUniqueId();
	
	std::list<std::pair<unsigned int, AstNodeExpression *>> expressions;
	for (auto expression: GetNode()->GetExpressions()){
		unsigned int en = GetUniqueId();
		source += expression->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->Generate(context, d, en);
		
		unsigned int ecn = GetUniqueId();
		std::string copycode = expression->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateCopy(context, d, en, ecn);
		source += copycode;
		
		expressions.push_back(std::pair<unsigned int, AstNodeExpression *>(copycode.size() ? ecn : en, expression));
	}
	
	source += GetNode()->GetLValue()->GetCodeGenerator<CodeGeneratorExpressionLValueLlvm>()->Generate(context, d, function);
	
	if (dynamic_cast<AstNodeTypeVoid *>(GetNode()->GetType())){
		//source += PrintString("%%.%u = void\n", n);
	}
	else {
		source += PrintString("%%.%u = ", n);
	}
	
	source += PrintString("call %s %s (", GetNode()->GetCCall() ? "ccc" : "fastcc", GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str());
	
	unsigned int i = 0;
	std::list<AstNodeType *> arg_types = GetNode()->GetArgumentTypes();
	for (auto type: arg_types){
		if (i){
			source += ", ";
		}
		source += type->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
		i++;
	}
	
	source += PrintString(")* %%.%u (", function);
	i = 0;
	for (auto expression: expressions){
		if (i){
			source += ", ";
		}
		source += PrintString("%s %%.%u", expression.second->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), expression.first);
		i++;
	}
	
	source += ")";
	source += GetDebug(context, d);
	source += "; Function call\n";
	
	
	return source;
}

std::string CodeGeneratorExpressionFunctionCallLlvm::GenerateFnBegin(CompileContext &context, unsigned int d)
{
	std::string source;
	for (auto expression: GetNode()->GetExpressions()){
		source += expression->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateFnBegin(context, d);
	}
	
	return source;
}

std::string CodeGeneratorExpressionFunctionCallLlvm::GenerateFnEnd(CompileContext &context, unsigned int d)
{
	std::string source;
	for (auto expression: GetNode()->GetExpressions()){
		source += expression->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateFnEnd(context, d);
	}
	
	return source;
}

std::string CodeGeneratorExpressionFunctionCallLlvm::GenerateTopLevel(CompileContext &context, unsigned int d)
{
	std::string source;
	for (auto expression: GetNode()->GetExpressions()){
		source += expression->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateTopLevel(context, d);
	}
	
	return source;
}

bool CodeGeneratorExpressionFunctionCallLlvm::CopyIfCopyable()
{
	return false;
}
