/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../../ast/astnode.hpp"
#include "codegen.hpp"

CodeGeneratorExpressionBinaryLlvm::CodeGeneratorExpressionBinaryLlvm(AstNodeExpressionBinary *node) :
	CodeGeneratorExpressionLlvm(node)
{
}

CodeGeneratorExpressionCastLlvm::CodeGeneratorExpressionCastLlvm(AstNodeExpressionCast *node) :
	CodeGeneratorExpressionLlvm(node)
{
}

CodeGeneratorExpressionCastLlvm *CodeGeneratorExpressionCastLlvm::Copy()
{
	return new CodeGeneratorExpressionCastLlvm(GetNode());
}

std::string CodeGeneratorExpressionCastLlvm::Generate(CompileContext &context, unsigned int d, unsigned int n)
{
	AstNodeTypeInt *cast_type_int = dynamic_cast<AstNodeTypeInt *>(GetNode()->GetType());
	AstNodeTypeFloat *cast_type_float = dynamic_cast<AstNodeTypeFloat *>(GetNode()->GetType());
	
	AstNodeTypeInt *expr_type_int = dynamic_cast<AstNodeTypeInt *>(GetNode()->GetExpression()->GetType());
	AstNodeTypeFloat *expr_type_float = dynamic_cast<AstNodeTypeFloat *>(GetNode()->GetExpression()->GetType());
	
	unsigned int en = GetUniqueId();
	std::string source = GetNode()->GetExpression()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->Generate(context, d, en);
	
	if (cast_type_int){ //we're casting to an int
		if (expr_type_int){ //we're casting from an int to an int
			if (cast_type_int->GetWidth(context) < expr_type_int->GetWidth(context)){
				source += PrintString("%%.%u = trunc %s %%.%u to %s ; cast from int to int\n",
					n, GetNode()->GetExpression()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), en,
					GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str());
				return source;
			}
			else if (cast_type_int->GetWidth(context) > expr_type_int->GetWidth(context)){
				source += PrintString("%%.%u = %cext %s %%.%u to %s ; cast from int to int\n",
					n, expr_type_int->GetSigned() ? 's' : 'z', GetNode()->GetExpression()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(),
					en,GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str());
				//I wonder if this is how C converts int16_t to uint32_t
				return source;
			}
			else {
				std::string type = GetNode()->GetExpression()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
				source += PrintString("%%.%u = bitcast %s %%.%u to %s ; actually, they're the same type\n", n, type.c_str(), en, type.c_str());
				return source;
			}
		}
		else if (expr_type_float){ //we're casting from a float to an int
			source += PrintString("%%.%u = fpto%ci %s %%.%u to %s ; cast from float to int\n",
				n, cast_type_int->GetSigned() ? 's' : 'u', GetNode()->GetExpression()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(),
				en, GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str());
			return source;
		}
	}
	else if (cast_type_float){ //we're casting to a float
		if (expr_type_int){ //we're casting from an int to a float
			source += PrintString("%%.%u = %citofp %s %%.%u to %s ; cast from int to float\n",
				n, expr_type_int->GetSigned() ? 's' : 'u', GetNode()->GetExpression()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(),
				en, GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str());
			return source;
		}
		else if (expr_type_float){ //we're casting from a float to a float
			if (expr_type_float->GetWidth() < cast_type_float->GetWidth()){
				source += PrintString("%%.%u = %%.%u ; actually, they're the same type (float)\n", n, en);
				return source;
			}
			else {
				source += PrintString("%%.%u = fp%s %s %%.%u to %s ; cast from float to float\n",
					n, (expr_type_float->GetWidth() < cast_type_float->GetWidth()) ? "ext" : "trunc",
					GetNode()->GetExpression()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), en,
					GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str());
				return source;
			}
		}
	}
	throw FormatCompilerError(GetNode(), "Cannot cast from type %s (%s) to type %s (%s)",
		GetNode()->GetExpression()->GetType()->GetName().c_str(), GetNode()->GetExpression()->GetType()->GetSourceLocation().c_str(),
		GetNode()->GetType()->GetName().c_str(), GetNode()->GetType()->GetSourceLocation().c_str());
}

std::string CodeGeneratorExpressionCastLlvm::GenerateFnBegin(CompileContext &context, unsigned int d)
{
	return GetNode()->GetExpression()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateFnBegin(context, d);
}

std::string CodeGeneratorExpressionCastLlvm::GenerateFnEnd(CompileContext &context, unsigned int d)
{
	return GetNode()->GetExpression()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateFnEnd(context, d);
}

std::string CodeGeneratorExpressionCastLlvm::GenerateTopLevel(CompileContext &context, unsigned int d)
{
	return GetNode()->GetExpression()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateTopLevel(context, d);
}

std::string CodeGeneratorExpressionBinaryLlvm::Generate(CompileContext &context, unsigned int d, unsigned int n)
{
	std::string source;
	unsigned int a = GetUniqueId();
	unsigned int b = GetUniqueId();
	source += GetNode()->GetA()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->Generate(context, d, a);
	source += GetNode()->GetB()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->Generate(context, d, b);
	
	return source + Generate(context, d, GetNode()->GetType(), n, a, b, GetNode()->GetId());
}

std::string CodeGeneratorExpressionBinaryLlvm::GenerateFnBegin(CompileContext &context, unsigned int d)
{
	return GetNode()->GetA()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateFnBegin(context, d) +
		GetNode()->GetB()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateFnBegin(context, d);
}

std::string CodeGeneratorExpressionBinaryLlvm::GenerateFnEnd(CompileContext &context, unsigned int d)
{
	return GetNode()->GetA()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateFnEnd(context, d) +
		GetNode()->GetB()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateFnEnd(context, d);
}

std::string CodeGeneratorExpressionBinaryLlvm::GenerateTopLevel(CompileContext &context, unsigned int d)
{
	return GetNode()->GetA()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateTopLevel(context, d) +
		GetNode()->GetB()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateTopLevel(context, d);
}

std::string CodeGeneratorExpressionBinaryLlvm::GetFloatFlags(AstNodeType *type)
{
	return tracked_cast<AstNodeTypeFloat>(type)->GetIsStrict() ? "" : " ";
}



CodeGeneratorExpressionBinaryAdditionLlvm::CodeGeneratorExpressionBinaryAdditionLlvm(AstNodeExpressionBinaryAddition *node) :
	CodeGeneratorExpressionBinaryLlvm(node)
{
}

CodeGeneratorExpressionBinaryAdditionLlvm *CodeGeneratorExpressionBinaryAdditionLlvm::Copy()
{
	return new CodeGeneratorExpressionBinaryAdditionLlvm(GetNode());
}

std::string CodeGeneratorExpressionBinaryAdditionLlvm::Generate(CompileContext &context, unsigned int d, AstNodeType *type, unsigned int n, unsigned int a, unsigned int b, unsigned int id)
{
	if (dynamic_cast<AstNodeTypeFloat *>(type)){
		return PrintString("%%.%u = fadd%s %s %%.%u, %%.%u ; floating point add\n",
			n, GetFloatFlags(type).c_str(), type->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), a, b);
	}
	else if (dynamic_cast<AstNodeTypeInt *>(type)){
		return PrintString("%%.%u = add %s %%.%u, %%.%u ; integer add\n", n, type->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), a, b);
	}
	else {
		throw FormatCompilerError(GetNode(), "Unable to perform operation on %s\n", GetNode()->GetType()->GetName().c_str());
	}
}


CodeGeneratorExpressionBinarySubtractionLlvm::CodeGeneratorExpressionBinarySubtractionLlvm(AstNodeExpressionBinarySubtraction *node) :
	CodeGeneratorExpressionBinaryLlvm(node)
{
}

CodeGeneratorExpressionBinarySubtractionLlvm *CodeGeneratorExpressionBinarySubtractionLlvm::Copy()
{
	return new CodeGeneratorExpressionBinarySubtractionLlvm(GetNode());
}

std::string CodeGeneratorExpressionBinarySubtractionLlvm::Generate(CompileContext &context, unsigned int d, AstNodeType *type, unsigned int n, unsigned int a, unsigned int b, unsigned int id)
{
	if (dynamic_cast<AstNodeTypeFloat *>(type)){
		return PrintString("%%.%u = fsub %s %%.%u, %%.%u\n", n, type->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), a, b);
	}
	else if (dynamic_cast<AstNodeTypeInt *>(type)){
		return PrintString("%%.%u = sub %s %%.%u, %%.%u\n", n, type->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), a, b);
	}
	else {
		throw FormatCompilerError(GetNode(), "Unable to perform operation on %s\n", GetNode()->GetType()->GetName().c_str());
	}
}


CodeGeneratorExpressionBinaryMultiplicationLlvm::CodeGeneratorExpressionBinaryMultiplicationLlvm(AstNodeExpressionBinaryMultiplication *node) :
	CodeGeneratorExpressionBinaryLlvm(node)
{
}

CodeGeneratorExpressionBinaryMultiplicationLlvm *CodeGeneratorExpressionBinaryMultiplicationLlvm::Copy()
{
	return new CodeGeneratorExpressionBinaryMultiplicationLlvm(GetNode());
}

std::string CodeGeneratorExpressionBinaryMultiplicationLlvm::Generate(CompileContext &context, unsigned int d, AstNodeType *type, unsigned int n, unsigned int a, unsigned int b, unsigned int id)
{
	if (dynamic_cast<AstNodeTypeFloat *>(type)){
		return PrintString("%%.%u = fmul %s %%.%u, %%.%u\n", n, type->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), a, b);
	}
	else if (dynamic_cast<AstNodeTypeInt *>(type)){
		return PrintString("%%.%u = mul %s %%.%u, %%.%u\n", n, type->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), a, b);
	}
	else {
		throw FormatCompilerError(GetNode(), "Unable to perform operation on %s\n", GetNode()->GetType()->GetName().c_str());
	}
}


CodeGeneratorExpressionBinaryDivisionLlvm::CodeGeneratorExpressionBinaryDivisionLlvm(AstNodeExpressionBinaryDivision *node) :
	CodeGeneratorExpressionBinaryLlvm(node)
{
}

CodeGeneratorExpressionBinaryDivisionLlvm *CodeGeneratorExpressionBinaryDivisionLlvm::Copy()
{
	return new CodeGeneratorExpressionBinaryDivisionLlvm(GetNode());
}

std::string CodeGeneratorExpressionBinaryDivisionLlvm::Generate(CompileContext &context, unsigned int d, AstNodeType *type, unsigned int n, unsigned int a, unsigned int b, unsigned int id)
{
	if (dynamic_cast<AstNodeTypeFloat *>(type)){
		return PrintString("%%.%u = fdiv %s %%.%u, %%.%u\n", n, type->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), a, b);
	}
	else if (dynamic_cast<AstNodeTypeInt *>(type)){
		return PrintString("%%.%u = udiv %s %%.%u, %%.%u\n", n, type->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), a, b);
	}
	else {
		throw FormatCompilerError(GetNode(), "Unable to perform operation on %s\n", GetNode()->GetType()->GetName().c_str());
	}
}


CodeGeneratorExpressionBinaryModuloLlvm::CodeGeneratorExpressionBinaryModuloLlvm(AstNodeExpressionBinaryModulo *node) :
	CodeGeneratorExpressionBinaryLlvm(node)
{
}

CodeGeneratorExpressionBinaryModuloLlvm *CodeGeneratorExpressionBinaryModuloLlvm::Copy()
{
	return new CodeGeneratorExpressionBinaryModuloLlvm(GetNode());
}

std::string CodeGeneratorExpressionBinaryModuloLlvm::Generate(CompileContext &context, unsigned int d, AstNodeType *type, unsigned int n, unsigned int a, unsigned int b, unsigned int id)
{
	if (dynamic_cast<AstNodeTypeFloat *>(type)){
		return PrintString("%%.%u = frem %s %%.%u, %%.%u\n", n, type->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), a, b);
	}
	else if (dynamic_cast<AstNodeTypeInt *>(type)){
		return PrintString("%%.%u = urem %s %%.%u, %%.%u\n", n, type->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), a, b);
	}
	else {
		throw FormatCompilerError(GetNode(), "Unable to perform operation on %s\n", GetNode()->GetType()->GetName().c_str());
	}
}


CodeGeneratorExpressionBinaryAndLlvm::CodeGeneratorExpressionBinaryAndLlvm(AstNodeExpressionBinaryAnd *node) :
	CodeGeneratorExpressionBinaryLlvm(node)
{
}

CodeGeneratorExpressionBinaryAndLlvm *CodeGeneratorExpressionBinaryAndLlvm::Copy()
{
	return new CodeGeneratorExpressionBinaryAndLlvm(GetNode());
}

std::string CodeGeneratorExpressionBinaryAndLlvm::Generate(CompileContext &context, unsigned int d, AstNodeType *type, unsigned int n, unsigned int a, unsigned int b, unsigned int id)
{
	if (dynamic_cast<AstNodeTypeBool *>(type)){
		return PrintString("%%.%u = and i1 %%.%u, %%.%u\n", n, a, b);
	}
	else if (dynamic_cast<AstNodeTypeInt *>(type)){
		return PrintString("%%.%u = and %s %%.%u, %%.%u\n", n, type->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), a, b);
	}
	else {
		throw FormatCompilerError(GetNode(), "Unable to perform and operation on %s\n", GetNode()->GetType()->GetName().c_str());
	}
}


CodeGeneratorExpressionBinaryOrLlvm::CodeGeneratorExpressionBinaryOrLlvm(AstNodeExpressionBinaryOr *node) :
	CodeGeneratorExpressionBinaryLlvm(node)
{
}

CodeGeneratorExpressionBinaryOrLlvm *CodeGeneratorExpressionBinaryOrLlvm::Copy()
{
	return new CodeGeneratorExpressionBinaryOrLlvm(GetNode());
}

std::string CodeGeneratorExpressionBinaryOrLlvm::Generate(CompileContext &context, unsigned int d, AstNodeType *type, unsigned int n, unsigned int a, unsigned int b, unsigned int id)
{
	if (dynamic_cast<AstNodeTypeBool *>(type)){
		return PrintString("%%.%u = or i1 %%.%u, %%.%u\n", n, a, b);
	}
	else if (dynamic_cast<AstNodeTypeInt *>(type)){
		return PrintString("%%.%u = or %s %%.%u, %%.%u\n", n, type->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), a, b);
	}
	else {
		throw FormatCompilerError(GetNode(), "Unable to perform or operation on %s\n", GetNode()->GetType()->GetName().c_str());
	}
}


CodeGeneratorExpressionBinaryXorLlvm::CodeGeneratorExpressionBinaryXorLlvm(AstNodeExpressionBinaryXor *node) :
	CodeGeneratorExpressionBinaryLlvm(node)
{
}

CodeGeneratorExpressionBinaryXorLlvm *CodeGeneratorExpressionBinaryXorLlvm::Copy()
{
	return new CodeGeneratorExpressionBinaryXorLlvm(GetNode());
}

std::string CodeGeneratorExpressionBinaryXorLlvm::Generate(CompileContext &context, unsigned int d, AstNodeType *type, unsigned int n, unsigned int a, unsigned int b, unsigned int id)
{
	if (dynamic_cast<AstNodeTypeBool *>(type)){
		return PrintString("%%.%u = xor i1 %%.%u, %%.%u\n", n, a, b);
	}
	else if (dynamic_cast<AstNodeTypeInt *>(type)){
		return PrintString("%%.%u = xor %s %%.%u, %%.%u\n", n, type->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), a, b);
	}
	else {
		throw FormatCompilerError(GetNode(), "Unable to perform xor operation on %s\n", GetNode()->GetType()->GetName().c_str());
	}
}


CodeGeneratorExpressionBinaryShlLlvm::CodeGeneratorExpressionBinaryShlLlvm(AstNodeExpressionBinaryShl *node) :
	CodeGeneratorExpressionBinaryLlvm(node)
{
}

CodeGeneratorExpressionBinaryShlLlvm *CodeGeneratorExpressionBinaryShlLlvm::Copy()
{
	return new CodeGeneratorExpressionBinaryShlLlvm(GetNode());
}

std::string CodeGeneratorExpressionBinaryShlLlvm::Generate(CompileContext &context, unsigned int d, AstNodeType *type, unsigned int n, unsigned int a, unsigned int b, unsigned int id)
{
	if (dynamic_cast<AstNodeTypeInt *>(type)){
		return PrintString("%%.%u = shl %s %%.%u, %%.%u\n", n, type->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), a, b);
	}
	else {
		throw FormatCompilerError(GetNode(), "Unable to perform shl operation on %s\n", GetNode()->GetType()->GetName().c_str());
	}
}


CodeGeneratorExpressionBinaryShrLlvm::CodeGeneratorExpressionBinaryShrLlvm(AstNodeExpressionBinaryShr *node) :
	CodeGeneratorExpressionBinaryLlvm(node)
{
}

CodeGeneratorExpressionBinaryShrLlvm *CodeGeneratorExpressionBinaryShrLlvm::Copy()
{
	return new CodeGeneratorExpressionBinaryShrLlvm(GetNode());
}

std::string CodeGeneratorExpressionBinaryShrLlvm::Generate(CompileContext &context, unsigned int d, AstNodeType *type, unsigned int n, unsigned int a, unsigned int b, unsigned int id)
{
	if (dynamic_cast<AstNodeTypeInt *>(type)){
		return PrintString("%%.%u = lshr %s %%.%u, %%.%u\n", n, type->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), a, b);
	}
	else {
		throw FormatCompilerError(GetNode(), "Unable to perform shr operation on %s\n", GetNode()->GetType()->GetName().c_str());
	}
}


CodeGeneratorExpressionBinaryCompareLlvm::CodeGeneratorExpressionBinaryCompareLlvm(AstNodeExpressionBinaryCompare *node, std::string llvmop, bool uses_sign) :
	CodeGeneratorExpressionBinaryLlvm(node), m_llvmop(llvmop), m_uses_sign(uses_sign)
{
}

CodeGeneratorExpressionBinaryCompareLlvm *CodeGeneratorExpressionBinaryCompareLlvm::Copy()
{
	return new CodeGeneratorExpressionBinaryCompareLlvm(GetNode(), m_llvmop);
}

std::string CodeGeneratorExpressionBinaryCompareLlvm::Generate(CompileContext &context, unsigned int d, unsigned int n)
{
	std::string source;
	
	AstNodeTypeBool *a_type_bool = dynamic_cast<AstNodeTypeBool *>(GetNode()->GetA()->GetType());
	AstNodeTypeInt *a_type_int = dynamic_cast<AstNodeTypeInt *>(GetNode()->GetA()->GetType());
	AstNodeTypeFloat *a_type_float = dynamic_cast<AstNodeTypeFloat *>(GetNode()->GetA()->GetType());
	
	AstNodeTypeBool *b_type_bool = dynamic_cast<AstNodeTypeBool *>(GetNode()->GetB()->GetType());
	AstNodeTypeInt *b_type_int = dynamic_cast<AstNodeTypeInt *>(GetNode()->GetB()->GetType());
	AstNodeTypeFloat *b_type_float = dynamic_cast<AstNodeTypeFloat *>(GetNode()->GetB()->GetType());
	
	unsigned int a = GetUniqueId();
	unsigned int b = GetUniqueId();
	
	source += GetNode()->GetA()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->Generate(context, d, a);
	source += GetNode()->GetB()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->Generate(context, d, b);
	
	const char *signed_string = m_uses_sign ? "s" : "";
	const char *unsigned_string = m_uses_sign ? "u" : "";
	
	if (a_type_int && b_type_int){
		source += PrintString("%%.%u = icmp %s%s %s %%.%u, %%.%u ; integer comparison: %s\n",
			n, a_type_int->GetSigned() ? signed_string : unsigned_string, m_llvmop.c_str(),
			a_type_int->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), a, b, GetNode()->GetName().c_str());
		return source;
	}
	else if (a_type_float && b_type_float){
		source += PrintString("%%.%u = fcmp o%s %s %%.%u, %%.%u ; floating point comparison: %s\n",
			n, m_llvmop.c_str(), a_type_float->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), a, b, GetNode()->GetName().c_str());
		return source;
	}
	else if (a_type_bool && b_type_bool){
		source += PrintString("%%.%u = icmp u%s %s %%.%u, %%.%u ; integer comparison: %s\n",
			n, m_llvmop.c_str(), a_type_bool->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), a, b, GetNode()->GetName().c_str());
		return source;
	}
	throw FormatCompilerError(GetNode(), "Comparison of %s and %s not supported",
		GetNode()->GetA()->GetType()->GetName().c_str(), GetNode()->GetB()->GetType()->GetName().c_str());
}

std::string CodeGeneratorExpressionBinaryCompareLlvm::Generate(CompileContext &context, unsigned int d, AstNodeType *type, unsigned int n, unsigned int a, unsigned int b, unsigned int id)
{
	throw FormatCompilerError(GetNode(), "Vector comparison not yet supported");
}


