/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../../ast/astnode.hpp"
#include "codegen.hpp"

CodeGeneratorStatementParallelLlvm::CodeGeneratorStatementParallelLlvm(AstNodeStatementParallel *node) :
	CodeGeneratorStatementLlvm(node)
{
}

CodeGeneratorStatementParallelLlvm *CodeGeneratorStatementParallelLlvm::Copy()
{
	return new CodeGeneratorStatementParallelLlvm(GetNode());
}

std::string CodeGeneratorStatementParallelLlvm::Generate(CompileContext &context, unsigned int d)
{
	return GetNode()->GetStatement()->GetCodeGenerator<CodeGeneratorStatementLlvm>()->Generate(context, d);
}

std::string CodeGeneratorStatementParallelLlvm::GenerateFnBegin(CompileContext &context, unsigned int d)
{
	std::string source = GetNode()->GetStatement()->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateFnBegin(context, d);
	return PrintString("; Parallel statement at line %u\n", GetNode()->GetLine()) + Indent(source, 1);
}

std::string CodeGeneratorStatementParallelLlvm::GenerateFnEnd(CompileContext &context, unsigned int d)
{
	std::string source = GetNode()->GetStatement()->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateFnEnd(context, d);
	return PrintString("; Parallel statement at line %u\n", GetNode()->GetLine()) + Indent(source, 1);
}

std::string CodeGeneratorStatementParallelLlvm::GenerateTopLevel(CompileContext &context, unsigned int d)
{
	std::string source = GetNode()->GetStatement()->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateTopLevel(context, d);
	return PrintString("; Parallel statement at line %u\n", GetNode()->GetLine()) + Indent(source, 1);
}


CodeGeneratorStatementParallelSyncLlvm::CodeGeneratorStatementParallelSyncLlvm(AstNodeStatementParallelSync *node) :
	CodeGeneratorStatementParallelLlvm(node)
{
}

CodeGeneratorStatementParallelSyncLlvm *CodeGeneratorStatementParallelSyncLlvm::Copy()
{
	return new CodeGeneratorStatementParallelSyncLlvm(tracked_cast<AstNodeStatementParallelSync>(GetNode()));
}

std::string CodeGeneratorStatementParallelSyncLlvm::Generate(CompileContext &context, unsigned int d)
{
	std::string source = PrintString("call fastcc void @.sync_start() ; call the sync start function\n"
		"%s\n"
		"call fastcc void @.sync_end() ; call the sync end function\n",
		GetNode()->GetStatement()->GetCodeGenerator<CodeGeneratorStatementLlvm>()->Generate(context, d).c_str());
	return PrintString("; Sync statement at line %u\n", GetNode()->GetLine()) + Indent(source, 1);
}

std::string CodeGeneratorStatementParallelSyncLlvm::GenerateFnBegin(CompileContext &context, unsigned int d)
{
	std::string source = GetNode()->GetStatement()->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateFnBegin(context, d);
	return PrintString("; Sync statement at line %u\n", GetNode()->GetLine()) + Indent(source, 1);
}

std::string CodeGeneratorStatementParallelSyncLlvm::GenerateFnEnd(CompileContext &context, unsigned int d)
{
	std::string source = GetNode()->GetStatement()->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateFnEnd(context, d);
	return PrintString("; Sync statement at line %u\n", GetNode()->GetLine()) + Indent(source, 1);
}

std::string CodeGeneratorStatementParallelSyncLlvm::GenerateTopLevel(CompileContext &context, unsigned int d)
{
	std::string source = GetNode()->GetStatement()->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateTopLevel(context, d);
	return PrintString("; Sync statement at line %u\n", GetNode()->GetLine()) + Indent(source, 1);
}



CodeGeneratorStatementVariableDeclarationForSpawnLlvm::CodeGeneratorStatementVariableDeclarationForSpawnLlvm(AstNodeStatementVariableDeclaration *node) :
	CodeGeneratorStatementVariableDeclarationLlvm(node)
{
}

CodeGeneratorStatementVariableDeclarationForSpawnLlvm *CodeGeneratorStatementVariableDeclarationForSpawnLlvm::Copy()
{
	return new CodeGeneratorStatementVariableDeclarationForSpawnLlvm(GetNode());
}

std::string CodeGeneratorStatementVariableDeclarationForSpawnLlvm::Generate(CompileContext &context, unsigned int d)
{
	return "";
}

std::string CodeGeneratorStatementVariableDeclarationForSpawnLlvm::GenerateFnBegin(CompileContext &context, unsigned int d)
{
	return "";
}

std::string CodeGeneratorStatementVariableDeclarationForSpawnLlvm::GenerateFnEnd(CompileContext &context, unsigned int d)
{
	return "";
}

std::string CodeGeneratorStatementVariableDeclarationForSpawnLlvm::GenerateTopLevel(CompileContext &context, unsigned int d)
{
	return "";
}



CodeGeneratorStatementParallelSpawnExtractedLlvm::CodeGeneratorStatementParallelSpawnExtractedLlvm
	(AstNodeStatementParallelSpawn *node, std::list<AstNodeStatementVariableDeclaration *> variables, unsigned int linkid) :
	CodeGeneratorStatementParallelLlvm(node), m_variables(variables), m_linkid(linkid)
{
}

CodeGeneratorStatementParallelSpawnExtractedLlvm *CodeGeneratorStatementParallelSpawnExtractedLlvm::Copy()
{
	return new CodeGeneratorStatementParallelSpawnExtractedLlvm(tracked_cast<AstNodeStatementParallelSpawn>(GetNode()), m_variables, m_linkid);
}

std::string CodeGeneratorStatementParallelSpawnExtractedLlvm::Generate(CompileContext &context, unsigned int d)
{
	std::string source;
	
	
	/* Create and store the structure */
	unsigned int i = 0;
	for (auto variable: m_variables){
		unsigned int variable_id = variable->GetId();
		std::string identifier = variable->GetIdentifier();
		std::string typecode = variable->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
		
		//Get a pointer to the element of the structure
		source += PrintString("%%spawn.structure_element.%u.%s = getelementptr %%spawn.type.%u* %%spawn.structure_storage.%u, i32 0, i32 %u\n",
			m_linkid, identifier.c_str(), m_linkid, m_linkid, i);
		
		/* Copy all the variables that need copying into an SSA register */
		if (variable->GetCopyOnSpawn()){
			source += PrintString("%%spawn.structure_copy.%u.%s = load %s* %%.%u\n",
				m_linkid, identifier.c_str(),
				typecode.c_str(),
				variable_id);
			
			source += PrintString("store %s %%spawn.structure_copy.%u.%s, %s* %%spawn.structure_element.%u.%s\n",
				typecode.c_str(), m_linkid, identifier.c_str(), typecode.c_str(), m_linkid, identifier.c_str());
		}
		else {
			source += PrintString("store %s* %%.%u, %s** %%spawn.structure_element.%u.%s\n",
				typecode.c_str(), variable_id, typecode.c_str(), m_linkid, identifier.c_str());
		}
		
		i++;
	}
	
	/* Perform the library call */
	source += PrintString("%%spawn.structure_size_pointer.%u = getelementptr %%spawn.type.%u* null, i32 1\n",
		m_linkid, m_linkid); //Get the length of the structure (as a pointer)
	source += PrintString("%%spawn.structure_size_value.%u = ptrtoint %%spawn.type.%u* %%spawn.structure_size_pointer.%u to %%t.size\n",
		m_linkid, m_linkid, m_linkid); //Get the length of the structure as an integer
	source += PrintString("%%spawn.structure_i8.%u = bitcast %%spawn.type.%u* %%spawn.structure_storage.%u to i8*\n",
		m_linkid, m_linkid, m_linkid); //Get the pointer as an i8* pointer
	source += PrintString("call fastcc void (void (i8*)*, i8*, %%t.size)* "
		"@.enqueue(void (i8*)* @spawn.%u, i8* %%spawn.structure_i8.%u, %%t.size %%spawn.structure_size_value.%u)\n",
		m_linkid, m_linkid, m_linkid);
	
	return source;
}

std::string CodeGeneratorStatementParallelSpawnExtractedLlvm::GenerateFnBegin(CompileContext &context, unsigned int d)
{
	return PrintString("%%spawn.structure_storage.%u = alloca %%spawn.type.%u\n", m_linkid, m_linkid);
}

std::string CodeGeneratorStatementParallelSpawnExtractedLlvm::GenerateFnEnd(CompileContext &context, unsigned int d)
{
	return "";
}

std::string CodeGeneratorStatementParallelSpawnExtractedLlvm::GenerateTopLevel(CompileContext &context, unsigned int d)
{
	return "";
}
	



CodeGeneratorStatementExtractedLlvm::CodeGeneratorStatementExtractedLlvm(AstNodeStatementExtracted *node) :
	CodeGeneratorStatementLlvm(node)
{
}

CodeGeneratorStatementExtractedLlvm *CodeGeneratorStatementExtractedLlvm::Copy()
{
	std::list<AstNodeStatementVariableDeclaration *> variables;
	for (AstNodeStatementVariableDeclaration *variable: GetNode()->GetVariables()){
		variables.push_back(variable->Copy());
	}
	
	//remember with this copy, we really only use the GetNode()->GetLinkId() in place of the function name, so it should be OK (also, we don't generate GetNode()->GetLinkId() in the constructor)
	return new CodeGeneratorStatementExtractedLlvm(GetNode());
}

std::string CodeGeneratorStatementExtractedLlvm::Generate(CompileContext &context, unsigned int d)
{
	return GetNode()->GetStatement()->GetCodeGenerator<CodeGeneratorStatementLlvm>()->Generate(context, d);
}

std::string CodeGeneratorStatementExtractedLlvm::GenerateFnBegin(CompileContext &context, unsigned int d)
{
	std::string source;
	
	source += PrintString("%%.variables = bitcast i8* %%.spawn.argument.%u to %%spawn.type.%u*\n", GetNode()->GetLinkId(), GetNode()->GetLinkId());
	
	//load the variables into suitable places
	unsigned int index = 0;
	for (auto variable: GetNode()->GetVariables()){
		if (variable->GetCopyOnSpawn()){
			source += PrintString("%%.%u = getelementptr %%spawn.type.%u* %%.variables, i32 0, i32 %u\n", variable->GetId(), GetNode()->GetLinkId(), index);
		}
		else {
			source += PrintString("%%spawn.argument_ptr.%u.%s = getelementptr %%spawn.type.%u* %%.variables, i32 0, i32 %u\n",
				GetNode()->GetLinkId(), variable->GetIdentifier().c_str(), GetNode()->GetLinkId(), index);
			source += PrintString("%%.%u = load %s** %%spawn.argument_ptr.%u.%s\n", 
				variable->GetId(),
				variable->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(),
				GetNode()->GetLinkId(),
				variable->GetIdentifier().c_str());
		}
		index++;
	}
	source += "\n\n";
	
	source += GetNode()->GetStatement()->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateFnBegin(context, d);
	
	return source;
}

std::string CodeGeneratorStatementExtractedLlvm::GenerateFnEnd(CompileContext &context, unsigned int d)
{
	return GetNode()->GetStatement()->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateFnEnd(context, d);
}

std::string CodeGeneratorStatementExtractedLlvm::GenerateTopLevel(CompileContext &context, unsigned int d)
{
	unsigned int type_count = 0;
	std::string source = PrintString("; Generated by the spawn optimisation\n%%spawn.type.%u = type { ", GetNode()->GetLinkId());
	for (auto variable: GetNode()->GetVariables()){
		if (type_count){
			source += ", ";
		}
		
		source += variable->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
		if (!variable->GetCopyOnSpawn()){ //this must be a pointer to that variable for it not to have been copied
			source += "*";
		}
		
		type_count++;
	}
	source += " }\n";
	
	source += GetNode()->GetStatement()->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateTopLevel(context, d);
	
	return source;
}



CodeGeneratorTypeLockLlvm::CodeGeneratorTypeLockLlvm(AstNodeTypeLock *node) :
	CodeGeneratorTypeLlvm(node)
{
}

CodeGeneratorTypeLockLlvm *CodeGeneratorTypeLockLlvm::Copy()
{
	return new CodeGeneratorTypeLockLlvm(GetNode());
}

std::string CodeGeneratorTypeLockLlvm::Generate(CompileContext &context, unsigned int d)
{
	/* Ideally, this would be %s, not %s*, but that would mean changing it so we can load/store only part of an lvalue */
	return PrintString("{%%t.lock, %s*}", GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str());
}

unsigned int CodeGeneratorTypeLockLlvm::GetSize(CompileContext &context)
{
	return context.GetComputeSpec().GetCpu().GetSizeT() + GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GetSize(context);
}

std::string CodeGeneratorTypeLockLlvm::GenerateDebug(CompileContext &context, unsigned int d)
{
	return "";
}

std::string CodeGeneratorTypeLockLlvm::GenerateCreate(CompileContext &context, unsigned int d, unsigned int n)
{
	unsigned int r1 = GetUniqueId();
	unsigned int value = GetUniqueId();
	unsigned int lock = GetUniqueId();
	unsigned int memory_i8 = GetUniqueId();
	unsigned int memory_type = GetUniqueId();
	
	std::string lvalue_type = GetNode()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
	std::string child_type = GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
	std::string create_code = GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateCreate(context, d, value);
	
	std::string source = PrintString(
		"; Create lock and data %s\n"
			"\t%%.%u = call fastcc %%t.lock ()* @.lock.create()%s ; Create the lock\n"
			"\t%%.%u = call fastcc i8* (%%t.size, %%t.size)* @.malloc(%%t.size 1, %%t.size %u)%s ; Allocate memory for the value\n"
			"\t%%.%u = bitcast i8* %%.%u to %s* ; Cast the pointer\n"
			
			"\t%%.%u = insertvalue %s undef, %%t.lock %%.%u, 0 ; Insert the lock\n"
			"\t%%.%u = insertvalue %s %%.%u, %s* %%.%u, 1 ; Insert the pointer\n",
			
		GetNode()->GetSourceLocation().c_str(),
			lock, GetDebug(context, d).c_str(),
			memory_i8, GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GetSize(context), GetDebug(context, d).c_str(),
			memory_type, memory_i8, child_type.c_str(),
			
			r1, lvalue_type.c_str(), lock,
			n, lvalue_type.c_str(), r1, child_type.c_str(), memory_type
	);
	
	if (create_code.size()){
		source += PrintString(
			"\t; Create the value\n"
				"%s"
			"\tstore %s %%.%u, %s* %%.%u%s ; Store the value\n",
			
				Indent(create_code, 2).c_str(),
			child_type.c_str(), value, child_type.c_str(), memory_type, GetDebug(context, d).c_str()
		);
	}
	
	return source;
}

std::string CodeGeneratorTypeLockLlvm::GenerateDestroy(CompileContext &context, unsigned int d, unsigned int n)
{
	unsigned int value = GetUniqueId();
	unsigned int lock = GetUniqueId();
	unsigned int memory_i8 = GetUniqueId();
	unsigned int memory_type = GetUniqueId();
	
	std::string destroy_code = GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateDestroy(context, d, value);
	std::string lvalue_type = GetNode()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
	std::string child_type = GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
	
	return PrintString(
		"; Destroy the lock and data %s\n"
			"\t%%.%u = extractvalue %s %%.%u, 0 ; Extract the lock\n"
			"\t%%.%u = extractvalue %s %%.%u, 1 ; Extract the pointer\n"
			"\t%%.%u = bitcast %s* %%.%u to i8* ; Cast the pointer\n"
			
			"\tcall fastcc void @.lock.destroy(%%t.lock %%.%u)%s ; Destroy the lock\n"
			"\t%%.%u = load %s* %%.%u%s ; Load the value so we can destroy it\n"
			"\t; Destroy the value\n"
				"%s"
			"\tcall fastcc void @.free(i8* %%.%u) ; Free the memory\n",
		
		GetNode()->GetSourceLocation().c_str(),
			lock, lvalue_type.c_str(), n,
			memory_type, lvalue_type.c_str(), n,
			memory_i8, child_type.c_str(), memory_type,
			
			lock, GetDebug(context, d).c_str(),
			value, child_type.c_str(), memory_type, GetDebug(context, d).c_str(),
				Indent(destroy_code, 2).c_str(),
			memory_i8
	);
}

std::string CodeGeneratorTypeLockLlvm::GenerateCopy(CompileContext &context, unsigned int d, unsigned int src, unsigned int dst)
{
	throw FormatCompilerError(GetNode(), "Internal error: We shouldn't be copying a lock!");
}




CodeGeneratorStatementParallelAtomicLlvm::CodeGeneratorStatementParallelAtomicLlvm(AstNodeStatementParallelAtomic *node) :
	CodeGeneratorStatementParallelLlvm(node), m_unlocked_by_atomic(false)
{
}

//TODO: Add to header.ll the code to lock and unlock a list of locks
/*
 * I guess functions have an implicit bool covering whether they were called from in an atomic or not
 * as well as a thingiemablob to get a list of all the locks that aren't local to the function (e.g from references, etc),
 * and atomics pass their lists up to parent atomics
 */
CodeGeneratorStatementParallelAtomicLlvm *CodeGeneratorStatementParallelAtomicLlvm::Copy()
{
	CodeGeneratorStatementParallelAtomicLlvm *copy = new CodeGeneratorStatementParallelAtomicLlvm(GetNode());
	copy->m_unlocked_by_atomic = m_unlocked_by_atomic;
	return copy;
}

std::string CodeGeneratorStatementParallelAtomicLlvm::Generate(CompileContext &context, unsigned int d)
{
	std::string source;
	unsigned int tsize_list = GetUniqueId();
	unsigned int lock_list = GetUniqueId();
	
	if (!m_unlocked_by_atomic){
		source += PrintString(
			"%%.%u = call fastcc %%t.size* ()* @.get_tsize() ; Get some memory of %%t.size (pointer size)\n"
			"%%.%u = bitcast %%t.size* %%.%u to %%t.lock.list** ; Bitcast the pointer\n"
			"store %%t.lock.list* null, %%t.lock.list** %%.%u ; Store null in the list\n",
			
			tsize_list,
			lock_list, tsize_list,
			lock_list
		);
		
		std::list<AstNode *> children;
		GetNode()->GetStatement()->GetChildren(children, true, false);
		
		for (AstNode *node: children){
			if (dynamic_cast<AstNodeStatementParallelAtomic *>(node)){ //We're dealing with this node's children, so it doesn't need to do anything
				node->GetCodeGenerator<CodeGeneratorStatementParallelAtomicLlvm>()->SetUnlockedByAtomic(true);
			}
			if (dynamic_cast<AstNodeExpressionLValueLock *>(node)){
				unsigned int lock = GetUniqueId();
				unsigned int old_list = GetUniqueId();
				unsigned int new_list = GetUniqueId();
				
				node->GetCodeGenerator<CodeGeneratorExpressionLValueLockLlvm>()->SetUnlockedByAtomic(true);
				
				source += node->GetCodeGenerator<CodeGeneratorExpressionLValueLockLlvm>()->GenerateLock(context, d, lock);
				source += PrintString(
					"%%.%u = load %%t.lock.list** %%.%u%s ; Load the list\n"
					"%%.%u = call fastcc %%t.lock.list* (%%t.lock, %%t.lock.list*)* @.lock.list.malloc(%%t.lock %%.%u, %%t.lock.list* %%.%u)%s ; Add the lock to the list\n"
					"store %%t.lock.list* %%.%u, %%t.lock.list** %%.%u%s ; Store the new list pointer\n",
					
					old_list, lock_list, GetDebug(context, d).c_str(),
					new_list, lock, old_list, GetDebug(context, d).c_str(),
					new_list, lock_list, GetDebug(context, d).c_str()
				);
			}
			if (dynamic_cast<AstNodeFunctionCall *>(node)){
				//AstNodeFunctionCall *call = tracked_cast<AstNodeFunctionCall>(node);
				//....TODO
			}
		}
		
		unsigned int list = GetUniqueId();
		source += PrintString(
			"%%.%u = load %%t.lock.list** %%.%u%s ; Load the list\n"
			"call fastcc void @.lock.list.lock(%%t.lock.list* %%.%u)%s ; Lock the locks\n",
			
			list, lock_list, GetDebug(context, d).c_str(),
			list, GetDebug(context, d).c_str()
		);
	}
	
	source += GetNode()->GetStatement()->GetCodeGenerator<CodeGeneratorStatementLlvm>()->Generate(context, d);
	
	if (!m_unlocked_by_atomic){
		unsigned int list = GetUniqueId();
		source += PrintString(
			"%%.%u = load %%t.lock.list** %%.%u%s ; Load the list\n"
			"call fastcc void @.lock.list.unlock(%%t.lock.list* %%.%u)%s ; Unlock the locks\n"
			"call fastcc void @.lock.list.free(%%t.lock.list* %%.%u)%s ; Unlock the locks\n"
			"call fastcc void @.release_tsize(%%t.size* %%.%u)\n",
			
			list, lock_list, GetDebug(context, d).c_str(),
			list, GetDebug(context, d).c_str(),
			list, GetDebug(context, d).c_str(),
			tsize_list
		);
	}
	
	return PrintString("; Atomic %s\n", GetNode()->GetSourceLocation().c_str()) + Indent(source, 1);
}

std::string CodeGeneratorStatementParallelAtomicLlvm::GenerateFnBegin(CompileContext &context, unsigned int d)
{
	return GetNode()->GetStatement()->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateFnBegin(context, d);
}

std::string CodeGeneratorStatementParallelAtomicLlvm::GenerateFnEnd(CompileContext &context, unsigned int d)
{
	return GetNode()->GetStatement()->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateFnEnd(context, d);
}

std::string CodeGeneratorStatementParallelAtomicLlvm::GenerateTopLevel(CompileContext &context, unsigned int d)
{
	return GetNode()->GetStatement()->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateTopLevel(context, d);
}

bool CodeGeneratorStatementParallelAtomicLlvm::GetUnlockedByAtomic()
{
	return m_unlocked_by_atomic;
}

void CodeGeneratorStatementParallelAtomicLlvm::SetUnlockedByAtomic(bool unlocked_by_atomic)
{
	m_unlocked_by_atomic = unlocked_by_atomic;
}



CodeGeneratorExpressionLValueLockLlvm::CodeGeneratorExpressionLValueLockLlvm(AstNodeExpressionLValueLock *node) :
	CodeGeneratorExpressionLValueLlvm(node), m_unlocked_by_atomic(false)
{
}

CodeGeneratorExpressionLValueLockLlvm *CodeGeneratorExpressionLValueLockLlvm::Copy()
{
	CodeGeneratorExpressionLValueLockLlvm *copy = new CodeGeneratorExpressionLValueLockLlvm(GetNode());
	copy->m_unlocked_by_atomic = m_unlocked_by_atomic;
	return copy;
}

std::string CodeGeneratorExpressionLValueLockLlvm::GenerateReference(CompileContext &context, unsigned int d, unsigned int n)
{
	return GetNode()->GetLValue()->GetCodeGenerator<CodeGeneratorExpressionLValueLlvm>()->GenerateReference(context, d, n);
}

std::string CodeGeneratorExpressionLValueLockLlvm::GenerateStore(CompileContext &context, unsigned int d, unsigned int n)
{
	return GenerateLoadStore(context, d, n, true);
}

std::string CodeGeneratorExpressionLValueLockLlvm::Generate(CompileContext &context, unsigned int d, unsigned int n)
{
	return GenerateLoadStore(context, d, n, false);
}

std::string CodeGeneratorExpressionLValueLockLlvm::GenerateLoadStore(CompileContext &context, unsigned int d, unsigned int n, bool store)
{
	std::string source;
	std::string lvalue_type = GetNode()->GetLValue()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
	std::string child_type = tracked_cast<AstNodeTypeLock>(GetNode()->GetLValue()->GetType())->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
	
	unsigned int lock_pointer = GetUniqueId();
	unsigned int lock = GetUniqueId();
	unsigned int pointer = GetUniqueId();
	
	source += PrintString("; Generate Lock/Pointer %s\n", GetNode()->GetSourceLocation().c_str());
	source += Indent(GetNode()->GetLValue()->GetCodeGenerator<CodeGeneratorExpressionLValueLlvm>()->Generate(context, d, lock_pointer), 1);
	
	source += PrintString(
		"%%.%u = extractvalue %s %%.%u, 0 ; Extract the lock\n"
		"%%.%u = extractvalue %s %%.%u, 1 ; Extract a pointer to the data\n",
		
		lock, lvalue_type.c_str(), lock_pointer,
		pointer, lvalue_type.c_str(), lock_pointer
	);
	
	
	if (!m_unlocked_by_atomic){
		source += PrintString("call fastcc void @.lock.lock(%%t.lock %%.%u)%s ; Lock the lock\n", lock, GetDebug(context, d).c_str());
	}
	
	if (store){
		source += PrintString(
			"store %s %%.%u, %s* %%.%u ; Load the value\n",
			child_type.c_str(), n, child_type.c_str(), pointer
		);
	}
	else {
		source += PrintString(
			"%%.%u = load %s* %%.%u ; Load the value\n",
			n, child_type.c_str(), pointer
		);
	}
	
	if (!m_unlocked_by_atomic){
		source += PrintString("call fastcc void @.lock.unlock(%%t.lock %%.%u)%s ; Unlock the lock\n", lock, GetDebug(context, d).c_str());
	}
	
	return PrintString("; Generate Lock %s\n", GetNode()->GetSourceLocation().c_str()) + Indent(source, 1);
}

std::string CodeGeneratorExpressionLValueLockLlvm::GenerateLock(CompileContext &context, unsigned int d, unsigned int n)
{
	std::string source;
	std::string lvalue_type = GetNode()->GetLValue()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
	
	unsigned int lock_pointer = GetUniqueId();
	
	source += PrintString("; Generate Lock %s\n", GetNode()->GetSourceLocation().c_str());
	source += Indent(GetNode()->GetLValue()->GetCodeGenerator<CodeGeneratorExpressionLValueLlvm>()->Generate(context, d, lock_pointer), 1);
	
	source += PrintString(
		"%%.%u = extractvalue %s %%.%u, 0 ; Extract the lock\n",
		n, lvalue_type.c_str(), lock_pointer
	);
	
	return source;
}

std::string CodeGeneratorExpressionLValueLockLlvm::GenerateFnBegin(CompileContext &context, unsigned int d)
{
	return GetNode()->GetLValue()->GetCodeGenerator<CodeGeneratorExpressionLValueLlvm>()->GenerateFnBegin(context, d);
}

std::string CodeGeneratorExpressionLValueLockLlvm::GenerateFnEnd(CompileContext &context, unsigned int d)
{
	return GetNode()->GetLValue()->GetCodeGenerator<CodeGeneratorExpressionLValueLlvm>()->GenerateFnEnd(context, d);
}

std::string CodeGeneratorExpressionLValueLockLlvm::GenerateTopLevel(CompileContext &context, unsigned int d)
{
	return GetNode()->GetLValue()->GetCodeGenerator<CodeGeneratorExpressionLValueLlvm>()->GenerateTopLevel(context, d);
}

bool CodeGeneratorExpressionLValueLockLlvm::CopyIfCopyable()
{
	return GetNode()->GetLValue()->GetCodeGenerator<CodeGeneratorExpressionLValueLlvm>()->CopyIfCopyable();
}

bool CodeGeneratorExpressionLValueLockLlvm::GetUnlockedByAtomic()
{
	return m_unlocked_by_atomic;
}

void CodeGeneratorExpressionLValueLockLlvm::SetUnlockedByAtomic(bool unlocked_by_atomic)
{
	m_unlocked_by_atomic = unlocked_by_atomic;
}






