/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../../ast/astnode.hpp"
#include "codegen.hpp"

CodeGeneratorStatementLlvm::CodeGeneratorStatementLlvm(AstNodeStatement *node) :
	CodeGeneratorLlvm(node)
{
}

CodeGeneratorStatementReturnLlvm::CodeGeneratorStatementReturnLlvm(AstNodeStatementReturn *node) :
	CodeGeneratorStatementLlvm(node)
{
}

CodeGeneratorStatementReturnLlvm *CodeGeneratorStatementReturnLlvm::Copy()
{
	return new CodeGeneratorStatementReturnLlvm(GetNode());
}

std::string CodeGeneratorStatementReturnLlvm::Generate(CompileContext &context, unsigned int d)
{
	std::string source;
	std::string debug = GetDebug(context, d);
	if (GetNode()->GetExpression()){
		unsigned int i = GetUniqueId();
		source += GetNode()->GetExpression()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->Generate(context, d, i);
		
		std::string type = GetNode()->GetExpression()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
		
		source += PrintString("store %s %%.%u, %s* %%.function_return_value%s ; store the return value\n"
			"br label %%.function_end%s ; jump to the return\n",
			type.c_str(), i, type.c_str(), debug.c_str(), debug.c_str());
	}
	else {
		source += PrintString("br label %%.function_end%s ; Execute the return (void)\n", debug.c_str());
	}
	return PrintString("; Return statement at line %u\n", GetNode()->GetLine()) + Indent(source, 1);
}

std::string CodeGeneratorStatementReturnLlvm::GenerateFnBegin(CompileContext &context, unsigned int d)
{
	AstNodeExpression *expression = GetNode()->GetExpression();
	std::string source = expression ? expression->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateFnBegin(context, d) : "";
	return PrintString("; Return statement at line %u\n", GetNode()->GetLine()) + Indent(source, 1);
}

std::string CodeGeneratorStatementReturnLlvm::GenerateFnEnd(CompileContext &context, unsigned int d)
{
	AstNodeExpression *expression = GetNode()->GetExpression();
	std::string source = expression ? expression->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateFnEnd(context, d) : "";
	return PrintString("; Return statement at line %u\n", GetNode()->GetLine()) + Indent(source, 1);
}

std::string CodeGeneratorStatementReturnLlvm::GenerateTopLevel(CompileContext &context, unsigned int d)
{
	AstNodeExpression *expression = GetNode()->GetExpression();
	return PrintString("; Return statement at line %u\n", GetNode()->GetLine()) +
		Indent(expression ? expression->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateTopLevel(context, d) : "", 1);
}

CodeGeneratorStatementVariableAssignmentLlvm::CodeGeneratorStatementVariableAssignmentLlvm(AstNodeStatementVariableAssignment *node) :
	CodeGeneratorStatementLlvm(node)
{
}

CodeGeneratorStatementVariableAssignmentLlvm *CodeGeneratorStatementVariableAssignmentLlvm::Copy()
{
	return new CodeGeneratorStatementVariableAssignmentLlvm(GetNode());
}

std::string CodeGeneratorStatementVariableAssignmentLlvm::Generate(CompileContext &context, unsigned int d)
{
	unsigned int expression = GetUniqueId();
	unsigned int expression_copy = GetUniqueId();
	unsigned int lvalue_old = GetUniqueId();
	
	std::string copycode = GetNode()->GetExpression()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateCopy(context, d, expression, expression_copy);
	bool should_copy = (copycode.size() && GetNode()->GetExpression()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->CopyIfCopyable());
	
	return PrintString(
		"; Variable assignment at line %u:\n"
			"\t; Get the old lvalue value\n"
				"%s\n"
			"\t; Generate new expression\n"
				"%s\n"
			"\t; Destroy the old expression\n"
				"%s\n"
			"\t; Copy the new expression if necessary\n"
				"%s\n"
			"\t; Store the new expression\n"
				"%s\n",
		GetNode()->GetLine(),
		Indent(GetNode()->GetLValue()->GetCodeGenerator<CodeGeneratorExpressionLValueLlvm>()->Generate(context, d, lvalue_old), 2).c_str(),
		Indent(GetNode()->GetExpression()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->Generate(context, d, expression), 2).c_str(),
		Indent(GetNode()->GetLValue()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateDestroy(context, d, lvalue_old), 2).c_str(),
		should_copy ? Indent(copycode, 2).c_str() : "",
		Indent(GetNode()->GetLValue()->GetCodeGenerator<CodeGeneratorExpressionLValueLlvm>()->GenerateStore(context, d, should_copy ? expression_copy : expression), 2).c_str()
	);
}

std::string CodeGeneratorStatementVariableAssignmentLlvm::GenerateFnBegin(CompileContext &context, unsigned int d)
{
	return GetNode()->GetExpression()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateFnBegin(context, d);
}

std::string CodeGeneratorStatementVariableAssignmentLlvm::GenerateFnEnd(CompileContext &context, unsigned int d)
{
	return GetNode()->GetExpression()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateFnEnd(context, d);
}

std::string CodeGeneratorStatementVariableAssignmentLlvm::GenerateTopLevel(CompileContext &context, unsigned int d)
{
	return GetNode()->GetExpression()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateTopLevel(context, d);
}

CodeGeneratorStatementVariableDeclarationLlvm::CodeGeneratorStatementVariableDeclarationLlvm(AstNodeStatementVariableDeclaration *node) :
	CodeGeneratorStatementLlvm(node)
{
}

CodeGeneratorStatementVariableDeclarationLlvm *CodeGeneratorStatementVariableDeclarationLlvm::Copy()
{
	return new CodeGeneratorStatementVariableDeclarationLlvm(GetNode());
}

//TODO/WARNING: I suspect this breaking because we aren't taking into account the fact that a variable may be declared many times in the same function - e.g: in a loop - or even never (in an if)
//Do we need BlockBegin and BlockEnd on statements?
std::string CodeGeneratorStatementVariableDeclarationLlvm::Generate(CompileContext &context, unsigned int d)
{
	std::string source;
	std::string debug = GetDebug(context, d);
	
	if (GetNode()->GetArgument()){
		source += PrintString("%s %%.%s", GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), GetNode()->GetIdentifier().c_str());
	}
	else {
		std::string type = GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
		
		if (GetNode()->GetGlobal()){
			source += PrintString("; == Variable Declaration ; ==\n@%s = global %s\n",
				GetNode()->GetIdentifier().c_str(),
				GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str());
			throw FormatCompilerError(GetNode(), "Global variables not implemented yet");
		}
	}
	
	return PrintString("; Variable declaration at line %u\n", GetNode()->GetLine()) + Indent(source, 1);
}

std::string CodeGeneratorStatementVariableDeclarationLlvm::GenerateFnBegin(CompileContext &context, unsigned int d)
{
	std::string name = GetNode()->GetIdentifier();
	std::string type = GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
	std::string debug = GetDebug(context, d);
	std::string source;
	
	if (GetNode()->GetArgument()){
		source += PrintString("%%.%u = alloca %s%s\nstore %s %%.%s, %s* %%.%u%s\n",
			GetNode()->GetId(), type.c_str(), debug.c_str(), type.c_str(), name.c_str(), type.c_str(), GetNode()->GetId(), debug.c_str());
		
		unsigned int old_value = GetUniqueId();
		source += PrintString("%%.%u = load %s* %%.%u%s ; variable declaration assignment\n",
			old_value, type.c_str(), GetNode()->GetId(), debug.c_str());
	}
	else {
		source += PrintString("%%.%u = alloca %s%s ; allocate space for a variable\n",
			GetNode()->GetId(),
			GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(),
			debug.c_str());
		
		unsigned int new_value = GetUniqueId();
		std::string created_value = GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateCreate(context, d, new_value);
		if (created_value.size()){
			source += PrintString(
				"; Create the default value\n"
				"%s\n"
				"; Store the new expression\n"
				"store %s %%.%u, %s* %%.%u%s ; variable declaration assignment\n",
				Indent(created_value, 1).c_str(),
				type.c_str(), new_value, type.c_str(), GetNode()->GetId(), debug.c_str()
			);
		}
	}
	
	if (context.GetDebug()){
		source += PrintString(
			"call void @llvm.dbg.declare(metadata !{%s* %%.%u}, metadata !{\n"
				"\ti32 %u, ; Tag\n"
				"\tmetadata !%u, ; Context\n"
				"\tmetadata !\"%s\", ; Name\n"
				"\tmetadata !1, ; Reference to file where defined\n"
				"\ti32 %u, ; 24 bit - Line number where defined, 8 bit - Argument number. 1 indicates 1st argument.\n"
				"%s, ; Type descriptor\n"
				"\ti32 0, ; flags\n"
				"\ti32 0\n"
			"})%s\n",
			GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(),
			GetNode()->GetId(),
			
			(GetNode()->GetArgument() ? 257 : 256) + LlvmDebugVersion,
			d,
			GetNode()->GetIdentifier().c_str(),
			(GetNode()->GetLine() & 0xFFFFFF) | (GetNode()->GetArgument() << 24),
			Indent(GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateDebug(context, d), 1).c_str(),
			
			debug.c_str()
		);
	}
	
	return PrintString("; Variable declaration at line %u\n", GetNode()->GetLine()) + Indent(source, 1);
}

std::string CodeGeneratorStatementVariableDeclarationLlvm::GenerateFnEnd(CompileContext &context, unsigned int d)
{
	std::string source;
	std::string type = GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
	std::string debug = GetDebug(context, d);
	source += PrintString("; Variable declaration at line %u\n", GetNode()->GetLine());
	
	unsigned int old_value = GetUniqueId();
	source += Indent(PrintString("%%.%u = load %s* %%.%u%s ; variable declaration assignment\n",
		old_value, type.c_str(), GetNode()->GetId(), debug.c_str()), 1);
	source += Indent(GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateDestroy(context, d, old_value), 1);
	
	return source;
}

std::string CodeGeneratorStatementVariableDeclarationLlvm::GenerateTopLevel(CompileContext &context, unsigned int d)
{
	std::string source;
	source += GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateTopLevel(context, d);
	
	return PrintString("; Variable declaration at line %u\n", GetNode()->GetLine()) + Indent(source, 1);
}

CodeGeneratorStatementVariableDeclarationLlvm_Highlighter::CodeGeneratorStatementVariableDeclarationLlvm_Highlighter(AstNodeStatementVariableDeclaration *node) :
	CodeGeneratorStatementVariableDeclarationLlvm(node)
{
}

std::string CodeGeneratorStatementVariableDeclarationLlvm_Highlighter::GenerateFnBegin(CompileContext &context, unsigned int d)
{
	std::string source;
	std::string name = GetNode()->GetIdentifier();
	std::string type = GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
	std::string debug = GetDebug(context, d);
	
	if (GetNode()->GetArgument()){
		if (context.GetDebug()){
			source += PrintString(
				"call void @llvm.dbg.value(metadata !{%s %%.%s}, i64 0, metadata !{\n"
					"\ti32 %u, ; Tag\n"
					"\tmetadata !%u, ; Context\n"
					"\tmetadata !\"%s\", ; Name\n"
					"\tmetadata !1, ; Reference to file where defined\n"
					"\ti32 %u, ; 24 bit - Line number where defined, 8 bit - Argument number. 1 indicates 1st argument.\n"
					"%s, ; Type descriptor\n"
					"\ti32 0, ; flags\n"
					"\ti32 0\n"
				"})%s\n",
				GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(),
				name.c_str(),
				
				(GetNode()->GetArgument() ? 257 : 256) + LlvmDebugVersion,
				d,
				GetNode()->GetIdentifier().c_str(),
				(GetNode()->GetLine() & 0xFFFFFF) | (GetNode()->GetArgument() << 24),
				Indent(GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateDebug(context, d), 1).c_str(),
				
				debug.c_str()
			);
		}
	}
	
	return PrintString("; Variable declaration highlighter at line %u\n", GetNode()->GetLine()) + Indent(source, 1);
}

std::string CodeGeneratorStatementVariableDeclarationLlvm_Highlighter::GenerateFnEnd(CompileContext &context, unsigned int d)
{
	return "";
}

CodeGeneratorStatementExpressionLlvm::CodeGeneratorStatementExpressionLlvm(AstNodeStatementExpression *node) :
	CodeGeneratorStatementLlvm(node)
{
}

CodeGeneratorStatementExpressionLlvm *CodeGeneratorStatementExpressionLlvm::Copy()
{
	return new CodeGeneratorStatementExpressionLlvm(GetNode());
}

std::string CodeGeneratorStatementExpressionLlvm::Generate(CompileContext &context, unsigned int d)
{
	std::string source;
	unsigned int expression = GetUniqueId();
	source += GetNode()->GetExpression()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->Generate(context, d, expression);
	source += GetNode()->GetExpression()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateDestroy(context, d, expression);
	return source;
}

std::string CodeGeneratorStatementExpressionLlvm::GenerateFnBegin(CompileContext &context, unsigned int d)
{
	return GetNode()->GetExpression()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateFnBegin(context, d);
}

std::string CodeGeneratorStatementExpressionLlvm::GenerateFnEnd(CompileContext &context, unsigned int d)
{
	return GetNode()->GetExpression()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateFnEnd(context, d);
}

std::string CodeGeneratorStatementExpressionLlvm::GenerateTopLevel(CompileContext &context, unsigned int d)
{
	return GetNode()->GetExpression()->GetCodeGenerator<CodeGeneratorExpressionLlvm>()->GenerateTopLevel(context, d);
}

CodeGeneratorStatementBlockLlvm::CodeGeneratorStatementBlockLlvm(AstNodeStatementBlock *node) :
	CodeGeneratorStatementLlvm(node)
{
}

CodeGeneratorStatementBlockLlvm *CodeGeneratorStatementBlockLlvm::Copy()
{
	return new CodeGeneratorStatementBlockLlvm(GetNode());
}

std::string CodeGeneratorStatementBlockLlvm::Generate(CompileContext &context, unsigned int d)
{
	std::string source = PrintString("; ==== Statement Block at line %u ==== ;\n", GetNode()->GetLine());
	for (auto statement: GetNode()->GetStatements()){
		source += Indent(statement->GetCodeGenerator<CodeGeneratorStatementLlvm>()->Generate(context, GetNode()->GetId()), 1);
	}
	return source + "\n";
}

std::string CodeGeneratorStatementBlockLlvm::GenerateFnBegin(CompileContext &context, unsigned int d)
{
	std::string source = PrintString("; ==== Statement Block at line %u, beginning of function ==== ;\n", GetNode()->GetLine());
	for (auto statement: GetNode()->GetStatements()){
		source += Indent(statement->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateFnBegin(context, GetNode()->GetId()), 1);
	}
	return source + "\n";
}

std::string CodeGeneratorStatementBlockLlvm::GenerateFnEnd(CompileContext &context, unsigned int d)
{
	std::string source = PrintString("; ==== Statement Block at line %u, end of function ==== ;\n", GetNode()->GetLine());
	for (auto statement: GetNode()->GetStatements()){
		source += Indent(statement->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateFnEnd(context, GetNode()->GetId()), 1);
	}
	return source + "\n";
}

std::string CodeGeneratorStatementBlockLlvm::GenerateTopLevel(CompileContext &context, unsigned int d)
{
	std::string source;
	
	if (context.GetDebug()){
		source += PrintString(
			"!%u = metadata !{\n"
				"\ti32 %u, ; Tag = 11 + LLVMDebugVersion (DW_TAG_lexical_block)\n"
				"\tmetadata !%u, ; Reference to context descriptor\n"
				"\ti32 %u, ; Line number\n"
				"\ti32 %u, ; Column number\n"
				"\tmetadata !1, ; Reference to source file\n"
				"\ti32 %u ; Unique ID to identify blocks from a template function\n"
			"}\n",
			GetNode()->GetId(),
			LlvmDebugVersion + 11,
			d, //remove the 1, and uncomment this, and LLVM will segfault
			GetNode()->GetLine(),
			GetNode()->GetColumn(),
			GetUniqueId()
		);
	}
	
	for (auto statement: GetNode()->GetStatements()){
		if (!statement->GetCodeGenerator<CodeGeneratorStatementLlvm>()){
			throw FormatCompilerError(statement, ":/");
		}
		source += statement->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateTopLevel(context, GetNode()->GetId());
	}
	return source + "\n";
}

CodeGeneratorStatementMarkedGroupLlvm::CodeGeneratorStatementMarkedGroupLlvm(AstNodeStatementMarkedGroup *node) :
	CodeGeneratorStatementLlvm(node)
{
}

CodeGeneratorStatementMarkedGroupLlvm *CodeGeneratorStatementMarkedGroupLlvm::Copy()
{
	return new CodeGeneratorStatementMarkedGroupLlvm(GetNode());
}

std::string CodeGeneratorStatementMarkedGroupLlvm::Generate(CompileContext &context, unsigned int d)
{
	std::string source = PrintString("; ==== Statement Marked Group at line %u ==== ;\n", GetNode()->GetLine());
	for (auto statement: GetNode()->GetStatements()){
		source += Indent(statement->GetCodeGenerator<CodeGeneratorStatementLlvm>()->Generate(context, d), 1);
	}
	return source + "\n";
}

std::string CodeGeneratorStatementMarkedGroupLlvm::GenerateFnBegin(CompileContext &context, unsigned int d)
{
	std::string source = PrintString("; ==== Statement Marked Group at line %u, beginning of function ==== ;\n", GetNode()->GetLine());
	for (auto statement: GetNode()->GetStatements()){
		source += Indent(statement->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateFnBegin(context, d), 1);
	}
	return source + "\n";
}

std::string CodeGeneratorStatementMarkedGroupLlvm::GenerateFnEnd(CompileContext &context, unsigned int d)
{
	std::string source = PrintString("; ==== Statement Marked Group at line %u, end of function ==== ;\n", GetNode()->GetLine());
	for (auto statement: GetNode()->GetStatements()){
		source += Indent(statement->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateFnEnd(context, d), 1);
	}
	return source + "\n";
}

std::string CodeGeneratorStatementMarkedGroupLlvm::GenerateTopLevel(CompileContext &context, unsigned int d)
{
	std::string source;
	//Difference between a block and a marked group is that the marked group is not a new debug context
	for (auto statement: GetNode()->GetStatements()){
		source += statement->GetCodeGenerator<CodeGeneratorStatementLlvm>()->GenerateTopLevel(context, d);
	}
	return source + "\n";
}
