/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../../ast/astnode.hpp"
#include "codegen.hpp"


CodeGeneratorStructDeclarationLlvm::CodeGeneratorStructDeclarationLlvm(AstNodeStructDeclaration *node) :
	CodeGeneratorLlvm(node)
{
}

CodeGeneratorStructDeclarationLlvm *CodeGeneratorStructDeclarationLlvm::Copy()
{
	return new CodeGeneratorStructDeclarationLlvm(GetNode());
}

std::string CodeGeneratorStructDeclarationLlvm::Generate(CompileContext &context, unsigned int d)
{
	std::string source;
	std::string elements;
	std::list<AstNodeStructItem *> items = GetNode()->GetItems();
	for (auto item: items){
		if (elements.size()){
			elements += ", ";
		}
		elements += item->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
	}
	
	source += PrintString("%%%s = type<{ %s }>\n", GetNode()->GetIdentifier().c_str(), elements.c_str());
	
	
	if (!context.GetDebug()){
		return source;
	}
	
	AstNodeStructDeclaration *struct_declaration = GetNode();
	
	std::string struct_members;
	
	unsigned int offset = 0;
	for (auto item: items){
		if (struct_members.size()){
			struct_members += ", ";
		}
		
		unsigned int size = item->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GetSize(context);
		struct_members += PrintString(
			"metadata !{\n"
				"\ti32 %u, ; Tag (DW_TAG_member)\n"
				"\tmetadata !%u, ; Reference to context\n"
				"\tmetadata !\"%s\", ; Name (may be "" for anonymous types)\n"
				"\tmetadata !1, ; Reference to file where defined (may be NULL)\n"
				"\ti32 0, ; Line number where defined (may be 0)\n"
				"\ti64 %u, ; Size in bits\n"
				"\ti64 8, ; Alignment in bits\n"
				"\ti64 %u, ; Offset in bits\n"
				"\ti32 0, ; Flags to encode attributes, e.g. private\n"
				"%s"
			"}\n",
			
			LlvmDebugVersion + 13,
			GetNode()->GetId(),
			item->GetIdentifier().c_str(),
			size * 8,
			offset * 8,
			Indent(item->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateDebug(context, d), 1).c_str()
		);
		offset += size;
		
		source += item->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateTopLevel(context, d);
	}
	
	source += PrintString(
		"!%u = metadata !{\n"
			"\ti32 %u, ; Tag (DW_TAG_structure_type)\n"
			"\tnull, ; Reference to context\n"
			"\tmetadata !\"%s\", ; Name (may be "" for anonymous types)\n"
			"\tmetadata !1, ; Reference to file where defined (may be NULL)\n"
			"\ti32 0, ; Line number where defined (may be 0)\n"
			"\ti64 %u, ; Size in bits\n"
			"\ti64 8, ; Alignment in bits\n"
			"\ti32 0, ; Offset in bits\n"
			"\ti32 0, ; Flags\n"
			"\tnull, ; Reference to type derived from\n"
			
			"\tmetadata !{ ; Reference to array of member descriptors\n"
				
				"%s"
				
			"\t},\n"
			
			"\ti32 0,\n"
			"\ti32 0,\n"
			"\ti32 0\n"
		"}\n",
		GetNode()->GetId(),
		LlvmDebugVersion + 19,
		struct_declaration->GetIdentifier().c_str(),
		context.GetComputeSpec().GetCpu().GetSizeT() * 4,
		Indent(struct_members, 2).c_str()
	);
	
	return source;
}



CodeGeneratorTypeStructLlvm::CodeGeneratorTypeStructLlvm(AstNodeTypeStruct *node) :
	CodeGeneratorTypeLlvm(node)
{
}

CodeGeneratorTypeStructLlvm *CodeGeneratorTypeStructLlvm::Copy()
{
	return new CodeGeneratorTypeStructLlvm(GetNode());
}

std::string CodeGeneratorTypeStructLlvm::Generate(CompileContext &context, unsigned int d)
{
	return PrintString("%%%s", GetNode()->GetIdentifier().c_str());
}

std::string CodeGeneratorTypeStructLlvm::GenerateDebug(CompileContext &context, unsigned int d)
{
	return PrintString("metadata !%u", GetNode()->GetLink()->GetId());
}

std::string CodeGeneratorTypeStructLlvm::GenerateCreate(CompileContext &context, unsigned int d, unsigned int dst)
{
	std::string source;
	std::string struct_type = Generate(context, d);
	std::list<AstNodeStructItem *> items = GetNode()->GetLink()->GetItems();
	
	unsigned int current_struct = 0;
	
	unsigned int i = 0;
	for (auto item: items){
		unsigned int new_struct = (i < items.size() - 1) ? GetUniqueId() : dst;
		unsigned int new_item = GetUniqueId();
		std::string loop_code = item->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateCreate(context, d, new_item);
		std::string type_string = item->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
		
		if (!loop_code.size()){
			loop_code = PrintString("%%.%u = bitcast %s undef to %s\n", new_item, type_string.c_str(), type_string.c_str());
		}
		
		source += loop_code;
		
		if (i){
			source += PrintString("%%.%u = insertvalue %s %%.%u, %s %%.%u, %u ; insert the element into a new struct\n",
				new_struct, struct_type.c_str(), current_struct, type_string.c_str(), new_item, i);
			current_struct = new_struct;
		}
		else {
			source += PrintString("%%.%u = insertvalue %s undef, %s %%.%u, 0 ; insert the element into a new struct\n",
				new_struct, struct_type.c_str(), type_string.c_str(), new_item);
			current_struct = new_struct;
		}
		
		i++;
	}
	
	return source;
}

std::string CodeGeneratorTypeStructLlvm::GenerateDestroy(CompileContext &context, unsigned int d, unsigned int src)
{
	std::string source;
	std::string struct_type = Generate(context, d);
	std::list<AstNodeStructItem *> items = GetNode()->GetLink()->GetItems();
	
	unsigned int i = 0;
	for (auto item: items){
		unsigned int extracted_item = GetUniqueId();
		std::string loop_code = item->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateDestroy(context, d, extracted_item);
		
		if (loop_code.size()){
			source += PrintString("%%.%u = extractvalue %s %%.%u, %u\n",
				extracted_item, struct_type.c_str(), src, i);
			source += loop_code;
		}
		
		i++;
	}
	
	return source;
}

std::string CodeGeneratorTypeStructLlvm::GenerateCopy(CompileContext &context, unsigned int d, unsigned int src, unsigned int dst)
{
	std::string source;
	std::string struct_type = Generate(context, d);
	std::list<AstNodeStructItem *> items = GetNode()->GetLink()->GetItems();
	
	unsigned int current_struct = 0;
	
	unsigned int i = 0;
	unsigned int items_needing_copying = 0;
	for (auto item: items){
		unsigned int new_struct = (i < items.size() - 1) ? GetUniqueId() : dst;
		unsigned int old_item = GetUniqueId();
		unsigned int new_item = GetUniqueId();
		
		std::string loop_code = item->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateCopy(context, d, old_item, new_item);
		std::string type_string = item->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
		
		source += PrintString("%%.%u = extractvalue %s %%.%u, %u\n",
			old_item, struct_type.c_str(), src, i);
		
		if (loop_code.size()){
			items_needing_copying++;
		}
		else {
			new_item = old_item;
		}
		
		source += loop_code;
		
		if (i){
			source += PrintString("%%.%u = insertvalue %s %%.%u, %s %%.%u, %u ; insert the element into a new struct\n",
				new_struct, struct_type.c_str(), current_struct, type_string.c_str(), new_item, i);
			current_struct = new_struct;
		}
		else {
			source += PrintString("%%.%u = insertvalue %s undef, %s %%.%u, 0 ; insert the element into a new struct\n",
				new_struct, struct_type.c_str(), type_string.c_str(), new_item);
			current_struct = new_struct;
		}
		
		i++;
	}
	
	return items_needing_copying ? source : "";
}






unsigned int CodeGeneratorTypeStructLlvm::GetSize(CompileContext &context)
{
	unsigned int total = 0;
	
	std::list<AstNodeStructItem *> items = GetNode()->GetLink()->GetItems();
	for (auto item: items){
		total += item->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GetSize(context);
	}
	
	return total;
}

std::string CodeGeneratorTypeStructLlvm::GenerateTopLevel(CompileContext &context, unsigned int d)
{
	return "";
}

CodeGeneratorExpressionLValueStructElementLlvm::CodeGeneratorExpressionLValueStructElementLlvm(AstNodeExpressionLValueStructElement *node) :
	CodeGeneratorExpressionLValueLlvm(node)
{
}

CodeGeneratorExpressionLValueStructElementLlvm *CodeGeneratorExpressionLValueStructElementLlvm::Copy()
{
	return new CodeGeneratorExpressionLValueStructElementLlvm(GetNode());
}

std::string CodeGeneratorExpressionLValueStructElementLlvm::Generate(CompileContext &context, unsigned int d, unsigned int n)
{
	std::string source;
	unsigned int lvalue = GetUniqueId();
	
	//lvalue is a struct
	source += GetNode()->GetLValue()->GetCodeGenerator<CodeGeneratorExpressionLValueLlvm>()->Generate(context, d, lvalue);
	
	unsigned int element_n = 0;
	std::list<AstNodeStructItem *> items = GetNode()->GetStructType()->GetLink()->GetItems();
	
	for (auto item: items){
		if (GetNode()->GetIdentifier() == item->GetIdentifier()){
			source += PrintString("%%.%u = extractvalue %%%s %%.%u, %u ; Get %s of the struct\n",
				n, GetNode()->GetStructType()->GetIdentifier().c_str(), lvalue, element_n, GetNode()->GetIdentifier().c_str());
			return source;
		}
		element_n++;
	}
	
	throw FormatError("\"%s\" not in \"%s\"", GetNode()->GetIdentifier().c_str(), GetNode()->GetStructType()->GetIdentifier().c_str());
}

std::string CodeGeneratorExpressionLValueStructElementLlvm::GenerateStore(CompileContext &context, unsigned int d, unsigned int n)
{
	std::string source;
	unsigned int lvalue = GetUniqueId();
	unsigned int new_value = GetUniqueId();
	
	//lvalue is a struct
	source += "; Load struct\n";
	source += Indent(GetNode()->GetLValue()->GetCodeGenerator<CodeGeneratorExpressionLValueLlvm>()->Generate(context, d, lvalue), 1);
	
	unsigned int element_n = 0;
	std::list<AstNodeStructItem *> items = GetNode()->GetStructType()->GetLink()->GetItems();
	
	for (auto item: items){
		if (GetNode()->GetIdentifier() == item->GetIdentifier()){
			source += "; Insert Element\n";
			source += Indent(PrintString("%%.%u = insertvalue %%%s %%.%u, %s %%.%u, %u ; Get %s of the struct\n",
				new_value, GetNode()->GetStructType()->GetIdentifier().c_str(), lvalue,
				item->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), n,
				element_n, GetNode()->GetIdentifier().c_str()), 1);
			
			//Now store the value
			source += "; Store struct\n";
			source += Indent(GetNode()->GetLValue()->GetCodeGenerator<CodeGeneratorExpressionLValueLlvm>()->GenerateStore(context, d, new_value), 1);
			
			return source;
		}
		element_n++;
	}
	
	throw FormatError("\"%s\" not in \"%s\"", GetNode()->GetIdentifier().c_str(), GetNode()->GetStructType()->GetIdentifier().c_str());
}

std::string CodeGeneratorExpressionLValueStructElementLlvm::GenerateReference(CompileContext &context, unsigned int d, unsigned int n)
{
	/* This works because the reference creation/destruction stuff needs to keep track of references that exist */
	std::string debug = GetDebug(context, d);
	std::string type = GetNode()->GetLValue()->GetType()->GetCodeGenerator<CodeGeneratorTypeStructLlvm>()->Generate(context, d);
	
	unsigned int lvalue_reference = GetUniqueId();
	unsigned int i8_struct = GetUniqueId();
	unsigned int type_struct = GetUniqueId();
	unsigned int i8_element = GetUniqueId();
	unsigned int type_element = GetUniqueId();
	
	//Find the element number
	unsigned int element_n = 0;
	std::list<AstNodeStructItem *> items = GetNode()->GetStructType()->GetLink()->GetItems();
	std::string element_type;
	
	for (auto item: items){
		if (GetNode()->GetIdentifier() == item->GetIdentifier()){
			element_type = item->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
			break;
		}
		element_n++;
	}
	
	if (!element_type.size()){
		throw FormatError("\"%s\" not in \"%s\"", GetNode()->GetIdentifier().c_str(), GetNode()->GetStructType()->GetIdentifier().c_str());
	}
	
	//Now produce the code
	return PrintString(
		"; Get reference to structure element %s\n"
			"%s"
			
			"\t%%.%u = call fastcc i8* (%%t.ref)* @.ref.read(%%t.ref %%.%u)%s ; Get the pointer we need for the reference\n"
			"\t%%.%u = bitcast i8* %%.%u to %s* ; Cast struct pointer\n"
			"\t%%.%u = getelementptr %s* %%.%u, i32 0, i32 %u ; Calculate the pointer to the element\n"
			"\t%%.%u = bitcast %s* %%.%u to i8* ; Cast the element pointer\n"
			
			"\tcall fastcc void (%%t.ref)* @.ref.commit.before(%%t.ref %%.%u)%s ; Pre commit\n"
			"\t%%.%u = call fastcc %%t.ref (i8*)* @.ref.create(i8* %%.%u)%s ; Create the new reference\n"
			"\tcall fastcc void (%%t.ref)* @.ref.commit.after(%%t.ref %%.%u)%s ; Post commit\n",
		
		GetNode()->GetSourceLocation().c_str(),
			Indent(GetNode()->GetLValue()->GetCodeGenerator<CodeGeneratorExpressionLValueLlvm>()->GenerateReference(context, d, lvalue_reference), 1).c_str(),
			
			i8_struct, lvalue_reference, debug.c_str(),
			type_struct, i8_struct, type.c_str(),
			type_element, type.c_str(), type_struct, element_n,
			i8_element, element_type.c_str(), type_element,
			
			lvalue_reference, debug.c_str(),
			n, i8_element, debug.c_str(),
			lvalue_reference, debug.c_str()
	);
}

std::string CodeGeneratorExpressionLValueStructElementLlvm::GenerateFnBegin(CompileContext &context, unsigned int d)
{
	return GetNode()->GetLValue()->GetCodeGenerator<CodeGeneratorExpressionLValueLlvm>()->GenerateFnBegin(context, d);
}

std::string CodeGeneratorExpressionLValueStructElementLlvm::GenerateFnEnd(CompileContext &context, unsigned int d)
{
	return GetNode()->GetLValue()->GetCodeGenerator<CodeGeneratorExpressionLValueLlvm>()->GenerateFnEnd(context, d);
}

