/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../../ast/astnode.hpp"
#include "codegen.hpp"

CodeGeneratorTypeLlvm::CodeGeneratorTypeLlvm(AstNodeType *node) :
	CodeGeneratorLlvm(node)
{
}

std::string CodeGeneratorTypeLlvm::GenerateTopLevel(CompileContext &context, unsigned int d)
{
	return "";
}

std::string CodeGeneratorTypeLlvm::GenerateCreate(CompileContext &context, unsigned int d, unsigned int value)
{
	return "";
}

std::string CodeGeneratorTypeLlvm::GenerateDestroy(CompileContext &context, unsigned int d, unsigned int value)
{
	return "";
}

std::string CodeGeneratorTypeLlvm::GenerateCopy(CompileContext &context, unsigned int d, unsigned int src, unsigned int dst)
{
	return "";
}




CodeGeneratorTypeVoidLlvm::CodeGeneratorTypeVoidLlvm(AstNodeTypeVoid *node) :
	CodeGeneratorTypeLlvm(node)
{
}

CodeGeneratorTypeVoidLlvm *CodeGeneratorTypeVoidLlvm::Copy()
{
	return new CodeGeneratorTypeVoidLlvm(GetNode());
}


std::string CodeGeneratorTypeVoidLlvm::Generate(CompileContext &context, unsigned int d)
{
	return "void";
}

std::string CodeGeneratorTypeVoidLlvm::GenerateDebug(CompileContext &context, unsigned int d)
{
	return "null";
}

unsigned int CodeGeneratorTypeVoidLlvm::GetSize(CompileContext &context)
{
	return 0;
}

CodeGeneratorTypeBoolLlvm::CodeGeneratorTypeBoolLlvm(AstNodeTypeBool *node) :
	CodeGeneratorTypeLlvm(node)
{
}

std::string CodeGeneratorTypeBoolLlvm::GenerateDebug(CompileContext &context, unsigned int d)
{
	return PrintString(
		"metadata !{\n"
			"\ti32 %u, ; Tag = 36 + LLVMDebugVersion (DW_TAG_base_type)\n"
			"\tnull, ; Reference to context\n"
			"\tmetadata !\"%s\", ; Name (may be "" for anonymous types)\n"
			"\tnull, ; Reference to file where defined (may be NULL)\n"
			"\ti32 0, ; Line number where defined (may be 0)\n"
			"\ti64 8, ; Size in bits\n"
			"\ti64 8, ; Alignment in bits\n"
			"\ti64 0, ; Offset in bits\n"
			"\ti32 0, ; Flags\n"
			"\ti32 2 ; DWARF type encoding\n"
		"}",
		LlvmDebugVersion + 36,
		GetNode()->GetName().c_str()
	);
}


CodeGeneratorTypeBoolLlvm *CodeGeneratorTypeBoolLlvm::Copy()
{
	return new CodeGeneratorTypeBoolLlvm(GetNode());
}

std::string CodeGeneratorTypeBoolLlvm::Generate(CompileContext &context, unsigned int d)
{
	return "i1";
}

unsigned int CodeGeneratorTypeBoolLlvm::GetSize(CompileContext &context)
{
	return 1;
}

CodeGeneratorTypeIntLlvm::CodeGeneratorTypeIntLlvm(AstNodeTypeInt *node) :
	CodeGeneratorTypeLlvm(node)
{
}

CodeGeneratorTypeIntLlvm *CodeGeneratorTypeIntLlvm::Copy()
{
	return new CodeGeneratorTypeIntLlvm(GetNode());
}

std::string CodeGeneratorTypeIntLlvm::Generate(CompileContext &context, unsigned int d)
{
	return GetNode()->GetWidth() ? PrintString("i%u", GetNode()->GetWidth()) : "%t.size";
}

std::string CodeGeneratorTypeIntLlvm::GenerateDebugGeneric(std::string name, unsigned int size, bool is_signed)
{
	return PrintString(
		"metadata !{\n"
			"\ti32 %u, ; Tag = 36 + LLVMDebugVersion (DW_TAG_base_type)\n"
			"\tnull, ; Reference to context\n"
			"\tmetadata !\"%s\", ; Name (may be "" for anonymous types)\n"
			"\tnull, ; Reference to file where defined (may be NULL)\n"
			"\ti32 0, ; Line number where defined (may be 0)\n"
			"\ti64 %u, ; Size in bits\n"
			"\ti64 8, ; Alignment in bits\n"
			"\ti64 0, ; Offset in bits\n"
			"\ti32 0, ; Flags\n"
			"\ti32 %u ; DWARF type encoding\n"
		"}",
		LlvmDebugVersion + 36,
		name.c_str(),
		size * 8,
		is_signed ? 5 : 7
	);
}

std::string CodeGeneratorTypeIntLlvm::GenerateDebug(CompileContext &context, unsigned int d)
{
	return GenerateDebugGeneric(GetNode()->GetName(), GetSize(context), GetNode()->GetSigned());
}



unsigned int CodeGeneratorTypeIntLlvm::GetSize(CompileContext &context)
{
	unsigned int width = GetNode()->GetWidth(context);
	if (width % 8){
		return (width/8) + 1;
	}
	else {
		return width/8;
	}
}

CodeGeneratorTypeFloatLlvm::CodeGeneratorTypeFloatLlvm(AstNodeTypeFloat *node) :
	CodeGeneratorTypeLlvm(node)
{
}

CodeGeneratorTypeFloatLlvm *CodeGeneratorTypeFloatLlvm::Copy()
{
	return new CodeGeneratorTypeFloatLlvm(GetNode());
}

std::string CodeGeneratorTypeFloatLlvm::Generate(CompileContext &context, unsigned int d)
{
	switch (GetNode()->GetWidth())
	{
		case 16:
			return "half";
		case 32:
			return "float";
		case 64:
			return "double";
		case 128:
			return "fp128";
	}
	
	//type not supported by the current implementation of the LLVM code generator
	throw CodeGenerationException(CODEGEN_LLVM, GetNode(), "Floating point width %u not supported", GetNode()->GetWidth());
}

std::string CodeGeneratorTypeFloatLlvm::GenerateDebug(CompileContext &context, unsigned int d)
{
	return PrintString(
		"metadata !{\n"
			"\ti32 %u, ; Tag = 36 + LLVMDebugVersion (DW_TAG_base_type)\n"
			"\tnull, ; Reference to context\n"
			"\tmetadata !\"%s\", ; Name (may be "" for anonymous types)\n"
			"\tnull, ; Reference to file where defined (may be NULL)\n"
			"\ti32 0, ; Line number where defined (may be 0)\n"
			"\ti64 %u, ; Size in bits\n"
			"\ti64 8, ; Alignment in bits\n"
			"\ti64 0, ; Offset in bits\n"
			"\ti32 0, ; Flags\n"
			"\ti32 4 ; DWARF type encoding\n"
		"}",
		LlvmDebugVersion + 36,
		GetNode()->GetName().c_str(),
		GetSize(context) * 8
	);
}


unsigned int CodeGeneratorTypeFloatLlvm::GetSize(CompileContext &context)
{
	if (GetNode()->GetWidth() % 8){
		return (GetNode()->GetWidth()/8) + 1;
	}
	else {
		return GetNode()->GetWidth()/8;
	}
}

CodeGeneratorTypePointerLlvm::CodeGeneratorTypePointerLlvm(AstNodeTypePointer *node) :
	CodeGeneratorTypeLlvm(node)
{
}

CodeGeneratorTypePointerLlvm *CodeGeneratorTypePointerLlvm::Copy()
{
	return new CodeGeneratorTypePointerLlvm(GetNode());
}

unsigned int CodeGeneratorTypePointerLlvm::GetSize(CompileContext &context)
{
	return context.GetComputeSpec().GetCpu().GetSizeT()/8;
}

std::string CodeGeneratorTypePointerLlvm::Generate(CompileContext &context, unsigned int d)
{
	return PrintString("%s*", GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str());
}

std::string CodeGeneratorTypePointerLlvm::GenerateDebug(CompileContext &context, unsigned int d)
{
	return PrintString(
		"metadata !{\n"
			"\ti32 %u, ; Tag = 15 + LLVMDebugVersion (DW_TAG_pointer_type)\n"
			"\tnull, ; Reference to context\n"
			"\tmetadata !\"\", ; Name (may be "" for anonymous types)\n"
			"\tnull, ; Reference to file where defined (may be NULL)\n"
			"\ti32 0, ; Line number where defined (may be 0)\n"
			"\ti64 %u, ; Size in bits\n"
			"\ti64 8, ; Alignment in bits\n"
			"\ti64 0, ; Offset in bits\n"
			"\ti32 0, ; Flags\n"
			"%s ; Reference to type derived from\n"
		"}",
		LlvmDebugVersion + 15,
		context.GetComputeSpec().GetCpu().GetSizeT(),
		Indent(GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateDebug(context, d), 1).c_str()
	);
}


CodeGeneratorTypeStaticVectorLlvm::CodeGeneratorTypeStaticVectorLlvm(AstNodeTypeStaticVector *node) :
	CodeGeneratorTypeLlvm(node)
{
}

CodeGeneratorTypeStaticVectorLlvm *CodeGeneratorTypeStaticVectorLlvm::Copy()
{
	return new CodeGeneratorTypeStaticVectorLlvm(GetNode());
}

std::string CodeGeneratorTypeStaticVectorLlvm::Generate(CompileContext &context, unsigned int d)
{
	return PrintString("<%u x %s>", GetNode()->GetCount(), GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str());
}

std::string CodeGeneratorTypeStaticVectorLlvm::GenerateDebug(CompileContext &context, unsigned int d)
{
	return PrintString(
		"metadata !{\n"
			"\ti32 %u, ; Tag = 1 + LLVMDebugVersion (DW_TAG_array_type)\n"
			"\tnull, ; Reference to context\n"
			"\tmetadata !\"%s\", ; Name (may be "" for anonymous types)\n"
			"\tnull, ; Reference to file where defined (may be NULL)\n"
			"\ti32 0, ; Line number where defined (may be 0)\n"
			"\ti64 %u, ; Size in bits\n"
			"\ti64 8, ; Alignment in bits\n"
			"\ti64 0, ; Offset in bits\n"
			"\ti32 0, ; Flags\n"
			"%s, ; Reference to type derived from\n"
			"\tmetadata !{i32 %u, i64 0, i64 %u} ; Reference to array of member descriptors\n"
		"}",
		LlvmDebugVersion + 1,
		GetNode()->GetName().c_str(),
		GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GetSize(context) * GetNode()->GetCount() * 8,
		Indent(GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateDebug(context, d), 1).c_str(),
		
		LlvmDebugVersion + 40, GetNode()->GetCount()
	);
}

std::string CodeGeneratorTypeStaticVectorLlvm::GenerateCreate(CompileContext &context, unsigned int d, unsigned int dst)
{
	std::string source;
	
	std::string vector_type = Generate(context, d);
	std::string element_type = GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
	
	unsigned int current_vector = GetUniqueId();
	source += PrintString("%%.%u = bitcast %s undef to %s\n", current_vector, vector_type.c_str(), vector_type.c_str());
	
	for (unsigned int i = 0; i < GetNode()->GetCount(); i++){
		unsigned int new_vector = GetUniqueId();
		unsigned int new_element = GetUniqueId();
		
		std::string child_op = GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateCreate(context, d, new_element);
		
		if (!child_op.size()){
			return "";
		}
		
		
		source += child_op;
		source += PrintString("%%.%u = insertelement %s %%.%u, %s %%.%u, i32 %u\n",
			new_vector, vector_type.c_str(), current_vector, element_type.c_str(), new_element, i);
		
		current_vector = new_vector;
	}
	
	source += PrintString("%%.%u = bitcast %s %%.%u to %s\n", dst, vector_type.c_str(), current_vector, vector_type.c_str());
	return source;
}

std::string CodeGeneratorTypeStaticVectorLlvm::GenerateDestroy(CompileContext &context, unsigned int d, unsigned int src)
{
	std::string source;
	std::string vector_type = Generate(context, d);
	std::string element_type = GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
	
	
	for (unsigned int i = 0; i < GetNode()->GetCount(); i++){
		unsigned int element = GetUniqueId();
		std::string child_op = GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GenerateDestroy(context, d, element);
		
		if (!child_op.size()){
			return "";
		}
		
		source += PrintString("%%.%u = extractelement %s %%.%u, i32 %u\n",
			element, vector_type.c_str(), src, i);
		
		source += child_op;
	}
	
	return source;
}


unsigned int CodeGeneratorTypeStaticVectorLlvm::GetSize(CompileContext &context)
{
	return GetNode()->GetType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->GetSize(context) * GetNode()->GetCount();
}


CodeGeneratorTypeFunctionPointerLlvm::CodeGeneratorTypeFunctionPointerLlvm(AstNodeTypeFunctionPointer *node) :
	CodeGeneratorTypeLlvm(node)
{
}

CodeGeneratorTypeFunctionPointerLlvm *CodeGeneratorTypeFunctionPointerLlvm::Copy()
{
	return new CodeGeneratorTypeFunctionPointerLlvm(GetNode());
}

std::string CodeGeneratorTypeFunctionPointerLlvm::Generate(CompileContext &context, unsigned int d)
{
	std::string arguments;
	std::list<AstNodeType *> types = GetNode()->GetArguments();
	for (auto type: types){
		if (arguments.size()){
			arguments += ", ";
		}
		arguments += type->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d);
	}
	
	return PrintString("%s (%s)*", GetNode()->GetReturnType()->GetCodeGenerator<CodeGeneratorTypeLlvm>()->Generate(context, d).c_str(), arguments.c_str());
}

std::string CodeGeneratorTypeFunctionPointerLlvm::GenerateDebug(CompileContext &context, unsigned int d)
{
	return
		PrintString("metadata !{\n"
			"\ti32 %u, ; Tag (see below)\n"
			"\tnull, ; Reference to context\n"
			"\tmetadata !\"\", ; Name (may be "" for anonymous types)\n"
			"\tnull, ; Reference to file where defined (may be NULL)\n"
			"\ti32 0, ; Line number where defined (may be 0)\n"
			"\ti64 %u, ; Size in bits\n"
			"\ti64 8, ; Alignment in bits\n"
			"\ti64 0, ; Offset in bits\n"
			"\ti32 0, ; Flags to encode attributes, e.g. private\n"
			"\tnull ; Reference to type derived from\n"
		"}\n",
		LlvmDebugVersion + 15,
		context.GetComputeSpec().GetCpu().GetSizeT());
}

unsigned int CodeGeneratorTypeFunctionPointerLlvm::GetSize(CompileContext &context)
{
	return context.GetComputeSpec().GetCpu().GetSizeT()/8;
}

