/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef CODEGEN_LLVM_TYPE_HPP
#define CODEGEN_LLVM_TYPE_HPP

#include "codegen.hpp"


class CodeGeneratorTypeLlvm : public CodeGeneratorLlvm
{
	public:
		CodeGeneratorTypeLlvm(AstNodeType *node);
		
		virtual AstNodeType *GetNode() = 0;
		
		virtual std::string Generate(CompileContext &context, unsigned int d) = 0; //this returns the target representation type, e.g int 32 would be i32 in LLVM
		virtual std::string GenerateDebug(CompileContext &context, unsigned int d) = 0;
		virtual std::string GenerateTopLevel(CompileContext &context, unsigned int d);
		
		/*
		 * WARNING: This comment is out of date
		 * These are recursively called. Attention should be paid to if one of these returns an empty string.
		 * There is no point in looping over an array doing nothing (in the case of create, this would result in uninitialized stuff).
		 * 
		 * Create is responsible for performing any required initialisation of the value itself, and its children.
		 * Value should be set by this method.
		 * 
		 * Destroy performs any free operation that is required on value and its children.
		 * Destroy should only be destructive if the reference count has reached 0.
		 * 
		 * It's important to increment the reference count of value A before performing the
		 * destruction of value B if A and B might be the same value
		 * 
		 * 
		 * 
		 * Generate(Create|IncRefCount|DecRefCount|Destroy) are for managing reference counts, and freeing things when necessary
		 * An expresion's generate is responsible for actually calculating or returning a value
		 * An lvalue's generate store is responsible for storing a given value into the lvalue - but not incrementing references and such like
		 * An expression's GenerateStatementBegin is for things like incrementing reference counts that are used (so others can't assume it's OK to write before th expression is calculated)
		 * An expression's GenerateStatementEnd is for things like decrementing reference counts incremented by GenerateStatementBegin
		 * A variable assignment or such like is responsible for calling the Generate, GenerateStatementBegin, and GenerateStatementEnd of an expression
		 * 
		 * What happens if an expression creates a dynamic vector that is then not stored anywhere later?
		 * We have to be careful to make sure that function arguments increment and decrement their references (so I guess function calls don't need to do this?).
		 * Also be careful about what return does in terms of decrementation and its implications for freeing (e.g: if we return a variable)
		 */
		virtual std::string GenerateCreate(CompileContext &context, unsigned int d, unsigned int value);
		virtual std::string GenerateDestroy(CompileContext &context, unsigned int d, unsigned int value);
		virtual std::string GenerateCopy(CompileContext &context, unsigned int d, unsigned int src, unsigned int dst);
		
		//Perhaps a pair of methods: one that retrieves an element given item (e.g: retrieves the second element, given a struct value)
		//The other to create an item having inserted that element
		
		virtual unsigned int GetSize(CompileContext &context) = 0;
		
	private:
};

class CodeGeneratorTypeVoidLlvm : public CodeGeneratorTypeLlvm
{
	public:
		CodeGeneratorTypeVoidLlvm(AstNodeTypeVoid *node);
		virtual CodeGeneratorTypeVoidLlvm *Copy();
		virtual std::string Generate(CompileContext &context, unsigned int d);
		virtual unsigned int GetSize(CompileContext &context);
		virtual std::string GenerateDebug(CompileContext &context, unsigned int d);
		
		AstType(AstNodeTypeVoid);
};

class CodeGeneratorTypeBoolLlvm : public CodeGeneratorTypeLlvm
{
	public:
		CodeGeneratorTypeBoolLlvm(AstNodeTypeBool *node);
		virtual CodeGeneratorTypeBoolLlvm *Copy();
		virtual std::string Generate(CompileContext &context, unsigned int d);
		virtual unsigned int GetSize(CompileContext &context);
		virtual std::string GenerateDebug(CompileContext &context, unsigned int d);
		
		AstType(AstNodeTypeBool);
};

class CodeGeneratorTypeIntLlvm : public CodeGeneratorTypeLlvm
{
	public:
		CodeGeneratorTypeIntLlvm(AstNodeTypeInt *node);
		virtual CodeGeneratorTypeIntLlvm *Copy();
		virtual std::string Generate(CompileContext &context, unsigned int d);
		virtual unsigned int GetSize(CompileContext &context);
		virtual std::string GenerateDebug(CompileContext &context, unsigned int d);
		static std::string GenerateDebugGeneric(std::string name, unsigned int size, bool is_signed);
		
		AstType(AstNodeTypeInt);
};

class CodeGeneratorTypeFloatLlvm : public CodeGeneratorTypeLlvm
{
	public:
		CodeGeneratorTypeFloatLlvm(AstNodeTypeFloat *node);
		virtual CodeGeneratorTypeFloatLlvm *Copy();
		virtual std::string Generate(CompileContext &context, unsigned int d);
		virtual unsigned int GetSize(CompileContext &context);
		virtual std::string GenerateDebug(CompileContext &context, unsigned int d);
		
		AstType(AstNodeTypeFloat);
};

class CodeGeneratorTypePointerLlvm : public CodeGeneratorTypeLlvm
{
	public:
		CodeGeneratorTypePointerLlvm(AstNodeTypePointer *node);
		virtual CodeGeneratorTypePointerLlvm *Copy();
		virtual std::string Generate(CompileContext &context, unsigned int d);
		virtual unsigned int GetSize(CompileContext &context);
		virtual std::string GenerateDebug(CompileContext &context, unsigned int d);
		
		AstType(AstNodeTypePointer);
};

class CodeGeneratorTypeStaticVectorLlvm : public CodeGeneratorTypeLlvm
{
	public:
		CodeGeneratorTypeStaticVectorLlvm(AstNodeTypeStaticVector *node);
		virtual CodeGeneratorTypeStaticVectorLlvm *Copy();
		virtual std::string Generate(CompileContext &context, unsigned int d);
		virtual unsigned int GetSize(CompileContext &context);
		virtual std::string GenerateDebug(CompileContext &context, unsigned int d);
		
		virtual std::string GenerateCreate(CompileContext &context, unsigned int d, unsigned int value);
		virtual std::string GenerateDestroy(CompileContext &context, unsigned int d, unsigned int value);
		
		AstType(AstNodeTypeStaticVector);
};

class CodeGeneratorTypeFunctionPointerLlvm : public CodeGeneratorTypeLlvm
{
	public:
		CodeGeneratorTypeFunctionPointerLlvm(AstNodeTypeFunctionPointer *node);
		virtual CodeGeneratorTypeFunctionPointerLlvm *Copy();
		virtual std::string Generate(CompileContext &context, unsigned int d);
		virtual unsigned int GetSize(CompileContext &context);
		virtual std::string GenerateDebug(CompileContext &context, unsigned int d);
		
		AstType(AstNodeTypeFunctionPointer);
};

//I think I've decided that struct will be a simple structure class (with a packed variant (packed keyword?))
//class will be more complicated with the ability to have associated metadata
//a sizeof like thing would probably be useful for use with struct

#endif
