/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <list>

#include "error.hpp"
#include "util.hpp"

#include "parser/parser.hpp"
#include "parser/polycute.yy.hpp"
#include "parser/polycute.ll.hpp"

#include "optimiser/optimiser.hpp"

#include "compilecontext.hpp"

#include "codegen/llvm/codegen.hpp"
#include "codegen/pseudopc/codegen.hpp"

extern int yyparse(yyscan_t);

bool Compile(
	/* Inputs and stuff */
	std::string input_filename,
	
	/* Program outputs */
	std::string llvm_filename,
	std::string pseudopc_filename,
	std::string computespec_output_filename,
	
	/* Options */
	std::string polycute_library_directory,
	
	/* Optimisation stuff */
	std::list<std::string> optimisers,
	std::string optimisation_report_filename,
	std::string astdot_filename,
	std::string asttext_filename,
	
	/* Debug stuff */
	bool debug,
	bool llvm_header,
	bool llvm_lines
)
{
	CompileContext context(debug, llvm_header, llvm_lines);
	
	if (false){ //If we are to use an externally provided specification
	}
	else { //The default is to autodetect
		context.GetComputeSpec().Autodetect();
	}
	
	/* Normal compile */
	
	yyscan_t scanner;
	polycute_parser_lex_init(&scanner);
	
	FILE *f = fopen(input_filename.c_str(), "r");
	if (!f){
		throw FormatError("Could not open \"%s\"", input_filename.c_str());
	}
	
	polycute_parser_set_in(f, scanner);
	polycute_parser_set_out(stdout, scanner);
	
	ParserData parserdata;
	parserdata.filename = input_filename.c_str();
	parserdata.scanner = &scanner;
	parserdata.top = 0;
	parserdata.error = false;
	
	if (polycute_parser_parse((void *)&parserdata)){
		throw FormatError("Could not parse");
	}
	
	polycute_parser_lex_destroy(scanner);
	fclose(f);
	
	if (parserdata.error){
		throw FormatError("An error occured parsing");
	}
	
	AstNodeRoot *asttop = parserdata.top;
	if (!asttop){
		throw FormatError("No AST from parsing");
	}
	
	try {
		asttop->Link();
		asttop = asttop->Normalize(context);
		
		//Perform the optimisations:
		std::string report = Optimiser::PerformOptimisations(optimisers, asttop, context);
		
		//After this point, optimisers contains only dead pointers. DO NOT USE.
		
		//output
		if (computespec_output_filename.size()){
			WriteToFile(context.GetComputeSpec().GetSpecSource(), computespec_output_filename);
		}
		
		if (optimisation_report_filename.size()){
			WriteToFile(report, optimisation_report_filename);
		}
		
		if (astdot_filename.size()){
			WriteToFile(PrintString("digraph\n{%s\n}\n", asttop->GetDot().c_str()), astdot_filename);
		}
		
		if (asttext_filename.size()){
			WriteToFile(asttop->GetTextTree(), asttext_filename);
		}
		
		if (llvm_filename.size()){
			WriteToFile(asttop->GetCodeGenerator<CodeGeneratorRootLlvm>()->Generate(polycute_library_directory, context), llvm_filename);
		}
		
		if (pseudopc_filename.size()){
			asttop->SetTreeCodeGenerator(CODEGEN_PSEUDOPC);
			WriteToFile(asttop->GetCodeGenerator<CodeGeneratorRootPseudopc>()->Generate(context), pseudopc_filename);
		}
	}
	catch (CompilerError &e){
		e.OutputCompilerMessage(asttop, context);
		if (asttop){
			delete asttop;
		}
		return false;
	}
	
	delete asttop;
	return true;
}
