/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cinttypes>
#include <unistd.h>
#include <cstring>
#include <list>
#include <map>

#include "computespec.hpp"
#include "../util.hpp"

ComputeSpecMachineCpu::ComputeSpecMachineCpu() :
	m_size_t(0), m_core_count(0), m_simd_width(0), m_memory_size(0)
{
}

unsigned int ComputeSpecMachineCpu::GetSizeT()
{
	return m_size_t;
}

unsigned int ComputeSpecMachineCpu::GetCoreCount()
{
	return m_core_count;
}

unsigned int ComputeSpecMachineCpu::GetSimdWidth()
{
	return m_simd_width;
}

uint64_t ComputeSpecMachineCpu::GetMemorySize()
{
	return m_memory_size;
}


void ComputeSpecMachineCpu::Autodetect()
{
	//Read the CPU information in
	std::string cpuinfo;
	FILE *f = fopen("/proc/cpuinfo", "r");
	if (!f){
		FormatError("Could not open /proc/cpuinfo: %s; Note: autodetection of compute specification does not work on non-Linux systems", strerror(errno));
	}
	while (!feof(f)){
		char c = 0;
		fread(&c, 1, 1, f);
		cpuinfo += c;
	}
	fclose(f);
	
	//split into individual cores
	std::list<std::string> cpuinfo_core_strings;
	while (1){
		size_t p = cpuinfo.find("\n\n");
		if (std::string::npos == p){ //no more cores
			break;
		}
		cpuinfo_core_strings.push_back(cpuinfo.substr(0, p));
		cpuinfo = cpuinfo.substr(p+2, std::string::npos);
	}
	
	//process it into key/values
	std::list<std::map<std::string, std::string>> cpuinfo_core_keyval;
	for (auto core: cpuinfo_core_strings){
		core = core + "\n";
		
		std::map<std::string, std::string> keyvalues;
		while (1){
			size_t p = core.find("\n");
			if (std::string::npos == p){ //no more key/value pairs
				break;
			}
			
			
			std::string key_value = core.substr(0, p);
			core = core.substr(p+1, std::string::npos);
			
			std::string key;
			std::string value;
			
			p = key_value.find(":");
			if (std::string::npos == p){
				key = core;
			}
			else {
				key = key_value.substr(0, p);
				if (p + 2 < key_value.size()){
					value = key_value.substr(p + 2, std::string::npos);
				}
			}
			
			p = key.find_last_not_of(" \t");
			if ((std::string::npos != p) && (key.size() != p)){
				key = key.substr(0, p+1);
			}
			
			keyvalues[key] = value;
		}
		cpuinfo_core_keyval.push_back(keyvalues);
	}
	
	//Get features
	std::list<std::set<std::string>> cores_features;
	std::set<std::string> features_union;
	for (auto core: cpuinfo_core_keyval){
		std::string features_string = core["flags"];
		std::set<std::string> features_single_core;
		while (features_string.size()){
			while (!features_string.find(" ")){
				features_string = features_string.substr(1, std::string::npos);
			}
			size_t p = features_string.find(" ");
			std::string feature = features_string.substr(0, p);
			features_single_core.insert(feature);
			features_union.insert(feature);
			
			if ((std::string::npos == p) || (features_string.size() == p)){
				features_string = "";
			}
			else {
				features_string = features_string.substr(p, std::string::npos);
			}
		}
		cores_features.push_back(features_single_core);
	}
	
	//Get the common features
	std::set<std::string> common_features;
	for (std::string feature: features_union){
		bool present = true;
		for (auto single_core_features: cores_features){
			if (!single_core_features.count(feature)){
				present = false;
			}
		}
		if (present){
			common_features.insert(feature);
		}
	}
	
	//Now we have the list of features!
	m_core_count = cores_features.size(); //the number of cores
	
	//determine the width of the SIMD instructions
	if (common_features.count("avx")){
		m_simd_width = 256;
	}
	else if (common_features.count("sse2")){
		m_simd_width = 128;
	}
	
	m_size_t = 8 * sizeof(void *);
	
	//determine the memory size
	long sysconf_result = sysconf(_SC_PAGE_SIZE);
	if (sysconf_result < 0){
		throw FormatError("Cannot detect page size");
	}
	uint64_t page_size = sysconf_result;
	
	sysconf_result = sysconf(_SC_PHYS_PAGES);
	if (sysconf_result < 0){
		throw FormatError("Cannot detect page count");
	}
	uint64_t page_count = sysconf_result;
	
	m_memory_size = page_size * page_count;
}

std::string ComputeSpecMachineCpu::GetSpecSource()
{
	return PrintString("machine cpu\n"
		"{\n"
			"\tsize_t: %u\n"
			"\tcores: %u\n"
			"\tsimd_width: %u\n"
			"\tmemory: %" PRIu64 "\n"
		"}\n",
		
		m_size_t,
		m_core_count,
		m_simd_width,
		m_memory_size
	);
} 
