/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <exception>

#include "error.hpp"

#include "ast/astnode.hpp"
#include "codegen/pseudopc/codegen.hpp"

#include "compilecontext.hpp"

/* WARNING: These methods may be intercepted by GDB - Update test/tools/polycute_debug.gdb if these lines are moved */

CompilerError::CompilerError(const char *file, int line, AstNode *node, const char *format, ...) :
	Error(file, line), m_node(node)
{
	char b[4096];
	
	va_list argp;
	va_start(argp, format);
	vsnprintf(b, 4096, format, argp);
	va_end(argp);
	
	SetMessage(b);
}

Error::Error(const char *file, int line, const char *format, ...) :
	m_file(file), m_line(line)
{
	char b[4096];
	
	va_list argp;
	va_start(argp, format);
	vsnprintf(b, 4096, format, argp);
	va_end(argp);
	
	s_current = this;
	m_message = b;
}


/* End of GDB Methods */

std::list<AstNode *> GetRouteToNode(AstNode *src, AstNode *dst);

static std::string GetNodeNormalName(AstNode *node)
{
	try {
		return node->GetNormalName();
	}
	catch (Error e){
		return PrintString("Error getting node description: %s", e.GetMessage().c_str());
	}
}

void CompilerError::OutputCompilerMessage(AstNodeRoot *root, CompileContext &context, bool pos, FILE *f)
{
	#if DEBUG
		
		root->SetTreeCodeGenerator(CODEGEN_PSEUDOPC);
		fprintf(stderr, "\n[*] Current pseudocode:\n\n%s\n", Indent(root->GetCodeGenerator<CodeGeneratorRootPseudopc>()->Generate(context), 1).c_str());
		
	#endif
	
	std::string backtrace;
	std::list<AstNode *> route = GetRouteToNode(root, m_node);
	
	if (route.size()){
		for (auto node: route){
			backtrace += PrintString("[*] %s -- %s\n", node->GetSourceLocation().c_str(), GetNodeNormalName(node).c_str());
		}
	}
	else {
		backtrace = PrintString("Error: Node (%p, %s, %s) does not exist in AST!",
			(void *)m_node, m_node->GetSourceLocation().c_str(), GetNodeNormalName(m_node).c_str());
	}
	
	
	if (pos){
		fprintf(f, "[*] Error (in compiler file %s at line %i): %s\n%s\n",
			GetFile(), GetLine(), GetMessage().c_str(), Indent(backtrace, 1).c_str());
	}
	else {
		fprintf(f, "[*] Error: %s\n%s\n", GetMessage().c_str(), Indent(backtrace, 1).c_str());
	}
}




void PrintLocatedError(const char *filename, int line, const char *format, ...)
{
	char str[ERROR_MSG_SIZE];
	
	va_list argp;
	va_start(argp, format);
	vsnprintf(str, ERROR_MSG_SIZE, format, argp);
	va_end(argp);
	
	fprintf(stderr, "%s:%i: %s\n", filename, line, str);
}

Error::Error(const char *file, int line) :
	m_file(file), m_line(line)
{
	s_current = this;
}

void Error::OutputMessage(bool pos, FILE *f)
{
	if (pos){
		fprintf(f, "Error in file %s at line %i: %s\n", m_file, m_line, m_message.c_str());
	}
	else {
		fprintf(f, "Error: %s\n", m_message.c_str());
	}
}

std::string Error::GetMessage()
{
	return m_message;
}

void Error::SetMessage(std::string message)
{
	m_message = message;
}

const char *Error::GetFile()
{
	return m_file;
}

int Error::GetLine()
{
	return m_line;
}

void Error::TerminateHandler()
{
	if (s_current){
		s_current->OutputMessage();
	}
	else {
		fprintf(stderr, "Current error is null");
	}
}

Error *Error::InitialiseCurrent()
{
	std::set_terminate(TerminateHandler);
	return 0;
}
Error *Error::Error::s_current = InitialiseCurrent();


std::list<AstNode *> GetRouteToNode(AstNode *src, AstNode *dst)
{
	if (src == dst){
		return {dst};
	}
	
	std::list<AstNode *> children = src->GetChildren();
	for (auto node: children){
		if (node){
			std::list<AstNode *> subpath = GetRouteToNode(node, dst);
			if (subpath.size()){
				subpath.push_front(src);
				return subpath;
			}
		}
	}
	
	return {};
}




#ifdef DEBUG
#include <map>
#include <string>
#include <typeinfo>

#ifdef __GNUG__
#include <cxxabi.h>
#endif

TrackedClass::TrackedClass()
{
	s_undeleted.insert(this);
}

TrackedClass::~TrackedClass()
{
	s_undeleted.erase(this);
}

unsigned int TrackedClass::GetCount()
{
	return s_undeleted.size();
}

void TrackedClass::PrintStatusMessage()
{
	fprintf(stderr, "%u TrackedClass objects unfreed\n", GetCount());
	
	std::map<std::string, unsigned int> infos;
	for (TrackedClass *object: s_undeleted){
		std::string info = object->GetClassName();
		
		if (!infos.count(info)){
			infos[info] = 0;
		}
		
		infos[info]++;
	}
	
	for (auto it: infos){
		fprintf(stderr, "\t%u object of type %s\n", it.second, it.first.c_str());
	}
}

std::string TrackedClass::GetClassName()
{
	return TypeIdToString(typeid(*this));
}

std::string TrackedClass::TypeIdToString(const std::type_info &info)
{
	char b[1024];
	#ifdef __GNUG__
		char *nice = abi::__cxa_demangle(info.name(), 0, 0, 0);
		snprintf(b, 1024, "%s (%s)", info.name(), nice);
		free(nice);
	#else
		snprintf(b, 1024, "%s", info.name());
	#endif
	
	return b;
}

std::set<TrackedClass *> TrackedClass::s_undeleted;
Tracking TrackedClass::s_tracking;


Tracking::~Tracking()
{
	fprintf(stderr, " ======== Exiting with status ========\n");
	TrackedClass::PrintStatusMessage();
}


#else


TrackedClass::~TrackedClass()
{
}

#endif
