/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ERROR_HPP
#define ERROR_HPP

#include <cstdio>
#include <stdarg.h>
#include <string>

#include "util.hpp"

#ifdef DEBUG
	#define ERROR_POS_DEFAULT true
#else
	#define ERROR_POS_DEFAULT false
#endif

#define ERROR_MSG_SIZE 4096
#define FormatError(...) Error(__FILE__, __LINE__, __VA_ARGS__)
#define FormatCompilerError(...) CompilerError(__FILE__, __LINE__, __VA_ARGS__)
#define TestRealValue(v) if ((isnan(v)) || (isinf(v)))\
	{throw FormatError("Real value is not valid: %f!", v);}

void PrintLocatedError(const char *filename, int line, const char *format, ...);
#define PrintCompilerErrorYyl(filename, yyl, ...)\
	PrintLocatedError(filename, yyl.last_line, __VA_ARGS__);

#ifdef DEBUG
	#define AssertError(condition, ...) if (!(condition)) throw Error(__FILE__, __LINE__, __VA_ARGS__)
#else
	#define AssertError(condition, ...) /*Do nothing*/
#endif

class CompileContext;
class Error
{
	public:
		Error(const char *file, int line);
		
		#ifdef __GNUC__
			Error(const char *file, int line, const char *format, ...)  __attribute__((format(printf, 4, 5)));
		#else
			Error(const char *file, int line, const char *format, ...);
		#endif
		void OutputMessage(bool pos = ERROR_POS_DEFAULT, FILE *f = stderr);
		std::string GetMessage();
		
		const char *GetFile();
		int GetLine();
		
	protected:
		void SetMessage(std::string message);
		
	private:
		static void TerminateHandler();
		static Error *InitialiseCurrent();
		
		std::string m_message;
		const char *m_file;
		int m_line;
		
		static Error *s_current;
};

class AstNode;
class AstNodeRoot;
class CompilerError : public Error
{
	public:
		#ifdef __GNUC__
			CompilerError(const char *file, int line, AstNode *node, const char *format, ...)  __attribute__((format(printf, 5, 6)));
		#else
			CompilerError(const char *file, int line, AstNode *node, const char *format, ...);
		#endif
		
		void OutputCompilerMessage(AstNodeRoot *root, CompileContext &context, bool pos = ERROR_POS_DEFAULT, FILE *f = stderr);
		
	private:
		AstNode *m_node;
};

#ifdef DEBUG
#include <set>
#include <typeinfo>

class Tracking
{
	public:
		virtual ~Tracking();
		
	private:
};

class TrackedClass
{
	public:
		TrackedClass();
		virtual ~TrackedClass(); // = default here would cause looser throw specifier errors
		
		template <typename T> static std::string TypeToString()
		{
			return TypeIdToString(typeid(T));
		}
		
		static std::string TypeIdToString(const std::type_info &info);
		static unsigned int GetCount();
		static void PrintStatusMessage();
		std::string GetClassName();
		
	private:
		static std::set<TrackedClass *> s_undeleted;
		static Tracking s_tracking;

};
#else
class TrackedClass
{
	public:
		virtual ~TrackedClass(); // = default here would cause looser throw specifier errors

};
#endif

#if DEBUG
template <typename T> T *tracked_cast(TrackedClass *p)
{
	if (!p){
		throw FormatError("Cannot cast null pointer to %s", TrackedClass::TypeToString<T>().c_str());
	}
	
	T *r = dynamic_cast<T *>(p);
	if (!r){
		throw FormatError("Could not perform tracked_cast from %s to %s", p->GetClassName().c_str(), TrackedClass::TypeToString<T>().c_str());
	}
	return r;
}
#else
template <typename T> T *tracked_cast(TrackedClass *p)
{
	return (T *)p;
}
#endif

#endif
