/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stddef.h>
#include <stdint.h>
#include <unistd.h>

#include <stdarg.h>
#include <cstdio>
#include <cstdlib>
#include <cstring>

#include <errno.h>
#include <pthread.h>

#include <list>
#include <set>

/* anonymous namespace reduces the amount of junk in the symbol table */
namespace { namespace LibPolycute { /* Polycute namespace */

#define Error(...) _Error(__FILE__, __LINE__, __VA_ARGS__)

void _Error(const char *file, int line, const char *format, ...)
{
	char str[4096];
	
	va_list argp;
	va_start(argp, format);
	vsnprintf(str, 4096, format, argp);
	va_end(argp);
	
	
	fprintf(stderr, "File \"%s\", line %i: %s\n", file, line, str);
	exit(-1);
}

class Mutex
{
	public:
		Mutex() : m_locked(false)
		{
			int e = pthread_mutex_init(&m_mutex, 0);
			if (e){
				Error("libpolycute: Could not execute pthread_mutex_init(): %s", strerror(e));
			}
		}
		
		//The destructor assumes at most one thread has this locked/is waiting for the lock
		virtual ~Mutex()
		{
			if (m_locked){
				Unlock();
			}
			
			//TODO: Make sure it's not locked
			int e = pthread_mutex_destroy(&m_mutex);
			if (e){
				Error("libpolycute: Could not execute pthread_mutex_destroy(): %s", strerror(e));
			}
		}
		
		void Lock()
		{
			int e = pthread_mutex_lock(&m_mutex);
			if (e){
				Error("libpolycute: Could not execute pthread_mutex_lock(): %s, %p", strerror(e), (void *)this);
			}
			m_locked = true;
		}
		
		void Unlock()
		{
			m_locked = false;
			int e = pthread_mutex_unlock(&m_mutex);
			if (e){
				Error("libpolycute: Could not execute pthread_mutex_unlock(): %s", strerror(e));
			}
		}
		
	private:
		pthread_mutex_t m_mutex;
		bool m_locked;
};

template <typename T> class MutexItem
{
	public:
		MutexItem()
		{
		}
		
		MutexItem(T item) : m_item(item)
		{
		}
		
		void Lock()
		{
			m_mutex.Lock();
		}
		
		void Unlock()
		{
			m_mutex.Unlock();
		}
		
		T &Get()
		{
			return m_item;
		}
		
	private:
		Mutex m_mutex;
		T m_item;
};

class Condition
{
	public:
		Condition()
		{
			int e = pthread_mutex_init(&m_mutex, 0);
			if (e){
				Error("libpolycute: Could not execute pthread_mutex_init(): %s", strerror(e));
			}
			
			e = pthread_cond_init(&m_cond, 0);
			if (e){
				Error("libpolycute: Could not execute pthread_cond_init(): %s", strerror(e));
			}
		}
		
		virtual ~Condition()
		{
			//TODO: Make sure it's not locked
			int e = pthread_mutex_destroy(&m_mutex);
			if (e){
				Error("libpolycute: Could not execute pthread_mutex_destroy(): %s", strerror(e));
			}
			
			e = pthread_cond_destroy(&m_cond);
			if (e){
				Error("libpolycute: Could not execute pthread_cond_destroy(): %s", strerror(e));
			}
		}
		
		void Lock()
		{
			int e = pthread_mutex_lock(&m_mutex);
			if (e){
				Error("libpolycute: Could not execute pthread_mutex_lock(): %s, %p", strerror(e), (void *)this);
			}
		}
		
		void Unlock()
		{
			int e = pthread_mutex_unlock(&m_mutex);
			if (e){
				Error("libpolycute: Could not execute pthread_mutex_unlock(): %s", strerror(e));
			}
		}
		
		void Wait()
		{
			Lock();
			WaitLocked();
			Unlock();
		}
		
		void WaitLocked()
		{
			int e = pthread_cond_wait(&m_cond, &m_mutex);
			if (e){
				Error("libpolycute: Could not execute pthread_cond_wait(): %s", strerror(e));
			}
		}
		
		void Signal()
		{
			int e = pthread_cond_broadcast(&m_cond);
			if (e){
				Error("libpolycute: Could not execute pthread_cond_broadcast(): %s", strerror(e));
			}
		}
		
	private:
		pthread_mutex_t m_mutex;
		pthread_cond_t m_cond;
};

template <typename T> class ConditionItem
{
	public:
		ConditionItem()
		{
		}
		
		ConditionItem(T item) : m_item(item)
		{
		}
		
		void Lock()
		{
			m_condition.Lock();
		}
		
		void Unlock()
		{
			m_condition.Unlock();
		}
		
		void Wait()
		{
			m_condition.Wait();
		}
		
		void WaitLocked()
		{
			m_condition.WaitLocked();
		}
		
		void Signal()
		{
			m_condition.Signal();
		}
		
		T &Get()
		{
			return m_item;
		}
		
	private:
		Condition m_condition;
		T m_item;
};

class ThreadId
{
	public:
		ThreadId(pthread_t pid) : m_pid(pid)
		{
		}
		
		ThreadId(const ThreadId &src)
		{
			m_pid = src.m_pid;
		}
		
		ThreadId &operator=(const ThreadId &src)
		{
			m_pid = src.m_pid;
			return *this;
		}
		
		bool operator==(const ThreadId &rhs) const
		{
			return pthread_equal(m_pid, rhs.m_pid);
		}
		
		bool operator!=(const ThreadId &rhs) const
		{
			return !pthread_equal(m_pid, rhs.m_pid);
		}
		
		static ThreadId GetCurrentThreadId()
		{
			return ThreadId(pthread_self());
		}
		
	private:
		pthread_t m_pid;
};

class Thread
{
	public:
		Thread()
		{
			int e = pthread_create(&m_thread, 0, start_routine, (void *)this);
			if (e){
				Error("libpolycute: Could not execute pthread_create(): %s", strerror(e));
			}
		}
		
		virtual ~Thread()
		{
			int e = pthread_join(m_thread, 0);
			if (e){
				Error("libpolycute: Could not execute pthread_join(): %s", strerror(e));
			}
		}
		
		ThreadId GetThreadId()
		{
			return ThreadId(m_thread);
		}
		
	private:
		virtual void Main() = 0;
		
		static void *start_routine(void *arg)
		{
			((Thread *)arg)->Main();
			return 0;
		}
		
		pthread_t m_thread;
};

class SyncItem
{
	public:
		virtual ~SyncItem() = default;
		
	private:
};

class SyncSet : public SyncItem
{
	public:
		SyncSet(SyncSet *parent) : m_parent(parent)
		{
			if (m_parent){
				m_parent->AddSyncItem(this);
			}
		}
		
		virtual ~SyncSet()
		{
			if (m_parent){
				m_parent->RemoveSyncItem(this);
			}
		}
		
		void AddSyncItem(SyncItem *item)
		{
			m_items.Lock();
			m_items.Get().insert(item);
			m_items.Unlock();
		}
		
		void RemoveSyncItem(SyncItem *item)
		{
			m_items.Lock();
			m_items.Get().erase(item);
			m_items.Unlock();
		}
		
		bool IsEmpty()
		{
			bool is_empty = false;
			m_items.Lock();
			is_empty = !m_items.Get().size();
			m_items.Unlock();
			return is_empty;
		}
		
		SyncSet *GetParent()
		{
			return m_parent;
		}
		
	private:
		MutexItem<std::set<SyncItem *>> m_items;
		SyncSet *m_parent;
};

class Job : public SyncItem
{
	public:
		virtual ~Job() = default;
		
		//Returns true if the thread should terminate
		virtual bool Execute() = 0;
		
	private:
};

class JobTerminate : public Job
{
	public:
		virtual bool Execute()
		{
			return true;
		}
		
	private:
};

class JobItem : public Job
{
	public:
		JobItem(SyncSet *sync_set, void (*function)(void *), const void *arguments, size_t arguments_size) :
			m_sync_set(sync_set), m_function(function)
		{
			if (m_sync_set){
				m_sync_set->AddSyncItem(this);
			}
			
			if (arguments_size){
				m_arguments = malloc(arguments_size);
				memcpy(m_arguments, arguments, arguments_size);
			}
			else {
				m_arguments = 0;
			}
		}
		
		virtual ~JobItem()
		{
			if (m_arguments){
				free(m_arguments);
			}
			
			if (m_sync_set){
				m_sync_set->RemoveSyncItem(this);
			}
		}
		
		virtual bool Execute()
		{
			m_function(m_arguments);
			return false;
		}
		
	private:
		SyncSet *m_sync_set;
		void (*m_function)(void *);
		void *m_arguments;
};

class JobQueue
{
	public:
		Job *GetJob(bool wait)
		{
			Job *job = 0;
			
			m_jobs.Lock();
			while (!job){
				if (m_jobs.Get().size()){
					job = m_jobs.Get().front();
					m_jobs.Get().pop_front();
				}
				
				if (!wait){
					break; //We should never wait
				}
				
				if (!job){
					m_jobs.WaitLocked();
				}
			}
			m_jobs.Unlock();
			
			return job;
		}
		
		void AddJob(Job *job)
		{
			m_jobs.Lock();
			m_jobs.Get().push_back(job);
			m_jobs.Unlock();
			m_jobs.Signal();
		}
		
	private:
		ConditionItem<std::list<Job *>> m_jobs;
};

class CpuWorkerThread;
class CpuThreadManager
{
	public:
		static CpuThreadManager &GetGlobal()
		{
			return s_global;
		}
		
		JobQueue &GetJobQueue()
		{
			return m_queue;
		}
		
		SyncSet *GetSyncSetOfCurrentThread()
		{
			SyncSet *sync_set = 0;
			ThreadId id = ThreadId::GetCurrentThreadId();
			
			m_sync_sets.Lock();
			for (auto pair: m_sync_sets.Get()){
				if (pair.first == id){
					sync_set = pair.second;
				}
			}
			m_sync_sets.Unlock();
			
			return sync_set;
		}
		
		void SetSyncSetOfCurrentThread(SyncSet *sync_set)
		{
			ThreadId id = ThreadId::GetCurrentThreadId();
			m_sync_sets.Lock();
			
			/* Remove any sync set that might exist for the current thread */
			bool should_loop = true;
			while (should_loop){
				should_loop = false;
				for (auto pair: m_sync_sets.Get()){
					if (pair.first == id){
						m_sync_sets.Get().remove(pair);
						should_loop = true;
						break;
					}
				}
			}
			
			/* Add the new thread/sync_set */
			m_sync_sets.Get().push_back(std::pair<ThreadId, SyncSet *>(id, sync_set));
			
			m_sync_sets.Unlock();
		}
		
	private:
		CpuThreadManager()
		{
			unsigned int n = GetCpuThreadCount();
			for (unsigned int i = 0; i < n; i++){
				m_threads.insert(CreateWorkerThread());
			}
		}
		
		virtual ~CpuThreadManager();
		
		CpuWorkerThread *CreateWorkerThread();
		
		static unsigned int GetCpuThreadCount()
		{
			long sysconf_result = sysconf(_SC_NPROCESSORS_ONLN);
			if (sysconf_result < 0){
				Error("Cannot detect CPU core count");
			}
			return sysconf_result;
		}
		
		std::set<CpuWorkerThread *> m_threads;
		JobQueue m_queue;
		MutexItem<std::list<std::pair<ThreadId, SyncSet *>>> m_sync_sets;
		
		static CpuThreadManager s_global;
};
CpuThreadManager CpuThreadManager::s_global;


class CpuWorkerThread : public Thread
{
	public:
		CpuWorkerThread(CpuThreadManager &manager) : m_manager(manager)
		{
		}
		
	private:
		virtual void Main()
		{
			while (true){
				Job *job = m_manager.GetJobQueue().GetJob(true);
				bool terminate = job->Execute();
				delete job;
				
				if (terminate){
					return;
				}
			}
		}
		
		CpuThreadManager &m_manager;
};

CpuThreadManager::~CpuThreadManager()
{
	//Push termination jobs (one for each thread)
	//Note: threads don't process anymore jobs after processing a termination
	//so there's no possibility of a thread consuming two terminations
	for (unsigned int i = 0; i < m_threads.size(); i++){
		m_queue.AddJob(new JobTerminate);
	}
	
	for (auto thread: m_threads){
		delete thread;
	}
}

CpuWorkerThread *CpuThreadManager::CreateWorkerThread()
{
	return new CpuWorkerThread(*this);
}



}} extern "C" { /* Start of extern C */ using namespace LibPolycute;




void _libpolycute_sync_start()
{
	CpuThreadManager::GetGlobal().SetSyncSetOfCurrentThread(new SyncSet(CpuThreadManager::GetGlobal().GetSyncSetOfCurrentThread()));
}

void _libpolycute_sync_end()
{
	SyncSet *sync_set = CpuThreadManager::GetGlobal().GetSyncSetOfCurrentThread();
	
	while (!sync_set->IsEmpty()){
		Job *job = CpuThreadManager::GetGlobal().GetJobQueue().GetJob(false);
		
		if (!job){ //The queue is empty, therefore so is the sync_set
			continue; //(we could also break)
		}
		
		bool terminate = job->Execute();
		delete job;
		
		if (terminate){
			CpuThreadManager::GetGlobal().GetJobQueue().AddJob(job); //We need the thread to process this, not the sync_end
			return;
		}
		
		//delete job;
	}
	
	CpuThreadManager::GetGlobal().SetSyncSetOfCurrentThread(sync_set->GetParent());
	delete sync_set;
}

void _libpolycute_enqueue(void (*function)(void *), const void *arguments, size_t arguments_size)
{
	CpuThreadManager::GetGlobal().GetJobQueue().AddJob(new JobItem(CpuThreadManager::GetGlobal().GetSyncSetOfCurrentThread(), function, arguments, arguments_size));
}


Mutex *_libpolycute_allocate_lock()
{
	return new Mutex;
}

void _libpolycute_deallocate_lock(Mutex *lock)
{
	delete lock;
}

void _libpolycute_lock_lock(Mutex *lock)
{
	lock->Lock();
}

void _libpolycute_unlock_lock(Mutex *lock)
{
	lock->Unlock();
}

/*
 * I wonder if we can have OOP stuff that dicatates things like mutual exclusion, what threads things are executed from, etc
 * I wonder if we can avoid concurrency bugs such as the one you get if you don't call DoWorkItem from SyncTracker::EndSync()?
 */


/* End of extern C */ }
