/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../ast/astnode.hpp"
#include "optimiser.hpp"


Optimiser *CreateOptimiserDynamicVectorPointer();
Optimiser *CreateOptimiserSpawn();
Optimiser *CreateOptimiserSpawnMerge();
Optimiser *CreateOptimiserUnpackStruct();
Optimiser *CreateOptimiserUnusedArguments();


std::string Optimiser::PerformOptimisations(std::list<std::string> optimisations, AstNodeRoot *root, CompileContext &context)
{
	unsigned int count = 0;
	std::string report;
	std::list<Optimiser *> optimisers = GetOptimisers(optimisations);
	for (Optimiser *optimiser: optimisers){
		int64_t t1 = GetMilliTime();
		
		printf("[*] Running optimiser %s\n", optimiser->GetName().c_str());
		report += PrintString("[*] %s\n", optimiser->GetName().c_str());
		report += Indent(optimiser->Optimise(root, context), 1);
		report += "\n\n";
		delete optimiser;
		count++;
		
		int64_t t2 = GetMilliTime();
		printf("\tTook %0.03f seconds\n", (double)(t2 - t1)/1000.0);
	}
	printf("[*] Done optimising\n");
	
	report += PrintString("[*] Summary\n"
		"\t[*] Number of optimisations applied: %u\n",
		count);
	
	return report;
}

std::list<Optimiser *> Optimiser::GetOptimisers(std::list<std::string> optimisations)
{
	std::list<Optimiser *> optimisers;
	
	for (std::string optimisation: optimisations){
		
		/*
		 * Optimisation levels
		 */
		if ("0" == optimisation){ //do nothing
			optimisers.clear();
		}
		else if ("1" == optimisation){ //Enables optimisations that do not require all object modules to be compiled with the same compiler and options
			optimisers.clear();
		}
		else if ("2" == optimisation){ //-O1 + Without the limitation from allowing different compilers or options
			optimisers.clear();
		}
		else if ("3" == optimisation){ //-O2 + Optimisations that may not be recommended (but should be better than -O2)
			optimisers.clear();
			
			optimisers.push_back(CreateOptimiserSpawnMerge());
			
			optimisers.push_back(CreateOptimiserUnpackStruct()); //This SHOULD be run before -Ounused-arguments
			optimisers.push_back(CreateOptimiserUnusedArguments()); //This MUST be run before -Ospawn IF we use -Ospawn
			
			optimisers.push_back(CreateOptimiserDynamicVectorPointer());
			
			optimisers.push_back(CreateOptimiserSpawn());
			
			//TODO: An optimiser that makes the reference count decrementation happen earlier (I think this has to be after spawn)
		}
		
		/*
		 * Specific optimisers
		 */
		else if (!optimisation.compare("dynamic-vector-to-pointer")){
			optimisers.push_back(CreateOptimiserDynamicVectorPointer());
		}
		else if (!optimisation.compare("spawn")){
			optimisers.push_back(CreateOptimiserSpawn());
		}
		else if (!optimisation.compare("spawn-merge")){
			optimisers.push_back(CreateOptimiserSpawnMerge());
		}
		else if (!optimisation.compare("unpack-struct")){
			optimisers.push_back(CreateOptimiserUnpackStruct());
		}
		else if (!optimisation.compare("unused-arguments")){
			optimisers.push_back(CreateOptimiserUnusedArguments());
		}
		
		/*
		 * Optimisation flag unrecognised
		 */
		else {
			throw FormatError("Unrecognised optimisation: %s\n", optimisation.c_str());
		}
	}
	
	return optimisers;
}

std::list<AstNode *> Optimiser::FindPathToNode(AstNode *src, AstNode *dst)
{
	if (src == dst){
		return {dst};
	}
	
	std::list<AstNode *> children = src->GetChildren();
	for (auto node: children){
		if (node){
			std::list<AstNode *> subpath = FindPathToNode(node, dst);
			if (subpath.size()){
				subpath.push_front(src);
				return subpath;
			}
		}
	}
	
	return {};
}


