/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <tuple>

#include "../../util.hpp"
#include "../../codegen/llvm/codegen.hpp"
#include "../optimiser.hpp"
#include "../pattern.hpp"


DefineAstNodeStatementMarkedGroup(OptimiserDynamicVectorPointerGroup_Block, "Dynamic Vector Pointer Block");
DefineAstNodeStatementMarkedGroup(OptimiserDynamicVectorPointerGroup_Loop, "Dynamic Vector Pointer Loop");

class OptimiserDynamicVectorPointer : public Optimiser
{
	private:
		virtual std::string GetName()
		{
			return "Dynamic Vector Access to Pointer Access";
		}
		
		virtual std::string Optimise(AstNodeRoot *root, CompileContext &context)
		{
			std::string report;
			
			
			unsigned int dynamic_vector_pointer_count = 0;
			std::string dynamic_vector_pointer_report;
			std::tie(dynamic_vector_pointer_count, dynamic_vector_pointer_report) = OptimiseDynamicVectorPointer(context, root);
			report += "[*] Dynamic Vector Access to Pointer Access\n";
			report += Indent(dynamic_vector_pointer_report, 1);
			
			
			/* Return */
			report += PrintString(
				"[*] Summary:\n"
					"\t[*] Optimised %u instances of dynamic vector access to pointer access\n",
				dynamic_vector_pointer_count
			);
			return report;
		}
		
		/* ======== OptimiseLoopPointerLoop ======== */
		
		/*  This is the conversion that takes place
		
			For
				Stuff 1
				Group
					Preamble
					Loop
						For
					Postamble
				Stuff 2
			
			
			Group
				Preamble
				Loop
					For
						Stuff 1
						For
						Stuff 2
				Postamble
		*/
		
		/* ======== OptimiseDynamicVectorPointer ======== */
		std::tuple<unsigned int, std::string> OptimiseDynamicVectorPointer(CompileContext &context, AstNodeRoot *root)
		{
			/*
			 * Keep looking for new matches until we run out, as applications of
			 * this optimisations might
			 */
			
			unsigned int count = 0; //Not all instances of this can be optimised
			std::string report;
			std::list<AstNode *> processed_nodes;
			
			while (true){
				PatternMatcherCat matcher;
				matcher.AddMany();
				matcher.AddTypeNotNode<AstNodeExpressionLValueVector>(processed_nodes);
				matcher.AddMany();
				matcher.AddNodeMatcher(new NodeMatcherOr({new NodeMatcherType<AstNodeStatementFor>(), new NodeMatcherType<AstNodeStatementWhile>()}));
				matcher.AddType<AstNode>();
				matcher.AddMany();
				
				PatternMatch match = matcher.Match(root);
				if (!match){
					break;
				}
				
				AstNode *container = match[0];
				AstNodeStatement *loop = match[1];
				AstNodeExpressionLValueVector *vector_access = match[2];
				
				
				std::string report_result;
				std::string safety_result = OptimiseDynamicVectorPointer_TestForSafety(container, loop, vector_access);
				if (!safety_result.size()){
					std::string perform_result = OptimiseDynamicVectorPointer_PerformOptimisation(container, loop, vector_access);
					root->Link();
					
					report_result = PrintString("[*] Converting to pointer %s, %s, %s: %s\n",
						container->GetSourceLocation().c_str(), loop->GetSourceLocation().c_str(), vector_access->GetSourceLocation().c_str(), perform_result.c_str());
					count++;
				}
				else {
					report_result = PrintString("[*] Not converting to pointer because %s: %s, %s, %s\n", safety_result.c_str(),
						container->GetSourceLocation().c_str(), loop->GetSourceLocation().c_str(), vector_access->GetSourceLocation().c_str());
				}
				report += report_result;
				
				//In any case, we've processed this variable declaration
				processed_nodes.push_back(vector_access);
			}
			
			return std::tuple<unsigned int, std::string>(count, report);
		}
		
		std::string OptimiseDynamicVectorPointer_TestForSafety(AstNode *container, AstNodeStatement *loop, AstNodeExpressionLValueVector *vector_access)
		{
			//Check that the vector access is to a dynamic vector
			if (!dynamic_cast<AstNodeTypeDynamicVector *>(vector_access->GetLValue()->GetType())){
				return "Vector access is not to a dynamic vector";
			}
			
			if (!container->HasChild(loop, true, false)){
				return "Loop is not child";
			}
			
			if (!container->HasChild(vector_access, true, false)){
				return "Vector access is not child";
			}
			
			if (container->CompareChildren(loop, vector_access, true, false) < 0){
				return "Vector access is not in loop";
			}
			
			//TODO: Check that the original lvalue exists outside the loop
			return "";
		}
		
		std::string OptimiseDynamicVectorPointer_PerformOptimisation(AstNode *container, AstNodeStatement *loop, AstNodeExpressionLValueVector *vector_access)
		{
			unsigned int identifier_id = GetUniqueId();
			bool write = OptimiseDynamicVectorPointer_Write(loop, vector_access);
			
			AstNodeExpressionLValue *dynamic_vector = vector_access->GetLValue(); //This is the dynamic vector, as opposed to the vector access
			//dynamic_vector->SetDoRefCount(false);
			
			std::string get_pointer_identifier = PrintString("dynamic_vector_pointer.%u", identifier_id); //We need lvalues, so let's use this name
			AstNodeDynamicVectorGetPointer *get_pointer = new AstNodeDynamicVectorGetPointer(dynamic_vector->Copy(), true, write); //An expression that is a pointer to the dynamic vector memory
			AstNodeStatementVariableDeclaration *get_pointer_declaration = new AstNodeStatementVariableDeclaration(get_pointer->GetType()->Copy(), get_pointer_identifier); //Variable declaration
			AstNodeStatementVariableAssignment *get_pointer_assignment = new AstNodeStatementVariableAssignment(new AstNodeExpressionLValueIdentifier(get_pointer_identifier), get_pointer);
				//To store the pointer
			
			//AstNodeDynamicVectorCommit *precommit = new AstNodeDynamicVectorCommit(dynamic_vector->Copy()); //we need to commit the read before using the pointer
			//AstNodeDynamicVectorCommit *postcommit = new AstNodeDynamicVectorCommit(dynamic_vector->Copy()); //we need to commit the write after using the pointre
			
			//Create the block
			std::list<AstNodeStatement *> statements;
			statements.push_back(get_pointer_declaration);
			statements.push_back(get_pointer_assignment);
			//statements.push_back(precommit);
			statements.push_back(new OptimiserDynamicVectorPointerGroup_Loop({loop}));
			//statements.push_back(postcommit);
			
			OptimiserDynamicVectorPointerGroup_Block *block = new OptimiserDynamicVectorPointerGroup_Block(statements);
			container->ReplaceNode(loop, block, true, false, false, false);
			
			//Replaces the lvalue of vector_access with an lvalue identifier to get_pointer_declaration
			AstNodeExpressionLValue *original_lvalue = vector_access->GetLValue();
			AstNodeExpressionLValueIdentifier *new_lvalue = new AstNodeExpressionLValueIdentifier(get_pointer_identifier);
			vector_access->ReplaceNode(original_lvalue, new_lvalue, true, false, false, false);
			
			delete dynamic_vector; //All instances of this should be copies now, so we don't need the original
			
			return write ? "read/write" : "read only";
		}
		
		bool OptimiseDynamicVectorPointer_Write(AstNodeStatement *loop, AstNodeExpressionLValueVector *vector_access)
		{
			std::list<AstNode *> processed_nodes;
			while (true) {
				PatternMatcherCat matcher;
				matcher.AddMany(true);
				matcher.AddNode(vector_access, true);
				matcher.AddMany(true);
				matcher.AddTypeNotNode<AstNodeStatementVariableAssignment>(processed_nodes, true);
				matcher.AddMany(true);
				matcher.AddNode(loop, true);
				
				PatternMatch match = matcher.Match(loop);
				if (!match){
					break;
				}
				
				AstNodeStatementVariableAssignment *assignment = match[1];
				
				
				if (assignment->GetLValue() == vector_access){ //It's a write!
					return true;
				}
				
				processed_nodes.push_back(assignment);
			}
			return false;
		}
		
};

Optimiser *CreateOptimiserDynamicVectorPointer()
{
	return new OptimiserDynamicVectorPointer;
} 
 
