/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <tuple>

#include "../../util.hpp"
#include "../../codegen/llvm/codegen.hpp"
#include "../optimiser.hpp"
#include "../pattern.hpp"

DefineAstNodeStatementMarkedGroup(OptimiserSpawnMergeMarker, "Optimiser Spawn Merge Marker");

class OptimiserSpawnMerge : public Optimiser
{
	private:
		virtual std::string GetName()
		{
			return "Spawn Merge";
		}
		
		virtual std::string Optimise(AstNodeRoot *root, CompileContext &context)
		{
			std::string report;
			
			/* Unused Arguments */
			unsigned int for_spawn_count = 0;
			std::string for_spawn_report;
			std::tie(for_spawn_count, for_spawn_report) = OptimiseForSpawn(root, context);
			report += "[*] Unused Parameter Removal\n";
			report += Indent(for_spawn_report, 1);
			
			/* Return */
			report += PrintString("[*] Summary:\n\t[*] Optimised %u instances of unused arguments\n", for_spawn_count);
			return report;
		}
		
		std::tuple<unsigned int, std::string> OptimiseForSpawn(AstNodeRoot *root, CompileContext &context)
		{
			unsigned int count = 0;
			std::string report;
			
			std::list<AstNode *> unsafe;
			while (true){
				PatternMatcherCat matcher;
				matcher.AddMany();
				matcher.AddType<AstNodeStatementParallelSpawn>();
				matcher.AddType<AstNodeStatementFor>();
				matcher.AddTypeNotNode<AstNode>(unsafe);
				matcher.AddMany();
				
				PatternMatch match = matcher.Match(root);
				if (!match){
					break;
				}
				
				AstNode *container = match[0];
				AstNodeStatementFor *for_statement = match[1];
				AstNodeStatementParallelSpawn *spawn_statement = match[2];
				
				std::string safety_result = OptimiseForSpawn_CheckForSafety(container, for_statement, spawn_statement);
				std::string report_result;
				if ("" == safety_result){
					report_result = PrintString("[*] Optimising spawn merge for spawn\n");
					OptimiseForSpawn_PerformOptimisation(context, container, for_statement, spawn_statement);
					root->Link();
					root->Normalize(context);
					count++;
				}
				else {
					unsafe.push_back(container);
					report_result = PrintString("[*] Not optimising spawn merge for spawn: %s\n", safety_result.c_str());
				}
				report += report_result;
			}
			
			return std::tuple<unsigned int, std::string>(count, report);
		}
		
		std::string OptimiseForSpawn_CheckForSafety(AstNode *container, AstNodeStatementFor *for_statement, AstNodeStatementParallelSpawn *spawn_statement)
		{
			if (dynamic_cast<OptimiserSpawnMergeMarker *>(container)){
				return "Container is a spawn merge marker";
			}
			return "";
		}
		
		void OptimiseForSpawn_PerformOptimisation(CompileContext &context, AstNode *container, AstNodeStatementFor *for_statement, AstNodeStatementParallelSpawn *spawn_statement)
		{
			/* Variable names that we're going to use */
			unsigned int id = GetUniqueId();
			std::string original_identifier = for_statement->GetVariableDeclaration()->GetIdentifier();
			std::string core_for_identifier = PrintString("spawn_merge_for_spawn.%u.for_thread", id);
			std::string main_for_identifier = PrintString("spawn_merge_for_spawn.%u.for_main", id);
			std::string remainder_for_identifier = PrintString("spawn_merge_for_spawn.%u.for_remainder", id);
			
			AstNodeExpressionLValueIdentifier *core_for_lvalue = new AstNodeExpressionLValueIdentifier(core_for_identifier);
			AstNodeExpressionLValueIdentifier *main_for_lvalue = new AstNodeExpressionLValueIdentifier(main_for_identifier);
			AstNodeExpressionLValueIdentifier *remainder_for_lvalue = new AstNodeExpressionLValueIdentifier(remainder_for_identifier);
			
			/* Get the stuff we need to build the new control flow*/
			AstNodeStatement *statement = spawn_statement->GetStatement();
			AstNodeExpression *iteration_count = for_statement->GetExpression();
			
			/* Create the expressions we need */
			AstNodeExpression *number_of_threads = new AstNodeExpressionBinaryMultiplication(new AstNodeExpressionInt(context.GetComputeSpec().GetCpu().GetCoreCount()), new AstNodeExpressionInt(2));
			AstNodeExpression *iteration_per_thread = new AstNodeExpressionBinaryDivision(iteration_count->Copy(), number_of_threads->Copy());
			AstNodeExpression *iterations_remaining = new AstNodeExpressionBinaryModulo(iteration_count->Copy(), number_of_threads->Copy());
			
			/* Create the variable declarations */
			AstNodeExpression *main_iteration_count = new AstNodeExpressionBinaryAddition(new AstNodeExpressionBinaryMultiplication(iteration_per_thread->Copy(), core_for_lvalue), main_for_lvalue);
			AstNodeExpression *remainder_iteration_count = new AstNodeExpressionBinaryAddition(new AstNodeExpressionBinaryMultiplication(iteration_per_thread->Copy(), number_of_threads->Copy()), remainder_for_lvalue);
			
			AstNodeStatementVariableDeclaration *main_declaration = new AstNodeStatementVariableDeclaration(main_iteration_count->GetType()->Copy(), original_identifier);
			AstNodeStatementVariableDeclaration *remaining_declaration = new AstNodeStatementVariableDeclaration(remainder_iteration_count->GetType()->Copy(), original_identifier);
			AstNodeStatementVariableAssignment *main_assignment = new AstNodeStatementVariableAssignment(new AstNodeExpressionLValueIdentifier(original_identifier), main_iteration_count);
			AstNodeStatementVariableAssignment *remaining_assignment = new AstNodeStatementVariableAssignment(new AstNodeExpressionLValueIdentifier(original_identifier), remainder_iteration_count);
			
			/* Create the control flow */
			AstNodeStatementBlock *main_block = new AstNodeStatementBlock({main_declaration, main_assignment, statement->Copy()});
			AstNodeStatementBlock *remainder_block = new AstNodeStatementBlock({remaining_declaration, remaining_assignment, statement->Copy()});
			
			AstNodeStatementFor *main_for = new AstNodeStatementFor(iteration_per_thread->GetType()->Copy(), main_for_identifier, iteration_per_thread->Copy(), main_block);
			AstNodeStatementFor *remainder_for = new AstNodeStatementFor(iterations_remaining->GetType()->Copy(), remainder_for_identifier, iterations_remaining, remainder_block);
			
			AstNodeStatementParallelSpawn *main_spawn_statement = new AstNodeStatementParallelSpawn(main_for);
			AstNodeStatementFor *core_for = new AstNodeStatementFor(number_of_threads->GetType()->Copy(), core_for_identifier, number_of_threads->Copy(), main_spawn_statement);
			
			OptimiserSpawnMergeMarker *group = new OptimiserSpawnMergeMarker({core_for, remainder_for});
			
			/* Clean up */
			container->ReplaceNode(for_statement, group, true, false, false, false);
			delete for_statement; //Should delete the spawn statement too
			delete number_of_threads;
			delete iteration_per_thread;
		}
};

Optimiser *CreateOptimiserSpawnMerge()
{
	return new OptimiserSpawnMerge;
} 
 
 
