/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <tuple>

#include "../../util.hpp"
#include "../../codegen/llvm/codegen.hpp"
#include "../optimiser.hpp"
#include "../pattern.hpp"

class OptimiserSpawn : public Optimiser
{
	private:
		virtual std::string GetName()
		{
			return "Spawn";
		}
		
		virtual std::string Optimise(AstNodeRoot *root, CompileContext &context)
		{
			/*
			 * TODO
			 *   Optimise for many small spawnings
			 *   Make this safe in the way that we wanted (e.g: writes to variables/etc) - haven't decided how safe this should be
			 */
			
			
			/*
			 * Detect for spawn with no local data dependencies (so sync not needed to access variables on the stack)
			 * Detect sync for spawn that works even with local data dependencies
			 * 
			 * Add sync as needed to do for spawn without sync? (will this change the semantics?)
			 */
			
			
			std::string report;
			
			unsigned int spawn_count = 0;
			std::string spawn_report;
			std::tie(spawn_count, spawn_report) = OptimiseSpawn(context, root);
			report += "[*] Standard Spawn\n";
			report += Indent(spawn_report, 1);
			
			
			/* Return */
			report += PrintString("[*] Summary\n\t[*] Optimised %u standard spawn instances\n", spawn_count);
			return report;
		}
		
		/*
		 * This part detects and optimises the sync/for/spawn
		 */
		std::tuple<unsigned int, std::string> OptimiseSpawn(CompileContext &context, AstNodeRoot *root)
		{
			/*
			 * We go round getting new matches each time because we might move
			 * the matches we found, which might affect what we subsequently do
			 * with them
			 */
			std::string report;
			
			std::list<AstNode *> processed_spawns;
			
			while (true){
				PatternMatcherCat matcher;
				matcher.AddMany();
				matcher.AddTypeNotNode<AstNodeStatementParallelSpawn>(processed_spawns);
				matcher.AddMany();
				
				PatternMatch match = matcher.Match(root);
				if (!match){
					break;
				}
				
				AstNodeStatementParallelSpawn *spawn_statement = match[0];
				processed_spawns.push_back(spawn_statement);
				
				
				if (ExtractSpawn_TestSafety(root, spawn_statement)){
					ExtractSpawn(root, spawn_statement);
					root->Link();
					report += PrintString("[*] Optimised spawn statement: %s\n", spawn_statement->GetSourceLocation().c_str());
				}
				else {
					report += PrintString("[*] Did not optimise spawn statement: %s\n", spawn_statement->GetSourceLocation().c_str());
				}
			}
			
			return std::tuple<unsigned int, std::string>(processed_spawns.size(), report);
		}
		
		bool ExtractSpawn_TestSafety(AstNodeRoot *root, AstNodeStatementParallelSpawn *spawn_statement)
		{
			return true;
		}
		
		//TODO: Maybe change copy_on_spawn to copy_on_extraction, and put this and its code generators and AST nodes in optimiser for generalised extraction in optimisations
		void ExtractSpawn(AstNodeRoot *root, AstNodeStatementParallelSpawn *spawn_statement)
		{
			/* Get all the information we need */
			unsigned int id = GetUniqueId();
			AstNodeFunctionDeclaration *parent_function = FindParentOf<AstNodeFunctionDeclaration>(root, spawn_statement); //find the function we're in
			std::list<AstNodeStatementVariableDeclaration *> declarations_entire = parent_function->GetVariableDeclarations(spawn_statement); //find the variable declarations we need to care about
			
			/* Get a list of variable declarations that we need to replace */
			std::list<AstNodeStatementVariableDeclaration *> variables;
			std::list<AstNodeStatementVariableDeclaration *> declarations;
			std::map<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableDeclaration *> replacements;
			for (AstNodeStatementVariableDeclaration *declaration: declarations_entire){
				if (VariableIsUsed(spawn_statement, declaration)){
					declarations.push_back(declaration);
					AstNodeStatementVariableDeclaration *declaration_copy = declaration->Copy(); //copy the variable declaration
					declaration_copy->AddCodeGenerator(new CodeGeneratorStatementVariableDeclarationForSpawnLlvm(declaration_copy)); //Replace the code generator in the copy
					replacements[declaration] = declaration_copy; //add the variable declaration copy to the replacements map
					variables.push_back(declaration_copy); //add the declaration copy to the list of variables
				}
			}
			
			/* Modify the existing parts of the AST so it contains what it should for the optimised version */
			AstNodeStatement *spawn_child_statement = spawn_statement->GetStatement(); //Get the child statement of the spawn; it is this that we're pulling out into a new function
			spawn_statement->RemoveStatement(); //Avoid two links to the same child
			spawn_statement->AddCodeGenerator(new CodeGeneratorStatementParallelSpawnExtractedLlvm(spawn_statement, declarations, id)); //change the spawn statement's generator
			spawn_child_statement->ReplaceVariableReferences(replacements); //replace the variable declaration references with the variable declaration copies
			
			/* Prepare the new function */
			//Create the block for the function; contains variable declarations and the spawn child statement
			std::list<AstNodeStatement *> statements; //the list of statements for the block
			for (auto variable: variables){
				statements.push_back(variable); //Add all the variables to the block
			}
			statements.push_back(spawn_child_statement); //add the spawn child statement to the block
			AstNodeStatementBlock *block = new AstNodeStatementBlock(statements); //create the block
			
			//Create the extracted statement
			AstNodeStatementExtracted *extracted = new AstNodeStatementExtracted(block, variables, id);
			
			//Create the function and add it to the tree. Note "Implicit in generator" - stuff that the code generator fills in
			AstNodeStatementVariableDeclaration *extracted_argument = new AstNodeStatementVariableDeclaration(new AstNodeTypePointer(new AstNodeTypeInt(8, false, false)), PrintString("spawn.argument.%u", id));
			extracted_argument->SetArgument(1);
			AstNodeFunctionDeclaration *function_declaration = new AstNodeFunctionDeclaration(new AstNodeTypeVoid, PrintString("spawn.%u", id),
				{extracted_argument}, extracted, false, true, false);
			root->AddFunction(function_declaration); //add the new function to the tree
			
			/*
			 * Perhaps (this allows the extraction to be generic):
			 * Make a function out of the spawn statement's statement
			 * Set the function's generator to be the extracted one
			 * Replace the spawn statement's statement with a function call
			 * Set the function call's generator to be the extracted one
			 */
		}
		
		bool VariableIsUsed(AstNode *node, AstNodeStatementVariableDeclaration *declaration)
		{
			PatternMatcherCat matcher;
			matcher.AddMany(true);
			matcher.AddType<AstNodeType>(true);
			matcher.AddNode(declaration, true);
			matcher.AddType<AstNodeExpressionLValueIdentifier>(true);
			matcher.AddMany(true);
			matcher.AddNode(node, true);
			
			return matcher.Match(node);
		}
};


Optimiser *CreateOptimiserSpawn()
{
	return new OptimiserSpawn;
}
