/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <tuple>

#include "../../util.hpp"
#include "../../codegen/llvm/codegen.hpp"
#include "../optimiser.hpp"
#include "../pattern.hpp"

DefineAstNodeStatementMarkedGroup(OptimiseReadonlyStructElementGroup, "Optimiser Read-only Struct Element");

class OptimiserUnpackStruct : public Optimiser
{
	private:
		virtual std::string GetName()
		{
			return "Unpack Struct";
		}
		
		virtual std::string Optimise(AstNodeRoot *root, CompileContext &context)
		{
			std::string report;
			
			/*
			 * This optimisation is helpful because it reduces the amount of
			 * data that must be copied for things like argument passing (in
			 * combination with unused-arguments), and when copying the struct,
			 * further optimisations may be able to optimise things away more
			 * easily.
			 * Essentially, this optimiser is intended to allow other optimisers
			 * to work.
			 * 
			 * There are two instances of this:
			 *   * Where a parameter is a struct
			 *       * This might have a cycle where structs are passed between
			 *         mutually recursive functions.
			 *   * Where a struct is created within the function
			 */
			
			/* This avoids the need to copy the entire struct (possibly with reference counting operations) */
			unsigned int parameter_struct_count = 0;
			std::string parameter_struct_report;
			std::tie(parameter_struct_count, parameter_struct_report) = OptimiseParameterStruct(context, root);
			report += "[*] Parameter Struct Replacement\n";
			report += Indent(parameter_struct_report, 1);
			
			/* This achieves a similar effect under certain circumstances for spawn */
			unsigned int readonly_struct_count = 0;
			std::string readonly_struct_report;
			//std::tie(readonly_struct_count, readonly_struct_report) = OptimiseReadonlyStructElement(context, root);
			report += "[*] Read-Only Struct Element Replacement\n";
			report += Indent(readonly_struct_report, 1);
			
			/* Clean up */
			for (auto node: m_to_delete){
				delete node;
			}
			
			/* Return */
			report += PrintString("[*] Summary:\n"
				"\t[*] Optimised %u instances of parameter struct\n"
				"\t[*] Optimised %u instances of readonly struct element\n",
				parameter_struct_count, readonly_struct_count);
			return report;
		}
		
		std::tuple<unsigned int, std::string> OptimiseReadonlyStructElement(CompileContext &context, AstNodeRoot *root)
		{
			unsigned int count = 0; //Not all instances of this can be optimised
			std::string report;
			std::list<AstNode *> processed;
			
			while (true){
				PatternMatcherCat matcher;
				matcher.AddMany();
				matcher.AddType<AstNodeTypeStruct>();
				matcher.AddTypeNotNode<AstNodeStatementVariableDeclaration>(processed);
				matcher.AddMany();
				matcher.AddType<AstNodeFunctionDeclaration>();
				matcher.AddType<AstNodeRoot>();
				
				PatternMatch match = matcher.Match(root);
				if (!match){
					break;
				}
				
				AstNodeFunctionDeclaration *function_declaration = match[1];
				AstNodeStatementVariableDeclaration *variable_declaration = match[2];
				AstNodeTypeStruct *struct_type = match[3];
				
				AstNodeStatement *statement = function_declaration->GetStatement();
				
				/* Iterate over its elements */
				std::list<AstNodeStructItem *> items = struct_type->GetLink()->GetItems();
				for (auto item: items){
					std::string report_result;
					std::string safety_result = OptimiseReadonlyStructElement_TestForSafety(statement, variable_declaration, item);
					if (!safety_result.size()){
						report_result = PrintString("[*] Optimising read-only struct element \"%s.%s\" (%s)\n",
							variable_declaration->GetIdentifier().c_str(), item->GetIdentifier().c_str(), variable_declaration->GetSourceLocation().c_str()
						);
						
						OptimiseReadonlyStructElement_PerformOptimisation(function_declaration, statement, variable_declaration, item);
						root->Link();
						count++;
					}
					else {
						report_result = PrintString("[*] Not optimising read-only struct element \"%s.%s\" (%s) because: %s\n",
							variable_declaration->GetIdentifier().c_str(), item->GetIdentifier().c_str(), variable_declaration->GetSourceLocation().c_str(),
							safety_result.c_str()
						);
					}
					
					report += report_result;
				}
				
				//In any case, we've processed this variable declaration
				processed.push_back(variable_declaration);
			}
			
			return std::tuple<unsigned int, std::string>(count, report);
		}
		
		
		/* Optimise Struct Element */
		std::string OptimiseReadonlyStructElement_TestForSafety(AstNodeStatement *statement, AstNodeStatementVariableDeclaration *variable_declaration, AstNodeStructItem *item)
		{
			/* See if there are any write struct accesses (unsafe) */
			if (OptimiseReadonlyStructElement_TestForSafetyStructWrite(statement, variable_declaration, item)){
				return "Write access to struct found";
			}
			
			/* See if there are any write struct *element* accesses (unsafe)  */
			if (OptimiseReadonlyStructElement_TestForSafetyElementWrite(statement, variable_declaration, item)){
				return "Write access to element found";
			}
			
			/* See if there are any read struct element accesses (if not, there's no point in performing the optimisation) */
			if (!OptimiseReadonlyStructElement_TestForSafetyElementRead(statement, variable_declaration, item)){
				return "No read access to element found";
			}
			
			return "";
		}
		
		bool OptimiseReadonlyStructElement_TestForSafetyStructWrite(AstNodeStatement *statement, AstNodeStatementVariableDeclaration *variable_declaration, AstNodeStructItem *item)
		{
			PatternMatcherCat matcher;
			matcher.AddMany(true);
			matcher.AddNode(variable_declaration, true);
			matcher.AddType<AstNodeExpressionLValueIdentifier>(true);
			matcher.AddType<AstNodeStatementVariableAssignment>(true);
			matcher.AddMany(true);
			matcher.AddNode(statement, true);
			
			return !matcher.Match(statement);
		}
		
		bool OptimiseReadonlyStructElement_TestForSafetyElementWrite(AstNodeStatement *statement, AstNodeStatementVariableDeclaration *variable_declaration, AstNodeStructItem *item)
		{
			std::list<AstNode *> processed;
			while (true){
				PatternMatcherCat matcher;
				matcher.AddMany(true);
				matcher.AddNode(variable_declaration, true);
				matcher.AddType<AstNodeExpressionLValueIdentifier>(true);
				matcher.AddType<AstNodeExpressionLValueStructElement>(true);
				matcher.AddTypeNotNode<AstNodeStatementVariableAssignment>(processed, true);
				matcher.AddMany(true);
				matcher.AddNode(statement);
				
				PatternMatch match = matcher.Match(statement);
				if (!match){
					return false;
					break;
				}
				
				AstNodeStatementVariableAssignment *assignment = match[1];
				processed.push_back(assignment);
				
				if (!dynamic_cast<AstNodeExpressionLValueStructElement *>(assignment->GetLValue())){
					continue; //We're not assigning to any struct element
				}
				AstNodeExpressionLValueStructElement *assignment_lvalue = tracked_cast<AstNodeExpressionLValueStructElement>(assignment->GetLValue());
				if (assignment_lvalue->GetIdentifier() != item->GetIdentifier()){
					continue; //We're assigning to a different element
				}
				
				if (!dynamic_cast<AstNodeExpressionLValueIdentifier *>(assignment_lvalue->GetLValue())){
					continue; //We're not assigning to any direct struct (struct within struct isn't processed by this for now)
				}
				AstNodeExpressionLValueIdentifier *assignment_lvalue_identifier = tracked_cast<AstNodeExpressionLValueIdentifier>(assignment_lvalue->GetLValue());
				
				if (assignment_lvalue_identifier->GetLink() == variable_declaration){
					return true; //It's a write access!
				}
			}
		}
		
		bool OptimiseReadonlyStructElement_TestForSafetyElementRead(AstNodeStatement *statement, AstNodeStatementVariableDeclaration *variable_declaration, AstNodeStructItem *item)
		{
			std::list<AstNode *> processed;
			while (true){
				PatternMatcherCat matcher;
				matcher.AddMany(true);
				matcher.AddNode(variable_declaration, true);
				matcher.AddType<AstNodeExpressionLValueIdentifier>(true);
				matcher.AddType<AstNodeExpressionLValueStructElement>(true);
				matcher.AddMany(true);
				matcher.AddTypeNotNode<AstNodeStatementVariableAssignment>(processed, true);
				matcher.AddMany(true);
				matcher.AddNode(statement, true);
				
				PatternMatch match = matcher.Match(statement);
				if (!match){
					return false;
					break;
				}
				
				AstNodeStatementVariableAssignment *assignment = match[1];
				processed.push_back(assignment);
				
				if (assignment->GetExpression()->HasChild(variable_declaration, true, true)){ //This is slightly inaccurate
					return true;
				}
			}
		}
		
		void OptimiseReadonlyStructElement_PerformOptimisation(AstNodeFunctionDeclaration *function_declaration, AstNodeStatement *statement,
			AstNodeStatementVariableDeclaration *variable_declaration, AstNodeStructItem *item)
		{
			AstNodeStatementVariableDeclaration *new_variable_declaration = OptimiseReadonlyStructElement_CreateAccessed(function_declaration, statement, variable_declaration, item);
			OptimiseReadonlyStructElement_ReplaceAccesses(statement, variable_declaration, item, new_variable_declaration);
		}
		
		AstNodeStatementVariableDeclaration *OptimiseReadonlyStructElement_CreateAccessed(AstNodeFunctionDeclaration *function_declaration,
			AstNodeStatement *statement, AstNodeStatementVariableDeclaration *variable_declaration, AstNodeStructItem *item)
		{
			std::pair<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableAssignment *> new_declaration_assignment = OptimiseReadonlyStructElement_CreateDeclaration(variable_declaration, item);
			AstNodeStatementVariableDeclaration *new_variable_declaration = new_declaration_assignment.first;
			AstNodeStatementVariableAssignment *new_variable_assignment = new_declaration_assignment.second;
			
			if (OptimiseReadonlyStructElement_VariableIsDeclaredInStatement(statement, variable_declaration)){
				printf("[*] ReplaceDeclarationInStatement\n");
				OptimiseReadonlyStructElement_ReplaceDeclarationInStatement(statement, variable_declaration, new_variable_declaration, new_variable_assignment);
			}
			else {
				printf("[*] ReplaceDeclarationOutOfStatement %s\n", function_declaration->GetIdentifier().c_str());
				OptimiseReadonlyStructElement_ReplaceDeclarationOutOfStatement(function_declaration, statement, new_variable_declaration, new_variable_assignment);
			}
			
			return new_variable_declaration;
		}
		
		bool OptimiseReadonlyStructElement_VariableIsDeclaredInStatement(AstNodeStatement *statement, AstNodeStatementVariableDeclaration *variable_declaration)
		{
			return statement->HasChild(variable_declaration, true, false);
		}
		
		std::pair<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableAssignment *> OptimiseReadonlyStructElement_CreateDeclaration
			(AstNodeStatementVariableDeclaration *variable_declaration, AstNodeStructItem *item)
		{
			unsigned int id = GetUniqueId();
			std::string identifier = PrintString("optimiser_struct_element.%u", id);
			printf("Creating %s\n", identifier.c_str());
			
			return std::pair<AstNodeStatementVariableDeclaration *, AstNodeStatementVariableAssignment *>(
				new AstNodeStatementVariableDeclaration(item->GetType()->Copy(), identifier),
				new AstNodeStatementVariableAssignment(new AstNodeExpressionLValueIdentifier(identifier),
					new AstNodeExpressionLValueStructElement(new AstNodeExpressionLValueIdentifier(variable_declaration->GetIdentifier()), item->GetIdentifier()))
			);
		}
		
		void OptimiseReadonlyStructElement_ReplaceDeclarationInStatement(AstNodeStatement *statement,
			AstNodeStatementVariableDeclaration *variable_declaration,
			AstNodeStatementVariableDeclaration *new_variable_declaration, AstNodeStatementVariableAssignment *new_variable_assignment)
		{
			OptimiseReadonlyStructElementGroup *group = new OptimiseReadonlyStructElementGroup({variable_declaration, new_variable_assignment, new_variable_declaration});
			statement->ReplaceNode(variable_declaration, group, true, false, true, false);
		}
		
		void OptimiseReadonlyStructElement_ReplaceDeclarationOutOfStatement
			(AstNodeFunctionDeclaration *function_declaration, AstNodeStatement *statement,
			 AstNodeStatementVariableDeclaration *new_variable_declaration, AstNodeStatementVariableAssignment *new_variable_assignment)
		{
			OptimiseReadonlyStructElementGroup *group = new OptimiseReadonlyStructElementGroup({new_variable_declaration, new_variable_assignment, statement});
			function_declaration->ReplaceNode(statement, group, true, false, false, false);
		}
		
		
		void OptimiseReadonlyStructElement_ReplaceAccesses(AstNodeStatement *statement, AstNodeStatementVariableDeclaration *variable_declaration,
			AstNodeStructItem *item, AstNodeStatementVariableDeclaration *new_variable_declaration)
		{
			std::list<AstNode *> processed;
			while (true){
				PatternMatcherCat matcher;
				matcher.AddMany();
				matcher.AddTypeNotNode<AstNodeExpressionLValueStructElement>(processed);
				matcher.AddMany();
				matcher.AddNode(statement);
				
				PatternMatch match = matcher.Match(statement);
				if (!match){
					break;
				}
				
				AstNodeExpressionLValueStructElement *element_access = match[1];
				processed.push_back(element_access);
				
				if (!dynamic_cast<AstNodeExpressionLValueIdentifier *>(element_access->GetLValue())){
					continue; //It's not an identifier lvalue
				}
				AstNodeExpressionLValueIdentifier *lvalue = tracked_cast<AstNodeExpressionLValueIdentifier>(element_access->GetLValue());
				
				if (lvalue->GetLink() != variable_declaration){
					continue; //It's a different variable declaration
				}
				
				if (element_access->GetIdentifier() != item->GetIdentifier()){
					continue; //Different struct element
				}
				
				AstNodeExpressionLValueIdentifier *new_lvalue = new AstNodeExpressionLValueIdentifier(new_variable_declaration->GetIdentifier());
				statement->ReplaceNode(element_access, new_lvalue, true, false, true, false);
				delete element_access;
			}
		}
		
		
		/* Optimise Struct Parameter */
		
		std::tuple<unsigned int, std::string> OptimiseParameterStruct(CompileContext &context, AstNodeRoot *root)
		{
			/*
			 * Keep looking for new matches until we run out, as applications of
			 * this optimisations might 
			 */
			
			unsigned int count = 0; //Not all instances of this can be optimised
			std::string report;
			std::list<AstNode *> processed_struct_parameters;
			
			while (true){
				PatternMatcherCat matcher;
				matcher.AddMany();
				matcher.AddType<AstNodeTypeStruct>();
				matcher.AddTypeNotNode<AstNodeStatementVariableDeclaration>(processed_struct_parameters);
				matcher.AddType<AstNodeFunctionDeclaration>();
				matcher.AddType<AstNodeRoot>();
				
				PatternMatch match = matcher.Match(root);
				if (!match){
					break;
				}
				
				AstNodeFunctionDeclaration *function_declaration = match[1];
				AstNodeStatementVariableDeclaration *variable_declaration = match[2];
				
				
				std::string safety_result = OptimiseParameterStruct_TestForSafety(root, function_declaration, variable_declaration);
				if (!safety_result.size()){
					report += PrintString("[*] Replacing struct parameter \"%s\" in function \"%s\": %s\n",
						variable_declaration->GetIdentifier().c_str(), function_declaration->GetIdentifier().c_str(), variable_declaration->GetSourceLocation().c_str());
					
					OptimiseParameterStruct_PerformOptimisation(root, function_declaration, variable_declaration);
					processed_struct_parameters.clear(); //This instance of the  optimisation may have enabled other instances of the optimisation
						//This doesn't infinite loop because application of this optimisation removes the node
					root->Link();
					count++;
				}
				else {
					report += PrintString("[*] Not replacing struct parameter \"%s\" in function \"%s\" because %s: %s\n",
						variable_declaration->GetIdentifier().c_str(), function_declaration->GetIdentifier().c_str(), safety_result.c_str(), variable_declaration->GetSourceLocation().c_str());
				}
				
				//In any case, we've processed this variable declaration
				processed_struct_parameters.push_back(variable_declaration);
			}
			
			return std::tuple<unsigned int, std::string>(count, report);
		}
		
		std::string OptimiseParameterStruct_TestForSafety(AstNodeRoot *root, AstNodeFunctionDeclaration *function_declaration, AstNodeStatementVariableDeclaration *variable_declaration)
		{
			//TODO: Remove this restriction for -O3 (which will probably reduce this function to true)
			//We can't make the modification because it's imported
			if (!function_declaration->GetStatement()){
				return "function is imported";
			}
			
			//Similarly with exporting
			if (function_declaration->GetExport()){
				return "function is exported";
			}
			
			PatternMatcherCat matcher;
			matcher.AddMany(true);
			matcher.AddNode(variable_declaration, true);
			matcher.AddType<AstNodeExpressionLValue>(true);
			matcher.AddTypeNot<AstNodeExpressionLValueStructElement>(true);
			matcher.AddMany(true);
			matcher.AddType<AstNodeStatement>(true);
			matcher.AddNode(function_declaration, true);
			
			PatternMatch match = matcher.Match(function_declaration);
			if (match){
				return "non-struct accesses are used";
			}
			
			
			return "";
		}
		
		void OptimiseParameterStruct_PerformOptimisation(AstNodeRoot *root, AstNodeFunctionDeclaration *function_declaration, AstNodeStatementVariableDeclaration *variable_declaration)
		{
			/* Replace the argument variable declaration */
			auto replacements = OptimiseParameterStruct_ReplaceVariableDeclaration(root, function_declaration, variable_declaration);
			
			/* Replace all instances of AstNodeExpressionLValueStructElement that reference that variable declaration
			   (Hence why we accept this in hte safety test) */
			OptimiseParameterStruct_ReplaceStructAccesses(root, function_declaration, variable_declaration, replacements);
			
			/* Replace all calls to the function with ones that use the unpacked form */
			OptimiseParameterStruct_ReplaceFunctionCalls(root, function_declaration, tracked_cast<AstNodeTypeStruct>(variable_declaration->GetType())->GetLink());
		}
		
		std::map<std::string, AstNodeStatementVariableDeclaration *> OptimiseParameterStruct_ReplaceVariableDeclaration
			(AstNodeRoot *root, AstNodeFunctionDeclaration *function_declaration, AstNodeStatementVariableDeclaration *variable_declaration)
		{
			std::map<std::string, AstNodeStatementVariableDeclaration *> r;
			std::list<AstNodeStatementVariableDeclaration *> replacements;
			
			//Get the struct definition's items
			std::list<AstNodeStructItem *> struct_items = tracked_cast<AstNodeTypeStruct>(variable_declaration->GetType())->GetLink()->GetItems();
			
			/* Create the new variable declarations */
			for (auto item: struct_items){
				//Create the new declaration
				std::string new_identifier = PrintString("optimiser_parameter_struct.%u.%s", variable_declaration->GetId(), item->GetIdentifier().c_str());
				AstNodeStatementVariableDeclaration *new_declaration = new AstNodeStatementVariableDeclaration(item->GetType()->Copy(), new_identifier);
				r[item->GetIdentifier()] = new_declaration;
				replacements.push_back(new_declaration);
			}
			
			/* Replace the struct variable with the new declarations and renumber all the variable declarations */
			function_declaration->ReplaceArgument(variable_declaration, replacements);
			m_to_delete.push_back(variable_declaration);
			
			return r;
		}
		
		void OptimiseParameterStruct_ReplaceStructAccesses(AstNodeRoot *root, AstNodeFunctionDeclaration *function_declaration,
			AstNodeStatementVariableDeclaration *variable_declaration, std::map<std::string, AstNodeStatementVariableDeclaration *> replacements)
		{
			while (true){
				PatternMatcherCat matcher;
				matcher.AddMany(true);
				matcher.AddNode(variable_declaration, true);
				matcher.AddType<AstNodeExpressionLValue>(true);
				matcher.AddType<AstNodeExpressionLValueStructElement>(true);
				matcher.AddType<AstNode>(true);
				matcher.AddMany(true);
				matcher.AddNode(function_declaration, true);
				
				PatternMatch match = matcher.Match(function_declaration);
				if (!match){
					break;
				}
				
				AstNode *containing_node = match[1];
				AstNodeExpressionLValueStructElement *element_access = match[2];
				
				
				std::map<std::string, AstNode *> replacement;
				if (!replacements.count(element_access->GetIdentifier())){
					std::string map_string;
					
					for (auto r: replacements){
						map_string += PrintString("\t%s -> %p\n", r.first.c_str(), (void *)r.second);
					}
					
					throw FormatError("replacements[%s] does not exist!\n%s", element_access->GetIdentifier().c_str(), map_string.c_str());
				}
				AstNodeStatementVariableDeclaration *declaration = replacements[element_access->GetIdentifier()];
				
				replacement[declaration->GetIdentifier()] = declaration;
				AstNodeExpressionLValueIdentifier *lvalue = new AstNodeExpressionLValueIdentifier(declaration->GetIdentifier());
				lvalue->Link(replacement);
				
				containing_node->ReplaceNode(element_access, lvalue, true, false, false, false);
				m_to_delete.push_back(element_access);
			}
		}
		
		void OptimiseParameterStruct_ReplaceFunctionCalls(AstNodeRoot *root, AstNodeFunctionDeclaration *function_declaration, AstNodeStructDeclaration *struct_declaration)
		{
			std::list<AstNode *> function_calls;
			while (true){
				//Select a function call that links to the function declaration
				PatternMatcherCat matcher;
				matcher.AddMany();
				matcher.AddNode(function_declaration);
				matcher.AddType<AstNodeExpressionLValue>(true);
				matcher.AddTypeNotNode<AstNodeFunctionCall>(function_calls);
				matcher.AddType<AstNode>();
				matcher.AddMany();
				
				PatternMatch match = matcher.Match(root);
				if (!match){
					break;
				}
				
				AstNode *container = match[0];
				AstNodeFunctionCall *function_call = match[1];
				
				function_calls.push_back(function_call);
				
				//This is needed to make sure all expressions are evaluated in the correct order
				AstNodeExpressionBlock *new_expression = OptimiseParameterStruct_UnInlineFunctionCall(function_call);
				container->ReplaceNode(function_call, new_expression, true, false, false, false);
				
				//Now replace LValueIdentifiers that access elements of the struct we're replacing
				OptimiseParameterStruct_ReplaceFunctionCallStructLValues(function_call, struct_declaration);
			}
		}
		
		AstNodeExpressionBlock *OptimiseParameterStruct_UnInlineFunctionCall(AstNodeFunctionCall *function_call)
		{
			/* We want to uninline any expressions that aren't LValueIdentifiers */
			std::list<AstNodeStatement *> statements;
			
			std::list<AstNodeExpression *> expressions = function_call->GetExpressions();
			for (auto expression: expressions){
				//We don't do anything to LValueIdentifiers
				if (!dynamic_cast<AstNodeExpressionLValueIdentifier *>(expression)){
					//Create a variable declaration, and add it to the list
					std::string identifier = PrintString("OptimiseParameterStruct.%u", expression->GetId());
					AstNodeStatementVariableDeclaration *declaration = new AstNodeStatementVariableDeclaration(expression->GetType()->Copy(), identifier);
					AstNodeStatementVariableAssignment *assignment = new AstNodeStatementVariableAssignment(new AstNodeExpressionLValueIdentifier(identifier), expression);
					statements.push_back(declaration);
					statements.push_back(assignment);
					
					//Create the LValueIdentifier
					std::map<std::string, AstNode *> replacement;
					replacement[identifier] = declaration;
					
					AstNodeExpressionLValueIdentifier *lvalue = new AstNodeExpressionLValueIdentifier(identifier);
					lvalue->Link(replacement);
					
					//Replace the expression in the function call with an lvalue
					function_call->ReplaceNode(expression, lvalue, true, false, false, false);
				}
			}
			
			//Create the expression statement
			return new AstNodeExpressionBlock(statements, function_call);
		}
		
		void OptimiseParameterStruct_ReplaceFunctionCallStructLValues(AstNodeFunctionCall *function_call, AstNodeStructDeclaration *struct_declaration)
		{
			//By the time we get here, all the parameters of the function should be AstNodeExpressionLValueIdentifier nodes
			//Get a list of arguments
			std::list<AstNodeExpression *> arguments = function_call->GetExpressions();
			
			for (auto argument: arguments){
				AstNodeExpressionLValueIdentifier *lvalue = tracked_cast<AstNodeExpressionLValueIdentifier>(argument);
				AstNodeTypeStruct *type = dynamic_cast<AstNodeTypeStruct *>(argument->GetType());
				
				if (type){ //it *IS* a structure type
					if (type->GetLink() == struct_declaration){ //This is one we should replace
						std::list<AstNodeExpression *> replacements;
						std::list<AstNodeStructItem *> items = struct_declaration->GetItems();
						
						for (auto item: items){
							replacements.push_back(new AstNodeExpressionLValueStructElement(lvalue->Copy(), item->GetIdentifier()));
						}
						
						function_call->ReplaceArgument(argument, replacements);
						m_to_delete.push_back(argument);
					}
				}
			}
		}
		
		
		std::list<AstNode *> m_to_delete;
};

Optimiser *CreateOptimiserUnpackStruct()
{
	return new OptimiserUnpackStruct;
} 
