/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <tuple>

#include "../../util.hpp"
#include "../../codegen/llvm/codegen.hpp"
#include "../optimiser.hpp"
#include "../pattern.hpp"

class OptimiserUnusedArguments : public Optimiser
{
	private:
		virtual std::string GetName()
		{
			return "Unused Arguments";
		}
		
		virtual std::string Optimise(AstNodeRoot *root, CompileContext &context)
		{
			std::string report;
			
			/* Unused Arguments */
			unsigned int unused_arguments_count = 0;
			std::string unused_arguments_report;
			std::tie(unused_arguments_count, unused_arguments_report) = OptimiseUnusedArguments(root, context);
			report += "[*] Unused Parameter Removal\n";
			report += Indent(unused_arguments_report, 1);
			
			/* Return */
			report += PrintString("[*] Summary:\n\t[*] Optimised %u instances of unused arguments\n", unused_arguments_count);
			return report;
		}
		
		std::tuple<unsigned int, std::string> OptimiseUnusedArguments(AstNodeRoot *root, CompileContext &context)
		{
			unsigned int count = 0;
			std::string report;
			
			std::list<AstNode *> processed_parameters;
			while (true){
				PatternMatcherCat matcher;
				matcher.AddMany();
				matcher.AddTypeNotNode<AstNodeStatementVariableDeclaration>(processed_parameters);
				matcher.AddType<AstNodeFunctionDeclaration>();
				matcher.AddType<AstNodeRoot>();
				
				PatternMatch match = matcher.Match(root);
				if (!match){
					break;
				}
				
				AstNodeFunctionDeclaration *function_declaration = match[1];
				AstNodeStatementVariableDeclaration *variable_declaration = match[2];
				
				processed_parameters.push_back(variable_declaration);
				
				
				std::string safety_result = OptimiseUnusedArguments_CheckForSafety(function_declaration, variable_declaration);
				if ("" == safety_result){
					report += PrintString("[*] Optimising parameter %s in %s\n",
						variable_declaration->GetIdentifier().c_str(), function_declaration->GetIdentifier().c_str());
					OptimiseUnusedArguments_PerformOptimisation(root, function_declaration, variable_declaration);
					processed_parameters.clear(); //If a parameter is simply passed through, we might be able to optimise away more
					root->Link();
					count++;
				}
				else {
					report += PrintString("[*] Not optimising parameter %s in %s because: %s\n",
						variable_declaration->GetIdentifier().c_str(), function_declaration->GetIdentifier().c_str(), safety_result.c_str());
				}
			}
			
			return std::tuple<unsigned int, std::string>(count, report);
		}
		
		std::string OptimiseUnusedArguments_CheckForSafety(AstNodeFunctionDeclaration *function_declaration, AstNodeStatementVariableDeclaration *variable_declaration)
		{
			if (!function_declaration->GetStatement()){
				return "Function is imported";
			}
			
			if (function_declaration->GetExport()){
				return "Function is exported";
			}
			
			if (!function_declaration->GetIdentifier().find(".ast.")){
				return "Function is .ast.*";
			}
			
			PatternMatcherCat matcher;
			matcher.AddMany();
			matcher.AddNode(variable_declaration, true);
			matcher.AddMany(true);
			matcher.AddType<AstNodeStatement>();
			matcher.AddNode(function_declaration);
			
			if (matcher.Match(function_declaration)){
				return "Parameter is used";
			}
			
			return "";
		}
		
		void OptimiseUnusedArguments_PerformOptimisation(AstNodeRoot *root, AstNodeFunctionDeclaration *function_declaration, AstNodeStatementVariableDeclaration *variable_declaration)
		{
			unsigned int n = function_declaration->ReplaceArgument(variable_declaration, {}); //Remove the argument from the function declaration
			
			//Replace the function calls
			std::list<AstNode *> function_calls;
			while (true){
				//Select a function call that links to the function declaration
				PatternMatcherCat matcher;
				matcher.AddMany();
				matcher.AddNode(function_declaration);
				matcher.AddType<AstNodeExpressionLValue>(true);
				matcher.AddTypeNotNode<AstNodeFunctionCall>(function_calls);
				matcher.AddMany();
				
				PatternMatch match = matcher.Match(root);
				if (!match){
					break;
				}
				
				AstNodeFunctionCall *function_call = match[0];
				
				function_calls.push_back(function_call);
				
				OptimiseUnusedArguments_UpdateFunctionCall(function_call, n);
			}
		}
		
		
		void OptimiseUnusedArguments_UpdateFunctionCall(AstNodeFunctionCall *function_call, unsigned int n)
		{
			std::list<AstNodeExpression *> expressions = function_call->GetExpressions();
			AstNodeExpression *argument = 0;
			
			for (auto expression: expressions){
				if (!n){
					argument = expression;
				}
				n--;
			}
			
			function_call->ReplaceArgument(argument, {});
			delete argument;
		}
};

Optimiser *CreateOptimiserUnusedArguments()
{
	return new OptimiserUnusedArguments;
} 
 
