/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "pattern.hpp"



PatternMatcher::PatternMatcher() :
	m_refcount(0)
{
}

void PatternMatcher::IncRefCount()
{
	m_refcount++;
}

void PatternMatcher::DecRefCount()
{
	m_refcount--;
	if (!m_refcount){
		delete this;
	}
}

PatternMatch PatternMatcher::Match(AstNode *node)
{
	std::set<AstNode *> previous;
	return Match(node, previous);
}



PatternMatcherNode::PatternMatcherNode(NodeMatcher *node_matcher, std::list<std::list<PatternMatcher *>> children, bool index, bool recurse_children, bool recurse_links, bool accept_empty) :
	m_node_matcher(node_matcher), m_index(index), m_recurse_children(recurse_children), m_recurse_links(recurse_links), m_accept_empty(accept_empty)
{
	for (std::list<PatternMatcher *> &child_list: children){
		std::list<PatternMatcher *> m_child_list;
		for (PatternMatcher *child: child_list){
			if (child){ //Don't reference count such that we have circular references
				child->IncRefCount();
			}
			else {
				child = this;
			}
			m_child_list.push_back(child);
		}
		m_children.push_back(m_child_list);
	}
}

PatternMatcherNode::~PatternMatcherNode()
{
	for (std::list<PatternMatcher *> &child_list: m_children){
		for (PatternMatcher *child: child_list){
			if (this != child){
				child->DecRefCount();
			}
		}
	}
	
	delete m_node_matcher;
}

PatternMatch PatternMatcherNode::Match(AstNode *node, std::set<AstNode *> &previous)
{
	
	
	if (m_accept_empty){ //If this is so, try skipping this matcher altogether
		for (std::list<PatternMatcher *> &children: m_children){
			if (!children.size() && !node){ //We're at the end of the path, AND the matcher list
				return PatternMatch(0, {}, m_index);
			}
			else if (!node){ //We're at the end of the path, but not the end of the matcher
				continue;
			}
			
			if (children.size() != 1){
				continue;
			}
			
			if (this == children.front()){ //If we allow this, we'll get infinite recursion
				continue;
			}
			
			PatternMatch match = children.front()->Match(node, previous);
			if (match){ //We found a match that way!
				return match;
			}
		}
		
		if (!node){ //We're at the end of the path, but didn't find an empty list
			return PatternMatch();
		}
	}
	
	if (!m_node_matcher->Match(node)){
		return PatternMatch(); //Invalid match
	}
	
	std::list<AstNode *> child_nodes;
	GetChildren(node, child_nodes, previous);
	
	for (std::list<PatternMatcher *> &children: m_children){
		PatternMatch match = MatchFromList(node, children, child_nodes, previous);
		if (match){
			return match;
		}
	}
	return PatternMatch();
}

PatternMatch PatternMatcherNode::MatchFromList(AstNode *node, std::list<PatternMatcher *> &child_matchers, std::list<AstNode *> child_nodes, std::set<AstNode *> &previous)
{
	std::list<PatternMatch> child_matches;
	for (PatternMatcher *child_matcher: child_matchers){
		while (true){
			if (!child_nodes.size()){ //We've reached the end of the path
				PatternMatch match = child_matcher->Match(0);
				if (match){
					child_matches.push_back(match);
					break; //Stop eating nodes for this matcher
				}
				else { //We couldn't match all the child matchers
					return PatternMatch(); //Invalid match
				}
			}
			
			AstNode *child_node = child_nodes.front();
			child_nodes.pop_front();
			
			previous.insert(node);
			PatternMatch match = child_matcher->Match(child_node, previous);
			previous.erase(node);
			
			if (match){
				child_matches.push_back(match);
				break; //Stop eating nodes for this matcher
			}
		}
	}
	
	return PatternMatch(node, child_matches, m_index);
}


void PatternMatcherNode::GetChildren(AstNode *node, std::list<AstNode *> &children, std::set<AstNode *> &previous)
{
	if (!node){
		return;
	}
	
	if (previous.count(node)){
		return; //Prevent cycles
	}
	
	if (m_recurse_children){
		std::list<AstNode *> l = node->GetChildren();
		for (AstNode *i: l){
			if (i){
				children.push_back(i);
			}
		}
	}
	
	if (m_recurse_links){
		std::list<AstNode *> l = node->GetLinks();
		for (AstNode *i: l){
			if (i){
				children.push_back(i);
			}
		}
	}
}




PatternMatcherCat::PatternMatcherCat() :
	m_top_matcher(0)
{
}

PatternMatcherCat::~PatternMatcherCat()
{
	if (m_top_matcher){
		m_top_matcher->DecRefCount();
	}
}

PatternMatch PatternMatcherCat::Match(AstNode *node)
{
	std::set<AstNode *> previous;
	return Match(node, previous);
}

PatternMatch PatternMatcherCat::Match(AstNode *node, std::set<AstNode *> &previous)
{
	return m_top_matcher->Match(node, previous);
}

void PatternMatcherCat::AddNode(AstNode *node, bool recurse_links, bool recurse_children)
{
	AddNodeMatcher(new NodeMatcherNode(node), recurse_links, recurse_children);
}

void PatternMatcherCat::AddNodeMatcher(NodeMatcher *node_matcher, bool recurse_links, bool recurse_children)
{
	SetTopPatternMatcher(new PatternMatcherNode(node_matcher,
		m_top_matcher ? std::list<std::list<PatternMatcher *>>{{m_top_matcher}} : std::list<std::list<PatternMatcher *>>{},
		true, recurse_children, recurse_links, false));
}

void PatternMatcherCat::AddMany(bool recurse_links, bool recurse_children)
{
	SetTopPatternMatcher(new PatternMatcherNode(new NodeMatcherTrue(),
		m_top_matcher ? std::list<std::list<PatternMatcher *>>{{m_top_matcher}, {0}} : std::list<std::list<PatternMatcher *>>{{}, {0}},
		false, recurse_children, recurse_links, true));
}

void PatternMatcherCat::SetTopPatternMatcher(PatternMatcher *matcher)
{
	matcher->IncRefCount();
	m_top_matcher = matcher;
}





