/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef OPTIMISER_PATTERNMATCHER_HPP
#define OPTIMISER_PATTERNMATCHER_HPP

#include <list>
#include <set>

#include "../error.hpp"
#include "../ast/astnode.hpp"


class PatternMatcher : public TrackedClass
{
	public:
		PatternMatcher();
		
		PatternMatch Match(AstNode *node);
		virtual PatternMatch Match(AstNode *node, std::set<AstNode *> &previous) = 0;
		
		void IncRefCount();
		void DecRefCount();
		
	private:
		unsigned int m_refcount;
};

//TODO/NOTE: Perhaps add an add chidlren interface to the above?

class PatternMatcherNode : public PatternMatcher
{
	public:
		PatternMatcherNode(NodeMatcher *node_matcher, std::list<std::list<PatternMatcher *>> children, bool index, bool recurse_children, bool recurse_links, bool accept_empty);
		virtual ~PatternMatcherNode();
		
		/*
		 * This returns a valid match if CurrentMatch returns true and
		 * Each child matcher returns a valid match, in order
		 * 
		 * The resulting match is that gotten if each child matcher is applied
		 * to the earliest possible child node
		 * 
		 * The list of lists is because we only require that one of the child lists works.
		 * Returns the first match
		 */
		virtual PatternMatch Match(AstNode *node, std::set<AstNode *> &previous);
		
	private:
		PatternMatcherNode(const PatternMatcherNode &) = delete;
		PatternMatcherNode &operator=(const PatternMatcherNode &) = delete;
		
		PatternMatch MatchFromList(AstNode *node, std::list<PatternMatcher *> &child_matchers, std::list<AstNode *> child_nodes, std::set<AstNode *> &previous);
		void GetChildren(AstNode *node, std::list<AstNode *> &children, std::set<AstNode *> &previous);
		
		NodeMatcher *m_node_matcher;
		std::list<std::list<PatternMatcher *>> m_children;
		bool m_index;
		bool m_recurse_children;
		bool m_recurse_links;
		bool m_accept_empty;
};

class PatternMatcherCat : public PatternMatcher
{
	public:
		PatternMatcherCat();
		virtual ~PatternMatcherCat();
		
		PatternMatch Match(AstNode *node); //I do *NOT* understand why this is necessary...
		virtual PatternMatch Match(AstNode *node, std::set<AstNode *> &previous);
		
		
		/* Adds matchers to the FRONT */
		
		template <typename T> void AddTypeNotNode(std::list<AstNode *> nodes, bool recurse_links = false, bool recurse_children = true)
		{
			std::list<NodeMatcher *> node_matchers;
			for (AstNode *node: nodes){
				node_matchers.push_back(new NodeMatcherNot(new NodeMatcherNode(node)));
			}
			node_matchers.push_back(new NodeMatcherType<T>());
			
			SetTopPatternMatcher(new PatternMatcherNode(new NodeMatcherAnd(node_matchers),
				m_top_matcher ? std::list<std::list<PatternMatcher *>>{{m_top_matcher}} : std::list<std::list<PatternMatcher *>>{},
				true, recurse_children, recurse_links, false));
			
		}
		
		template <typename T> void AddTypeNot(bool recurse_links = false, bool recurse_children = true)
		{
			AddNodeMatcher(new NodeMatcherNot(new NodeMatcherType<T>()), recurse_links, recurse_children);
		}
		
		template <typename T> void AddType(bool recurse_links = false, bool recurse_children = true)
		{
			AddNodeMatcher(new NodeMatcherType<T>(), recurse_links, recurse_children);
		}
		
		void AddNode(AstNode *node, bool recurse_links = false, bool recurse_children = true);
		void AddNodeMatcher(NodeMatcher *node_matcher, bool recurse_links = false, bool recurse_children = true);
		void AddMany(bool recurse_links = false, bool recurse_children = true);
		
	private:
		void SetTopPatternMatcher(PatternMatcher *matcher);
		
		PatternMatcher *m_top_matcher;
};



#endif
