/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

%{
	#include "parser.hpp"
	#include "polycute.yy.hpp"
	#include "polycute.ll.hpp"
	#include "../error.hpp"
	
	#define DoParserError ((ParserData *)parserdata)->error = true;
	#define PrintCompilerError(...) PrintCompilerErrorYyl(((ParserData *)parserdata)->filename, yylloc, __VA_ARGS__) DoParserError
	
	void yyerror(const char *);
	
	#define YYPARSE_PARAM parserdata
	#define YYLEX_PARAM  ((yyscan_t *)(((ParserData *)parserdata)->scanner))[0]
	
	void yyerror (YYLTYPE *locp, void *parserdata, char const *msg)
	{
		YYLTYPE yylloc = *locp;
		PrintCompilerError("%s", msg);
	}
%}

/* Setup stuff */
%define api.prefix polycute_parser_
%glr-parser
%pure_parser
%lex-param {YYLEX_PARAM}
%locations
%error-verbose
%token_table
%parse-param {void *parserdata}

%start start

/* Tokens */
%token LTE "<="
%token GTE ">="
%token EQU "=="
%token NEQ "!="

%token SHL "<<"
%token SHR ">>"

%token SPLAT "->"

%token IF "if"
%token ELSE "else"
%token WHILE "while"
%token FOR "for"

%token SPAWN "spawn"
%token SYNC "sync"

%token TO "to"
%token LENGTH "length"

%token LOCK "lock"
%token ATOMIC "atomic"

%token VOID "void"
%token BOOL "bool"
%token INT "int"
%token FLOAT "float"
%token STRUCT "struct"

%token STRICT "strict"
%token SIGNED "signed"

%token POINTER "pointer"
%token REF "ref"
%token DEREF "deref"
%token ALLOC "alloc"
%token DEALLOC "dealloc"

%token EXPORT "export"
%token IMPORT "import"
%token CCALL "ccall"

%token RETURN "return"

%token TRUE "true"
%token FALSE "false"
%token NULL_REF "null"

%token IDENTIFIER "identifier"
%token INTEGER "integer constant"
%token DECIMAL "floating point constant"

%token END 0 "end of file"


/* Presedence */

%left '+' '-'
%left '*' '/'
%left '%'

/* Types */

%union {
	/* Lex Stuff */
	const char *text;
	
	/* Basic values */
	bool boolean;
	unsigned int integer;
	double floating_point;
	char identifier[1024];
	
	/* Imperative stuff */
	AstNodeRoot *root;
	AstNodeStatementBlock *block;
	std::list<AstNodeStatement *> *statement_list;
	AstNodeStatement *statement;
	AstNodeStatementVariableDeclaration *vardec;
	std::list<AstNodeStatementVariableDeclaration *> *vardec_list;
	AstNodeFunctionCall *function_call;
	
	/* Struct stuff */
	AstNodeStructItem *struct_member;
	std::list<AstNodeStructItem *> *struct_member_list;
	AstNodeStructDeclaration *struct_declaration;
	
	/* Types */
	AstNodeType *type;
	std::list<AstNodeType *> *type_list;
	
	/* Expressions */
	AstNodeExpression *expression;
	std::list<AstNodeExpression *> *expression_list;
	AstNodeExpressionLValue *lvalue;
	
	/* Top Level */
	AstNode *root_item;
	std::list<AstNode *> *root_item_list;
	
	/* Structures used in parsing */
	FunctionAttributes *function_attributes;
};

/* Top Level */
%type <root> start
%type <root_item_list> root_list
%type <root_item> root_item

/* Imperative Stuff */
%type <block> block
%type <vardec_list> vardec_list
%type <vardec> vardec
%type <statement> vardeca
%type <statement_list> statement_list
%type <statement> statement
%type <function_call> function_call

/* Struct stuff */
%type <struct_member> struct_member
%type <struct_member_list> struct_member_list

/* Types */
%type <type> type
%type <type_list> type_list

/* Expressions */
%type <expression> expression
%type <expression_list> expression_list
%type <lvalue> lvalue
%type <lvalue> lvalue_pre_lock

/* Basic Values */
%type <identifier> identifier
%type <boolean> boolean
%type <integer> integer
%type <floating_point> float

/* Terminals */
%type <identifier> IDENTIFIER
%type <integer> INTEGER
%type <floating_point> DECIMAL

/* Structures used in parsing */
%type <function_attributes> function_attributes;

/*
	Expect shift/reduce
	* Between function declarations and global variable declarations
 */
%expect 267
%expect-rr 9

%%

/* Top Level */
start: 
	root_list {
		AstNodeRoot *root = new AstNodeRoot(*$1);
		root->SetSourceLocation("", @$.first_line);
		delete $1;
		((ParserData *)parserdata)->top = root;
	};
root_list:
	root_item root_list {$2->push_front($1); $$ = $2;} |
	{$$ = new std::list<AstNode *>;};
	
root_item:
	//vardeca ';' {$$ = $1;} |
	function_attributes type identifier '(' vardec_list ')' statement
	{
		if ($1->ccall && !($1->imported || $1->exported)){
			yyerror (&yylloc, parserdata, "Function declared ccall but not import or export");
		}
		if ($1->imported){
			yyerror (&yylloc, parserdata, "Imported function provides implementation");
		}
		$$ = new AstNodeFunctionDeclaration($2, $3, *$5, $7, $1->exported, $1->ccall, $1->lock);
		delete $5;
		$$->SetSourceLocation("", @$.first_line);
	} |
	function_attributes type identifier '(' vardec_list ')' ';'
	{
		if (!$1->imported){
			yyerror (&yylloc, parserdata, "No implementation provided for non-imported function");
		}
		$$ = new AstNodeFunctionDeclaration($2, $3, *$5, 0, $1->exported, $1->ccall, $1->lock);
		delete $5;
		$$->SetSourceLocation("", @$.first_line);
	} |
	STRUCT identifier '{' struct_member_list '}' {
		$$ = new AstNodeStructDeclaration($2, *$4);
		delete $4;
		$$->SetSourceLocation("", @$.first_line);
	};

struct_member_list:
	{$$ = new std::list<AstNodeStructItem *>;} |
	struct_member struct_member_list {$2->push_front($1); $$ = $2;};

struct_member:
	type identifier ';' {$$ = new AstNodeStructItem($1, $2); $$->SetSourceLocation("", @$.first_line);};

function_attributes:
	CCALL function_attributes
	{
		$$ = $2;
		if ($$->ccall){
			yyerror (&yylloc, parserdata, "Function declared ccall more than once");
		}
		$$->ccall = true;
	} |
	EXPORT function_attributes
	{
		$$ = $2;
		if ($$->exported){
			yyerror (&yylloc, parserdata, "Function declared export more than once");
		}
		if ($$->imported){
			yyerror (&yylloc, parserdata, "Function declared import and export");
		}
		$$->exported = true;
	} |
	IMPORT function_attributes
	{
		$$ = $2;
		if ($$->imported){
			yyerror (&yylloc, parserdata, "Function declared import more than once");
		}
		if ($$->exported){
			yyerror (&yylloc, parserdata, "Function declared import and export");
		}
		$$->imported = true;
	} |
	LOCK function_attributes
	{
		$$ = $2;
		if ($$->lock){
			yyerror (&yylloc, parserdata, "Function declared lock more than once");
		}
		if ($$->ccall){
			yyerror (&yylloc, parserdata, "Function declared lock and ccall");
		}
		$$->lock = true;
	} |
	{
		$$ = new FunctionAttributes;
		$$->ccall = false;
		$$->exported = false;
		$$->imported = false;
		$$->lock = false;
	};

vardec:
	type identifier {$$ = new AstNodeStatementVariableDeclaration($1, $2); $$->SetSourceLocation("", @$.first_line);};
vardec_list:
	vardec ',' vardec_list {$3->push_front($1); $$=$3;} |
	vardec {$$ = new std::list<AstNodeStatementVariableDeclaration *>; $$->push_front($1);} |
	{$$ = new std::list<AstNodeStatementVariableDeclaration *>;};
vardeca:
	vardec {$$ = $1;} |
	type identifier '=' expression {
		AstNodeStatementVariableDeclaration *declaration = new AstNodeStatementVariableDeclaration($1, $2);
		AstNodeStatementVariableAssignment *assignment = new AstNodeStatementVariableAssignment(new AstNodeExpressionLValueLock(new AstNodeExpressionLValueIdentifier($2)), $4);
		
		declaration->SetSourceLocation("", @$.first_line);
		assignment->SetSourceLocation("", @$.first_line);
		
		$$ = new AstNodeStatementVariableDeclarationAssignment({declaration, assignment});
		$$->SetSourceLocation("", @$.first_line);
	};



/* Imperative stuff */
block:
	'{' statement_list '}' {
		$$ = new AstNodeStatementBlock(*$2); $$->SetSourceLocation("", @$.first_line);
		delete $2;
	};
statement_list:
	statement statement_list {$2->push_front($1); $$ = $2;} |
	{$$ = new std::list<AstNodeStatement *>;};
statement:
	vardeca ';' {$$ = $1;} |
	lvalue '=' expression ';' {$$ = new AstNodeStatementVariableAssignment($1, $3); $$->SetSourceLocation("", @$.first_line);} |
	RETURN ';' {$$ = new AstNodeStatementReturn(0); $$->SetSourceLocation("", @$.first_line);} |
	RETURN expression ';' {$$ = new AstNodeStatementReturn($2); $$->SetSourceLocation("", @$.first_line);} |
	DEALLOC expression ';' {$$ = new AstNodeStatementDealloc($2); $$->SetSourceLocation("", @$.first_line);} |
	function_call ';' {$$ = new AstNodeStatementExpression($1); $$->SetSourceLocation("", @$.first_line);} |
	block {$$ = $1;} |
	
	SPAWN statement {$$ = new AstNodeStatementParallelSpawn($2); $$->SetSourceLocation("", @$.first_line);} |
	SYNC statement {$$ = new AstNodeStatementParallelSync($2); $$->SetSourceLocation("", @$.first_line);} |
	ATOMIC statement {$$ = new AstNodeStatementParallelAtomic($2); $$->SetSourceLocation("", @$.first_line);} |
	
	SPAWN '(' expression ')' statement {
		AstNodeStatementParallelSpawn *spawn_statement = new AstNodeStatementParallelSpawn($5);
		spawn_statement->SetSourceLocation("", @$.first_line);
		
		AstNodeStatementIf *if_statement = new AstNodeStatementIf($3, spawn_statement, $5->Copy());
		if_statement->SetSourceLocation("", @$.first_line);
		
		$$ = if_statement;
	} |
	SYNC '(' expression ')' statement {
		AstNodeStatementParallelSync *sync_statement = new AstNodeStatementParallelSync($5);
		sync_statement->SetSourceLocation("", @$.first_line);
		
		AstNodeStatementIf *if_statement = new AstNodeStatementIf($3, sync_statement, $5->Copy());
		if_statement->SetSourceLocation("", @$.first_line);
		
		$$ = if_statement;
	} |
	ATOMIC '(' expression ')' statement {
		AstNodeStatementParallelAtomic *atomic_statement = new AstNodeStatementParallelAtomic($5);
		atomic_statement->SetSourceLocation("", @$.first_line);
		
		AstNodeStatementIf *if_statement = new AstNodeStatementIf($3, atomic_statement, $5->Copy());
		if_statement->SetSourceLocation("", @$.first_line);
		
		$$ = if_statement;
	} |
	
	
	
	FOR '(' type identifier TO expression ')' statement {$$ = new AstNodeStatementFor($3, $4, $6, $8); $$->SetSourceLocation("", @$.first_line);} |
	WHILE '(' expression ')' statement {$$ = new AstNodeStatementWhile($3, $5); $$->SetSourceLocation("", @$.first_line);} |
	IF '(' expression ')' statement {$$ = new AstNodeStatementIf($3, $5, 0); $$->SetSourceLocation("", @$.first_line);} |
	IF '(' expression ')' statement ELSE statement {$$ = new AstNodeStatementIf($3, $5, $7); $$->SetSourceLocation("", @$.first_line);};
function_call:
	lvalue '(' expression_list ')' {
		$$ = new AstNodeFunctionCall($1, *$3);
		delete $3;
		$$->SetSourceLocation("", @$.first_line);
	};


/* Types */
type:
	VOID {$$ = new AstNodeTypeVoid(); $$->SetSourceLocation("", @$.first_line);} |
	BOOL {$$ = new AstNodeTypeBool(); $$->SetSourceLocation("", @$.first_line);} |
	INT integer {$$ = new AstNodeTypeInt($2, false, false); $$->SetSourceLocation("", @$.first_line);} |
	STRICT INT integer {$$ = new AstNodeTypeInt($3, true, false); $$->SetSourceLocation("", @$.first_line);} |
	SIGNED INT integer {$$ = new AstNodeTypeInt($3, false, true); $$->SetSourceLocation("", @$.first_line);} |
	STRICT SIGNED INT integer {$$ = new AstNodeTypeInt($4, true, true); $$->SetSourceLocation("", @$.first_line);} |
	FLOAT integer {$$ = new AstNodeTypeFloat($2, false); $$->SetSourceLocation("", @$.first_line);} |
	STRICT FLOAT integer {$$ = new AstNodeTypeFloat($3, true); $$->SetSourceLocation("", @$.first_line);} |
	
	type '(' type_list ')' {
		$$ = new AstNodeTypeFunctionPointer($1, *$3, false);
		delete $3;
		$$->SetSourceLocation("", @$.first_line);
	} |
	CCALL type '(' type_list ')' {
		$$ = new AstNodeTypeFunctionPointer($2, *$4, true);
		delete $4;
		$$->SetSourceLocation("", @$.first_line);
	} |
	STRUCT identifier {$$ = new AstNodeTypeStruct($2); $$->SetSourceLocation("", @$.first_line);} |
	
	type LOCK {$$ = new AstNodeTypeLock($1); $$->SetSourceLocation("", @$.first_line);} |
	
	type POINTER {$$ = new AstNodeTypePointer($1); $$->SetSourceLocation("", @$.first_line);} | /* pointer to type */
	type REF {$$ = new AstNodeTypeReference($1); $$->SetSourceLocation("", @$.first_line);} | /* reference to type */
	type '[' ']' {$$ = new AstNodeTypeDynamicVector($1); $$->SetSourceLocation("", @$.first_line);} | /* Dynamic sized vector type */
	type '[' integer ']' {$$ = new AstNodeTypeStaticVector($1, $3); $$->SetSourceLocation("", @$.first_line);}; /* Static sized vector type */
	

type_list:
	type ',' type_list {$3->push_front($1); $$=$3;} |
	type {$$ = new std::list<AstNodeType *>; $$->push_front($1);} |
	{$$ = new std::list<AstNodeType *>;};

/* To initialize the dynamic vector type, we need constant values that are vectors (e.g: "int 32: (0, 10)" or whatever) */


/* Expressions */
expression:
	VOID {$$ = new AstNodeExpressionVoid(); $$->SetSourceLocation("", @$.first_line);} |
	boolean {$$ = new AstNodeExpressionBool($1); $$->SetSourceLocation("", @$.first_line);} |
	integer {$$ = new AstNodeExpressionInt($1); $$->SetSourceLocation("", @$.first_line);} |
	float {$$ = new AstNodeExpressionFloat($1); $$->SetSourceLocation("", @$.first_line);} |
	lvalue {$$ = $1;} |
	type '(' expression ')' {$$ = new AstNodeExpressionCast($1, $3); $$->SetSourceLocation("", @$.first_line);} |
	function_call {$$ = $1;} |
	type '[' expression ']' {$$ = new AstNodeExpressionDynamicVector($1, new AstNodeExpressionCast(new AstNodeTypeInt(0, true, false), $3)); $$->SetSourceLocation("", @$.first_line);} |
	LENGTH '(' expression ')' {$$ = new AstNodeExpressionLength($3); $$->SetSourceLocation("", @$.first_line);} |
	REF lvalue {$$ = new AstNodeExpressionReference($2); $$->SetSourceLocation("", @$.first_line);} |
	ALLOC type {$$ = new AstNodeExpressionAlloc($2); $$->SetSourceLocation("", @$.first_line);} |
	
	'(' expression ')' {$$ = $2;} |
	expression '+' expression {$$ = new AstNodeExpressionBinaryAddition($1, $3); $$->SetSourceLocation("", @$.first_line);} |
	expression '-' expression {$$ = new AstNodeExpressionBinarySubtraction($1, $3); $$->SetSourceLocation("", @$.first_line);} |
	expression '*' expression {$$ = new AstNodeExpressionBinaryMultiplication($1, $3); $$->SetSourceLocation("", @$.first_line);} |
	expression '/' expression {$$ = new AstNodeExpressionBinaryDivision($1, $3); $$->SetSourceLocation("", @$.first_line);} |
	expression '%' expression {$$ = new AstNodeExpressionBinaryModulo($1, $3); $$->SetSourceLocation("", @$.first_line);} |
	
	expression '&' expression {$$ = new AstNodeExpressionBinaryAnd($1, $3); $$->SetSourceLocation("", @$.first_line);} |
	expression '|' expression {$$ = new AstNodeExpressionBinaryOr($1, $3); $$->SetSourceLocation("", @$.first_line);} |
	expression '^' expression {$$ = new AstNodeExpressionBinaryXor($1, $3); $$->SetSourceLocation("", @$.first_line);} |
	expression SHL expression {$$ = new AstNodeExpressionBinaryShl($1, $3); $$->SetSourceLocation("", @$.first_line);} |
	expression SHR expression {$$ = new AstNodeExpressionBinaryShr($1, $3); $$->SetSourceLocation("", @$.first_line);} |
	
	expression '<' expression {$$ = new AstNodeExpressionBinaryLessThan($1, $3); $$->SetSourceLocation("", @$.first_line);} |
	expression '>' expression {$$ = new AstNodeExpressionBinaryGreaterThan($1, $3); $$->SetSourceLocation("", @$.first_line);} |
	expression LTE expression {$$ = new AstNodeExpressionBinaryLessThanEqual($1, $3); $$->SetSourceLocation("", @$.first_line);} |
	expression GTE expression {$$ = new AstNodeExpressionBinaryGreaterThanEqual($1, $3); $$->SetSourceLocation("", @$.first_line);} |
	expression EQU expression {$$ = new AstNodeExpressionBinaryEqual($1, $3); $$->SetSourceLocation("", @$.first_line);} |
	expression NEQ expression {$$ = new AstNodeExpressionBinaryNotEqual($1, $3); $$->SetSourceLocation("", @$.first_line);};
	
	
expression_list:
	expression ',' expression_list {$3->push_front($1); $$=$3;} |
	expression {$$ = new std::list<AstNodeExpression *>; $$->push_front($1);} |
	{$$ = new std::list<AstNodeExpression *>;};

/* The AstNodeExpressionLValueLock is normalized away if the lvalue type is not a lock */
lvalue:
	lvalue_pre_lock {$$ = new AstNodeExpressionLValueLock($1); $$->SetSourceLocation("", @$.first_line);};
	
lvalue_pre_lock:
	identifier {$$ = new AstNodeExpressionLValueIdentifier($1); $$->SetSourceLocation("", @$.first_line);} |
	lvalue '[' expression ']' {$$ = new AstNodeExpressionLValueVector($1, new AstNodeExpressionCast(new AstNodeTypeInt(0, true, false), $3)); $$->SetSourceLocation("", @$.first_line);} |
	lvalue '.' identifier {$$ = new AstNodeExpressionLValueStructElement($1, $3); $$->SetSourceLocation("", @$.first_line);} |
	lvalue SPLAT identifier {
		AstNodeExpressionLValueDereference *dereferenced_struct = new AstNodeExpressionLValueDereference($1); dereferenced_struct->SetSourceLocation("", @$.first_line);
		AstNodeExpressionLValueStructElement *struct_access = new AstNodeExpressionLValueStructElement(dereferenced_struct, $3); struct_access->SetSourceLocation("", @$.first_line);
		$$ = struct_access;
	} |
	DEREF expression {$$ = new AstNodeExpressionLValueDereference($2); $$->SetSourceLocation("", @$.first_line);} |
	'(' lvalue ')' {$$ = $2;};
	
/* Basic values */
identifier:
	IDENTIFIER {snprintf($$, 1024, "%s", $1);}; 
boolean:
	TRUE {$$ = true;} |
	FALSE {$$ = false;};
integer:
	INTEGER {$$ = $1;};
float:
	DECIMAL {$$ = $1;};



/*
	Standard word for "parallel" is "spawn", atomic "sync"
	
	Perhaps have atomic blocks that have the semantic that everything that
	happens in them appears to be atomic to other threads. Actually, statement
	not block. Likewise with the rest of the blocks. This would be useful for,
	for example:
	float 32 total = 0.0;
	foreach (i in 0 to sum.dim(0)) parallel {
		atomic total += sum[i]; //the compiler can use its knownedge of
			//associtivity of + to optimise this:D
	}
	This atomic semantic might be particularly necessary if you had function
	calls and so on.
	This could be rewritten
	float 32 total = 0.0;
	foreach (e in sum) parallel atomic total += e;
	
	Future block.
	Parallel block
	
	Vector type - fixed and dynamic, multimentional too
	Perhaps value.dim(2) to get the third dimension's element count
	foreach(some_vector)
	foreach(i in 0 to some_vector.dim(0))
*/

/*
	Unary vector operators:
	+vector    Adds all the elements of a vector together (returns scalar)
	*vector    Multiplies all the elements of a vector together (returns scalar)
	-vector    Negates all the elements of a vector (returns vector)
	/vector    Reciprocates all the elements of a vector (returns vector)
	~vector    Logical not of all the elements of a vector (returns vector)
	
	The choice of - and / may seem odd that they don't work the way + and * do,
	but subtraction and division are not associative as + and * are
	
	E.g:
	return 1.0 + +vector;
	
	scalar op vector (or vector op scalar)
	1.0 + vector    adds 1.0 to all the elements of the vector
	vector / 2.0    divides all the elements of the vector by two
	2.0 / vector    returns 2.0 * /vector
*/

/*
	OOP would be nice to have
*/
