/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "computespec/computespec.hpp" //This has to go before argp.h, otherwise we get many errors

#include <cstdio>
#include <list>
#include <argp.h>
#include "error.hpp"

#include "compile.hpp"

#include "ast/astnode.hpp"


enum {
	OPT_LLVM = 256,
	OPT_AST_DOT,
	OPT_AST_TEXT,
	OPT_PARSER_DOT,
	OPT_LIB,
	OPT_OPT_REPORT,
	OPT_GENERATE_COMPUTESPEC,
	OPT_DEBUG_HEADER_LL,
	OPT_DEBUG_LLVM_LINES,
	OPT_PSEUDOCODE,
	OPT_COMPUTESPEC,
	
	#ifdef OPENCL
		OPT_CLINFO,
	#endif
};

enum PROGRAM_MODE {
	PROGRAM_MODE_NONE,
	PROGRAM_MODE_COMPILE,
	PROGRAM_MODE_COMPUTESPEC,
	
	#ifdef OPENCL
		PROGRAM_MODE_CLINFO,
	#endif
};

class Arguments
{
	public:
		Arguments(int argc, char **argv) :
			m_mode(PROGRAM_MODE_NONE),
			m_polycute_library_directory("/usr/share/polycute/lib/"),
			m_generate_computespec(false),
			m_debug(false),
			m_debug_llvm_header(false),
			m_debug_llvm_lines(false)
		{
			argp_parse (&s_argp_i, argc, argv, 0, 0, this);
		}
		
		PROGRAM_MODE GetMode()
		{
			return m_mode;
		}
		
		std::string GetLlvmFilename()
		{
			return m_llvm_name;
		}
		
		std::string GetParserDotFilename()
		{
			return m_parser_dot_name;
		}
		
		std::string GetAstDotFilename()
		{
			return m_ast_dot_name;
		}
		
		std::string GetAstTextFilename()
		{
			return m_ast_text_name;
		}
		
		std::string GetOptimisationReportFilename()
		{
			return m_optimisation_report;
		}
		
		std::string GetInputFilename()
		{
			return m_input;
		}
		
		std::string GetPolycuteLibraryDirectory()
		{
			return m_polycute_library_directory;
		}
		
		std::list<std::string> GetOptimisers()
		{
			std::list<std::string> optimisations = m_optimisations;
			if (!optimisations.size()){
				optimisations.push_back("3");
			}
			return optimisations;
		}
		
		std::string GetPseudocode()
		{
			return m_pseudocode;
		}
		
		std::string GetComputespec()
		{
			return m_computespec;
		}
		
		bool GetGenerateComputespec()
		{
			return m_generate_computespec;
		}
		
		bool GetDebug()
		{
			return m_debug;
		}
		
		bool GetDebugLlvmHeader()
		{
			return m_debug_llvm_header;
		}
		
		bool GetDebugLlvmLines()
		{
			return m_debug_llvm_lines;
		}
		
	private:
		static error_t parse_opt (int key, char *arg, struct argp_state *state)
		{
			Arguments *arguments = (Arguments *)state->input;
			return arguments->ParseOpt(key, arg, state);
		}
		
		error_t ParseOpt (int key, char *arg, struct argp_state *state)
		{
			switch (key)
			{
				case OPT_GENERATE_COMPUTESPEC:
					ExpectedCompilerMode(PROGRAM_MODE_COMPUTESPEC, key);
					m_generate_computespec = true;
					break;
				
				#ifdef OPENCL
					case OPT_CLINFO:
						ExpectedCompilerMode(PROGRAM_MODE_CLINFO, key);
						break;
				#endif
				
				
				case OPT_LLVM:
					ExpectedCompilerMode(PROGRAM_MODE_COMPILE, key);
					m_llvm_name = arg;
					break;
				case 'O':
					ExpectedCompilerMode(PROGRAM_MODE_COMPILE, key);
					m_optimisations.push_back(arg);
					break;
				case 'g':
					ExpectedCompilerMode(PROGRAM_MODE_COMPILE, key);
					m_debug = true;
					break;
				case OPT_AST_DOT:
					ExpectedCompilerMode(PROGRAM_MODE_COMPILE, key);
					m_ast_dot_name = arg;
					break;
				case OPT_AST_TEXT:
					ExpectedCompilerMode(PROGRAM_MODE_COMPILE, key);
					m_ast_text_name = arg;
					break;
				case OPT_PARSER_DOT:
					ExpectedCompilerMode(PROGRAM_MODE_COMPILE, key);
					m_parser_dot_name = arg;
					break;
				case OPT_OPT_REPORT:
					ExpectedCompilerMode(PROGRAM_MODE_COMPILE, key);
					m_optimisation_report = arg;
					break;
				case OPT_LIB:
					ExpectedCompilerMode(PROGRAM_MODE_COMPILE, key);
					m_polycute_library_directory = arg;
					break;
				case OPT_COMPUTESPEC:
					ExpectedCompilerMode(PROGRAM_MODE_COMPILE, key);
					m_computespec = arg;
					break;
				case OPT_DEBUG_HEADER_LL:
					ExpectedCompilerMode(PROGRAM_MODE_COMPILE, key);
					m_debug_llvm_header = true;
					break;
				case OPT_DEBUG_LLVM_LINES:
					ExpectedCompilerMode(PROGRAM_MODE_COMPILE, key);
					m_debug_llvm_lines = true;
					break;
				case OPT_PSEUDOCODE:
					ExpectedCompilerMode(PROGRAM_MODE_COMPILE, key);
					m_pseudocode = arg;
					break;
				case ARGP_KEY_ARG:
					ExpectedCompilerMode(PROGRAM_MODE_COMPILE, key);
					if (state->arg_num != 0){
						argp_usage (state);
					}
					m_input = arg;
					break;
				case ARGP_KEY_END:
					if (!m_input.size() && ((PROGRAM_MODE_NONE == m_mode) || (PROGRAM_MODE_COMPILE == m_mode))){
						argp_usage (state);
					}
					break;
				default:
					return ARGP_ERR_UNKNOWN;
			}
			return 0;
		}
		
		void ExpectedCompilerMode(PROGRAM_MODE mode, int key)
		{
			std::string option = "<input>";
			for (unsigned int i = 0; s_options[i].name; i++){
				argp_option opt = s_options[i];
				if (key == opt.key){
					option = (opt.key < 255) ? PrintString("-%c", (char)key) : PrintString("--%s", opt.name);
					break;
				}
			}
			
			if (PROGRAM_MODE_NONE == m_mode){
				m_mode = mode;
				return;
			}
			else if (mode == m_mode){
				return;
			}
			throw FormatError("Cannot specify %s in %s mode: expected %s mode", option.c_str(),
				ModeToString(m_mode).c_str(), ModeToString(mode).c_str()
			);
		}
		
		static std::string ModeToString(PROGRAM_MODE mode)
		{
			return (mode == PROGRAM_MODE_COMPILE) ? "compile" :
				(mode == PROGRAM_MODE_COMPUTESPEC) ? "generate compute specification" :
				"unknown";
		}
		
		static argp_option s_options[];
		static argp s_argp_i;
		
		PROGRAM_MODE m_mode;
		
		std::string m_input;
		std::string m_parser_dot_name;
		std::string m_ast_dot_name;
		std::string m_ast_text_name;
		std::string m_llvm_name;
		std::string m_optimisation_report;
		std::string m_polycute_library_directory;
		std::string m_pseudocode;
		std::string m_computespec;
		std::list<std::string> m_optimisations;
		bool m_generate_computespec;
		bool m_debug;
		bool m_debug_llvm_header;
		bool m_debug_llvm_lines;
};

argp_option Arguments::s_options[] = {
	{"generate-computespec", OPT_GENERATE_COMPUTESPEC, 0, 0, "Generate compute specification", 0},
	#ifdef OPENCL
		{"clinfo", OPT_CLINFO, 0, 0, "Print OpenCL Information", 0},
	#endif
	
	
	{"ast-dot", OPT_AST_DOT, "FILE", 0, "Output the abstract syntax as a dot file", 0},
	{"ast-text", OPT_AST_TEXT, "FILE", 0, "Output the abstract syntax as a textual tree", 0},
	{"computespec", OPT_COMPUTESPEC, 0, 0, "Output the computespec to a file", 0},
	{"debug", 'g', 0, 0, "Enable debugging", 0},
	{"debug-llvm-header", OPT_DEBUG_HEADER_LL, 0, 0, "Add debugging support to the LLVM header", 0},
	{"debug-llvm-lines", OPT_DEBUG_LLVM_LINES, 0, 0, "Make the lines report LLVM source lines", 0},
	{"library-directory", OPT_LIB, "DIRECTORY", 0, "Directory containing the Polycute Library", 0},
	{"llvm", OPT_LLVM, "FILE", 0, "Output the LLVM code to a file", 0},
	{"optimisation", 'O', "OPTIMISATION", 0, "Enable optimisation", 0},
	{"optimisation-report", OPT_OPT_REPORT, "FILE", 0, "Output the optimisation report to a file", 0},
	{"parser-dot", OPT_PARSER_DOT, "FILE", 0, "Output the (first) parse tree as a dot file", 0},
	{"pseudocode", OPT_PSEUDOCODE, "FILE", 0, "Output a pseudocode representation of the optimised code", 0},
	
	
	{0, 0, 0, 0, 0, 0}
};
argp Arguments::s_argp_i = {s_options, parse_opt, 0, "Polycute Compiler", 0, 0, 0};

bool MainCompile(Arguments &arguments)
{
	return Compile(
		/* Inputs and stuff */
		arguments.GetInputFilename(),
		
		/* Program outputs */
		arguments.GetLlvmFilename(),
		arguments.GetPseudocode(),
		arguments.GetComputespec(),
		
		/* Options */
		arguments.GetPolycuteLibraryDirectory(),
		
		/* Optimisation stuff */
		arguments.GetOptimisers(),
		arguments.GetOptimisationReportFilename(),
		arguments.GetAstDotFilename(),
		arguments.GetAstTextFilename(),
		
		/* Debug stuff */
		arguments.GetDebug(),
		arguments.GetDebugLlvmHeader(),
		arguments.GetDebugLlvmLines()
	);
}

void MainComputeSpec()
{
	ComputeSpec spec;
	spec.Autodetect();
	printf("%s", spec.GetSpecSource().c_str());
}

#ifdef OPENCL
	void MainClInfo()
	{
		std::set<OpenCL::Platform> platforms = OpenCL::Platform::GetPlatforms();
		
		for (OpenCL::Platform platform: platforms){
			printf("%s\n", platform.GetHumanReadableString().c_str());
		}
	}
#endif

int main(int argc, char **argv)
{
	try
	{
		Arguments arguments(argc, argv);
		switch (arguments.GetMode())
		{
			case PROGRAM_MODE_COMPILE:
				MainCompile(arguments);
				break;
			case PROGRAM_MODE_COMPUTESPEC:
				MainComputeSpec();
				break;
			
			#ifdef OPENCL
				case PROGRAM_MODE_CLINFO:
					MainClInfo();
					break;
			#endif
			
			default:
				throw FormatError("Unknown program mode");
		}
	}
	catch (Error &e)
	{
		e.OutputMessage();
		return 1;
	}
	catch (CodeGenerationException &e)
	{
		fprintf(stderr, "Uncaught code generation exception: %s\n", e.GetMessage().c_str());
		return 1;
	}
}
