/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef DATASET_H
#define DATASET_H

#include "error.hpp"
#include "image.hpp"

#include <stdint.h>
#include <cstdio>

class Dataset
{
	public:
		Dataset(const char *DataFilename, const char *LabelFilename = 0)
		{
			count = 0;
			rows = 0;
			cols = 0;
			
			data = 0;
			labels = 0;
			
			FILE *df = fopen(DataFilename, "rb");
			if (df == 0){
				throw FormatError("Could not open file %s!", DataFilename);
			}
			
			uint32_t imgmagic = Read32bitBigEndianInt(df);
			count = Read32bitBigEndianInt(df);
			rows = Read32bitBigEndianInt(df);
			cols = Read32bitBigEndianInt(df);
			
			if (imgmagic != 2051){
				throw FormatError("File %s has a magic of %u not %u!", DataFilename, imgmagic, 2051);
			}
			
			int imgsize = count*rows*cols;
			data = new uint8_t[imgsize];
			int dr = fread(data, 1, imgsize, df);
			if (dr != imgsize){
				throw FormatError("Expected to read %i bytes (%ux%ux%u), but only read %i bytes!", imgsize, count, rows, cols, dr);
			}
			
			fclose(df);
			
			if (LabelFilename != 0){
				FILE *lf = fopen(LabelFilename, "rb");
				if (lf == 0){
					throw FormatError("Could not open file %s!", LabelFilename);
				}
				
				uint32_t labelmagic = Read32bitBigEndianInt(lf);
				uint32_t labelcount = Read32bitBigEndianInt(lf);
				
				if (labelmagic != 2049){
					throw FormatError("File %s has a magic of %u not %u!", LabelFilename, labelmagic, 2049);
				}
				
				if (labelcount != count){
					throw FormatError("There are %u images but %u labels!", count, labelcount);
				}
				
				labels = new uint8_t[count];
				int lr = fread(labels, 1, count, lf);
				if (lr != (int)count){
					throw FormatError("Expected to read %i bytes, but only read %i bytes!", count, lr);
				}
				
				fclose(lf);
			}
		}
		~Dataset()
		{
			if (data == 0) {
				delete []data;
			}
			
			if (labels == 0){
				delete []labels;
			}
		}
		
		unsigned int GetRows()
		{
			return rows;
		}
		
		unsigned int GetCols()
		{
			return cols;
		}

		unsigned int GetCount()
		{
			return count;
		}
		
		bool HasLabels()
		{
			return (labels != 0);
		}
		
		uint8_t GetLabel(unsigned int n)
		{
			if (n >= count){
				throw FormatError("Requested label %u from %u labels!", n, count);
			}
			else if (labels == 0){
				throw FormatError("There are no labels!");
			}
			else {
				return labels[n];
			}
		}
		
		Image GetImage(unsigned int n)
		{
			if (n >= count){
				throw FormatError("Requested image %u from %u images!", n, count);
			}
			else if (data == 0){
				throw FormatError("There is no data!");
			}
			
			Image img(cols, rows);
			for (unsigned int x = 0; x < cols; x++){
				for (unsigned int y = 0; y < rows; y++){
					img.SetPixel(/*255-*/data[(n*cols*rows) + (y*cols)+x], x, y);
				}
			}
			return img;
		}
		
	private:
		uint32_t Read32bitBigEndianInt(FILE *file)
		{
			uint32_t v = 0;
			fread(&(((uint8_t *)(&v))[3]), 1, 1, file);
			fread(&(((uint8_t *)(&v))[2]), 1, 1, file);
			fread(&(((uint8_t *)(&v))[1]), 1, 1, file);
			fread(&(((uint8_t *)(&v))[0]), 1, 1, file);
			
			return v;
		}

		
		uint32_t count;
		uint32_t rows;
		uint32_t cols;
		
		uint8_t *data;
		uint8_t *labels;
};

#endif
