/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <math.h>
#include <stdint.h>
#include <stdlib.h>

uint32_t GetRandomUInt32();
float *GetPointerToPixel(float *p, uint32_t i, uint32_t x, uint32_t y, uint32_t width, uint32_t height);
void DisplayImagePair(float *original, float *generated, uint32_t width, uint32_t height, uint32_t interval);
void PrintStatusMessage(uint32_t index);
uint32_t GetExit();
float Exp(float f);


typedef struct RBM
{
	float *weights;
	float *random;
	uint32_t visible_size;
	uint32_t hidden_size;
	uint32_t random_size;
	uint32_t random_offset;
	uint32_t visible_type;
	uint32_t hidden_type;
} RBM;

typedef struct BSresult {
	RBM rbm;
	float result;
} BSresult;
static inline BSresult BinaryStochastic(RBM rbm, float p)
{
	float f = rbm.random[rbm.random_offset];
	rbm.random_offset = (rbm.random_offset + 1) % rbm.random_size;
	
	BSresult result;
	result.rbm = rbm;
	result.result = f < p ? 1.0 : 0.0;
	
	return result;
}

static inline float Function(float p)
{
	return 1.0/(1.0 + Exp(-p));
}

static inline uint32_t Weight(RBM rbm, uint32_t visible, uint32_t hidden)
{
	return (visible * rbm.hidden_size + hidden);
}

static inline uint32_t GetVisibleBias(RBM rbm, uint32_t visible)
{
	return (rbm.visible_size * rbm.hidden_size + visible);
}

static inline uint32_t GetHiddenBias(RBM rbm, uint32_t hidden)
{
	return (rbm.visible_size * (rbm.hidden_size + 1) + hidden);
}

static RBM CalculateVisible(RBM rbm, float *visible, const float *hidden) //const
{
	float hidden_bin[rbm.hidden_size];
	
	if (1 == rbm.hidden_type){
		for (uint32_t i = 0; i < rbm.hidden_size; i++){
			BSresult result = BinaryStochastic(rbm, hidden[i]);
			rbm = result.rbm;
			hidden_bin[i] = result.result;
		}
		hidden = hidden_bin;
	}
	else if (2 == rbm.hidden_type){
		for (uint32_t i = 0; i < rbm.hidden_size; i++){
			hidden_bin[i] = (hidden[i] < 0.5 ? 0.0 : 1.0);
		}
		hidden = hidden_bin;
	}
	
	for (uint32_t i = 0; i < rbm.visible_size; i++){
		visible[i] = 0.0;
	}
	
	for (uint32_t v = 0; v < rbm.visible_size; v++){
		for (uint32_t h = 0; h < rbm.hidden_size; h++){
			visible[v] += rbm.weights[Weight(rbm, v, h)] * hidden[h];
		}
	}
	
	for (uint32_t i = 0; i < rbm.visible_size; i++){
		visible[i] = Function(visible[i] + rbm.weights[GetVisibleBias(rbm, i)]);
	}
	
	if (1 == rbm.visible_type){
		for (uint32_t i = 0; i < rbm.visible_size; i++){
			BSresult result = BinaryStochastic(rbm, visible[i]);
			rbm = result.rbm;
			visible[i] = result.result;
		}
	}
	else if (2 == rbm.visible_type){
		for (uint32_t i = 0; i < rbm.visible_size; i++){
			visible[i] = (visible[i] < 0.5 ? 0.0 : 1.0);
		}
	}
	
	return rbm;
}

static RBM CalculateHidden(RBM rbm, const float *visible, float *hidden)
{
	float visible_bin[rbm.visible_size];
	
	if (1 == rbm.visible_type){
		for (uint32_t i = 0; i < rbm.visible_size; i++){
			BSresult result = BinaryStochastic(rbm, visible[i]);
			rbm = result.rbm;
			visible_bin[i] = result.result;
		}
		visible = visible_bin;
	}
	else if (2 == rbm.visible_type){
		for (uint32_t i = 0; i < rbm.visible_size; i++){
			visible_bin[i] = (visible[i] < 0.5 ? 0.0 : 1.0);
		}
		visible = visible_bin;
	}
	
	for (uint32_t i = 0; i < rbm.hidden_size; i++){
		hidden[i] = 0.0;
	}
	
	for (uint32_t v = 0; v < rbm.visible_size; v++){
		for (uint32_t h = 0; h < rbm.hidden_size; h++){
			hidden[h] += rbm.weights[Weight(rbm, v, h)] * visible[v];
		}
	}
	
	for (uint32_t i = 0; i < rbm.hidden_size; i++){
		hidden[i] = Function(hidden[i] + rbm.weights[GetHiddenBias(rbm, i)]);
	}
	
	if (1 == rbm.hidden_type){
		for (uint32_t i = 0; i < rbm.hidden_size; i++){
			BSresult result = BinaryStochastic(rbm, hidden[i]);
			rbm = result.rbm;
			hidden[i] = result.result;
		}
	}
	else if (2 == rbm.hidden_type){
		for (uint32_t i = 0; i < rbm.visible_size; i++){
			hidden[i] = (hidden[i] < 0.5 ? 0.0 : 1.0);
		}
	}
	
	return rbm;
}

static RBM Train(RBM rbm, const float *visible, float rate)
{
	float v0[rbm.visible_size];
	float h0[rbm.hidden_size];
	float v1[rbm.visible_size];
	float h1[rbm.hidden_size];
	
	if (1 == rbm.visible_type){
		for (uint32_t i = 0; i < rbm.visible_size; i++){
			BSresult result = BinaryStochastic(rbm, visible[i]);
			rbm = result.rbm;
			v0[i] = result.result;
		}
	}
	else if (2 == rbm.visible_type){
		for (uint32_t i = 0; i < rbm.visible_size; i++){
			v0[i] = (visible[i] < 0.5 ? 0.0 : 1.0);
		}
	}
	else {
		for (uint32_t i = 0; i < rbm.visible_size; i++){
			v0[i] = visible[i];
		}
	}
	
	rbm = CalculateHidden(rbm, v0, h0);
	rbm = CalculateVisible(rbm, v1, h0);
	rbm = CalculateHidden(rbm, v1, h1);
	
	for (uint32_t i = 0; i < rbm.hidden_size; i++){
		rbm.weights[GetHiddenBias(rbm, i)] += rate * (h0[i] - h1[i]);
	}
	
	for (uint32_t i = 0; i < rbm.visible_size; i++){
		rbm.weights[GetVisibleBias(rbm, i)] += rate * (v0[i] - v1[i]);
	}
	
	for (uint32_t v = 0; v < rbm.visible_size; v++){
		for (uint32_t h = 0; h < rbm.hidden_size; h++){
			rbm.weights[Weight(rbm, v, h)] += rate * (v0[v] * h0[h] - v1[v] * h1[h]);
		}
	}
	
	return rbm;
}

static RBM InitialiseRbm(uint32_t width, uint32_t height, uint32_t count, uint32_t hidden_size, uint32_t random_size, float initial, float rate)
{
	RBM rbm;
	rbm.visible_size = width * height;
	rbm.hidden_size = hidden_size;
	rbm.random_size = random_size;
	rbm.random_offset = 0;
	rbm.visible_type = 0;
	rbm.hidden_type = 1;
	rbm.weights = (float *)malloc((rbm.visible_size * rbm.hidden_size + rbm.visible_size + rbm.hidden_size) * sizeof(float));
	rbm.random = (float *)malloc(rbm.random_size * sizeof(float));
	
	for (uint32_t i = 0; i < rbm.random_size; i++){
		uint32_t integer = GetRandomUInt32();
		float f = (float)integer / ((float)0x10000 * (float)0x10000);
		rbm.random[i] = f;
	}
	
	for (uint32_t i = 0; i < rbm.visible_size * rbm.hidden_size + rbm.visible_size + rbm.hidden_size; i++){
		uint32_t integer = GetRandomUInt32();
		float f = (float)integer / ((float)0x10000 * (float)0x10000);
		rbm.weights[i] = (f * 2.0 - 1.0) * initial;
	}
	
	return rbm;
}

static void DeinitialiseRbm(RBM rbm)
{
	free(rbm.weights);
	free(rbm.random);
}

static void DisplayRbmGeneration(float *image_buffer1, float *image_buffer2, RBM rbm, float *original_image, uint32_t width, uint32_t height, uint32_t interval)
{
	float hidden[rbm.hidden_size];
	float generated_image[rbm.visible_size];
	
	CalculateHidden(rbm, original_image, hidden);
	CalculateVisible(rbm, generated_image, hidden);
	
	
	for (uint32_t i = 0; i < width * height; i++){
		image_buffer1[i] = original_image[i];
		image_buffer2[i] = generated_image[i];
	}
	
	DisplayImagePair(image_buffer1, image_buffer2, width, height, interval);
}

#include <stdlib.h>
void DoRbm(float *image_buffer1, float *image_buffer2, float *image_floats, uint32_t width, uint32_t height, uint32_t count,
	uint32_t hidden_size, uint32_t random_size, float initial, float rate, uint32_t update_interval)
{
	PrintStatusMessage(0);
	RBM rbm = InitialiseRbm(width, height, count, hidden_size, random_size, initial, rate);
	PrintStatusMessage(1);
	float *current_img = (float *)malloc(width * height * sizeof(float));
	
	for (uint32_t i = 0; i < count; i++){
		float *image_pointer = GetPointerToPixel(image_floats, i, 0, 0, width, height);
		for (uint32_t j = 0; j < width * height; j++){
			current_img[j] = image_pointer[j];
		}
		
		if (!(i % update_interval)){
			DisplayRbmGeneration(image_buffer1, image_buffer2, rbm, current_img, width, height, update_interval);
			if (GetExit()){
				break;
			}
		}
		rbm = Train(rbm, current_img, rate);
	}
	
	free(current_img);
	PrintStatusMessage(2);
	DeinitialiseRbm(rbm);
}
