/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef IMAGE_H
#define IMAGE_H

#include <cstdio>
#include <stdint.h>
#include <cstring>
#include <cstdlib>
#include <SDL/SDL.h>
#include "error.hpp"

typedef struct BMP_HEADER {
	char magica;
	char magicb;
	uint32_t filesize;
	uint16_t ra;
	uint16_t rb;
	uint32_t dataoffs;
} __attribute__((packed)) BMP_HEADER;

typedef struct IMG_HEADER {
	uint32_t headersize;
	uint32_t width;
	uint32_t height;
	uint16_t colourplanes;
	uint16_t bpp;
	uint32_t compression;
	uint32_t rawsize;
	int32_t hres;
	int32_t vres;
	uint32_t colourcount;
	uint32_t impcolourcount;
} __attribute__((packed)) IMG_HEADER;


#define C_RED 0
#define C_GREEN 1
#define C_BLUE 2

class Image {
	public:
		Image()
		{
			width = 0;
			height = 0;
			data = 0;
			d_data = true;
		}
		
		Image(unsigned char *d, unsigned int w, unsigned int h, bool del)
		{
			width = w;
			height = h;
			data = d;
			d_data = del;
		}
		
		Image(unsigned int w, unsigned int h)
		{
			width = w;
			height = h;
			data = new unsigned char[width*height*3];
			d_data = true;
		}
		Image(const Image& image)
		{
			width = image.width;
			height = image.height;
			data = new unsigned char[width*height*3];
			memcpy (data, image.data, width*height*3);
			d_data = true;
		}
		
		Image &operator=(const Image& image)
		{
			if (this != &image){
				width = image.width;
				height = image.height;
				data = new unsigned char[width*height*3];
				memcpy (data, image.data, width*height*3);
				d_data = true;
			}
			return *this;
		}
		
		~Image()
		{
			if ((data != 0) && d_data){
				delete []data;
			}
		}
		
		void LoadBmp(const char *Filename)
		{
			FILE *file = fopen(Filename, "rb");
			if (file == 0){
				throw FormatError("Could not open the file!");
			}
			
			BMP_HEADER bmp_header;
			if (fread(&bmp_header, sizeof(BMP_HEADER), 1, file) != 1){
				throw FormatError("Could not read bitmap header!");
			}
			
			if ((bmp_header.magica != 'B') && (bmp_header.magica != 'M')){
				throw FormatError("This simple program does not recognise this file format!");
			}
			
			//printf("Size of bmp header: %u", sizeof(BMP_HEADER));
			//printf("Position in file: %u", ftell(file));
			
			IMG_HEADER img_header;
			if (fread(&img_header, sizeof(IMG_HEADER), 1, file) != 1){
				throw FormatError("Could not read image header!");
			}
			
			if (img_header.headersize != sizeof(IMG_HEADER)){
				throw FormatError("Image header size is incorrect (is %u, should be %u)!", img_header.headersize, sizeof(IMG_HEADER));
			}
			if (img_header.compression != 0){
				throw FormatError("This simple program does not decompress this compression format!");
			}
			if ((img_header.bpp != 24) && (img_header.bpp != 32)){
				throw FormatError("This simple program does not read this pixel format!");
			}
			
			width = img_header.width;
			height = img_header.height;
			
			if (fseek(file, bmp_header.dataoffs, SEEK_SET) != 0){
				throw FormatError("Could not seek!");
			}
			
			data = new unsigned char[width*height*3];
			for (unsigned int i = 0; i < height; i++){
				unsigned int rb = 0;
				char b;
				for (unsigned int j = 0; j < width; j++){
					if (fread(&data[((((height-(i+1))*width) + j)*3)+2], 1, 1, file) != 1){
						throw FormatError("Could not read blue pixel channel: i is %u, j is %u!", i, j);
					}
					if (fread(&data[((((height-(i+1))*width) + j)*3)+1], 1, 1, file) != 1){
						throw FormatError("Could not read green pixel channel: i is %u, j is %u!", i, j);
					}
					if (fread(&data[((((height-(i+1))*width) + j)*3)], 1, 1, file) != 1){
						throw FormatError("Could not read red pixel channel: i is %u, j is %u!", i, j);
					}
					rb += 3;
					if (img_header.bpp == 32){
						if (fread(&b, 1, 1, file) != 1){
							throw FormatError("Could not read fourth pixel: i is %u, j is %u!", i, j);
						}
						rb++;
					}
				}
				while (rb%4 != 0){
					rb += fread(&b, 1, 1, file);
				}
			}
			fclose(file);
		}
		
		void SaveBmp(const char *Filename) const
		{
			BMP_HEADER bmp_header;
			bmp_header.magica = 'B';
			bmp_header.magicb = 'M';
			bmp_header.filesize = sizeof(BMP_HEADER) + sizeof(IMG_HEADER) + (width*height*3);
			bmp_header.ra = 0;
			bmp_header.rb = 0;
			bmp_header.dataoffs = sizeof(BMP_HEADER) + sizeof(IMG_HEADER);
			
			IMG_HEADER img_header;
			img_header.headersize = sizeof(IMG_HEADER);
			img_header.width = width;
			img_header.height = height;
			img_header.colourplanes = 1;
			img_header.bpp = 24;
			img_header.compression = 0;
			img_header.rawsize = width*height*3;
			img_header.hres = 72;
			img_header.vres = 72;
			img_header.colourcount = 0;
			img_header.impcolourcount = 0;
			
			FILE *file = fopen(Filename, "wb");
			if (file == 0){
				throw FormatError("Could not open file!");
			}
			
			if (fwrite(&bmp_header, sizeof(BMP_HEADER), 1, file) != 1){
				throw FormatError("Could not write bitmap header");
			}
			if (fwrite(&img_header, sizeof(IMG_HEADER), 1, file) != 1){
				throw FormatError("Could not write bitmap header");
			}
			
			for (unsigned int i = 0; i < height; i++){
				unsigned int rb = 0;
				char b = 0;
				for (unsigned int j = 0; j < width; j++){
					if (fwrite(&data[((((height-(i+1))*width) + j)*3)+2], 1, 1, file) != 1){
						throw FormatError("Could not write blue pixel channel: i is %u, j is %u!", i, j);
					}
					if (fwrite(&data[((((height-(i+1))*width) + j)*3)+1], 1, 1, file) != 1){
						throw FormatError("Could not write green pixel channel: i is %u, j is %u!", i, j);
					}
					if (fwrite(&data[((((height-(i+1))*width) + j)*3)], 1, 1, file) != 1){
						throw FormatError("Could not write red pixel channel: i is %u, j is %u!", i, j);
					}
					rb += 3;
					if (img_header.bpp == 32){
						if (fwrite(&b, 1, 1, file) != 1){
							throw FormatError("Could not write fourth pixel: i is %u, j is %u!", i, j);
						}
						rb++;
					}
				}
				while (rb%4 != 0){
					rb += fwrite(&b, 1, 1, file);
				}
			}
			
			fclose(file);
		}
		
		unsigned int GetWidth() const
		{
			return width;
		}
		
		unsigned int GetHeight() const
		{
			return height;
		}
		
		unsigned char GetPixel(int x, int y) const
		{
			if ((x >= (int)width) || (y >= (int)height)){
				return 0;
			}
			unsigned int acc = 0;
			acc += data[(((y*width)+x)*3) + 0];
			acc += data[(((y*width)+x)*3) + 1];
			acc += data[(((y*width)+x)*3) + 2];
			return acc/3;
		}
		
		unsigned char GetPixel(int x, int y, int channel) const
		{
			if ((x >= (int)width) || (y >= (int)height) || (channel > 2)){
				return 0;
			}
			return data[(((y*width)+x)*3) + channel];
		}
		
		unsigned char GetPixel(int x, int y, int channel, int w, int h) const
		{
			if (((int)width <= w) && ((int)height <= h)){
				unsigned int n = 0;
				do {
					n++;
				} while (((int)n * (int)width < w) && ((int)n * (int)height < h));
				return GetPixel(x/n, y/n, channel);
			}
			else {
				int facx = (width % w == 0) ? (width/w) : ((width/w)+1);
				int facy = (height % h == 0) ? (height/h) : ((height/h)+1);
				int fac = (facx < facy) ? facy : facx;
				
				int pix = 0;
				int pixrc = 0;
				for (int py = y*fac; py < (y+1)*fac; py++){
					for (int px = x*fac; px < (x+1)*fac; px++){
						int ipixrc = 0;
						ipixrc = GetPixel(px, py, channel);
						pix += ipixrc;
						pixrc++;
					}
				}
				int r = pix/pixrc;
				return r;
			}
		}
		
		void SetPixel(unsigned char p, int x, int y)
		{
			if ((x >= (int)width) || (y >= (int)height)){
				return;
			}
			data[(((y*width)+x)*3) + 0] = p;
			data[(((y*width)+x)*3) + 1] = p;
			data[(((y*width)+x)*3) + 2] = p;
		}
		
		void SetPixel(unsigned char p, int x, int y, int channel)
		{
			if ((x >= (int)width) || (y >= (int)height) || (channel > 2)){
				return;
			}
			data[(((y*width)+x)*3) + channel] = p;
		}
		
		void CopyImage(const Image &img, int srcx, int srcy, int dstx, int dsty, int w, int h)
		{
			for (int i = 0; i < h; i++){
				for (int j = 0; j < w; j++){
					SetPixel(img.GetPixel(srcx + j, srcy + i, C_RED), dstx + j, dsty + i, C_RED);
					SetPixel(img.GetPixel(srcx + j, srcy + i, C_GREEN), dstx + j, dsty + i, C_GREEN);
					SetPixel(img.GetPixel(srcx + j, srcy + i, C_BLUE), dstx + j, dsty + i, C_BLUE);
				}
			}
		}
		
		unsigned char *GetInternalBuffer()
		{
			return data;
		}
		
		/**Display stuff**/
		static void WaitForKeyExit()
		{
			if (screen == 0){
				throw FormatError("Screen (and hence library) not initialized!");
			}
			
			SDL_Event event;
			
			while (true){
				SDL_WaitEvent(&event);
				switch (event.type)
				{
					case SDL_KEYDOWN:
						return;
						break;
					case SDL_QUIT:
						return;
				}
			}
		}
		
		static bool GetKeyExit()
		{
			if (screen == 0){
				throw FormatError("Screen (and hence library) not initialized!");
			}
			
			SDL_Event event;
			
			while (SDL_PollEvent(&event) == 1){
				switch (event.type)
				{
					case SDL_KEYDOWN:
						return true;
					case SDL_QUIT:
						return true;
				}
			}
			return false;
		}
		
		static void CreateScreen(int w, int h)
		{
			putenv((char *)"SDL_VIDEO_WINDOW_POS");
			putenv((char *)"SDL_VIDEO_CENTERED=1");

			if (screen != 0){
				throw FormatError("Screen already initialized!");
			}
			if (SDL_Init(SDL_INIT_VIDEO) < 0) {
				throw FormatError("Could not initialize SDL: %s\n", SDL_GetError());
			}
			
			screen = SDL_SetVideoMode(w, h, 32, SDL_SWSURFACE);
			if (screen == 0) {
				throw FormatError("Could not set %ix%i video: %s\n", 
					w, h, SDL_GetError());
			}
			if (screen->format->BytesPerPixel != 4){
				throw FormatError("Could not set 32 bits per pixel (%i)!",
					screen->format->BytesPerPixel);
			}
		}

		void Display()
		{
			if (screen == 0){
				throw FormatError("Screen not initialized!");
			}
			if (SDL_MUSTLOCK(screen)) {
				if (SDL_LockSurface(screen) < 0) {
					throw FormatError("Could not lock screen surface: %s",
						SDL_GetError());
				}
			}
			
			for (int y = 0; y < screen->h; y++){
				for (int x = 0; x < screen->w; x++){
					uint32_t pix = SDL_MapRGB(screen->format,
						GetPixel(x, y, C_RED, screen->w, screen->h),
						GetPixel(x, y, C_GREEN, screen->w, screen->h),
						GetPixel(x, y, C_BLUE, screen->w, screen->h));
					((uint32_t *)screen->pixels)[(y*screen->pitch/4) + x] = pix;
				}
			}
			
			if (SDL_MUSTLOCK(screen)) {
				SDL_UnlockSurface(screen);
			}
			SDL_UpdateRect(screen, 0, 0, screen->w, screen->h);
		}
		
	private:
		unsigned int width;
		unsigned int height;
		unsigned char *data;
		bool d_data;
		
		/**Display stuff**/
		static SDL_Surface *screen;
};

SDL_Surface *Image::screen = 0;

#endif
