/*
	Polycute - The reference Polycute implementation
	Copyright (C) 2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cstdio>
#include <cinttypes>
#include <random>

#include <math.h>
#include <sys/resource.h>
#include <stdint.h>
#include <sys/time.h>
#include <sys/timeb.h>
#include <sys/types.h>
#include <unistd.h>

#include "dataset.hpp"
#include "error.hpp"
#include "image.hpp"


static int64_t GetMilliTime()
{
	struct timeb t;
	ftime(&t);
	return (int64_t)t.time * (int64_t)1000 + (int64_t)t.millitm;
}

static void GetProcPidStatm(uint64_t &size, uint64_t &resident, uint64_t &share,
	uint64_t &text, uint64_t &lib, uint64_t &data, uint64_t &dt)
{
	int pid = getpid();
	char filename[1024];
	snprintf(filename, 1024, "/proc/%i/statm", pid);
	FILE *statm = fopen(filename, "r");
	if (!statm){
		throw FormatError("Could not open %s", filename);
	}
	if (fscanf(statm, "%" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64 "\n",
		&size, &resident, &share, &text, &lib, &data, &dt) != 7){
		throw FormatError("Could not read statm");
	}
	fclose(statm);
}

static size_t GetMemoryUsage()
{
	uint64_t size = 0; /* total program size */
	uint64_t resident = 0; /* resident set size */
	uint64_t share = 0; /* shared pages */
	uint64_t text = 0; /* text (code) */
	uint64_t lib = 0; /* library (unused in Linux 2.6) */
	uint64_t data = 0; /* data + stack */
	uint64_t dt = 0; /* dirty pages (unused in Linux 2.6) */
	GetProcPidStatm(size, resident, share, text, lib, data, dt);
	return size * 4;
}

static void ImageToReals(Image &img, float *realimg)
{
	for (uint32_t y = 0; y < img.GetHeight(); y++){
		for (uint32_t x = 0; x < img.GetWidth(); x++){
			realimg[y * img.GetWidth() + x] = ((float)img.GetPixel(x, y))/255.0;
		}
	}
}

static void RealsToImage(Image &img, float *realimg)
{
	for (uint32_t y = 0; y < img.GetHeight(); y++){
		for (uint32_t x = 0; x < img.GetWidth(); x++){
			img.SetPixel(255.0*(realimg[y * img.GetWidth() + x]), x, y);
		}
	}
}

static void DiffImg(Image a, Image b, Image &d1, Image &d2)
{
	for (uint32_t y = 0; y < 28; y++){
		for (uint32_t x = 0; x < 28; x++){
			unsigned char pa = a.GetPixel(x, y);
			unsigned char pb = b.GetPixel(x, y);
			
			uint32_t d = (pa < pb) ? (pb - pa) : (pa - pb);
			d1.SetPixel(d, x, y);
			d2.SetPixel((d < 16) ? d * 17 : 255, x, y);
		}
	}
}

static double CalculateSNR(Image img, Image diff)
{
	double rmsimg = 0.0;
	double rmsdiff = 0.0;
	for (uint32_t y = 0; y < 28; y++){
		for (uint32_t x = 0; x < 28; x++){
			unsigned char pimg = img.GetPixel(x, y);
			unsigned char pdiff = diff.GetPixel(x, y);
			
			rmsimg += pimg * pimg;
			rmsdiff += pdiff * pdiff;
		}
	}
	
	return rmsimg/rmsdiff;
}

extern "C" void PrintStatusMessage(uint32_t index)
{
	const char *message = 0;
	char b[1024];
	snprintf(b, 1024, "Message ID: %u", (unsigned int)index);
	switch (index)
	{
		case 0: message = "Initialising RBM"; break;
		case 1: message = "Beginning Training"; break;
		case 2: message = "Deinitializing"; break;
		case 3: message = "User termination"; break;
		default: message = b;
	}
	fprintf(stderr, "[*] %s\n", message);
}

extern "C" uint32_t GetRandomUInt32()
{
	static std::uniform_int_distribution<uint32_t> distribution(0, 0xFFFFFFFF);
	static std::mt19937 engine; // Mersenne twister MT19937
	return distribution(engine);
}

extern "C" float *GetPointerToPixel(float *p, uint32_t i, uint32_t x, uint32_t y, uint32_t width, uint32_t height)
{
	return p + (i * width * height + y * width + x);
}

extern "C" void DisplayImagePair(float *original, float *generated, uint32_t width, uint32_t height, uint32_t interval)
{
	static int64_t start = GetMilliTime();
	static unsigned int n = 0;
	int64_t current = GetMilliTime();
	double seconds = (double)(current - start)/1000.0;
	unsigned int count = n * interval;
	fprintf(stderr, "[-] Displaying %u at time %0.03f: %0.02f/s\n", count, seconds, (double)count/seconds);
	n++;
	
	Image original_img(width, height);
	Image generated_img(width, height);
	Image diff1(width, height);
	Image diff2(width, height);
	
	RealsToImage(original_img, original);
	RealsToImage(generated_img, generated);
	DiffImg(original_img, generated_img, diff1, diff2);
	
	Image display(width * 2, height * 2);
	display.CopyImage(original_img, 0, 0, 0, 0, width, height);
	display.CopyImage(generated_img, 0, 0, width, 0, width, height);
	display.CopyImage(diff1, 0, 0, 0, height, width, height);
	display.CopyImage(diff2, 0, 0, width, height, width, height);
	
	double snr = CalculateSNR(original_img, diff1);
	display.Display();
	fprintf(stderr, "[*] SNR: %f\n", snr);
	
	size_t usage = GetMemoryUsage();
	fprintf(stderr, "[*] Using %" PRIuPTR " kB (%0.02f MB) of memory\n", usage, (double)usage/1024.0);
}

extern "C" uint32_t GetExit()
{
	return Image::GetKeyExit() ? 1 : 0;
}

extern "C" float Exp(float f)
{
	return exp(f);
}

extern "C" void DoRbm(float *image_buffer1, float *image_buffer2, float *image_floats, uint32_t width, uint32_t height, uint32_t count,
	uint32_t hidden_size, uint32_t random_size, float initial, float rate, uint32_t update_interval);

int main(int argc, char **argv)
{
	const uint32_t display_factor = 4;
	const uint32_t hidden_size = 500;
	const uint32_t random_size = 1048576;
	const float initial = 0.1;
	const float rate = 0.01;
	const uint32_t update_interval = 1000;
	
	fprintf(stderr, "[*] Reading training data set\n");
	Dataset training_data("dataset/train-images-idx3-ubyte", "dataset/train-labels-idx1-ubyte");
	uint32_t width = training_data.GetCols();
	uint32_t height = training_data.GetRows();
	uint32_t count = training_data.GetCount();
	Image::CreateScreen(width * 2 * display_factor, height * 2 * display_factor);
	fprintf(stderr, "[*] Training data set contains %" PRIu32 " %" PRIu32 "x%" PRIu32 " images\n",
		count, width, height);
	float *buffer1 = new float[width * height];
	float *buffer2 = new float[width * height];
	
	fprintf(stderr, "[*] Allocating %" PRIu32 " bytes (%0.02f MB) of memory\n",
		width * height * count * (uint32_t)sizeof(float),
		(double)(width * height * count * (uint32_t)sizeof(float))/1048576.0);
	float *floats = new float[width * height * count];
	
	fprintf(stderr, "[*] Reading data images into floats\n");
	for (uint32_t i = 0; i < count; i++){
		Image img = training_data.GetImage(i);
		ImageToReals(img, floats + (i * width * height));
	}
	
	fprintf(stderr, "[+] Running program\n");
	DoRbm(buffer1, buffer2, floats, width, height, count, hidden_size, random_size, initial, rate, update_interval);
	delete []floats;
	delete []buffer1;
	delete []buffer2;
	fprintf(stderr, "[+] Program terminated\n");
	
	return 0;
}
