#!/usr/bin/python

#	Polycute - The reference Polycute implementation
#	Copyright (C) 2013  Nicholas Tomlinson
#	
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#	
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#	
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import select
import subprocess
import sys
import time



# ======== Utility Functions ========

original_directory = os.getcwd()
def Error(string):
	sys.stderr.write("[!] " + string + "\n")
	os.chdir(original_directory)
	exit(1)

def System(args):
	print ("[*] Executing: " + str(args))
	
	try:
		p = subprocess.Popen(args)
		return p.wait()
	except:
		Error("Could not execute command")

def Cd(new_directory):
	try:
		current_directory = os.getcwd()
		os.chdir(new_directory)
		return current_directory
	except:
		Error('Could not change to directory: "' + new_directory + '"')

def Rm(filename):
	if os.path.exists(filename):
		os.unlink(filename)

def GetTime():
	return time.time()

def GetSourceFromFilename(filename):
	prefix,extension = filename.rsplit(".", 1)
	if extension == "cpp":
		language = "c++"
		language_target = prefix + ".o"
		opt_target = None
	elif extension == "c":
		language = "c"
		language_target = prefix + ".o"
		opt_target = None
	elif extension == "pc":
		language = "polycute"
		language_target = prefix + ".ll"
		opt_target = prefix + ".opt.ll"
	object_target = prefix + ".o"
	return (language,language_target,opt_target,object_target,prefix)


# ======== Program Flow ========

def GetArguments():
	def Help():
		Error(sys.argv[0] + " <test directory> <test type> [-g] [-O0] [-l] [-v] [-e] [-u] [-s]:\n"
			"\t<test directory>: Directory of the test to run\n"
			"\t<test type>: Type of test to run (polycute|gcc|clang|gcc-openmp|auto|auto-short|clean)\n"
			"\t-g: Compile with debug information\n"
			"\t-O0: Compile with -O0 (-O3 if not specified)\n"
			"\t-l: Line numbers are LLVM line numbers\n"
			"\t-v: Run in valgrind\n"
			"\t-e: Debug Polycute\n"
			"\t-u: Run user command (e.g: to display a generated image)\n"
			"\t-m: Compile with OpenMP")
		
	if len(sys.argv) < 3:
		Help()
	
	arguments = dict()
	arguments["dir"] = sys.argv[1]
	arguments["type"] = sys.argv[2]
	arguments["debug"] = False
	arguments["optimise"] = True
	arguments["llvm lines"] = False
	arguments["valgrind"] = False
	arguments["debug polycute"] = False
	arguments["user command"] = False
	arguments["compile"] = True
	arguments["execute"] = True
	
	for arg in sys.argv[3:]:
		if "-g" == arg:
			arguments["debug"] = True
		elif "-O0" == arg:
			arguments["optimise"] = False
		elif "-l" == arg:
			arguments["llvm lines"] = True
		elif "-v" == arg:
			arguments["valgrind"] = True
		elif "-e" == arg:
			arguments["debug polycute"] = True
		elif "-u" == arg:
			arguments["user command"] = True
		elif "-r" == arg:
			arguments["execute"] = False
		elif "-c" == arg:
			arguments["compile"] = False
		else:
			Help()
	
	return arguments
	

def GetTestInfo():
	with open("test.info") as f:
		text = f.read()
		f.close()
	
	r = dict()
	lines = text.split("\n")
	
	for line in lines:
		line = line.rstrip("\n")
		key,value = line.split(":", 1)
		r[key] = value
	
	return r


def BuildGccSourceFile(arguments, filename, language_target, compiler):
	system_args = [compiler, "-Wall",  "--std=c++11" if (compiler == "g++") else "--std=c99", "-O3" if arguments["optimise"] else "-O0"]
	
	if (arguments["optimise"]):
		system_args += ['-march=native']
	
	if (arguments["type"] == 'gcc-openmp'):
		system_args += ['-fopenmp']
	
	if (arguments["debug"]):
		system_args += ['-g']
	
	system_args += ["-c", '-o', language_target, filename]
	
	if System(system_args) != 0:
		Error("Could not compile with GCC")


def BuildClangSourceFile(arguments, filename, language_target, compiler):
	system_args = [compiler, "-O3" if arguments["optimise"] else "-O0", "-c", "-o", language_target, filename]
	if System(system_args) != 0:
		Error("Could not compile with Clang")


def BuildCSourceFile(arguments, filename, language_target, exchange):
	if (not exchange) or (arguments["type"] == "gcc"):
		BuildGccSourceFile(arguments, filename, language_target, "gcc")
	elif (not exchange) or (arguments["type"] == "gcc-openmp"):
		BuildGccSourceFile(arguments, filename, language_target, "gcc")
	else:
		BuildClangSourceFile(arguments, filename, language_target, "clang")


def BuildCppSourceFile(arguments, filename, language_target, exchange):
	if (not exchange) or (arguments["type"] == "gcc"):
		BuildGccSourceFile(arguments, filename, language_target, "g++")
	else:
		BuildClangSourceFile(arguments, filename, language_target, "clang++")


def BuildPolycuteSourceFile(arguments, filename, language_target, prefix):
	system_args = [
		"../tools/polycute_debug.sh" if arguments["debug polycute"] else "../../polycute",
		"-O3" if arguments["optimise"] else "-O0",
		"--library-directory=../../src/lib/",
		"--optimisation-report=" + prefix + ".optimise.log",
		"--ast-text=" + prefix + ".tree.txt",
		"--pseudocode=" + prefix + ".ppc"
	]
	
	if (arguments["debug"]):
		system_args += ['-g']
	
	if arguments["llvm lines"]:
		system_args += ["--debug-llvm-lines"]
	
	system_args += [
		"--llvm=" + language_target,
		filename
	]
	
	if System(system_args) != 0:
		Error("Could not compile with Polycute")


def BuildLlvmIrToObjectFile(arguments, filename, language_target, prefix):
	system_args = ["llc", "-O3" if arguments["optimise"] else "-O0", "-o", prefix + ".s", filename]
	if System(system_args) != 0:
		Error("Could not compile with LLVM's IR Compiler")
	
	system_args = ["as"]
	
	if (arguments["debug"]):
		system_args += ["-g"]
	
	system_args += ["-o", language_target, prefix + ".s"]
	
	if System(system_args) != 0:
		Error("Could not assemble with as Compiler")
	
	pass


def BuildLlvmIrToOptLlvmIrFile(arguments, filename, language_target):
	system_args = ["opt", "-S", "-std-compile-opts", "-o", language_target, filename]
	if System(system_args) != 0:
		Error("Could not optimise with LLVM's IR Optimiser")


def BuildSourceFile(arguments, filename, exchange):
	language,language_target,opt_target,object_target,prefix = GetSourceFromFilename(filename)
	
	if "c" == language:
		BuildCSourceFile(arguments, filename, language_target, exchange)
	elif "c++" == language:
		BuildCppSourceFile(arguments, filename, language_target, exchange)
	elif "polycute" == language:
		BuildPolycuteSourceFile(arguments, filename, language_target, prefix)
		
		if arguments["optimise"]:
			BuildLlvmIrToOptLlvmIrFile(arguments, language_target, opt_target)
			BuildLlvmIrToObjectFile(arguments, opt_target, object_target, prefix)
		else:
			BuildLlvmIrToObjectFile(arguments, language_target, object_target, prefix)
	
	return object_target


def Link(arguments, test_info, objects):
	system_args = ["g++", "-g" if arguments["debug"] else "-s", "-o", test_info["name"]]
	
	if (arguments["type"] == 'gcc-openmp'):
		system_args += ['-fopenmp']
	
	system_args += objects
	
	if arguments["type"] == "polycute":
		system_args += ["../../libpolycute.a", "-lpthread"]
	
	if "libs" in test_info:
		for lib in test_info["libs"].split(","):
			system_args += ["-l" + lib]
	
	if System(system_args) != 0:
		Error("Could not link")


def BuildTest(arguments, test_info):
	objects = list()
	
	for source in test_info["common"].split(","):
		objects += [BuildSourceFile(arguments, source, False)]
	
	if arguments["type"] == "polycute":
		exchange_type = "polycute"
	elif arguments["type"] == "gcc":
		exchange_type = "c"
	elif arguments["type"] == "clang":
		exchange_type = "c"
	elif arguments["type"] == "gcc-openmp":
		exchange_type = "c"
	
	for source in test_info[exchange_type].split(","):
		objects += [BuildSourceFile(arguments, source, True)]
	
	Link(arguments, test_info, objects)


def RunTest(arguments, test_info):
	name = test_info["name"]
	
	if arguments["valgrind"]:
		system_args = ["valgrind", "--leak-check=full", "./" + name]
	elif arguments["debug"]:
		system_args = ["gdb", "-batch", "-x", name + ".gdb", name]
	else:
		system_args = ["./" + name]
	print ("[*] Executing: " + str(system_args))
	
	t = None
	timings = list()
	p = subprocess.Popen(system_args, stdout=open(name + ".stdout", "wb"), stderr=subprocess.PIPE, bufsize=0)
	
	stderr_open = True
	while stderr_open:
		line = ""
		while True:
			c = p.stderr.read(1).decode("utf-8")
			if c == "":
				stderr_open = False
				break
			if c == "\n":
				break
			else:
				line += c
		
		if (len(line) >= 3) and (line[:3] == "[+]"):
			if t == None:
				t = GetTime()
			else:
				timing = GetTime() - t
				timings += [timing]
				t = None
				line += ": %0.02f s" % timing
		if len(line) > 0:
			print (line)
	
	p.wait()
	
	if arguments["user command"]:
		System(["bash", "-c", test_info["usr"]])
	
	with open(name + ".stdout", "rb") as f:
		stdout = f.read()
	
	return (stdout,timings)



def Clean(test_info):
	Rm(test_info["name"])
	Rm(test_info["name"] + ".stdout")
	
	filenames = test_info["common"].split(",")
	if "c" in test_info:
		filenames += test_info["c"].split(",")
	if "polycute" in test_info:
		filenames += test_info["polycute"].split(",")
	
	for filename in filenames:
		prefix,extension = filename.rsplit(".", 1)
		for candidate in ["o", "s", "ll", "opt.ll", "ppc", "optimise.log", "tree.txt"]:
			Rm(prefix + "." + candidate)


def CompareStdoutByte(a, b, n):
	if len(a) != len(b):
		return False
	
	for i in range(len(a)):
		diff = int(a[i]) - int(b[i])
		
		if (diff > n) or (diff < -n):
			return False
	
	return True

def CompareStdout(arguments, test_info, a, b):
	auto = test_info["auto"].split(",")
	
	if (auto[0] == ""):
		return a == b
	elif (auto[0] == "byte"):
		return CompareStdoutByte(a, b, int(auto[1]))


def PrintInfo(arguments, test_info, a, b, a_name, b_name, short):
	def PrintToStream(stream, string):
		stream.write(string + "\n")
	
	stream = sys.stderr if short else sys.stdout
	if CompareStdout(arguments, test_info, a["stdout"], b["stdout"]):
		PrintToStream (stream, "[*] [   passed   ]: Standard out comparison of " + a_name + " and " + b_name)
	else:
		PrintToStream (stream, "[*] [ * FAILED * ]: Standard out comparison of " + a_name + " and " + b_name)
	
	if not short:
		if (len(a["timings"]) != len(b["timings"])):
			print ("[*] Timings do not match")
		else:
			print ("[*] Timings")
			for i in range(len(a["timings"])):
				index = i + 1
				time_a = a["timings"][i]
				time_b = b["timings"][i]
				
				print ("\t[*] Timing %u:" % index)
				print ("\t\t%s = %0.02f s" % (a_name, time_a))
				print ("\t\t%s = %0.02f s" % (b_name, time_b))
				print ("\t\t%s/%s = %0.02f%%" % (a_name, b_name, 100.0 * time_a/time_b))
				print ("\t\t%s/%s = %0.02f%%" % (b_name, a_name, 100.0 * time_b/time_a))


def Run(arguments, test_info):
	if (arguments["type"] == "auto") or (arguments["type"] == "auto-short"):
		if not "auto" in test_info:
			print ("[*] Automatic testing is not supported")
			return
		short = arguments["type"] == "auto-short"
		
		# Polycute
		arguments["type"] = "polycute"
		polycute_info = Run(arguments, test_info)
		
		# Clang
		arguments["type"] = "clang"
		clang_info = Run(arguments, test_info)
		
		if not short:
			# GCC
			arguments["type"] = "gcc"
			gcc_info = Run(arguments, test_info)
			
			# GCC OpenMP
			arguments["type"] = "gcc-openmp"
			gcc_openmp_info = Run(arguments, test_info)
		
		# Clean up
		arguments["type"] = "clean"
		Run(arguments, test_info)
		
		# Print Information
		PrintInfo(arguments, test_info, polycute_info, clang_info, "Polycute", "Clang", short)
		
		if not short:
			PrintInfo(arguments, test_info, polycute_info, gcc_info, "Polycute", "GCC", False)
			PrintInfo(arguments, test_info, polycute_info, gcc_openmp_info, "Polycute", "GCC-OpenMP", False)
			PrintInfo(arguments, test_info, clang_info, gcc_info, "Clang", "GCC", False)
			PrintInfo(arguments, test_info, clang_info, gcc_openmp_info, "Clang", "GCC-OpenMP", False)
		
		
	elif arguments["type"] == "clean":
		Clean(test_info)
	else:
		if (arguments["type"] == "polycute"):
			type_name = "Polycute"
		elif (arguments["type"] == "gcc"):
			type_name = "GCC"
		elif (arguments["type"] == "clang"):
			type_name = "Clang"
		elif (arguments["type"] == "gcc-openmp"):
			type_name = "GCC-OpenMP"
		else:
			Error("Unknown test type")
		print ("[*] Running " + type_name + " Test")
		
		
		info = dict()
		
		if arguments["compile"]:
			print ("[*] Building")
			t1 = GetTime()
			BuildTest(arguments, test_info)
			t2 = GetTime()
			build_time = t2 - t1
			print ("[*] Build time: %0.02f s" % build_time)
			info["build time"] = build_time
		
		if arguments["execute"]:
			print ("[*] Executing")
			t1 = GetTime()
			stdout,timings = RunTest(arguments, test_info)
			t2 = GetTime()
			run_time = t2 - t1
			print ("[*] Execution time: %0.02f s" % run_time)
			info["run time"] = run_time
			info["stdout"] = stdout
			info["timings"] = timings
		
		return info


# ======== Main ========
arguments = GetArguments()
original_directory = Cd(arguments["dir"])
test_info = GetTestInfo()
Run(arguments, test_info)
Cd(original_directory)


